import datetime
import numba
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.core.imputils import lower_constant
from numba.extending import NativeValue, box, intrinsic, lower_cast, make_attribute_wrapper, models, overload, overload_attribute, overload_method, register_jitable, register_model, typeof_impl, unbox
"""
Implementation is based on
https://github.com/python/cpython/blob/39a5c889d30d03a88102e56f03ee0c95db198fb3/Lib/datetime.py
"""


class DatetimeDatetimeType(types.Type):

    def __init__(self):
        super(DatetimeDatetimeType, self).__init__(name=
            'DatetimeDatetimeType()')


datetime_datetime_type = DatetimeDatetimeType()
types.datetime_datetime_type = datetime_datetime_type


@typeof_impl.register(datetime.datetime)
def typeof_datetime_datetime(val, c):
    return datetime_datetime_type


@register_model(DatetimeDatetimeType)
class DatetimeDateTimeModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        pkjw__mxr = [('year', types.int64), ('month', types.int64), ('day',
            types.int64), ('hour', types.int64), ('minute', types.int64), (
            'second', types.int64), ('microsecond', types.int64)]
        super(DatetimeDateTimeModel, self).__init__(dmm, fe_type, pkjw__mxr)


@box(DatetimeDatetimeType)
def box_datetime_datetime(typ, val, c):
    xsobv__rzvsw = cgutils.create_struct_proxy(typ)(c.context, c.builder,
        value=val)
    wqfip__wenm = c.pyapi.long_from_longlong(xsobv__rzvsw.year)
    mxmo__ohzec = c.pyapi.long_from_longlong(xsobv__rzvsw.month)
    xan__auz = c.pyapi.long_from_longlong(xsobv__rzvsw.day)
    zqunu__nlk = c.pyapi.long_from_longlong(xsobv__rzvsw.hour)
    vpuzx__kec = c.pyapi.long_from_longlong(xsobv__rzvsw.minute)
    rypx__eaqe = c.pyapi.long_from_longlong(xsobv__rzvsw.second)
    xrt__hsduj = c.pyapi.long_from_longlong(xsobv__rzvsw.microsecond)
    lqld__owxqo = c.pyapi.unserialize(c.pyapi.serialize_object(datetime.
        datetime))
    hwoey__agq = c.pyapi.call_function_objargs(lqld__owxqo, (wqfip__wenm,
        mxmo__ohzec, xan__auz, zqunu__nlk, vpuzx__kec, rypx__eaqe, xrt__hsduj))
    c.pyapi.decref(wqfip__wenm)
    c.pyapi.decref(mxmo__ohzec)
    c.pyapi.decref(xan__auz)
    c.pyapi.decref(zqunu__nlk)
    c.pyapi.decref(vpuzx__kec)
    c.pyapi.decref(rypx__eaqe)
    c.pyapi.decref(xrt__hsduj)
    c.pyapi.decref(lqld__owxqo)
    return hwoey__agq


@unbox(DatetimeDatetimeType)
def unbox_datetime_datetime(typ, val, c):
    wqfip__wenm = c.pyapi.object_getattr_string(val, 'year')
    mxmo__ohzec = c.pyapi.object_getattr_string(val, 'month')
    xan__auz = c.pyapi.object_getattr_string(val, 'day')
    zqunu__nlk = c.pyapi.object_getattr_string(val, 'hour')
    vpuzx__kec = c.pyapi.object_getattr_string(val, 'minute')
    rypx__eaqe = c.pyapi.object_getattr_string(val, 'second')
    xrt__hsduj = c.pyapi.object_getattr_string(val, 'microsecond')
    xsobv__rzvsw = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    xsobv__rzvsw.year = c.pyapi.long_as_longlong(wqfip__wenm)
    xsobv__rzvsw.month = c.pyapi.long_as_longlong(mxmo__ohzec)
    xsobv__rzvsw.day = c.pyapi.long_as_longlong(xan__auz)
    xsobv__rzvsw.hour = c.pyapi.long_as_longlong(zqunu__nlk)
    xsobv__rzvsw.minute = c.pyapi.long_as_longlong(vpuzx__kec)
    xsobv__rzvsw.second = c.pyapi.long_as_longlong(rypx__eaqe)
    xsobv__rzvsw.microsecond = c.pyapi.long_as_longlong(xrt__hsduj)
    c.pyapi.decref(wqfip__wenm)
    c.pyapi.decref(mxmo__ohzec)
    c.pyapi.decref(xan__auz)
    c.pyapi.decref(zqunu__nlk)
    c.pyapi.decref(vpuzx__kec)
    c.pyapi.decref(rypx__eaqe)
    c.pyapi.decref(xrt__hsduj)
    jzyvl__guwu = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(xsobv__rzvsw._getvalue(), is_error=jzyvl__guwu)


@lower_constant(DatetimeDatetimeType)
def constant_datetime(context, builder, ty, pyval):
    year = context.get_constant(types.int64, pyval.year)
    month = context.get_constant(types.int64, pyval.month)
    day = context.get_constant(types.int64, pyval.day)
    hour = context.get_constant(types.int64, pyval.hour)
    minute = context.get_constant(types.int64, pyval.minute)
    second = context.get_constant(types.int64, pyval.second)
    microsecond = context.get_constant(types.int64, pyval.microsecond)
    return lir.Constant.literal_struct([year, month, day, hour, minute,
        second, microsecond])


@overload(datetime.datetime, no_unliteral=True)
def datetime_datetime(year, month, day, hour=0, minute=0, second=0,
    microsecond=0):

    def impl_datetime(year, month, day, hour=0, minute=0, second=0,
        microsecond=0):
        return init_datetime(year, month, day, hour, minute, second,
            microsecond)
    return impl_datetime


@intrinsic
def init_datetime(typingctx, year, month, day, hour, minute, second,
    microsecond):

    def codegen(context, builder, signature, args):
        typ = signature.return_type
        xsobv__rzvsw = cgutils.create_struct_proxy(typ)(context, builder)
        xsobv__rzvsw.year = args[0]
        xsobv__rzvsw.month = args[1]
        xsobv__rzvsw.day = args[2]
        xsobv__rzvsw.hour = args[3]
        xsobv__rzvsw.minute = args[4]
        xsobv__rzvsw.second = args[5]
        xsobv__rzvsw.microsecond = args[6]
        return xsobv__rzvsw._getvalue()
    return DatetimeDatetimeType()(year, month, day, hour, minute, second,
        microsecond), codegen


make_attribute_wrapper(DatetimeDatetimeType, 'year', '_year')
make_attribute_wrapper(DatetimeDatetimeType, 'month', '_month')
make_attribute_wrapper(DatetimeDatetimeType, 'day', '_day')
make_attribute_wrapper(DatetimeDatetimeType, 'hour', '_hour')
make_attribute_wrapper(DatetimeDatetimeType, 'minute', '_minute')
make_attribute_wrapper(DatetimeDatetimeType, 'second', '_second')
make_attribute_wrapper(DatetimeDatetimeType, 'microsecond', '_microsecond')


@overload_attribute(DatetimeDatetimeType, 'year')
def datetime_get_year(dt):

    def impl(dt):
        return dt._year
    return impl


@overload_attribute(DatetimeDatetimeType, 'month')
def datetime_get_month(dt):

    def impl(dt):
        return dt._month
    return impl


@overload_attribute(DatetimeDatetimeType, 'day')
def datetime_get_day(dt):

    def impl(dt):
        return dt._day
    return impl


@overload_attribute(DatetimeDatetimeType, 'hour')
def datetime_get_hour(dt):

    def impl(dt):
        return dt._hour
    return impl


@overload_attribute(DatetimeDatetimeType, 'minute')
def datetime_get_minute(dt):

    def impl(dt):
        return dt._minute
    return impl


@overload_attribute(DatetimeDatetimeType, 'second')
def datetime_get_second(dt):

    def impl(dt):
        return dt._second
    return impl


@overload_attribute(DatetimeDatetimeType, 'microsecond')
def datetime_get_microsecond(dt):

    def impl(dt):
        return dt._microsecond
    return impl


@overload_method(DatetimeDatetimeType, 'date', no_unliteral=True)
def date(dt):

    def impl(dt):
        return datetime.date(dt.year, dt.month, dt.day)
    return impl


@register_jitable
def now_impl():
    with numba.objmode(d='datetime_datetime_type'):
        d = datetime.datetime.now()
    return d


@register_jitable
def today_impl():
    with numba.objmode(d='datetime_datetime_type'):
        d = datetime.datetime.today()
    return d


@register_jitable
def strptime_impl(date_string, dtformat):
    with numba.objmode(d='datetime_datetime_type'):
        d = datetime.datetime.strptime(date_string, dtformat)
    return d


@register_jitable
def _cmp(x, y):
    return 0 if x == y else 1 if x > y else -1


def create_cmp_op_overload(op):

    def overload_datetime_cmp(lhs, rhs):
        if lhs == datetime_datetime_type and rhs == datetime_datetime_type:

            def impl(lhs, rhs):
                y, ktbze__qtx = lhs.year, rhs.year
                sbazr__vtl, kcwrq__blwl = lhs.month, rhs.month
                d, rfuml__yfhw = lhs.day, rhs.day
                xrqd__stzai, cult__vyqc = lhs.hour, rhs.hour
                mmbi__cxvkd, gpqoe__viypa = lhs.minute, rhs.minute
                ymvr__cngsg, lyzqi__fkoo = lhs.second, rhs.second
                osysl__hxx, acwnc__mfys = lhs.microsecond, rhs.microsecond
                return op(_cmp((y, sbazr__vtl, d, xrqd__stzai, mmbi__cxvkd,
                    ymvr__cngsg, osysl__hxx), (ktbze__qtx, kcwrq__blwl,
                    rfuml__yfhw, cult__vyqc, gpqoe__viypa, lyzqi__fkoo,
                    acwnc__mfys)), 0)
            return impl
    return overload_datetime_cmp


def overload_sub_operator_datetime_datetime(lhs, rhs):
    if lhs == datetime_datetime_type and rhs == datetime_datetime_type:

        def impl(lhs, rhs):
            fbej__hzcq = lhs.toordinal()
            nkoe__hskit = rhs.toordinal()
            agcl__mskjt = lhs.second + lhs.minute * 60 + lhs.hour * 3600
            zqk__ixo = rhs.second + rhs.minute * 60 + rhs.hour * 3600
            qosjp__lelai = datetime.timedelta(fbej__hzcq - nkoe__hskit, 
                agcl__mskjt - zqk__ixo, lhs.microsecond - rhs.microsecond)
            return qosjp__lelai
        return impl


@lower_cast(types.Optional(numba.core.types.NPTimedelta('ns')), numba.core.
    types.NPTimedelta('ns'))
@lower_cast(types.Optional(numba.core.types.NPDatetime('ns')), numba.core.
    types.NPDatetime('ns'))
def optional_dt64_to_dt64(context, builder, fromty, toty, val):
    hwupn__pkd = context.make_helper(builder, fromty, value=val)
    dcett__yaq = cgutils.as_bool_bit(builder, hwupn__pkd.valid)
    with builder.if_else(dcett__yaq) as (then, orelse):
        with then:
            bpbe__jhihr = context.cast(builder, hwupn__pkd.data, fromty.
                type, toty)
            rtmm__sqjn = builder.block
        with orelse:
            vpvw__mud = numba.np.npdatetime.NAT
            ecm__czqc = builder.block
    hwoey__agq = builder.phi(bpbe__jhihr.type)
    hwoey__agq.add_incoming(bpbe__jhihr, rtmm__sqjn)
    hwoey__agq.add_incoming(vpvw__mud, ecm__czqc)
    return hwoey__agq
