"""Numba extension support for datetime.timedelta objects and their arrays.
"""
import datetime
import operator
from collections import namedtuple
import llvmlite.binding as ll
import numba
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.core.imputils import lower_constant
from numba.extending import NativeValue, box, intrinsic, make_attribute_wrapper, models, overload, overload_attribute, overload_method, register_jitable, register_model, typeof_impl, unbox
from numba.parfors.array_analysis import ArrayAnalysis
import bodo
from bodo.hiframes.datetime_datetime_ext import datetime_datetime_type
from bodo.libs import hdatetime_ext
from bodo.utils.indexing import get_new_null_mask_bool_index, get_new_null_mask_int_index, get_new_null_mask_slice_index, setitem_slice_index_null_bits
from bodo.utils.typing import BodoError, get_overload_const_str, is_iterable_type, is_list_like_index_type, is_overload_constant_str
ll.add_symbol('box_datetime_timedelta_array', hdatetime_ext.
    box_datetime_timedelta_array)
ll.add_symbol('unbox_datetime_timedelta_array', hdatetime_ext.
    unbox_datetime_timedelta_array)


class NoInput:
    pass


_no_input = NoInput()


class NoInputType(types.Type):

    def __init__(self):
        super(NoInputType, self).__init__(name='NoInput')


register_model(NoInputType)(models.OpaqueModel)


@typeof_impl.register(NoInput)
def _typ_no_input(val, c):
    return NoInputType()


@lower_constant(NoInputType)
def constant_no_input(context, builder, ty, pyval):
    return context.get_dummy_value()


class PDTimeDeltaType(types.Type):

    def __init__(self):
        super(PDTimeDeltaType, self).__init__(name='PDTimeDeltaType()')


pd_timedelta_type = PDTimeDeltaType()
types.pd_timedelta_type = pd_timedelta_type


@typeof_impl.register(pd.Timedelta)
def typeof_pd_timedelta(val, c):
    return pd_timedelta_type


@register_model(PDTimeDeltaType)
class PDTimeDeltaModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        rpbc__ean = [('value', types.int64)]
        super(PDTimeDeltaModel, self).__init__(dmm, fe_type, rpbc__ean)


@box(PDTimeDeltaType)
def box_pd_timedelta(typ, val, c):
    xuqgh__wrx = cgutils.create_struct_proxy(typ)(c.context, c.builder,
        value=val)
    iqyul__vnf = c.pyapi.long_from_longlong(xuqgh__wrx.value)
    wxgee__lotu = c.pyapi.unserialize(c.pyapi.serialize_object(pd.Timedelta))
    res = c.pyapi.call_function_objargs(wxgee__lotu, (iqyul__vnf,))
    c.pyapi.decref(iqyul__vnf)
    c.pyapi.decref(wxgee__lotu)
    return res


@unbox(PDTimeDeltaType)
def unbox_pd_timedelta(typ, val, c):
    iqyul__vnf = c.pyapi.object_getattr_string(val, 'value')
    dqi__pta = c.pyapi.long_as_longlong(iqyul__vnf)
    xuqgh__wrx = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    xuqgh__wrx.value = dqi__pta
    c.pyapi.decref(iqyul__vnf)
    acxpc__swrry = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(xuqgh__wrx._getvalue(), is_error=acxpc__swrry)


@lower_constant(PDTimeDeltaType)
def lower_constant_pd_timedelta(context, builder, ty, pyval):
    value = context.get_constant(types.int64, pyval.value)
    return lir.Constant.literal_struct([value])


@overload(pd.Timedelta, no_unliteral=True)
def pd_timedelta(value=_no_input, unit='ns', days=0, seconds=0,
    microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0):
    if value == _no_input:

        def impl_timedelta_kw(value=_no_input, unit='ns', days=0, seconds=0,
            microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0):
            days += weeks * 7
            hours += days * 24
            minutes += 60 * hours
            seconds += 60 * minutes
            milliseconds += 1000 * seconds
            microseconds += 1000 * milliseconds
            hts__lxm = 1000 * microseconds
            return init_pd_timedelta(hts__lxm)
        return impl_timedelta_kw
    if value == bodo.string_type or is_overload_constant_str(value):

        def impl_str(value=_no_input, unit='ns', days=0, seconds=0,
            microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0):
            with numba.objmode(res='pd_timedelta_type'):
                res = pd.Timedelta(value)
            return res
        return impl_str
    if value == pd_timedelta_type:
        return (lambda value=_no_input, unit='ns', days=0, seconds=0,
            microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0: value)
    if value == datetime_timedelta_type:

        def impl_timedelta_datetime(value=_no_input, unit='ns', days=0,
            seconds=0, microseconds=0, milliseconds=0, minutes=0, hours=0,
            weeks=0):
            days = value.days
            seconds = 60 * 60 * 24 * days + value.seconds
            microseconds = 1000 * 1000 * seconds + value.microseconds
            hts__lxm = 1000 * microseconds
            return init_pd_timedelta(hts__lxm)
        return impl_timedelta_datetime
    if not is_overload_constant_str(unit):
        raise BodoError('pd.to_timedelta(): unit should be a constant string')
    unit = pd._libs.tslibs.timedeltas.parse_timedelta_unit(
        get_overload_const_str(unit))
    pdqg__fvdp, sef__ngged = pd._libs.tslibs.conversion.precision_from_unit(
        unit)

    def impl_timedelta(value=_no_input, unit='ns', days=0, seconds=0,
        microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0):
        return init_pd_timedelta(value * pdqg__fvdp)
    return impl_timedelta


@intrinsic
def init_pd_timedelta(typingctx, value):

    def codegen(context, builder, signature, args):
        typ = signature.return_type
        timedelta = cgutils.create_struct_proxy(typ)(context, builder)
        timedelta.value = args[0]
        return timedelta._getvalue()
    return PDTimeDeltaType()(value), codegen


make_attribute_wrapper(PDTimeDeltaType, 'value', '_value')


@overload_attribute(PDTimeDeltaType, 'value')
@overload_attribute(PDTimeDeltaType, 'delta')
def pd_timedelta_get_value(td):

    def impl(td):
        return td._value
    return impl


@overload_attribute(PDTimeDeltaType, 'days')
def pd_timedelta_get_days(td):

    def impl(td):
        return td._value // (1000 * 1000 * 1000 * 60 * 60 * 24)
    return impl


@overload_attribute(PDTimeDeltaType, 'seconds')
def pd_timedelta_get_seconds(td):

    def impl(td):
        return td._value // (1000 * 1000 * 1000) % (60 * 60 * 24)
    return impl


@overload_attribute(PDTimeDeltaType, 'microseconds')
def pd_timedelta_get_microseconds(td):

    def impl(td):
        return td._value // 1000 % 1000000
    return impl


@overload_attribute(PDTimeDeltaType, 'nanoseconds')
def pd_timedelta_get_nanoseconds(td):

    def impl(td):
        return td._value % 1000
    return impl


@register_jitable
def _to_hours_pd_td(td):
    return td._value // (1000 * 1000 * 1000 * 60 * 60) % 24


@register_jitable
def _to_minutes_pd_td(td):
    return td._value // (1000 * 1000 * 1000 * 60) % 60


@register_jitable
def _to_seconds_pd_td(td):
    return td._value // (1000 * 1000 * 1000) % 60


@register_jitable
def _to_milliseconds_pd_td(td):
    return td._value // (1000 * 1000) % 1000


@register_jitable
def _to_microseconds_pd_td(td):
    return td._value // 1000 % 1000


Components = namedtuple('Components', ['days', 'hours', 'minutes',
    'seconds', 'milliseconds', 'microseconds', 'nanoseconds'], defaults=[0,
    0, 0, 0, 0, 0, 0])


@overload_attribute(PDTimeDeltaType, 'components', no_unliteral=True)
def pd_timedelta_get_components(td):

    def impl(td):
        a = Components(td.days, _to_hours_pd_td(td), _to_minutes_pd_td(td),
            _to_seconds_pd_td(td), _to_milliseconds_pd_td(td),
            _to_microseconds_pd_td(td), td.nanoseconds)
        return a
    return impl


@overload_method(PDTimeDeltaType, '__hash__', no_unliteral=True)
def pd_td___hash__(td):

    def impl(td):
        return hash(td._value)
    return impl


@overload_method(PDTimeDeltaType, 'to_numpy', no_unliteral=True)
@overload_method(PDTimeDeltaType, 'to_timedelta64', no_unliteral=True)
def pd_td_to_numpy(td):
    from bodo.hiframes.pd_timestamp_ext import integer_to_timedelta64

    def impl(td):
        return integer_to_timedelta64(td.value)
    return impl


@overload_method(PDTimeDeltaType, 'to_pytimedelta', no_unliteral=True)
def pd_td_to_pytimedelta(td):

    def impl(td):
        return datetime.timedelta(microseconds=np.int64(td._value / 1000))
    return impl


@overload_method(PDTimeDeltaType, 'total_seconds', no_unliteral=True)
def pd_td_total_seconds(td):

    def impl(td):
        return td._value // 1000 / 10 ** 6
    return impl


def overload_add_operator_datetime_timedelta(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            val = lhs.value + rhs.value
            return pd.Timedelta(val)
        return impl
    if lhs == pd_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            blus__fdzr = (rhs.microseconds + (rhs.seconds + rhs.days * 60 *
                60 * 24) * 1000 * 1000) * 1000
            val = lhs.value + blus__fdzr
            return pd.Timedelta(val)
        return impl
    if lhs == datetime_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            kjv__wfr = (lhs.microseconds + (lhs.seconds + lhs.days * 60 * 
                60 * 24) * 1000 * 1000) * 1000
            val = kjv__wfr + rhs.value
            return pd.Timedelta(val)
        return impl
    if lhs == pd_timedelta_type and rhs == datetime_datetime_type:
        from bodo.hiframes.pd_timestamp_ext import compute_pd_timestamp

        def impl(lhs, rhs):
            xjs__ttua = rhs.toordinal()
            xrfyc__rwe = rhs.second + rhs.minute * 60 + rhs.hour * 3600
            hmkos__nvw = rhs.microsecond
            fgdqr__kzxa = lhs.value // 1000
            qvql__akeph = lhs.nanoseconds
            wlqr__mmdd = hmkos__nvw + fgdqr__kzxa
            votdw__dcfim = 1000000 * (xjs__ttua * 86400 + xrfyc__rwe
                ) + wlqr__mmdd
            jhtje__sxyc = qvql__akeph
            return compute_pd_timestamp(votdw__dcfim, jhtje__sxyc)
        return impl
    if lhs == datetime_datetime_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return lhs + rhs.to_pytimedelta()
        return impl
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            d = lhs.days + rhs.days
            s = lhs.seconds + rhs.seconds
            us = lhs.microseconds + rhs.microseconds
            return datetime.timedelta(d, s, us)
        return impl
    if lhs == datetime_timedelta_type and rhs == datetime_datetime_type:

        def impl(lhs, rhs):
            weea__ysd = datetime.timedelta(rhs.toordinal(), hours=rhs.hour,
                minutes=rhs.minute, seconds=rhs.second, microseconds=rhs.
                microsecond)
            weea__ysd = weea__ysd + lhs
            emjso__ufm, gxefg__vqa = divmod(weea__ysd.seconds, 3600)
            kus__tarb, izn__zuvo = divmod(gxefg__vqa, 60)
            if 0 < weea__ysd.days <= _MAXORDINAL:
                d = bodo.hiframes.datetime_date_ext.fromordinal_impl(weea__ysd
                    .days)
                return datetime.datetime(d.year, d.month, d.day, emjso__ufm,
                    kus__tarb, izn__zuvo, weea__ysd.microseconds)
            raise OverflowError('result out of range')
        return impl
    if lhs == datetime_datetime_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            weea__ysd = datetime.timedelta(lhs.toordinal(), hours=lhs.hour,
                minutes=lhs.minute, seconds=lhs.second, microseconds=lhs.
                microsecond)
            weea__ysd = weea__ysd + rhs
            emjso__ufm, gxefg__vqa = divmod(weea__ysd.seconds, 3600)
            kus__tarb, izn__zuvo = divmod(gxefg__vqa, 60)
            if 0 < weea__ysd.days <= _MAXORDINAL:
                d = bodo.hiframes.datetime_date_ext.fromordinal_impl(weea__ysd
                    .days)
                return datetime.datetime(d.year, d.month, d.day, emjso__ufm,
                    kus__tarb, izn__zuvo, weea__ysd.microseconds)
            raise OverflowError('result out of range')
        return impl


def overload_sub_operator_datetime_timedelta(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            klfd__bmr = lhs.value - rhs.value
            return pd.Timedelta(klfd__bmr)
        return impl
    if lhs == pd_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            return lhs + -rhs
        return impl
    if lhs == datetime_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return lhs + -rhs
        return impl
    if lhs == datetime_datetime_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return lhs + -rhs
        return impl
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            d = lhs.days - rhs.days
            s = lhs.seconds - rhs.seconds
            us = lhs.microseconds - rhs.microseconds
            return datetime.timedelta(d, s, us)
        return impl
    if lhs == datetime_datetime_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            return lhs + -rhs
        return impl
    if lhs == datetime_timedelta_array_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            kfvf__lgst = lhs
            numba.parfors.parfor.init_prange()
            n = len(kfvf__lgst)
            A = alloc_datetime_timedelta_array(n)
            for vpmt__cxve in numba.parfors.parfor.internal_prange(n):
                A[vpmt__cxve] = kfvf__lgst[vpmt__cxve] - rhs
            return A
        return impl


def overload_mul_operator_timedelta(lhs, rhs):
    if lhs == pd_timedelta_type and isinstance(rhs, types.Integer):

        def impl(lhs, rhs):
            return pd.Timedelta(lhs.value * rhs)
        return impl
    elif isinstance(lhs, types.Integer) and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return pd.Timedelta(rhs.value * lhs)
        return impl
    if lhs == datetime_timedelta_type and isinstance(rhs, types.Integer):

        def impl(lhs, rhs):
            d = lhs.days * rhs
            s = lhs.seconds * rhs
            us = lhs.microseconds * rhs
            return datetime.timedelta(d, s, us)
        return impl
    elif isinstance(lhs, types.Integer) and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            d = lhs * rhs.days
            s = lhs * rhs.seconds
            us = lhs * rhs.microseconds
            return datetime.timedelta(d, s, us)
        return impl


def overload_floordiv_operator_pd_timedelta(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return lhs.value // rhs.value
        return impl
    elif lhs == pd_timedelta_type and isinstance(rhs, types.Integer):

        def impl(lhs, rhs):
            return pd.Timedelta(lhs.value // rhs)
        return impl


def overload_truediv_operator_pd_timedelta(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return lhs.value / rhs.value
        return impl
    elif lhs == pd_timedelta_type and isinstance(rhs, types.Integer):

        def impl(lhs, rhs):
            return pd.Timedelta(int(lhs.value / rhs))
        return impl


def overload_mod_operator_timedeltas(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return pd.Timedelta(lhs.value % rhs.value)
        return impl
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            lrqr__ydt = _to_microseconds(lhs) % _to_microseconds(rhs)
            return datetime.timedelta(0, 0, lrqr__ydt)
        return impl


def pd_create_cmp_op_overload(op):

    def overload_pd_timedelta_cmp(lhs, rhs):
        if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

            def impl(lhs, rhs):
                return op(lhs.value, rhs.value)
            return impl
        if lhs == pd_timedelta_type and rhs == bodo.timedelta64ns:
            return lambda lhs, rhs: op(bodo.hiframes.pd_timestamp_ext.
                integer_to_timedelta64(lhs.value), rhs)
        if lhs == bodo.timedelta64ns and rhs == pd_timedelta_type:
            return lambda lhs, rhs: op(lhs, bodo.hiframes.pd_timestamp_ext.
                integer_to_timedelta64(rhs.value))
    return overload_pd_timedelta_cmp


@overload(operator.neg, no_unliteral=True)
def pd_timedelta_neg(lhs):
    if lhs == pd_timedelta_type:

        def impl(lhs):
            return pd.Timedelta(-lhs.value)
        return impl


@overload(operator.pos, no_unliteral=True)
def pd_timedelta_pos(lhs):
    if lhs == pd_timedelta_type:

        def impl(lhs):
            return lhs
        return impl


@overload(divmod, no_unliteral=True)
def pd_timedelta_divmod(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            rkh__sqzr, lrqr__ydt = divmod(lhs.value, rhs.value)
            return rkh__sqzr, pd.Timedelta(lrqr__ydt)
        return impl


@overload(abs, no_unliteral=True)
def pd_timedelta_abs(lhs):
    if lhs == pd_timedelta_type:

        def impl(lhs):
            if lhs.value < 0:
                return -lhs
            else:
                return lhs
        return impl


class DatetimeTimeDeltaType(types.Type):

    def __init__(self):
        super(DatetimeTimeDeltaType, self).__init__(name=
            'DatetimeTimeDeltaType()')


datetime_timedelta_type = DatetimeTimeDeltaType()


@typeof_impl.register(datetime.timedelta)
def typeof_datetime_timedelta(val, c):
    return datetime_timedelta_type


@register_model(DatetimeTimeDeltaType)
class DatetimeTimeDeltaModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        rpbc__ean = [('days', types.int64), ('seconds', types.int64), (
            'microseconds', types.int64)]
        super(DatetimeTimeDeltaModel, self).__init__(dmm, fe_type, rpbc__ean)


@box(DatetimeTimeDeltaType)
def box_datetime_timedelta(typ, val, c):
    xuqgh__wrx = cgutils.create_struct_proxy(typ)(c.context, c.builder,
        value=val)
    gyv__fti = c.pyapi.long_from_longlong(xuqgh__wrx.days)
    zspg__tceh = c.pyapi.long_from_longlong(xuqgh__wrx.seconds)
    uoyio__tafw = c.pyapi.long_from_longlong(xuqgh__wrx.microseconds)
    wxgee__lotu = c.pyapi.unserialize(c.pyapi.serialize_object(datetime.
        timedelta))
    res = c.pyapi.call_function_objargs(wxgee__lotu, (gyv__fti, zspg__tceh,
        uoyio__tafw))
    c.pyapi.decref(gyv__fti)
    c.pyapi.decref(zspg__tceh)
    c.pyapi.decref(uoyio__tafw)
    c.pyapi.decref(wxgee__lotu)
    return res


@unbox(DatetimeTimeDeltaType)
def unbox_datetime_timedelta(typ, val, c):
    gyv__fti = c.pyapi.object_getattr_string(val, 'days')
    zspg__tceh = c.pyapi.object_getattr_string(val, 'seconds')
    uoyio__tafw = c.pyapi.object_getattr_string(val, 'microseconds')
    zurq__wii = c.pyapi.long_as_longlong(gyv__fti)
    rfum__bpy = c.pyapi.long_as_longlong(zspg__tceh)
    tfha__gox = c.pyapi.long_as_longlong(uoyio__tafw)
    xuqgh__wrx = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    xuqgh__wrx.days = zurq__wii
    xuqgh__wrx.seconds = rfum__bpy
    xuqgh__wrx.microseconds = tfha__gox
    c.pyapi.decref(gyv__fti)
    c.pyapi.decref(zspg__tceh)
    c.pyapi.decref(uoyio__tafw)
    acxpc__swrry = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(xuqgh__wrx._getvalue(), is_error=acxpc__swrry)


@lower_constant(DatetimeTimeDeltaType)
def lower_constant_datetime_timedelta(context, builder, ty, pyval):
    days = context.get_constant(types.int64, pyval.days)
    seconds = context.get_constant(types.int64, pyval.seconds)
    microseconds = context.get_constant(types.int64, pyval.microseconds)
    return lir.Constant.literal_struct([days, seconds, microseconds])


@overload(datetime.timedelta, no_unliteral=True)
def datetime_timedelta(days=0, seconds=0, microseconds=0, milliseconds=0,
    minutes=0, hours=0, weeks=0):

    def impl_timedelta(days=0, seconds=0, microseconds=0, milliseconds=0,
        minutes=0, hours=0, weeks=0):
        d = s = us = 0
        days += weeks * 7
        seconds += minutes * 60 + hours * 3600
        microseconds += milliseconds * 1000
        d = days
        days, seconds = divmod(seconds, 24 * 3600)
        d += days
        s += int(seconds)
        seconds, us = divmod(microseconds, 1000000)
        days, seconds = divmod(seconds, 24 * 3600)
        d += days
        s += seconds
        return init_timedelta(d, s, us)
    return impl_timedelta


@intrinsic
def init_timedelta(typingctx, d, s, us):

    def codegen(context, builder, signature, args):
        typ = signature.return_type
        timedelta = cgutils.create_struct_proxy(typ)(context, builder)
        timedelta.days = args[0]
        timedelta.seconds = args[1]
        timedelta.microseconds = args[2]
        return timedelta._getvalue()
    return DatetimeTimeDeltaType()(d, s, us), codegen


make_attribute_wrapper(DatetimeTimeDeltaType, 'days', '_days')
make_attribute_wrapper(DatetimeTimeDeltaType, 'seconds', '_seconds')
make_attribute_wrapper(DatetimeTimeDeltaType, 'microseconds', '_microseconds')


@overload_attribute(DatetimeTimeDeltaType, 'days')
def timedelta_get_days(td):

    def impl(td):
        return td._days
    return impl


@overload_attribute(DatetimeTimeDeltaType, 'seconds')
def timedelta_get_seconds(td):

    def impl(td):
        return td._seconds
    return impl


@overload_attribute(DatetimeTimeDeltaType, 'microseconds')
def timedelta_get_microseconds(td):

    def impl(td):
        return td._microseconds
    return impl


@overload_method(DatetimeTimeDeltaType, 'total_seconds', no_unliteral=True)
def total_seconds(td):

    def impl(td):
        return ((td._days * 86400 + td._seconds) * 10 ** 6 + td._microseconds
            ) / 10 ** 6
    return impl


@overload_method(DatetimeTimeDeltaType, '__hash__', no_unliteral=True)
def __hash__(td):

    def impl(td):
        return hash((td._days, td._seconds, td._microseconds))
    return impl


@register_jitable
def _to_nanoseconds(td):
    return np.int64(((td._days * 86400 + td._seconds) * 1000000 + td.
        _microseconds) * 1000)


@register_jitable
def _to_microseconds(td):
    return (td._days * (24 * 3600) + td._seconds) * 1000000 + td._microseconds


@register_jitable
def _cmp(x, y):
    return 0 if x == y else 1 if x > y else -1


@register_jitable
def _getstate(td):
    return td._days, td._seconds, td._microseconds


@register_jitable
def _divide_and_round(a, b):
    rkh__sqzr, lrqr__ydt = divmod(a, b)
    lrqr__ydt *= 2
    uagq__qkr = lrqr__ydt > b if b > 0 else lrqr__ydt < b
    if uagq__qkr or lrqr__ydt == b and rkh__sqzr % 2 == 1:
        rkh__sqzr += 1
    return rkh__sqzr


_MAXORDINAL = 3652059


def overload_floordiv_operator_dt_timedelta(lhs, rhs):
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            us = _to_microseconds(lhs)
            return us // _to_microseconds(rhs)
        return impl
    elif lhs == datetime_timedelta_type and rhs == types.int64:

        def impl(lhs, rhs):
            us = _to_microseconds(lhs)
            return datetime.timedelta(0, 0, us // rhs)
        return impl


def overload_truediv_operator_dt_timedelta(lhs, rhs):
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            us = _to_microseconds(lhs)
            return us / _to_microseconds(rhs)
        return impl
    elif lhs == datetime_timedelta_type and rhs == types.int64:

        def impl(lhs, rhs):
            us = _to_microseconds(lhs)
            return datetime.timedelta(0, 0, _divide_and_round(us, rhs))
        return impl


def create_cmp_op_overload(op):

    def overload_timedelta_cmp(lhs, rhs):
        if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

            def impl(lhs, rhs):
                ivq__zgans = _cmp(_getstate(lhs), _getstate(rhs))
                return op(ivq__zgans, 0)
            return impl
    return overload_timedelta_cmp


@overload(operator.neg, no_unliteral=True)
def timedelta_neg(lhs):
    if lhs == datetime_timedelta_type:

        def impl(lhs):
            return datetime.timedelta(-lhs.days, -lhs.seconds, -lhs.
                microseconds)
        return impl


@overload(operator.pos, no_unliteral=True)
def timedelta_pos(lhs):
    if lhs == datetime_timedelta_type:

        def impl(lhs):
            return lhs
        return impl


@overload(divmod, no_unliteral=True)
def timedelta_divmod(lhs, rhs):
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            rkh__sqzr, lrqr__ydt = divmod(_to_microseconds(lhs),
                _to_microseconds(rhs))
            return rkh__sqzr, datetime.timedelta(0, 0, lrqr__ydt)
        return impl


@overload(abs, no_unliteral=True)
def timedelta_abs(lhs):
    if lhs == datetime_timedelta_type:

        def impl(lhs):
            if lhs.days < 0:
                return -lhs
            else:
                return lhs
        return impl


@intrinsic
def cast_numpy_timedelta_to_int(typingctx, val=None):
    assert val in (types.NPTimedelta('ns'), types.int64)

    def codegen(context, builder, signature, args):
        return args[0]
    return types.int64(val), codegen


@overload(bool, no_unliteral=True)
def timedelta_to_bool(timedelta):
    if timedelta != datetime_timedelta_type:
        return
    zrauo__kcg = datetime.timedelta(0)

    def impl(timedelta):
        return timedelta != zrauo__kcg
    return impl


class DatetimeTimeDeltaArrayType(types.ArrayCompatible):

    def __init__(self):
        super(DatetimeTimeDeltaArrayType, self).__init__(name=
            'DatetimeTimeDeltaArrayType()')

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    @property
    def dtype(self):
        return datetime_timedelta_type

    def copy(self):
        return DatetimeTimeDeltaArrayType()


datetime_timedelta_array_type = DatetimeTimeDeltaArrayType()
types.datetime_timedelta_array_type = datetime_timedelta_array_type
days_data_type = types.Array(types.int64, 1, 'C')
seconds_data_type = types.Array(types.int64, 1, 'C')
microseconds_data_type = types.Array(types.int64, 1, 'C')
nulls_type = types.Array(types.uint8, 1, 'C')


@register_model(DatetimeTimeDeltaArrayType)
class DatetimeTimeDeltaArrayModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        rpbc__ean = [('days_data', days_data_type), ('seconds_data',
            seconds_data_type), ('microseconds_data',
            microseconds_data_type), ('null_bitmap', nulls_type)]
        models.StructModel.__init__(self, dmm, fe_type, rpbc__ean)


make_attribute_wrapper(DatetimeTimeDeltaArrayType, 'days_data', '_days_data')
make_attribute_wrapper(DatetimeTimeDeltaArrayType, 'seconds_data',
    '_seconds_data')
make_attribute_wrapper(DatetimeTimeDeltaArrayType, 'microseconds_data',
    '_microseconds_data')
make_attribute_wrapper(DatetimeTimeDeltaArrayType, 'null_bitmap',
    '_null_bitmap')


@overload_method(DatetimeTimeDeltaArrayType, 'copy', no_unliteral=True)
def overload_datetime_timedelta_arr_copy(A):
    return (lambda A: bodo.hiframes.datetime_timedelta_ext.
        init_datetime_timedelta_array(A._days_data.copy(), A._seconds_data.
        copy(), A._microseconds_data.copy(), A._null_bitmap.copy()))


@unbox(DatetimeTimeDeltaArrayType)
def unbox_datetime_timedelta_array(typ, val, c):
    n = bodo.utils.utils.object_length(c, val)
    rctq__sgh = types.Array(types.intp, 1, 'C')
    xoh__albd = bodo.utils.utils._empty_nd_impl(c.context, c.builder,
        rctq__sgh, [n])
    pjjc__uakzu = bodo.utils.utils._empty_nd_impl(c.context, c.builder,
        rctq__sgh, [n])
    lwcxe__sgqus = bodo.utils.utils._empty_nd_impl(c.context, c.builder,
        rctq__sgh, [n])
    yqq__jzzt = c.builder.udiv(c.builder.add(n, lir.Constant(lir.IntType(64
        ), 7)), lir.Constant(lir.IntType(64), 8))
    vma__qeel = bodo.utils.utils._empty_nd_impl(c.context, c.builder, types
        .Array(types.uint8, 1, 'C'), [yqq__jzzt])
    bns__ern = lir.FunctionType(lir.VoidType(), [lir.IntType(8).as_pointer(
        ), lir.IntType(64), lir.IntType(64).as_pointer(), lir.IntType(64).
        as_pointer(), lir.IntType(64).as_pointer(), lir.IntType(8).
        as_pointer()])
    rbl__cpm = cgutils.get_or_insert_function(c.builder.module, bns__ern,
        name='unbox_datetime_timedelta_array')
    c.builder.call(rbl__cpm, [val, n, xoh__albd.data, pjjc__uakzu.data,
        lwcxe__sgqus.data, vma__qeel.data])
    krmx__gjm = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    krmx__gjm.days_data = xoh__albd._getvalue()
    krmx__gjm.seconds_data = pjjc__uakzu._getvalue()
    krmx__gjm.microseconds_data = lwcxe__sgqus._getvalue()
    krmx__gjm.null_bitmap = vma__qeel._getvalue()
    acxpc__swrry = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(krmx__gjm._getvalue(), is_error=acxpc__swrry)


@box(DatetimeTimeDeltaArrayType)
def box_datetime_timedelta_array(typ, val, c):
    kfvf__lgst = cgutils.create_struct_proxy(typ)(c.context, c.builder, val)
    xoh__albd = c.context.make_array(types.Array(types.int64, 1, 'C'))(c.
        context, c.builder, kfvf__lgst.days_data)
    pjjc__uakzu = c.context.make_array(types.Array(types.int64, 1, 'C'))(c.
        context, c.builder, kfvf__lgst.seconds_data).data
    lwcxe__sgqus = c.context.make_array(types.Array(types.int64, 1, 'C'))(c
        .context, c.builder, kfvf__lgst.microseconds_data).data
    krise__qsnlg = c.context.make_array(types.Array(types.uint8, 1, 'C'))(c
        .context, c.builder, kfvf__lgst.null_bitmap).data
    n = c.builder.extract_value(xoh__albd.shape, 0)
    bns__ern = lir.FunctionType(c.pyapi.pyobj, [lir.IntType(64), lir.
        IntType(64).as_pointer(), lir.IntType(64).as_pointer(), lir.IntType
        (64).as_pointer(), lir.IntType(8).as_pointer()])
    duujw__qls = cgutils.get_or_insert_function(c.builder.module, bns__ern,
        name='box_datetime_timedelta_array')
    qjzb__bfhnl = c.builder.call(duujw__qls, [n, xoh__albd.data,
        pjjc__uakzu, lwcxe__sgqus, krise__qsnlg])
    c.context.nrt.decref(c.builder, typ, val)
    return qjzb__bfhnl


@intrinsic
def init_datetime_timedelta_array(typingctx, days_data, seconds_data,
    microseconds_data, nulls=None):
    assert days_data == types.Array(types.int64, 1, 'C')
    assert seconds_data == types.Array(types.int64, 1, 'C')
    assert microseconds_data == types.Array(types.int64, 1, 'C')
    assert nulls == types.Array(types.uint8, 1, 'C')

    def codegen(context, builder, signature, args):
        xpt__igu, empoa__tmat, jjuy__wcpn, twgbc__nltxl = args
        irmpj__jxvnt = cgutils.create_struct_proxy(signature.return_type)(
            context, builder)
        irmpj__jxvnt.days_data = xpt__igu
        irmpj__jxvnt.seconds_data = empoa__tmat
        irmpj__jxvnt.microseconds_data = jjuy__wcpn
        irmpj__jxvnt.null_bitmap = twgbc__nltxl
        context.nrt.incref(builder, signature.args[0], xpt__igu)
        context.nrt.incref(builder, signature.args[1], empoa__tmat)
        context.nrt.incref(builder, signature.args[2], jjuy__wcpn)
        context.nrt.incref(builder, signature.args[3], twgbc__nltxl)
        return irmpj__jxvnt._getvalue()
    fer__jfgs = datetime_timedelta_array_type(days_data, seconds_data,
        microseconds_data, nulls)
    return fer__jfgs, codegen


@lower_constant(DatetimeTimeDeltaArrayType)
def lower_constant_datetime_timedelta_arr(context, builder, typ, pyval):
    n = len(pyval)
    xoh__albd = np.empty(n, np.int64)
    pjjc__uakzu = np.empty(n, np.int64)
    lwcxe__sgqus = np.empty(n, np.int64)
    chu__rtnzg = np.empty(n + 7 >> 3, np.uint8)
    for vpmt__cxve, s in enumerate(pyval):
        oqfdp__gln = pd.isna(s)
        bodo.libs.int_arr_ext.set_bit_to_arr(chu__rtnzg, vpmt__cxve, int(
            not oqfdp__gln))
        if not oqfdp__gln:
            xoh__albd[vpmt__cxve] = s.days
            pjjc__uakzu[vpmt__cxve] = s.seconds
            lwcxe__sgqus[vpmt__cxve] = s.microseconds
    gqljn__exq = context.get_constant_generic(builder, days_data_type,
        xoh__albd)
    obbeo__fghny = context.get_constant_generic(builder, seconds_data_type,
        pjjc__uakzu)
    zylc__jdimv = context.get_constant_generic(builder,
        microseconds_data_type, lwcxe__sgqus)
    mgqo__gbea = context.get_constant_generic(builder, nulls_type, chu__rtnzg)
    return lir.Constant.literal_struct([gqljn__exq, obbeo__fghny,
        zylc__jdimv, mgqo__gbea])


@numba.njit(no_cpython_wrapper=True)
def alloc_datetime_timedelta_array(n):
    xoh__albd = np.empty(n, dtype=np.int64)
    pjjc__uakzu = np.empty(n, dtype=np.int64)
    lwcxe__sgqus = np.empty(n, dtype=np.int64)
    nulls = np.full(n + 7 >> 3, 255, np.uint8)
    return init_datetime_timedelta_array(xoh__albd, pjjc__uakzu,
        lwcxe__sgqus, nulls)


def alloc_datetime_timedelta_array_equiv(self, scope, equiv_set, loc, args, kws
    ):
    assert len(args) == 1 and not kws
    return ArrayAnalysis.AnalyzeResult(shape=args[0], pre=[])


(ArrayAnalysis.
    _analyze_op_call_bodo_hiframes_datetime_timedelta_ext_alloc_datetime_timedelta_array
    ) = alloc_datetime_timedelta_array_equiv


@overload(operator.getitem, no_unliteral=True)
def dt_timedelta_arr_getitem(A, ind):
    if A != datetime_timedelta_array_type:
        return
    if isinstance(ind, types.Integer):

        def impl_int(A, ind):
            return datetime.timedelta(days=A._days_data[ind], seconds=A.
                _seconds_data[ind], microseconds=A._microseconds_data[ind])
        return impl_int
    if is_list_like_index_type(ind) and ind.dtype == types.bool_:

        def impl_bool(A, ind):
            pml__wycfm = bodo.utils.conversion.coerce_to_ndarray(ind)
            yuing__knpu = A._null_bitmap
            otvu__zbnme = A._days_data[pml__wycfm]
            oml__yqss = A._seconds_data[pml__wycfm]
            blfwb__teqyq = A._microseconds_data[pml__wycfm]
            n = len(otvu__zbnme)
            znunq__siqy = get_new_null_mask_bool_index(yuing__knpu, ind, n)
            return init_datetime_timedelta_array(otvu__zbnme, oml__yqss,
                blfwb__teqyq, znunq__siqy)
        return impl_bool
    if is_list_like_index_type(ind) and isinstance(ind.dtype, types.Integer):

        def impl(A, ind):
            pml__wycfm = bodo.utils.conversion.coerce_to_ndarray(ind)
            yuing__knpu = A._null_bitmap
            otvu__zbnme = A._days_data[pml__wycfm]
            oml__yqss = A._seconds_data[pml__wycfm]
            blfwb__teqyq = A._microseconds_data[pml__wycfm]
            n = len(otvu__zbnme)
            znunq__siqy = get_new_null_mask_int_index(yuing__knpu,
                pml__wycfm, n)
            return init_datetime_timedelta_array(otvu__zbnme, oml__yqss,
                blfwb__teqyq, znunq__siqy)
        return impl
    if isinstance(ind, types.SliceType):

        def impl_slice(A, ind):
            n = len(A._days_data)
            yuing__knpu = A._null_bitmap
            otvu__zbnme = np.ascontiguousarray(A._days_data[ind])
            oml__yqss = np.ascontiguousarray(A._seconds_data[ind])
            blfwb__teqyq = np.ascontiguousarray(A._microseconds_data[ind])
            znunq__siqy = get_new_null_mask_slice_index(yuing__knpu, ind, n)
            return init_datetime_timedelta_array(otvu__zbnme, oml__yqss,
                blfwb__teqyq, znunq__siqy)
        return impl_slice
    raise BodoError(
        f'getitem for DatetimeTimedeltaArray with indexing type {ind} not supported.'
        )


@overload(operator.setitem, no_unliteral=True)
def dt_timedelta_arr_setitem(A, ind, val):
    if A != datetime_timedelta_array_type:
        return
    if val == types.none or isinstance(val, types.optional):
        return
    swko__zgc = (
        f"setitem for DatetimeTimedeltaArray with indexing type {ind} received an incorrect 'value' type {val}."
        )
    if isinstance(ind, types.Integer):
        if types.unliteral(val) == datetime_timedelta_type:

            def impl(A, ind, val):
                A._days_data[ind] = val._days
                A._seconds_data[ind] = val._seconds
                A._microseconds_data[ind] = val._microseconds
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap, ind, 1)
            return impl
        else:
            raise BodoError(swko__zgc)
    if not (is_iterable_type(val) and val.dtype == bodo.
        datetime_timedelta_type or types.unliteral(val) ==
        datetime_timedelta_type):
        raise BodoError(swko__zgc)
    if is_list_like_index_type(ind) and isinstance(ind.dtype, types.Integer):
        if types.unliteral(val) == datetime_timedelta_type:

            def impl_arr_ind_scalar(A, ind, val):
                n = len(A)
                for vpmt__cxve in range(n):
                    A._days_data[ind[vpmt__cxve]] = val._days
                    A._seconds_data[ind[vpmt__cxve]] = val._seconds
                    A._microseconds_data[ind[vpmt__cxve]] = val._microseconds
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        ind[vpmt__cxve], 1)
            return impl_arr_ind_scalar
        else:

            def impl_arr_ind(A, ind, val):
                val = bodo.utils.conversion.coerce_to_array(val,
                    use_nullable_array=True)
                n = len(val._days_data)
                for vpmt__cxve in range(n):
                    A._days_data[ind[vpmt__cxve]] = val._days_data[vpmt__cxve]
                    A._seconds_data[ind[vpmt__cxve]] = val._seconds_data[
                        vpmt__cxve]
                    A._microseconds_data[ind[vpmt__cxve]
                        ] = val._microseconds_data[vpmt__cxve]
                    exk__appwr = bodo.libs.int_arr_ext.get_bit_bitmap_arr(val
                        ._null_bitmap, vpmt__cxve)
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        ind[vpmt__cxve], exk__appwr)
            return impl_arr_ind
    if is_list_like_index_type(ind) and ind.dtype == types.bool_:
        if types.unliteral(val) == datetime_timedelta_type:

            def impl_bool_ind_mask_scalar(A, ind, val):
                n = len(ind)
                for vpmt__cxve in range(n):
                    if not bodo.libs.array_kernels.isna(ind, vpmt__cxve
                        ) and ind[vpmt__cxve]:
                        A._days_data[vpmt__cxve] = val._days
                        A._seconds_data[vpmt__cxve] = val._seconds
                        A._microseconds_data[vpmt__cxve] = val._microseconds
                        bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                            vpmt__cxve, 1)
            return impl_bool_ind_mask_scalar
        else:

            def impl_bool_ind_mask(A, ind, val):
                val = bodo.utils.conversion.coerce_to_array(val,
                    use_nullable_array=True)
                n = len(ind)
                ckj__npdaw = 0
                for vpmt__cxve in range(n):
                    if not bodo.libs.array_kernels.isna(ind, vpmt__cxve
                        ) and ind[vpmt__cxve]:
                        A._days_data[vpmt__cxve] = val._days_data[ckj__npdaw]
                        A._seconds_data[vpmt__cxve] = val._seconds_data[
                            ckj__npdaw]
                        A._microseconds_data[vpmt__cxve
                            ] = val._microseconds_data[ckj__npdaw]
                        exk__appwr = bodo.libs.int_arr_ext.get_bit_bitmap_arr(
                            val._null_bitmap, ckj__npdaw)
                        bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                            vpmt__cxve, exk__appwr)
                        ckj__npdaw += 1
            return impl_bool_ind_mask
    if isinstance(ind, types.SliceType):
        if types.unliteral(val) == datetime_timedelta_type:

            def impl_slice_scalar(A, ind, val):
                zhxpi__bulm = numba.cpython.unicode._normalize_slice(ind,
                    len(A))
                for vpmt__cxve in range(zhxpi__bulm.start, zhxpi__bulm.stop,
                    zhxpi__bulm.step):
                    A._days_data[vpmt__cxve] = val._days
                    A._seconds_data[vpmt__cxve] = val._seconds
                    A._microseconds_data[vpmt__cxve] = val._microseconds
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        vpmt__cxve, 1)
            return impl_slice_scalar
        else:

            def impl_slice_mask(A, ind, val):
                val = bodo.utils.conversion.coerce_to_array(val,
                    use_nullable_array=True)
                n = len(A._days_data)
                A._days_data[ind] = val._days_data
                A._seconds_data[ind] = val._seconds_data
                A._microseconds_data[ind] = val._microseconds_data
                zqj__qus = val._null_bitmap.copy()
                setitem_slice_index_null_bits(A._null_bitmap, zqj__qus, ind, n)
            return impl_slice_mask
    raise BodoError(
        f'setitem for DatetimeTimedeltaArray with indexing type {ind} not supported.'
        )


@overload(len, no_unliteral=True)
def overload_len_datetime_timedelta_arr(A):
    if A == datetime_timedelta_array_type:
        return lambda A: len(A._days_data)


@overload_attribute(DatetimeTimeDeltaArrayType, 'shape')
def overload_datetime_timedelta_arr_shape(A):
    return lambda A: (len(A._days_data),)


@overload_attribute(DatetimeTimeDeltaArrayType, 'nbytes')
def timedelta_arr_nbytes_overload(A):
    return (lambda A: A._days_data.nbytes + A._seconds_data.nbytes + A.
        _microseconds_data.nbytes + A._null_bitmap.nbytes)


def overload_datetime_timedelta_arr_sub(arg1, arg2):
    if (arg1 == datetime_timedelta_array_type and arg2 ==
        datetime_timedelta_type):

        def impl(arg1, arg2):
            kfvf__lgst = arg1
            numba.parfors.parfor.init_prange()
            n = len(kfvf__lgst)
            A = alloc_datetime_timedelta_array(n)
            for vpmt__cxve in numba.parfors.parfor.internal_prange(n):
                A[vpmt__cxve] = kfvf__lgst[vpmt__cxve] - arg2
            return A
        return impl


def create_cmp_op_overload_arr(op):

    def overload_date_arr_cmp(lhs, rhs):
        if op == operator.ne:
            iixkm__lonmy = True
        else:
            iixkm__lonmy = False
        if (lhs == datetime_timedelta_array_type and rhs ==
            datetime_timedelta_array_type):

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                n = len(lhs)
                kgi__mgrus = bodo.libs.bool_arr_ext.alloc_bool_array(n)
                for vpmt__cxve in numba.parfors.parfor.internal_prange(n):
                    dime__iyu = bodo.libs.array_kernels.isna(lhs, vpmt__cxve)
                    wsc__ewxlr = bodo.libs.array_kernels.isna(rhs, vpmt__cxve)
                    if dime__iyu or wsc__ewxlr:
                        bknkw__uqgpa = iixkm__lonmy
                    else:
                        bknkw__uqgpa = op(lhs[vpmt__cxve], rhs[vpmt__cxve])
                    kgi__mgrus[vpmt__cxve] = bknkw__uqgpa
                return kgi__mgrus
            return impl
        elif lhs == datetime_timedelta_array_type:

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                n = len(lhs)
                kgi__mgrus = bodo.libs.bool_arr_ext.alloc_bool_array(n)
                for vpmt__cxve in numba.parfors.parfor.internal_prange(n):
                    exk__appwr = bodo.libs.array_kernels.isna(lhs, vpmt__cxve)
                    if exk__appwr:
                        bknkw__uqgpa = iixkm__lonmy
                    else:
                        bknkw__uqgpa = op(lhs[vpmt__cxve], rhs)
                    kgi__mgrus[vpmt__cxve] = bknkw__uqgpa
                return kgi__mgrus
            return impl
        elif rhs == datetime_timedelta_array_type:

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                n = len(rhs)
                kgi__mgrus = bodo.libs.bool_arr_ext.alloc_bool_array(n)
                for vpmt__cxve in numba.parfors.parfor.internal_prange(n):
                    exk__appwr = bodo.libs.array_kernels.isna(rhs, vpmt__cxve)
                    if exk__appwr:
                        bknkw__uqgpa = iixkm__lonmy
                    else:
                        bknkw__uqgpa = op(lhs, rhs[vpmt__cxve])
                    kgi__mgrus[vpmt__cxve] = bknkw__uqgpa
                return kgi__mgrus
            return impl
    return overload_date_arr_cmp


timedelta_unsupported_attrs = ['asm8', 'resolution_string', 'freq',
    'is_populated']
timedelta_unsupported_methods = ['isoformat']


def _intstall_pd_timedelta_unsupported():
    from bodo.utils.typing import create_unsupported_overload
    for qbsz__teda in timedelta_unsupported_attrs:
        yovf__ttlfe = 'pandas.Timedelta.' + qbsz__teda
        overload_attribute(PDTimeDeltaType, qbsz__teda)(
            create_unsupported_overload(yovf__ttlfe))
    for alk__pcyrl in timedelta_unsupported_methods:
        yovf__ttlfe = 'pandas.Timedelta.' + alk__pcyrl
        overload_method(PDTimeDeltaType, alk__pcyrl)(
            create_unsupported_overload(yovf__ttlfe + '()'))


_intstall_pd_timedelta_unsupported()
