import enum
import operator
import numba
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.core.imputils import lower_constant
from numba.extending import NativeValue, box, intrinsic, lower_cast, make_attribute_wrapper, models, overload, overload_attribute, overload_method, register_jitable, register_model, typeof_impl, unbox
from numba.parfors.array_analysis import ArrayAnalysis
import bodo
from bodo.utils.typing import NOT_CONSTANT, BodoError, MetaType, check_unsupported_args, dtype_to_array_type, get_literal_value, get_overload_const, get_overload_const_bool, is_common_scalar_dtype, is_iterable_type, is_list_like_index_type, is_literal_type, is_overload_constant_bool, is_overload_none, is_overload_true, is_scalar_type, raise_bodo_error


class PDCategoricalDtype(types.Opaque):

    def __init__(self, categories, elem_type, ordered, data=None, int_type=None
        ):
        self.categories = categories
        self.elem_type = elem_type
        self.ordered = ordered
        self.data = _get_cat_index_type(elem_type) if data is None else data
        self.int_type = int_type
        ebln__xaxo = (
            f'PDCategoricalDtype({self.categories}, {self.elem_type}, {self.ordered}, {self.data}, {self.int_type})'
            )
        super(PDCategoricalDtype, self).__init__(name=ebln__xaxo)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


@typeof_impl.register(pd.CategoricalDtype)
def _typeof_pd_cat_dtype(val, c):
    vlzyw__ygu = tuple(val.categories.values)
    elem_type = None if len(vlzyw__ygu) == 0 else bodo.typeof(val.
        categories.values).dtype
    int_type = getattr(val, '_int_type', None)
    return PDCategoricalDtype(vlzyw__ygu, elem_type, val.ordered, bodo.
        typeof(val.categories), int_type)


def _get_cat_index_type(elem_type):
    elem_type = bodo.string_type if elem_type is None else elem_type
    return bodo.utils.typing.get_index_type_from_dtype(elem_type)


@lower_constant(PDCategoricalDtype)
def lower_constant_categorical_type(context, builder, typ, pyval):
    categories = context.get_constant_generic(builder, bodo.typeof(pyval.
        categories), pyval.categories)
    ordered = context.get_constant(types.bool_, pyval.ordered)
    return lir.Constant.literal_struct([categories, ordered])


@register_model(PDCategoricalDtype)
class PDCategoricalDtypeModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        yrre__oqi = [('categories', fe_type.data), ('ordered', types.bool_)]
        models.StructModel.__init__(self, dmm, fe_type, yrre__oqi)


make_attribute_wrapper(PDCategoricalDtype, 'categories', 'categories')
make_attribute_wrapper(PDCategoricalDtype, 'ordered', 'ordered')


@intrinsic
def init_cat_dtype(typingctx, categories_typ, ordered_typ, int_type,
    cat_vals_typ=None):
    assert bodo.hiframes.pd_index_ext.is_index_type(categories_typ
        ), 'init_cat_dtype requires index type for categories'
    assert is_overload_constant_bool(ordered_typ
        ), 'init_cat_dtype requires constant ordered flag'
    ttip__muc = None if is_overload_none(int_type) else int_type.dtype
    assert is_overload_none(cat_vals_typ) or isinstance(cat_vals_typ, types
        .TypeRef), 'init_cat_dtype requires constant category values'
    avswk__uzt = None if is_overload_none(cat_vals_typ
        ) else cat_vals_typ.instance_type.meta

    def codegen(context, builder, sig, args):
        categories, ordered, yso__foy, yso__foy = args
        cat_dtype = cgutils.create_struct_proxy(sig.return_type)(context,
            builder)
        cat_dtype.categories = categories
        context.nrt.incref(builder, sig.args[0], categories)
        context.nrt.incref(builder, sig.args[1], ordered)
        cat_dtype.ordered = ordered
        return cat_dtype._getvalue()
    seo__xgls = PDCategoricalDtype(avswk__uzt, categories_typ.dtype,
        is_overload_true(ordered_typ), categories_typ, ttip__muc)
    return seo__xgls(categories_typ, ordered_typ, int_type, cat_vals_typ
        ), codegen


@unbox(PDCategoricalDtype)
def unbox_cat_dtype(typ, obj, c):
    cat_dtype = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    aalc__gkbux = c.pyapi.object_getattr_string(obj, 'ordered')
    cat_dtype.ordered = c.pyapi.to_native_value(types.bool_, aalc__gkbux).value
    c.pyapi.decref(aalc__gkbux)
    azu__iramf = c.pyapi.object_getattr_string(obj, 'categories')
    cat_dtype.categories = c.pyapi.to_native_value(typ.data, azu__iramf).value
    c.pyapi.decref(azu__iramf)
    uygha__xzgde = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(cat_dtype._getvalue(), is_error=uygha__xzgde)


@box(PDCategoricalDtype)
def box_cat_dtype(typ, val, c):
    cat_dtype = cgutils.create_struct_proxy(typ)(c.context, c.builder, val)
    aalc__gkbux = c.pyapi.from_native_value(types.bool_, cat_dtype.ordered,
        c.env_manager)
    c.context.nrt.incref(c.builder, typ.data, cat_dtype.categories)
    nsilc__ohz = c.pyapi.from_native_value(typ.data, cat_dtype.categories,
        c.env_manager)
    tka__mbk = c.context.insert_const_string(c.builder.module, 'pandas')
    vukc__vdf = c.pyapi.import_module_noblock(tka__mbk)
    ieo__ozlfo = c.pyapi.call_method(vukc__vdf, 'CategoricalDtype', (
        nsilc__ohz, aalc__gkbux))
    c.pyapi.decref(aalc__gkbux)
    c.pyapi.decref(nsilc__ohz)
    c.pyapi.decref(vukc__vdf)
    c.context.nrt.decref(c.builder, typ, val)
    return ieo__ozlfo


@overload_attribute(PDCategoricalDtype, 'nbytes')
def pd_categorical_nbytes_overload(A):
    return lambda A: A.categories.nbytes + bodo.io.np_io.get_dtype_size(types
        .bool_)


class CategoricalArrayType(types.ArrayCompatible):

    def __init__(self, dtype):
        self.dtype = dtype
        super(CategoricalArrayType, self).__init__(name=
            f'CategoricalArrayType({dtype})')

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    def copy(self):
        return CategoricalArrayType(self.dtype)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


@typeof_impl.register(pd.Categorical)
def _typeof_pd_cat(val, c):
    return CategoricalArrayType(bodo.typeof(val.dtype))


@register_model(CategoricalArrayType)
class CategoricalArrayModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        ypbl__kls = get_categories_int_type(fe_type.dtype)
        yrre__oqi = [('dtype', fe_type.dtype), ('codes', types.Array(
            ypbl__kls, 1, 'C'))]
        super(CategoricalArrayModel, self).__init__(dmm, fe_type, yrre__oqi)


make_attribute_wrapper(CategoricalArrayType, 'codes', 'codes')
make_attribute_wrapper(CategoricalArrayType, 'dtype', 'dtype')


@unbox(CategoricalArrayType)
def unbox_categorical_array(typ, val, c):
    hue__hyxr = c.pyapi.object_getattr_string(val, 'codes')
    dtype = get_categories_int_type(typ.dtype)
    codes = c.pyapi.to_native_value(types.Array(dtype, 1, 'C'), hue__hyxr
        ).value
    c.pyapi.decref(hue__hyxr)
    ieo__ozlfo = c.pyapi.object_getattr_string(val, 'dtype')
    pdp__zmoo = c.pyapi.to_native_value(typ.dtype, ieo__ozlfo).value
    c.pyapi.decref(ieo__ozlfo)
    fax__jox = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    fax__jox.codes = codes
    fax__jox.dtype = pdp__zmoo
    return NativeValue(fax__jox._getvalue())


@lower_constant(CategoricalArrayType)
def lower_constant_categorical_array(context, builder, typ, pyval):
    bcl__udmw = get_categories_int_type(typ.dtype)
    uagrx__itwou = context.get_constant_generic(builder, types.Array(
        bcl__udmw, 1, 'C'), pyval.codes)
    cat_dtype = context.get_constant_generic(builder, typ.dtype, pyval.dtype)
    return lir.Constant.literal_struct([cat_dtype, uagrx__itwou])


def get_categories_int_type(cat_dtype):
    dtype = types.int64
    if cat_dtype.int_type is not None:
        return cat_dtype.int_type
    if cat_dtype.categories is None:
        return types.int64
    ytg__lkalg = len(cat_dtype.categories)
    if ytg__lkalg < np.iinfo(np.int8).max:
        dtype = types.int8
    elif ytg__lkalg < np.iinfo(np.int16).max:
        dtype = types.int16
    elif ytg__lkalg < np.iinfo(np.int32).max:
        dtype = types.int32
    return dtype


@box(CategoricalArrayType)
def box_categorical_array(typ, val, c):
    dtype = typ.dtype
    tka__mbk = c.context.insert_const_string(c.builder.module, 'pandas')
    vukc__vdf = c.pyapi.import_module_noblock(tka__mbk)
    ypbl__kls = get_categories_int_type(dtype)
    smx__zqpn = cgutils.create_struct_proxy(typ)(c.context, c.builder, val)
    fuit__zwsw = types.Array(ypbl__kls, 1, 'C')
    c.context.nrt.incref(c.builder, fuit__zwsw, smx__zqpn.codes)
    hue__hyxr = c.pyapi.from_native_value(fuit__zwsw, smx__zqpn.codes, c.
        env_manager)
    c.context.nrt.incref(c.builder, dtype, smx__zqpn.dtype)
    ieo__ozlfo = c.pyapi.from_native_value(dtype, smx__zqpn.dtype, c.
        env_manager)
    ktps__zqc = c.pyapi.borrow_none()
    kxtiw__rcohq = c.pyapi.object_getattr_string(vukc__vdf, 'Categorical')
    jkh__jyu = c.pyapi.call_method(kxtiw__rcohq, 'from_codes', (hue__hyxr,
        ktps__zqc, ktps__zqc, ieo__ozlfo))
    c.pyapi.decref(kxtiw__rcohq)
    c.pyapi.decref(hue__hyxr)
    c.pyapi.decref(ieo__ozlfo)
    c.pyapi.decref(vukc__vdf)
    c.context.nrt.decref(c.builder, typ, val)
    return jkh__jyu


def _to_readonly(t):
    from bodo.hiframes.pd_index_ext import DatetimeIndexType, NumericIndexType, TimedeltaIndexType
    if isinstance(t, CategoricalArrayType):
        return CategoricalArrayType(_to_readonly(t.dtype))
    if isinstance(t, PDCategoricalDtype):
        return PDCategoricalDtype(t.categories, t.elem_type, t.ordered,
            _to_readonly(t.data), t.int_type)
    if isinstance(t, types.Array):
        return types.Array(t.dtype, t.ndim, 'C', True)
    if isinstance(t, NumericIndexType):
        return NumericIndexType(t.dtype, t.name_typ, _to_readonly(t.data))
    if isinstance(t, (DatetimeIndexType, TimedeltaIndexType)):
        return t.__class__(t.name_typ, _to_readonly(t.data))
    return t


@lower_cast(CategoricalArrayType, CategoricalArrayType)
def cast_cat_arr(context, builder, fromty, toty, val):
    if _to_readonly(toty) == fromty:
        return val
    raise BodoError(f'Cannot cast from {fromty} to {toty}')


def create_cmp_op_overload(op):

    def overload_cat_arr_cmp(A, other):
        if not isinstance(A, CategoricalArrayType):
            return
        if A.dtype.categories and is_literal_type(other) and types.unliteral(
            other) == A.dtype.elem_type:
            val = get_literal_value(other)
            benu__mwp = list(A.dtype.categories).index(val
                ) if val in A.dtype.categories else -2

            def impl_lit(A, other):
                qsiec__xso = op(bodo.hiframes.pd_categorical_ext.
                    get_categorical_arr_codes(A), benu__mwp)
                return qsiec__xso
            return impl_lit

        def impl(A, other):
            benu__mwp = get_code_for_value(A.dtype, other)
            qsiec__xso = op(bodo.hiframes.pd_categorical_ext.
                get_categorical_arr_codes(A), benu__mwp)
            return qsiec__xso
        return impl
    return overload_cat_arr_cmp


def _install_cmp_ops():
    for op in [operator.eq, operator.ne]:
        lbpc__mcem = create_cmp_op_overload(op)
        overload(op, inline='always', no_unliteral=True)(lbpc__mcem)


_install_cmp_ops()


@register_jitable
def get_code_for_value(cat_dtype, val):
    smx__zqpn = cat_dtype.categories
    n = len(smx__zqpn)
    for hkrl__ibdmf in range(n):
        if smx__zqpn[hkrl__ibdmf] == val:
            return hkrl__ibdmf
    return -2


@overload_method(CategoricalArrayType, 'astype', inline='always',
    no_unliteral=True)
def overload_cat_arr_astype(A, dtype, copy=True, _bodo_nan_to_str=True):
    if dtype == types.unicode_type:
        raise_bodo_error(
            "CategoricalArray.astype(): 'dtype' when passed as string must be a constant value"
            )
    vwy__iijui = bodo.utils.typing.parse_dtype(dtype, 'CategoricalArray.astype'
        )
    if vwy__iijui != A.dtype.elem_type and vwy__iijui != types.unicode_type:
        raise BodoError(
            f'Converting categorical array {A} to dtype {dtype} not supported yet'
            )
    if vwy__iijui == types.unicode_type:

        def impl(A, dtype, copy=True, _bodo_nan_to_str=True):
            codes = bodo.hiframes.pd_categorical_ext.get_categorical_arr_codes(
                A)
            categories = A.dtype.categories
            n = len(codes)
            qsiec__xso = bodo.libs.str_arr_ext.pre_alloc_string_array(n, -1)
            for hkrl__ibdmf in numba.parfors.parfor.internal_prange(n):
                xeg__nom = codes[hkrl__ibdmf]
                if xeg__nom == -1:
                    if _bodo_nan_to_str:
                        bodo.libs.str_arr_ext.str_arr_setitem_NA_str(qsiec__xso
                            , hkrl__ibdmf)
                    else:
                        bodo.libs.array_kernels.setna(qsiec__xso, hkrl__ibdmf)
                    continue
                qsiec__xso[hkrl__ibdmf] = str(bodo.utils.conversion.
                    unbox_if_timestamp(categories[xeg__nom]))
            return qsiec__xso
        return impl
    fuit__zwsw = dtype_to_array_type(vwy__iijui)

    def impl(A, dtype, copy=True, _bodo_nan_to_str=True):
        codes = bodo.hiframes.pd_categorical_ext.get_categorical_arr_codes(A)
        categories = A.dtype.categories
        n = len(codes)
        qsiec__xso = bodo.utils.utils.alloc_type(n, fuit__zwsw, (-1,))
        for hkrl__ibdmf in numba.parfors.parfor.internal_prange(n):
            xeg__nom = codes[hkrl__ibdmf]
            if xeg__nom == -1:
                bodo.libs.array_kernels.setna(qsiec__xso, hkrl__ibdmf)
                continue
            qsiec__xso[hkrl__ibdmf] = bodo.utils.conversion.unbox_if_timestamp(
                categories[xeg__nom])
        return qsiec__xso
    return impl


@overload(pd.api.types.CategoricalDtype, no_unliteral=True)
def cat_overload_dummy(val_list):
    return lambda val_list: 1


@intrinsic
def init_categorical_array(typingctx, codes, cat_dtype=None):
    assert isinstance(codes, types.Array) and isinstance(codes.dtype, types
        .Integer)

    def codegen(context, builder, signature, args):
        uxbzu__izvf, pdp__zmoo = args
        smx__zqpn = cgutils.create_struct_proxy(signature.return_type)(context,
            builder)
        smx__zqpn.codes = uxbzu__izvf
        smx__zqpn.dtype = pdp__zmoo
        context.nrt.incref(builder, signature.args[0], uxbzu__izvf)
        context.nrt.incref(builder, signature.args[1], pdp__zmoo)
        return smx__zqpn._getvalue()
    chaw__ugmv = CategoricalArrayType(cat_dtype)
    sig = chaw__ugmv(codes, cat_dtype)
    return sig, codegen


def init_categorical_array_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 2 and not kws
    foy__fdzmi = args[0]
    if equiv_set.has_shape(foy__fdzmi):
        return ArrayAnalysis.AnalyzeResult(shape=foy__fdzmi, pre=[])
    return None


(ArrayAnalysis.
    _analyze_op_call_bodo_hiframes_pd_categorical_ext_init_categorical_array
    ) = init_categorical_array_equiv


def alloc_categorical_array(n, cat_dtype):
    pass


@overload(alloc_categorical_array, no_unliteral=True)
def _alloc_categorical_array(n, cat_dtype):
    ypbl__kls = get_categories_int_type(cat_dtype)

    def impl(n, cat_dtype):
        codes = np.empty(n, ypbl__kls)
        return init_categorical_array(codes, cat_dtype)
    return impl


def alloc_categorical_array_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 2 and not kws
    return ArrayAnalysis.AnalyzeResult(shape=args[0], pre=[])


(ArrayAnalysis.
    _analyze_op_call_bodo_hiframes_pd_categorical_ext_alloc_categorical_array
    ) = alloc_categorical_array_equiv


@numba.generated_jit(nopython=True, no_cpython_wrapper=True)
def get_categorical_arr_codes(A):
    return lambda A: A.codes


def alias_ext_dummy_func(lhs_name, args, alias_map, arg_aliases):
    assert len(args) >= 1
    numba.core.ir_utils._add_alias(lhs_name, args[0].name, alias_map,
        arg_aliases)


numba.core.ir_utils.alias_func_extensions['init_categorical_array',
    'bodo.hiframes.pd_categorical_ext'] = alias_ext_dummy_func
numba.core.ir_utils.alias_func_extensions['get_categorical_arr_codes',
    'bodo.hiframes.pd_categorical_ext'] = alias_ext_dummy_func


@overload_method(CategoricalArrayType, 'copy', no_unliteral=True)
def cat_arr_copy_overload(arr):
    return lambda arr: init_categorical_array(arr.codes.copy(), arr.dtype)


def build_replace_dicts(to_replace, value, categories):
    return dict(), np.empty(len(categories) + 1), 0


@overload(build_replace_dicts, no_unliteral=True)
def _build_replace_dicts(to_replace, value, categories):
    if isinstance(to_replace, types.Number) or to_replace == bodo.string_type:

        def impl(to_replace, value, categories):
            return build_replace_dicts([to_replace], value, categories)
        return impl
    else:

        def impl(to_replace, value, categories):
            n = len(categories)
            rhi__ihz = {}
            uagrx__itwou = np.empty(n + 1, np.int64)
            qdeg__spie = {}
            gxz__pyoo = []
            wfdy__axabq = {}
            for hkrl__ibdmf in range(n):
                wfdy__axabq[categories[hkrl__ibdmf]] = hkrl__ibdmf
            for lgm__cap in to_replace:
                if lgm__cap != value:
                    if lgm__cap in wfdy__axabq:
                        if value in wfdy__axabq:
                            rhi__ihz[lgm__cap] = lgm__cap
                            jmqlb__cdwph = wfdy__axabq[lgm__cap]
                            qdeg__spie[jmqlb__cdwph] = wfdy__axabq[value]
                            gxz__pyoo.append(jmqlb__cdwph)
                        else:
                            rhi__ihz[lgm__cap] = value
                            wfdy__axabq[value] = wfdy__axabq[lgm__cap]
            top__dkl = np.sort(np.array(gxz__pyoo))
            cyza__vrmg = 0
            jgo__mrcly = []
            for kppdv__qma in range(-1, n):
                while cyza__vrmg < len(top__dkl) and kppdv__qma > top__dkl[
                    cyza__vrmg]:
                    cyza__vrmg += 1
                jgo__mrcly.append(cyza__vrmg)
            for gyzu__zumdm in range(-1, n):
                lmvrs__cvtu = gyzu__zumdm
                if gyzu__zumdm in qdeg__spie:
                    lmvrs__cvtu = qdeg__spie[gyzu__zumdm]
                uagrx__itwou[gyzu__zumdm + 1] = lmvrs__cvtu - jgo__mrcly[
                    lmvrs__cvtu + 1]
            return rhi__ihz, uagrx__itwou, len(top__dkl)
        return impl


@numba.njit
def python_build_replace_dicts(to_replace, value, categories):
    return build_replace_dicts(to_replace, value, categories)


@register_jitable
def reassign_codes(new_codes_arr, old_codes_arr, codes_map_arr):
    for hkrl__ibdmf in range(len(new_codes_arr)):
        new_codes_arr[hkrl__ibdmf] = codes_map_arr[old_codes_arr[
            hkrl__ibdmf] + 1]


@overload_method(CategoricalArrayType, 'replace', inline='always',
    no_unliteral=True)
def overload_replace(arr, to_replace, value):

    def impl(arr, to_replace, value):
        return bodo.hiframes.pd_categorical_ext.cat_replace(arr, to_replace,
            value)
    return impl


def cat_replace(arr, to_replace, value):
    return


@overload(cat_replace, no_unliteral=True)
def cat_replace_overload(arr, to_replace, value):
    mwjk__welq = arr.dtype.ordered
    amiiz__ibgv = arr.dtype.elem_type
    tpf__dbxhn = get_overload_const(to_replace)
    jlo__vnm = get_overload_const(value)
    if (arr.dtype.categories is not None and tpf__dbxhn is not NOT_CONSTANT and
        jlo__vnm is not NOT_CONSTANT):
        yluxg__cntgs, codes_map_arr, yso__foy = python_build_replace_dicts(
            tpf__dbxhn, jlo__vnm, arr.dtype.categories)
        if len(yluxg__cntgs) == 0:
            return lambda arr, to_replace, value: arr.copy()
        mxjl__dxr = []
        for caoar__jsvlv in arr.dtype.categories:
            if caoar__jsvlv in yluxg__cntgs:
                cwm__chv = yluxg__cntgs[caoar__jsvlv]
                if cwm__chv != caoar__jsvlv:
                    mxjl__dxr.append(cwm__chv)
            else:
                mxjl__dxr.append(caoar__jsvlv)
        sjhm__pfaz = pd.CategoricalDtype(mxjl__dxr, mwjk__welq
            ).categories.values
        evqv__tnnc = MetaType(tuple(sjhm__pfaz))

        def impl_dtype(arr, to_replace, value):
            strp__zfflh = init_cat_dtype(bodo.utils.conversion.
                index_from_array(sjhm__pfaz), mwjk__welq, None, evqv__tnnc)
            smx__zqpn = alloc_categorical_array(len(arr.codes), strp__zfflh)
            reassign_codes(smx__zqpn.codes, arr.codes, codes_map_arr)
            return smx__zqpn
        return impl_dtype
    amiiz__ibgv = arr.dtype.elem_type
    if amiiz__ibgv == types.unicode_type:

        def impl_str(arr, to_replace, value):
            categories = arr.dtype.categories
            rhi__ihz, codes_map_arr, sug__gvoe = build_replace_dicts(to_replace
                , value, categories.values)
            if len(rhi__ihz) == 0:
                return init_categorical_array(arr.codes.copy().astype(np.
                    int64), init_cat_dtype(categories.copy(), mwjk__welq,
                    None, None))
            n = len(categories)
            sjhm__pfaz = bodo.libs.str_arr_ext.pre_alloc_string_array(n -
                sug__gvoe, -1)
            jfh__xghk = 0
            for kppdv__qma in range(n):
                bdrr__jimmc = categories[kppdv__qma]
                if bdrr__jimmc in rhi__ihz:
                    hsqp__hzrz = rhi__ihz[bdrr__jimmc]
                    if hsqp__hzrz != bdrr__jimmc:
                        sjhm__pfaz[jfh__xghk] = hsqp__hzrz
                        jfh__xghk += 1
                else:
                    sjhm__pfaz[jfh__xghk] = bdrr__jimmc
                    jfh__xghk += 1
            smx__zqpn = alloc_categorical_array(len(arr.codes),
                init_cat_dtype(bodo.utils.conversion.index_from_array(
                sjhm__pfaz), mwjk__welq, None, None))
            reassign_codes(smx__zqpn.codes, arr.codes, codes_map_arr)
            return smx__zqpn
        return impl_str
    jsw__stoyh = dtype_to_array_type(amiiz__ibgv)

    def impl(arr, to_replace, value):
        categories = arr.dtype.categories
        rhi__ihz, codes_map_arr, sug__gvoe = build_replace_dicts(to_replace,
            value, categories.values)
        if len(rhi__ihz) == 0:
            return init_categorical_array(arr.codes.copy().astype(np.int64),
                init_cat_dtype(categories.copy(), mwjk__welq, None, None))
        n = len(categories)
        sjhm__pfaz = bodo.utils.utils.alloc_type(n - sug__gvoe, jsw__stoyh,
            None)
        jfh__xghk = 0
        for hkrl__ibdmf in range(n):
            bdrr__jimmc = categories[hkrl__ibdmf]
            if bdrr__jimmc in rhi__ihz:
                hsqp__hzrz = rhi__ihz[bdrr__jimmc]
                if hsqp__hzrz != bdrr__jimmc:
                    sjhm__pfaz[jfh__xghk] = hsqp__hzrz
                    jfh__xghk += 1
            else:
                sjhm__pfaz[jfh__xghk] = bdrr__jimmc
                jfh__xghk += 1
        smx__zqpn = alloc_categorical_array(len(arr.codes), init_cat_dtype(
            bodo.utils.conversion.index_from_array(sjhm__pfaz), mwjk__welq,
            None, None))
        reassign_codes(smx__zqpn.codes, arr.codes, codes_map_arr)
        return smx__zqpn
    return impl


@overload(len, no_unliteral=True)
def overload_cat_arr_len(A):
    if isinstance(A, CategoricalArrayType):
        return lambda A: len(A.codes)


@overload_attribute(CategoricalArrayType, 'shape')
def overload_cat_arr_shape(A):
    return lambda A: (len(A.codes),)


@overload_attribute(CategoricalArrayType, 'ndim')
def overload_cat_arr_ndim(A):
    return lambda A: 1


@overload_attribute(CategoricalArrayType, 'nbytes')
def cat_arr_nbytes_overload(A):
    return lambda A: A.codes.nbytes + A.dtype.nbytes


@register_jitable
def get_label_dict_from_categories(vals):
    sykj__obp = dict()
    ujins__ihup = 0
    for hkrl__ibdmf in range(len(vals)):
        val = vals[hkrl__ibdmf]
        if val in sykj__obp:
            continue
        sykj__obp[val] = ujins__ihup
        ujins__ihup += 1
    return sykj__obp


@register_jitable
def get_label_dict_from_categories_no_duplicates(vals):
    sykj__obp = dict()
    for hkrl__ibdmf in range(len(vals)):
        val = vals[hkrl__ibdmf]
        sykj__obp[val] = hkrl__ibdmf
    return sykj__obp


@overload(pd.Categorical, no_unliteral=True)
def pd_categorical_overload(values, categories=None, ordered=None, dtype=
    None, fastpath=False):
    woo__cpxi = dict(fastpath=fastpath)
    dnvzq__hcl = dict(fastpath=False)
    check_unsupported_args('pd.Categorical', woo__cpxi, dnvzq__hcl)
    if isinstance(dtype, bodo.hiframes.pd_categorical_ext.PDCategoricalDtype):

        def impl_dtype(values, categories=None, ordered=None, dtype=None,
            fastpath=False):
            data = bodo.utils.conversion.coerce_to_array(values)
            return bodo.utils.conversion.fix_arr_dtype(data, dtype)
        return impl_dtype
    if not is_overload_none(categories):
        uswd__srk = get_overload_const(categories)
        if uswd__srk is not NOT_CONSTANT and get_overload_const(ordered
            ) is not NOT_CONSTANT:
            if is_overload_none(ordered):
                hplv__wjqx = False
            else:
                hplv__wjqx = get_overload_const_bool(ordered)
            oes__jfxl = pd.CategoricalDtype(uswd__srk, hplv__wjqx
                ).categories.values
            obp__fvn = MetaType(tuple(oes__jfxl))

            def impl_cats_const(values, categories=None, ordered=None,
                dtype=None, fastpath=False):
                data = bodo.utils.conversion.coerce_to_array(values)
                strp__zfflh = init_cat_dtype(bodo.utils.conversion.
                    index_from_array(oes__jfxl), hplv__wjqx, None, obp__fvn)
                return bodo.utils.conversion.fix_arr_dtype(data, strp__zfflh)
            return impl_cats_const

        def impl_cats(values, categories=None, ordered=None, dtype=None,
            fastpath=False):
            ordered = bodo.utils.conversion.false_if_none(ordered)
            data = bodo.utils.conversion.coerce_to_array(values)
            vlzyw__ygu = bodo.utils.conversion.convert_to_index(categories)
            cat_dtype = bodo.hiframes.pd_categorical_ext.init_cat_dtype(
                vlzyw__ygu, ordered, None, None)
            return bodo.utils.conversion.fix_arr_dtype(data, cat_dtype)
        return impl_cats
    elif is_overload_none(ordered):

        def impl_auto(values, categories=None, ordered=None, dtype=None,
            fastpath=False):
            data = bodo.utils.conversion.coerce_to_array(values)
            return bodo.utils.conversion.fix_arr_dtype(data, 'category')
        return impl_auto
    raise BodoError(
        f'pd.Categorical(): argument combination not supported yet: {values}, {categories}, {ordered}, {dtype}'
        )


@overload(operator.getitem, no_unliteral=True)
def categorical_array_getitem(arr, ind):
    if not isinstance(arr, CategoricalArrayType):
        return
    if isinstance(ind, types.Integer):

        def categorical_getitem_impl(arr, ind):
            eho__xmd = arr.codes[ind]
            return arr.dtype.categories[max(eho__xmd, 0)]
        return categorical_getitem_impl
    if is_list_like_index_type(ind) or isinstance(ind, types.SliceType):

        def impl_bool(arr, ind):
            return init_categorical_array(arr.codes[ind], arr.dtype)
        return impl_bool
    raise BodoError(
        f'getitem for CategoricalArrayType with indexing type {ind} not supported.'
        )


class CategoricalMatchingValues(enum.Enum):
    DIFFERENT_TYPES = -1
    DONT_MATCH = 0
    MAY_MATCH = 1
    DO_MATCH = 2


def categorical_arrs_match(arr1, arr2):
    if not (isinstance(arr1, CategoricalArrayType) and isinstance(arr2,
        CategoricalArrayType)):
        return CategoricalMatchingValues.DIFFERENT_TYPES
    if arr1.dtype.categories is None or arr2.dtype.categories is None:
        return CategoricalMatchingValues.MAY_MATCH
    return (CategoricalMatchingValues.DO_MATCH if arr1.dtype.categories ==
        arr2.dtype.categories and arr1.dtype.ordered == arr2.dtype.ordered else
        CategoricalMatchingValues.DONT_MATCH)


@register_jitable
def cat_dtype_equal(dtype1, dtype2):
    if dtype1.ordered != dtype2.ordered or len(dtype1.categories) != len(dtype2
        .categories):
        return False
    arr1 = dtype1.categories.values
    arr2 = dtype2.categories.values
    for hkrl__ibdmf in range(len(arr1)):
        if arr1[hkrl__ibdmf] != arr2[hkrl__ibdmf]:
            return False
    return True


@overload(operator.setitem, no_unliteral=True)
def categorical_array_setitem(arr, ind, val):
    if not isinstance(arr, CategoricalArrayType):
        return
    if val == types.none or isinstance(val, types.optional):
        return
    ailmj__ahc = is_scalar_type(val) and is_common_scalar_dtype([types.
        unliteral(val), arr.dtype.elem_type]) and not (isinstance(arr.dtype
        .elem_type, types.Integer) and isinstance(val, types.Float))
    jilla__lnck = not isinstance(val, CategoricalArrayType
        ) and is_iterable_type(val) and is_common_scalar_dtype([val.dtype,
        arr.dtype.elem_type]) and not (isinstance(arr.dtype.elem_type,
        types.Integer) and isinstance(val.dtype, types.Float))
    iydj__zygyt = categorical_arrs_match(arr, val)
    ycxvz__yph = (
        f"setitem for CategoricalArrayType of dtype {arr.dtype} with indexing type {ind} received an incorrect 'value' type {val}."
        )
    dvtd__smn = (
        'Cannot set a Categorical with another, without identical categories')
    if isinstance(ind, types.Integer):
        if not ailmj__ahc:
            raise BodoError(ycxvz__yph)

        def impl_scalar(arr, ind, val):
            if val not in arr.dtype.categories:
                raise ValueError(
                    'Cannot setitem on a Categorical with a new category, set the categories first'
                    )
            eho__xmd = arr.dtype.categories.get_loc(val)
            arr.codes[ind] = eho__xmd
        return impl_scalar
    if is_list_like_index_type(ind) and isinstance(ind.dtype, types.Integer):
        if not (ailmj__ahc or jilla__lnck or iydj__zygyt !=
            CategoricalMatchingValues.DIFFERENT_TYPES):
            raise BodoError(ycxvz__yph)
        if iydj__zygyt == CategoricalMatchingValues.DONT_MATCH:
            raise BodoError(dvtd__smn)
        if ailmj__ahc:

            def impl_scalar(arr, ind, val):
                if val not in arr.dtype.categories:
                    raise ValueError(
                        'Cannot setitem on a Categorical with a new category, set the categories first'
                        )
                xfm__nysw = arr.dtype.categories.get_loc(val)
                n = len(ind)
                for kppdv__qma in range(n):
                    arr.codes[ind[kppdv__qma]] = xfm__nysw
            return impl_scalar
        if iydj__zygyt == CategoricalMatchingValues.DO_MATCH:

            def impl_arr_ind_mask(arr, ind, val):
                n = len(val.codes)
                for hkrl__ibdmf in range(n):
                    arr.codes[ind[hkrl__ibdmf]] = val.codes[hkrl__ibdmf]
            return impl_arr_ind_mask
        if iydj__zygyt == CategoricalMatchingValues.MAY_MATCH:

            def impl_arr_ind_mask(arr, ind, val):
                if not cat_dtype_equal(arr.dtype, val.dtype):
                    raise ValueError(dvtd__smn)
                n = len(val.codes)
                for hkrl__ibdmf in range(n):
                    arr.codes[ind[hkrl__ibdmf]] = val.codes[hkrl__ibdmf]
            return impl_arr_ind_mask
        if jilla__lnck:

            def impl_arr_ind_mask_cat_values(arr, ind, val):
                n = len(val)
                categories = arr.dtype.categories
                for kppdv__qma in range(n):
                    ijrme__qviuu = bodo.utils.conversion.unbox_if_timestamp(val
                        [kppdv__qma])
                    if ijrme__qviuu not in categories:
                        raise ValueError(
                            'Cannot setitem on a Categorical with a new category, set the categories first'
                            )
                    eho__xmd = categories.get_loc(ijrme__qviuu)
                    arr.codes[ind[kppdv__qma]] = eho__xmd
            return impl_arr_ind_mask_cat_values
    if is_list_like_index_type(ind) and ind.dtype == types.bool_:
        if not (ailmj__ahc or jilla__lnck or iydj__zygyt !=
            CategoricalMatchingValues.DIFFERENT_TYPES):
            raise BodoError(ycxvz__yph)
        if iydj__zygyt == CategoricalMatchingValues.DONT_MATCH:
            raise BodoError(dvtd__smn)
        if ailmj__ahc:

            def impl_scalar(arr, ind, val):
                if val not in arr.dtype.categories:
                    raise ValueError(
                        'Cannot setitem on a Categorical with a new category, set the categories first'
                        )
                xfm__nysw = arr.dtype.categories.get_loc(val)
                n = len(ind)
                for kppdv__qma in range(n):
                    if ind[kppdv__qma]:
                        arr.codes[kppdv__qma] = xfm__nysw
            return impl_scalar
        if iydj__zygyt == CategoricalMatchingValues.DO_MATCH:

            def impl_bool_ind_mask(arr, ind, val):
                n = len(ind)
                djahc__ejg = 0
                for hkrl__ibdmf in range(n):
                    if ind[hkrl__ibdmf]:
                        arr.codes[hkrl__ibdmf] = val.codes[djahc__ejg]
                        djahc__ejg += 1
            return impl_bool_ind_mask
        if iydj__zygyt == CategoricalMatchingValues.MAY_MATCH:

            def impl_bool_ind_mask(arr, ind, val):
                if not cat_dtype_equal(arr.dtype, val.dtype):
                    raise ValueError(dvtd__smn)
                n = len(ind)
                djahc__ejg = 0
                for hkrl__ibdmf in range(n):
                    if ind[hkrl__ibdmf]:
                        arr.codes[hkrl__ibdmf] = val.codes[djahc__ejg]
                        djahc__ejg += 1
            return impl_bool_ind_mask
        if jilla__lnck:

            def impl_bool_ind_mask_cat_values(arr, ind, val):
                n = len(ind)
                djahc__ejg = 0
                categories = arr.dtype.categories
                for kppdv__qma in range(n):
                    if ind[kppdv__qma]:
                        ijrme__qviuu = (bodo.utils.conversion.
                            unbox_if_timestamp(val[djahc__ejg]))
                        if ijrme__qviuu not in categories:
                            raise ValueError(
                                'Cannot setitem on a Categorical with a new category, set the categories first'
                                )
                        eho__xmd = categories.get_loc(ijrme__qviuu)
                        arr.codes[kppdv__qma] = eho__xmd
                        djahc__ejg += 1
            return impl_bool_ind_mask_cat_values
    if isinstance(ind, types.SliceType):
        if not (ailmj__ahc or jilla__lnck or iydj__zygyt !=
            CategoricalMatchingValues.DIFFERENT_TYPES):
            raise BodoError(ycxvz__yph)
        if iydj__zygyt == CategoricalMatchingValues.DONT_MATCH:
            raise BodoError(dvtd__smn)
        if ailmj__ahc:

            def impl_scalar(arr, ind, val):
                if val not in arr.dtype.categories:
                    raise ValueError(
                        'Cannot setitem on a Categorical with a new category, set the categories first'
                        )
                xfm__nysw = arr.dtype.categories.get_loc(val)
                zzfyc__ijeq = numba.cpython.unicode._normalize_slice(ind,
                    len(arr))
                for kppdv__qma in range(zzfyc__ijeq.start, zzfyc__ijeq.stop,
                    zzfyc__ijeq.step):
                    arr.codes[kppdv__qma] = xfm__nysw
            return impl_scalar
        if iydj__zygyt == CategoricalMatchingValues.DO_MATCH:

            def impl_arr(arr, ind, val):
                arr.codes[ind] = val.codes
            return impl_arr
        if iydj__zygyt == CategoricalMatchingValues.MAY_MATCH:

            def impl_arr(arr, ind, val):
                if not cat_dtype_equal(arr.dtype, val.dtype):
                    raise ValueError(dvtd__smn)
                arr.codes[ind] = val.codes
            return impl_arr
        if jilla__lnck:

            def impl_slice_cat_values(arr, ind, val):
                categories = arr.dtype.categories
                zzfyc__ijeq = numba.cpython.unicode._normalize_slice(ind,
                    len(arr))
                djahc__ejg = 0
                for kppdv__qma in range(zzfyc__ijeq.start, zzfyc__ijeq.stop,
                    zzfyc__ijeq.step):
                    ijrme__qviuu = bodo.utils.conversion.unbox_if_timestamp(val
                        [djahc__ejg])
                    if ijrme__qviuu not in categories:
                        raise ValueError(
                            'Cannot setitem on a Categorical with a new category, set the categories first'
                            )
                    eho__xmd = categories.get_loc(ijrme__qviuu)
                    arr.codes[kppdv__qma] = eho__xmd
                    djahc__ejg += 1
            return impl_slice_cat_values
    raise BodoError(
        f'setitem for CategoricalArrayType with indexing type {ind} not supported.'
        )
