"""
Support for Series.dt attributes and methods
"""
import datetime
import operator
import numba
import numpy as np
from numba.core import cgutils, types
from numba.extending import intrinsic, make_attribute_wrapper, models, overload_attribute, overload_method, register_model
import bodo
from bodo.hiframes.pd_series_ext import SeriesType
from bodo.utils.typing import BodoError, check_unsupported_args, create_unsupported_overload, raise_bodo_error
dt64_dtype = np.dtype('datetime64[ns]')
timedelta64_dtype = np.dtype('timedelta64[ns]')


class SeriesDatetimePropertiesType(types.Type):

    def __init__(self, stype):
        self.stype = stype
        gcp__crwv = 'SeriesDatetimePropertiesType({})'.format(stype)
        super(SeriesDatetimePropertiesType, self).__init__(gcp__crwv)


@register_model(SeriesDatetimePropertiesType)
class SeriesDtModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        qkz__ywtr = [('obj', fe_type.stype)]
        super(SeriesDtModel, self).__init__(dmm, fe_type, qkz__ywtr)


make_attribute_wrapper(SeriesDatetimePropertiesType, 'obj', '_obj')


@intrinsic
def init_series_dt_properties(typingctx, obj=None):

    def codegen(context, builder, signature, args):
        aoqqn__oytkw, = args
        dgkjf__cofwn = signature.return_type
        fvcch__mtp = cgutils.create_struct_proxy(dgkjf__cofwn)(context, builder
            )
        fvcch__mtp.obj = aoqqn__oytkw
        context.nrt.incref(builder, signature.args[0], aoqqn__oytkw)
        return fvcch__mtp._getvalue()
    return SeriesDatetimePropertiesType(obj)(obj), codegen


@overload_attribute(SeriesType, 'dt')
def overload_series_dt(s):
    if not (bodo.hiframes.pd_series_ext.is_dt64_series_typ(s) or bodo.
        hiframes.pd_series_ext.is_timedelta64_series_typ(s)):
        raise_bodo_error('Can only use .dt accessor with datetimelike values.')
    return lambda s: bodo.hiframes.series_dt_impl.init_series_dt_properties(s)


def create_date_field_overload(field):

    def overload_field(S_dt):
        if not S_dt.stype.dtype == types.NPDatetime('ns'):
            return
        nfrk__dzabl = 'def impl(S_dt):\n'
        nfrk__dzabl += '    S = S_dt._obj\n'
        nfrk__dzabl += (
            '    arr = bodo.hiframes.pd_series_ext.get_series_data(S)\n')
        nfrk__dzabl += (
            '    index = bodo.hiframes.pd_series_ext.get_series_index(S)\n')
        nfrk__dzabl += (
            '    name = bodo.hiframes.pd_series_ext.get_series_name(S)\n')
        nfrk__dzabl += '    numba.parfors.parfor.init_prange()\n'
        nfrk__dzabl += '    n = len(arr)\n'
        if field in ('is_leap_year', 'is_month_start', 'is_month_end',
            'is_quarter_start', 'is_quarter_end', 'is_year_start',
            'is_year_end'):
            nfrk__dzabl += '    out_arr = np.empty(n, np.bool_)\n'
        else:
            nfrk__dzabl += (
                '    out_arr = bodo.libs.int_arr_ext.alloc_int_array(n, np.int64)\n'
                )
        nfrk__dzabl += (
            '    for i in numba.parfors.parfor.internal_prange(n):\n')
        nfrk__dzabl += '        if bodo.libs.array_kernels.isna(arr, i):\n'
        nfrk__dzabl += (
            '            bodo.libs.array_kernels.setna(out_arr, i)\n')
        nfrk__dzabl += '            continue\n'
        nfrk__dzabl += (
            '        dt64 = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(arr[i])\n'
            )
        if field in ('year', 'month', 'day'):
            nfrk__dzabl += """        dt, year, days = bodo.hiframes.pd_timestamp_ext.extract_year_days(dt64)
"""
            if field in ('month', 'day'):
                nfrk__dzabl += """        month, day = bodo.hiframes.pd_timestamp_ext.get_month_day(year, days)
"""
            nfrk__dzabl += '        out_arr[i] = {}\n'.format(field)
        elif field in ('dayofyear', 'day_of_year', 'dayofweek',
            'day_of_week', 'weekday'):
            wnzx__gkwou = {'dayofyear': 'get_day_of_year', 'day_of_year':
                'get_day_of_year', 'dayofweek': 'get_day_of_week',
                'day_of_week': 'get_day_of_week', 'weekday': 'get_day_of_week'}
            nfrk__dzabl += """        dt, year, days = bodo.hiframes.pd_timestamp_ext.extract_year_days(dt64)
"""
            nfrk__dzabl += """        month, day = bodo.hiframes.pd_timestamp_ext.get_month_day(year, days)
"""
            nfrk__dzabl += (
                """        out_arr[i] = bodo.hiframes.pd_timestamp_ext.{}(year, month, day)
"""
                .format(wnzx__gkwou[field]))
        elif field == 'is_leap_year':
            nfrk__dzabl += """        dt, year, days = bodo.hiframes.pd_timestamp_ext.extract_year_days(dt64)
"""
            nfrk__dzabl += """        out_arr[i] = bodo.hiframes.pd_timestamp_ext.is_leap_year(year)
"""
        elif field in ('daysinmonth', 'days_in_month'):
            wnzx__gkwou = {'days_in_month': 'get_days_in_month',
                'daysinmonth': 'get_days_in_month'}
            nfrk__dzabl += """        dt, year, days = bodo.hiframes.pd_timestamp_ext.extract_year_days(dt64)
"""
            nfrk__dzabl += """        month, day = bodo.hiframes.pd_timestamp_ext.get_month_day(year, days)
"""
            nfrk__dzabl += (
                '        out_arr[i] = bodo.hiframes.pd_timestamp_ext.{}(year, month)\n'
                .format(wnzx__gkwou[field]))
        else:
            nfrk__dzabl += """        ts = bodo.hiframes.pd_timestamp_ext.convert_datetime64_to_timestamp(dt64)
"""
            nfrk__dzabl += '        out_arr[i] = ts.' + field + '\n'
        nfrk__dzabl += (
            '    return bodo.hiframes.pd_series_ext.init_series(out_arr, index, name)\n'
            )
        jxmr__hulx = {}
        exec(nfrk__dzabl, {'bodo': bodo, 'numba': numba, 'np': np}, jxmr__hulx)
        impl = jxmr__hulx['impl']
        return impl
    return overload_field


def _install_date_fields():
    for field in bodo.hiframes.pd_timestamp_ext.date_fields:
        gel__lwhk = create_date_field_overload(field)
        overload_attribute(SeriesDatetimePropertiesType, field)(gel__lwhk)


_install_date_fields()


def create_date_method_overload(method):
    hfq__jglqk = method in ['day_name', 'month_name']
    if hfq__jglqk:
        nfrk__dzabl = 'def overload_method(S_dt, locale=None):\n'
        nfrk__dzabl += '    unsupported_args = dict(locale=locale)\n'
        nfrk__dzabl += '    arg_defaults = dict(locale=None)\n'
        nfrk__dzabl += '    bodo.utils.typing.check_unsupported_args(\n'
        nfrk__dzabl += f"        'Series.dt.{method}',\n"
        nfrk__dzabl += '        unsupported_args,\n'
        nfrk__dzabl += '        arg_defaults,\n'
        nfrk__dzabl += "        package_name='pandas',\n"
        nfrk__dzabl += "        module_name='Series',\n"
        nfrk__dzabl += '    )\n'
    else:
        nfrk__dzabl = 'def overload_method(S_dt):\n'
    nfrk__dzabl += '    if not S_dt.stype.dtype == bodo.datetime64ns:\n'
    nfrk__dzabl += '        return\n'
    if hfq__jglqk:
        nfrk__dzabl += '    def impl(S_dt, locale=None):\n'
    else:
        nfrk__dzabl += '    def impl(S_dt):\n'
    nfrk__dzabl += '        S = S_dt._obj\n'
    nfrk__dzabl += (
        '        arr = bodo.hiframes.pd_series_ext.get_series_data(S)\n')
    nfrk__dzabl += (
        '        index = bodo.hiframes.pd_series_ext.get_series_index(S)\n')
    nfrk__dzabl += (
        '        name = bodo.hiframes.pd_series_ext.get_series_name(S)\n')
    nfrk__dzabl += '        numba.parfors.parfor.init_prange()\n'
    nfrk__dzabl += '        n = len(arr)\n'
    if hfq__jglqk:
        nfrk__dzabl += """        out_arr = bodo.utils.utils.alloc_type(n, bodo.string_array_type, (-1,))
"""
    else:
        nfrk__dzabl += (
            "        out_arr = np.empty(n, np.dtype('datetime64[ns]'))\n")
    nfrk__dzabl += (
        '        for i in numba.parfors.parfor.internal_prange(n):\n')
    nfrk__dzabl += '            if bodo.libs.array_kernels.isna(arr, i):\n'
    nfrk__dzabl += (
        '                bodo.libs.array_kernels.setna(out_arr, i)\n')
    nfrk__dzabl += '                continue\n'
    nfrk__dzabl += """            ts = bodo.hiframes.pd_timestamp_ext.convert_datetime64_to_timestamp(arr[i])
"""
    nfrk__dzabl += f'            method_val = ts.{method}()\n'
    if hfq__jglqk:
        nfrk__dzabl += '            out_arr[i] = method_val\n'
    else:
        nfrk__dzabl += """            out_arr[i] = bodo.hiframes.pd_timestamp_ext.integer_to_dt64(method_val.value)
"""
    nfrk__dzabl += (
        '        return bodo.hiframes.pd_series_ext.init_series(out_arr, index, name)\n'
        )
    nfrk__dzabl += '    return impl\n'
    jxmr__hulx = {}
    exec(nfrk__dzabl, {'bodo': bodo, 'numba': numba, 'np': np}, jxmr__hulx)
    overload_method = jxmr__hulx['overload_method']
    return overload_method


def _install_date_methods():
    for method in bodo.hiframes.pd_timestamp_ext.date_methods:
        gel__lwhk = create_date_method_overload(method)
        overload_method(SeriesDatetimePropertiesType, method, inline='always')(
            gel__lwhk)


_install_date_methods()


@overload_attribute(SeriesDatetimePropertiesType, 'date')
def series_dt_date_overload(S_dt):
    if not S_dt.stype.dtype == types.NPDatetime('ns'):
        return

    def impl(S_dt):
        clu__eocg = S_dt._obj
        cupg__fozk = bodo.hiframes.pd_series_ext.get_series_data(clu__eocg)
        ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(clu__eocg)
        gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(clu__eocg)
        numba.parfors.parfor.init_prange()
        ahhzf__rwrhu = len(cupg__fozk)
        lgwae__mka = bodo.hiframes.datetime_date_ext.alloc_datetime_date_array(
            ahhzf__rwrhu)
        for skne__jfjdw in numba.parfors.parfor.internal_prange(ahhzf__rwrhu):
            owzj__znf = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                cupg__fozk[skne__jfjdw])
            tjt__ula = (bodo.hiframes.pd_timestamp_ext.
                convert_datetime64_to_timestamp(owzj__znf))
            lgwae__mka[skne__jfjdw] = datetime.date(tjt__ula.year, tjt__ula
                .month, tjt__ula.day)
        return bodo.hiframes.pd_series_ext.init_series(lgwae__mka,
            ljb__yzswj, gcp__crwv)
    return impl


def create_series_dt_df_output_overload(attr):

    def series_dt_df_output_overload(S_dt):
        if not (attr == 'components' and S_dt.stype.dtype == types.
            NPTimedelta('ns') or attr == 'isocalendar' and S_dt.stype.dtype ==
            types.NPDatetime('ns')):
            return
        if attr == 'components':
            jqwgl__vreh = ['days', 'hours', 'minutes', 'seconds',
                'milliseconds', 'microseconds', 'nanoseconds']
            xehcq__zhgi = 'convert_numpy_timedelta64_to_pd_timedelta'
            fio__cwmm = 'np.empty(n, np.int64)'
            sqb__nlf = attr
        elif attr == 'isocalendar':
            jqwgl__vreh = ['year', 'week', 'day']
            xehcq__zhgi = 'convert_datetime64_to_timestamp'
            fio__cwmm = 'bodo.libs.int_arr_ext.alloc_int_array(n, np.uint32)'
            sqb__nlf = attr + '()'
        nfrk__dzabl = 'def impl(S_dt):\n'
        nfrk__dzabl += '    S = S_dt._obj\n'
        nfrk__dzabl += (
            '    arr = bodo.hiframes.pd_series_ext.get_series_data(S)\n')
        nfrk__dzabl += (
            '    index = bodo.hiframes.pd_series_ext.get_series_index(S)\n')
        nfrk__dzabl += '    numba.parfors.parfor.init_prange()\n'
        nfrk__dzabl += '    n = len(arr)\n'
        for field in jqwgl__vreh:
            nfrk__dzabl += '    {} = {}\n'.format(field, fio__cwmm)
        nfrk__dzabl += (
            '    for i in numba.parfors.parfor.internal_prange(n):\n')
        nfrk__dzabl += '        if bodo.libs.array_kernels.isna(arr, i):\n'
        for field in jqwgl__vreh:
            nfrk__dzabl += (
                '            bodo.libs.array_kernels.setna({}, i)\n'.format
                (field))
        nfrk__dzabl += '            continue\n'
        xagh__hme = '(' + '[i], '.join(jqwgl__vreh) + '[i])'
        nfrk__dzabl += (
            '        {} = bodo.hiframes.pd_timestamp_ext.{}(arr[i]).{}\n'.
            format(xagh__hme, xehcq__zhgi, sqb__nlf))
        ypqc__izsjc = '(' + ', '.join(jqwgl__vreh) + ')'
        opypv__kpzt = "('" + "', '".join(jqwgl__vreh) + "')"
        nfrk__dzabl += (
            '    return bodo.hiframes.pd_dataframe_ext.init_dataframe({}, index, {})\n'
            .format(ypqc__izsjc, opypv__kpzt))
        jxmr__hulx = {}
        exec(nfrk__dzabl, {'bodo': bodo, 'numba': numba, 'np': np}, jxmr__hulx)
        impl = jxmr__hulx['impl']
        return impl
    return series_dt_df_output_overload


def _install_df_output_overload():
    ejs__xpz = [('components', overload_attribute), ('isocalendar',
        overload_method)]
    for attr, ktswj__khhq in ejs__xpz:
        gel__lwhk = create_series_dt_df_output_overload(attr)
        ktswj__khhq(SeriesDatetimePropertiesType, attr, inline='always')(
            gel__lwhk)


_install_df_output_overload()


def create_timedelta_field_overload(field):

    def overload_field(S_dt):
        if not S_dt.stype.dtype == types.NPTimedelta('ns'):
            return
        nfrk__dzabl = 'def impl(S_dt):\n'
        nfrk__dzabl += '    S = S_dt._obj\n'
        nfrk__dzabl += (
            '    A = bodo.hiframes.pd_series_ext.get_series_data(S)\n')
        nfrk__dzabl += (
            '    index = bodo.hiframes.pd_series_ext.get_series_index(S)\n')
        nfrk__dzabl += (
            '    name = bodo.hiframes.pd_series_ext.get_series_name(S)\n')
        nfrk__dzabl += '    numba.parfors.parfor.init_prange()\n'
        nfrk__dzabl += '    n = len(A)\n'
        nfrk__dzabl += (
            '    B = bodo.libs.int_arr_ext.alloc_int_array(n, np.int64)\n')
        nfrk__dzabl += (
            '    for i in numba.parfors.parfor.internal_prange(n):\n')
        nfrk__dzabl += '        if bodo.libs.array_kernels.isna(A, i):\n'
        nfrk__dzabl += '            bodo.libs.array_kernels.setna(B, i)\n'
        nfrk__dzabl += '            continue\n'
        nfrk__dzabl += """        td64 = bodo.hiframes.pd_timestamp_ext.timedelta64_to_integer(A[i])
"""
        if field == 'nanoseconds':
            nfrk__dzabl += '        B[i] = td64 % 1000\n'
        elif field == 'microseconds':
            nfrk__dzabl += '        B[i] = td64 // 1000 % 1000000\n'
        elif field == 'seconds':
            nfrk__dzabl += (
                '        B[i] = td64 // (1000 * 1000000) % (60 * 60 * 24)\n')
        elif field == 'days':
            nfrk__dzabl += (
                '        B[i] = td64 // (1000 * 1000000 * 60 * 60 * 24)\n')
        else:
            assert False, 'invalid timedelta field'
        nfrk__dzabl += (
            '    return bodo.hiframes.pd_series_ext.init_series(B, index, name)\n'
            )
        jxmr__hulx = {}
        exec(nfrk__dzabl, {'numba': numba, 'np': np, 'bodo': bodo}, jxmr__hulx)
        impl = jxmr__hulx['impl']
        return impl
    return overload_field


def create_timedelta_method_overload(method):

    def overload_method(S_dt):
        if not S_dt.stype.dtype == types.NPTimedelta('ns'):
            return
        nfrk__dzabl = 'def impl(S_dt):\n'
        nfrk__dzabl += '    S = S_dt._obj\n'
        nfrk__dzabl += (
            '    A = bodo.hiframes.pd_series_ext.get_series_data(S)\n')
        nfrk__dzabl += (
            '    index = bodo.hiframes.pd_series_ext.get_series_index(S)\n')
        nfrk__dzabl += (
            '    name = bodo.hiframes.pd_series_ext.get_series_name(S)\n')
        nfrk__dzabl += '    numba.parfors.parfor.init_prange()\n'
        nfrk__dzabl += '    n = len(A)\n'
        if method == 'total_seconds':
            nfrk__dzabl += '    B = np.empty(n, np.float64)\n'
        else:
            nfrk__dzabl += """    B = bodo.hiframes.datetime_timedelta_ext.alloc_datetime_timedelta_array(n)
"""
        nfrk__dzabl += (
            '    for i in numba.parfors.parfor.internal_prange(n):\n')
        nfrk__dzabl += '        if bodo.libs.array_kernels.isna(A, i):\n'
        nfrk__dzabl += '            bodo.libs.array_kernels.setna(B, i)\n'
        nfrk__dzabl += '            continue\n'
        nfrk__dzabl += """        td64 = bodo.hiframes.pd_timestamp_ext.timedelta64_to_integer(A[i])
"""
        if method == 'total_seconds':
            nfrk__dzabl += '        B[i] = td64 / (1000.0 * 1000000.0)\n'
        elif method == 'to_pytimedelta':
            nfrk__dzabl += (
                '        B[i] = datetime.timedelta(microseconds=td64 // 1000)\n'
                )
        else:
            assert False, 'invalid timedelta method'
        if method == 'total_seconds':
            nfrk__dzabl += (
                '    return bodo.hiframes.pd_series_ext.init_series(B, index, name)\n'
                )
        else:
            nfrk__dzabl += '    return B\n'
        jxmr__hulx = {}
        exec(nfrk__dzabl, {'numba': numba, 'np': np, 'bodo': bodo,
            'datetime': datetime}, jxmr__hulx)
        impl = jxmr__hulx['impl']
        return impl
    return overload_method


def _install_S_dt_timedelta_fields():
    for field in bodo.hiframes.pd_timestamp_ext.timedelta_fields:
        gel__lwhk = create_timedelta_field_overload(field)
        overload_attribute(SeriesDatetimePropertiesType, field)(gel__lwhk)


_install_S_dt_timedelta_fields()


def _install_S_dt_timedelta_methods():
    for method in bodo.hiframes.pd_timestamp_ext.timedelta_methods:
        gel__lwhk = create_timedelta_method_overload(method)
        overload_method(SeriesDatetimePropertiesType, method, inline='always')(
            gel__lwhk)


_install_S_dt_timedelta_methods()


@overload_method(SeriesDatetimePropertiesType, 'strftime', inline='always',
    no_unliteral=True)
def dt_strftime(S_dt, date_format):
    if S_dt.stype.dtype != types.NPDatetime('ns'):
        return
    if types.unliteral(date_format) != types.unicode_type:
        raise BodoError(
            "Series.str.strftime(): 'date_format' argument must be a string")

    def impl(S_dt, date_format):
        clu__eocg = S_dt._obj
        fmfar__wui = bodo.hiframes.pd_series_ext.get_series_data(clu__eocg)
        ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(clu__eocg)
        gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(clu__eocg)
        numba.parfors.parfor.init_prange()
        ahhzf__rwrhu = len(fmfar__wui)
        whgq__sii = bodo.libs.str_arr_ext.pre_alloc_string_array(ahhzf__rwrhu,
            -1)
        for maj__nqyl in numba.parfors.parfor.internal_prange(ahhzf__rwrhu):
            if bodo.libs.array_kernels.isna(fmfar__wui, maj__nqyl):
                bodo.libs.array_kernels.setna(whgq__sii, maj__nqyl)
                continue
            whgq__sii[maj__nqyl
                ] = bodo.hiframes.pd_timestamp_ext.convert_datetime64_to_timestamp(
                fmfar__wui[maj__nqyl]).strftime(date_format)
        return bodo.hiframes.pd_series_ext.init_series(whgq__sii,
            ljb__yzswj, gcp__crwv)
    return impl


def create_timedelta_freq_overload(method):

    def freq_overload(S_dt, freq, ambiguous='raise', nonexistent='raise'):
        if S_dt.stype.dtype != types.NPTimedelta('ns'
            ) and S_dt.stype.dtype != types.NPDatetime('ns'):
            return
        nif__zhyfp = dict(ambiguous=ambiguous, nonexistent=nonexistent)
        tgum__pfbn = dict(ambiguous='raise', nonexistent='raise')
        check_unsupported_args(f'Series.dt.{method}', nif__zhyfp,
            tgum__pfbn, package_name='pandas', module_name='Series')
        nfrk__dzabl = (
            "def impl(S_dt, freq, ambiguous='raise', nonexistent='raise'):\n")
        nfrk__dzabl += '    S = S_dt._obj\n'
        nfrk__dzabl += (
            '    A = bodo.hiframes.pd_series_ext.get_series_data(S)\n')
        nfrk__dzabl += (
            '    index = bodo.hiframes.pd_series_ext.get_series_index(S)\n')
        nfrk__dzabl += (
            '    name = bodo.hiframes.pd_series_ext.get_series_name(S)\n')
        nfrk__dzabl += '    numba.parfors.parfor.init_prange()\n'
        nfrk__dzabl += '    n = len(A)\n'
        if S_dt.stype.dtype == types.NPTimedelta('ns'):
            nfrk__dzabl += "    B = np.empty(n, np.dtype('timedelta64[ns]'))\n"
        else:
            nfrk__dzabl += "    B = np.empty(n, np.dtype('datetime64[ns]'))\n"
        nfrk__dzabl += (
            '    for i in numba.parfors.parfor.internal_prange(n):\n')
        nfrk__dzabl += '        if bodo.libs.array_kernels.isna(A, i):\n'
        nfrk__dzabl += '            bodo.libs.array_kernels.setna(B, i)\n'
        nfrk__dzabl += '            continue\n'
        if S_dt.stype.dtype == types.NPTimedelta('ns'):
            jxhn__ludff = (
                'bodo.hiframes.pd_timestamp_ext.convert_numpy_timedelta64_to_pd_timedelta'
                )
            dhbl__jfy = 'bodo.hiframes.pd_timestamp_ext.integer_to_timedelta64'
        else:
            jxhn__ludff = (
                'bodo.hiframes.pd_timestamp_ext.convert_datetime64_to_timestamp'
                )
            dhbl__jfy = 'bodo.hiframes.pd_timestamp_ext.integer_to_dt64'
        nfrk__dzabl += '        B[i] = {}({}(A[i]).{}(freq).value)\n'.format(
            dhbl__jfy, jxhn__ludff, method)
        nfrk__dzabl += (
            '    return bodo.hiframes.pd_series_ext.init_series(B, index, name)\n'
            )
        jxmr__hulx = {}
        exec(nfrk__dzabl, {'numba': numba, 'np': np, 'bodo': bodo}, jxmr__hulx)
        impl = jxmr__hulx['impl']
        return impl
    return freq_overload


def _install_S_dt_timedelta_freq_methods():
    jowt__ijenv = ['ceil', 'floor', 'round']
    for method in jowt__ijenv:
        gel__lwhk = create_timedelta_freq_overload(method)
        overload_method(SeriesDatetimePropertiesType, method, inline='always')(
            gel__lwhk)


_install_S_dt_timedelta_freq_methods()


def create_bin_op_overload(op):

    def overload_series_dt_binop(lhs, rhs):
        if bodo.hiframes.pd_series_ext.is_dt64_series_typ(lhs
            ) and bodo.hiframes.pd_series_ext.is_dt64_series_typ(rhs):
            ioak__ojueu = lhs.dtype('NaT')

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                avtco__lemj = bodo.hiframes.pd_series_ext.get_series_data(lhs)
                ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(lhs)
                gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(lhs)
                ujood__vaudz = bodo.hiframes.pd_series_ext.get_series_data(rhs)
                ahhzf__rwrhu = len(avtco__lemj)
                clu__eocg = np.empty(ahhzf__rwrhu, timedelta64_dtype)
                kffy__lkmg = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                    ioak__ojueu)
                for skne__jfjdw in numba.parfors.parfor.internal_prange(
                    ahhzf__rwrhu):
                    hoqgl__zgkb = (bodo.hiframes.pd_timestamp_ext.
                        dt64_to_integer(avtco__lemj[skne__jfjdw]))
                    vfwbk__ilxc = (bodo.hiframes.pd_timestamp_ext.
                        dt64_to_integer(ujood__vaudz[skne__jfjdw]))
                    if hoqgl__zgkb == kffy__lkmg or vfwbk__ilxc == kffy__lkmg:
                        vpd__lbtmm = kffy__lkmg
                    else:
                        vpd__lbtmm = op(hoqgl__zgkb, vfwbk__ilxc)
                    clu__eocg[skne__jfjdw
                        ] = bodo.hiframes.pd_timestamp_ext.integer_to_timedelta64(
                        vpd__lbtmm)
                return bodo.hiframes.pd_series_ext.init_series(clu__eocg,
                    ljb__yzswj, gcp__crwv)
            return impl
        if bodo.hiframes.pd_series_ext.is_dt64_series_typ(lhs
            ) and bodo.hiframes.pd_series_ext.is_timedelta64_series_typ(rhs):
            ioak__ojueu = lhs.dtype('NaT')

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                cupg__fozk = bodo.hiframes.pd_series_ext.get_series_data(lhs)
                ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(lhs)
                gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(lhs)
                ujood__vaudz = bodo.hiframes.pd_series_ext.get_series_data(rhs)
                ahhzf__rwrhu = len(cupg__fozk)
                clu__eocg = np.empty(ahhzf__rwrhu, dt64_dtype)
                kffy__lkmg = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                    ioak__ojueu)
                for skne__jfjdw in numba.parfors.parfor.internal_prange(
                    ahhzf__rwrhu):
                    ndvw__vgoj = (bodo.hiframes.pd_timestamp_ext.
                        dt64_to_integer(cupg__fozk[skne__jfjdw]))
                    mmesb__keih = (bodo.hiframes.pd_timestamp_ext.
                        timedelta64_to_integer(ujood__vaudz[skne__jfjdw]))
                    if ndvw__vgoj == kffy__lkmg or mmesb__keih == kffy__lkmg:
                        vpd__lbtmm = kffy__lkmg
                    else:
                        vpd__lbtmm = op(ndvw__vgoj, mmesb__keih)
                    clu__eocg[skne__jfjdw
                        ] = bodo.hiframes.pd_timestamp_ext.integer_to_dt64(
                        vpd__lbtmm)
                return bodo.hiframes.pd_series_ext.init_series(clu__eocg,
                    ljb__yzswj, gcp__crwv)
            return impl
        if bodo.hiframes.pd_series_ext.is_dt64_series_typ(rhs
            ) and bodo.hiframes.pd_series_ext.is_timedelta64_series_typ(lhs):
            ioak__ojueu = rhs.dtype('NaT')

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                cupg__fozk = bodo.hiframes.pd_series_ext.get_series_data(rhs)
                ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(rhs)
                gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(rhs)
                ujood__vaudz = bodo.hiframes.pd_series_ext.get_series_data(lhs)
                ahhzf__rwrhu = len(cupg__fozk)
                clu__eocg = np.empty(ahhzf__rwrhu, dt64_dtype)
                kffy__lkmg = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                    ioak__ojueu)
                for skne__jfjdw in numba.parfors.parfor.internal_prange(
                    ahhzf__rwrhu):
                    ndvw__vgoj = (bodo.hiframes.pd_timestamp_ext.
                        dt64_to_integer(cupg__fozk[skne__jfjdw]))
                    mmesb__keih = (bodo.hiframes.pd_timestamp_ext.
                        timedelta64_to_integer(ujood__vaudz[skne__jfjdw]))
                    if ndvw__vgoj == kffy__lkmg or mmesb__keih == kffy__lkmg:
                        vpd__lbtmm = kffy__lkmg
                    else:
                        vpd__lbtmm = op(ndvw__vgoj, mmesb__keih)
                    clu__eocg[skne__jfjdw
                        ] = bodo.hiframes.pd_timestamp_ext.integer_to_dt64(
                        vpd__lbtmm)
                return bodo.hiframes.pd_series_ext.init_series(clu__eocg,
                    ljb__yzswj, gcp__crwv)
            return impl
        if bodo.hiframes.pd_series_ext.is_dt64_series_typ(lhs
            ) and rhs == bodo.hiframes.pd_timestamp_ext.pd_timestamp_type:
            ioak__ojueu = lhs.dtype('NaT')

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                cupg__fozk = bodo.hiframes.pd_series_ext.get_series_data(lhs)
                ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(lhs)
                gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(lhs)
                ahhzf__rwrhu = len(cupg__fozk)
                clu__eocg = np.empty(ahhzf__rwrhu, timedelta64_dtype)
                kffy__lkmg = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                    ioak__ojueu)
                sqm__lbx = rhs.value
                for skne__jfjdw in numba.parfors.parfor.internal_prange(
                    ahhzf__rwrhu):
                    ndvw__vgoj = (bodo.hiframes.pd_timestamp_ext.
                        dt64_to_integer(cupg__fozk[skne__jfjdw]))
                    if ndvw__vgoj == kffy__lkmg or sqm__lbx == kffy__lkmg:
                        vpd__lbtmm = kffy__lkmg
                    else:
                        vpd__lbtmm = op(ndvw__vgoj, sqm__lbx)
                    clu__eocg[skne__jfjdw
                        ] = bodo.hiframes.pd_timestamp_ext.integer_to_timedelta64(
                        vpd__lbtmm)
                return bodo.hiframes.pd_series_ext.init_series(clu__eocg,
                    ljb__yzswj, gcp__crwv)
            return impl
        if bodo.hiframes.pd_series_ext.is_dt64_series_typ(rhs
            ) and lhs == bodo.hiframes.pd_timestamp_ext.pd_timestamp_type:
            ioak__ojueu = rhs.dtype('NaT')

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                cupg__fozk = bodo.hiframes.pd_series_ext.get_series_data(rhs)
                ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(rhs)
                gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(rhs)
                ahhzf__rwrhu = len(cupg__fozk)
                clu__eocg = np.empty(ahhzf__rwrhu, timedelta64_dtype)
                kffy__lkmg = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                    ioak__ojueu)
                sqm__lbx = lhs.value
                for skne__jfjdw in numba.parfors.parfor.internal_prange(
                    ahhzf__rwrhu):
                    ndvw__vgoj = (bodo.hiframes.pd_timestamp_ext.
                        dt64_to_integer(cupg__fozk[skne__jfjdw]))
                    if sqm__lbx == kffy__lkmg or ndvw__vgoj == kffy__lkmg:
                        vpd__lbtmm = kffy__lkmg
                    else:
                        vpd__lbtmm = op(sqm__lbx, ndvw__vgoj)
                    clu__eocg[skne__jfjdw
                        ] = bodo.hiframes.pd_timestamp_ext.integer_to_timedelta64(
                        vpd__lbtmm)
                return bodo.hiframes.pd_series_ext.init_series(clu__eocg,
                    ljb__yzswj, gcp__crwv)
            return impl
        if (bodo.hiframes.pd_series_ext.is_dt64_series_typ(lhs) and rhs ==
            bodo.hiframes.datetime_timedelta_ext.datetime_timedelta_type):
            ioak__ojueu = lhs.dtype('NaT')

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                cupg__fozk = bodo.hiframes.pd_series_ext.get_series_data(lhs)
                ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(lhs)
                gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(lhs)
                ahhzf__rwrhu = len(cupg__fozk)
                clu__eocg = np.empty(ahhzf__rwrhu, dt64_dtype)
                kffy__lkmg = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                    ioak__ojueu)
                tfqy__aph = (bodo.hiframes.pd_timestamp_ext.
                    datetime_timedelta_to_timedelta64(rhs))
                mmesb__keih = (bodo.hiframes.pd_timestamp_ext.
                    timedelta64_to_integer(tfqy__aph))
                for skne__jfjdw in numba.parfors.parfor.internal_prange(
                    ahhzf__rwrhu):
                    ndvw__vgoj = (bodo.hiframes.pd_timestamp_ext.
                        dt64_to_integer(cupg__fozk[skne__jfjdw]))
                    if ndvw__vgoj == kffy__lkmg or mmesb__keih == kffy__lkmg:
                        vpd__lbtmm = kffy__lkmg
                    else:
                        vpd__lbtmm = op(ndvw__vgoj, mmesb__keih)
                    clu__eocg[skne__jfjdw
                        ] = bodo.hiframes.pd_timestamp_ext.integer_to_dt64(
                        vpd__lbtmm)
                return bodo.hiframes.pd_series_ext.init_series(clu__eocg,
                    ljb__yzswj, gcp__crwv)
            return impl
        if (bodo.hiframes.pd_series_ext.is_dt64_series_typ(rhs) and lhs ==
            bodo.hiframes.datetime_timedelta_ext.datetime_timedelta_type):
            ioak__ojueu = rhs.dtype('NaT')

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                cupg__fozk = bodo.hiframes.pd_series_ext.get_series_data(rhs)
                ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(rhs)
                gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(rhs)
                ahhzf__rwrhu = len(cupg__fozk)
                clu__eocg = np.empty(ahhzf__rwrhu, dt64_dtype)
                kffy__lkmg = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                    ioak__ojueu)
                tfqy__aph = (bodo.hiframes.pd_timestamp_ext.
                    datetime_timedelta_to_timedelta64(lhs))
                mmesb__keih = (bodo.hiframes.pd_timestamp_ext.
                    timedelta64_to_integer(tfqy__aph))
                for skne__jfjdw in numba.parfors.parfor.internal_prange(
                    ahhzf__rwrhu):
                    ndvw__vgoj = (bodo.hiframes.pd_timestamp_ext.
                        dt64_to_integer(cupg__fozk[skne__jfjdw]))
                    if ndvw__vgoj == kffy__lkmg or mmesb__keih == kffy__lkmg:
                        vpd__lbtmm = kffy__lkmg
                    else:
                        vpd__lbtmm = op(ndvw__vgoj, mmesb__keih)
                    clu__eocg[skne__jfjdw
                        ] = bodo.hiframes.pd_timestamp_ext.integer_to_dt64(
                        vpd__lbtmm)
                return bodo.hiframes.pd_series_ext.init_series(clu__eocg,
                    ljb__yzswj, gcp__crwv)
            return impl
        if (bodo.hiframes.pd_series_ext.is_dt64_series_typ(lhs) and rhs ==
            bodo.hiframes.datetime_datetime_ext.datetime_datetime_type):
            ioak__ojueu = lhs.dtype('NaT')

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                cupg__fozk = bodo.hiframes.pd_series_ext.get_series_data(lhs)
                ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(lhs)
                gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(lhs)
                ahhzf__rwrhu = len(cupg__fozk)
                clu__eocg = np.empty(ahhzf__rwrhu, timedelta64_dtype)
                kffy__lkmg = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                    ioak__ojueu)
                owzj__znf = (bodo.hiframes.pd_timestamp_ext.
                    datetime_datetime_to_dt64(rhs))
                ndvw__vgoj = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                    owzj__znf)
                for skne__jfjdw in numba.parfors.parfor.internal_prange(
                    ahhzf__rwrhu):
                    hfxg__pqyye = (bodo.hiframes.pd_timestamp_ext.
                        dt64_to_integer(cupg__fozk[skne__jfjdw]))
                    if hfxg__pqyye == kffy__lkmg or ndvw__vgoj == kffy__lkmg:
                        vpd__lbtmm = kffy__lkmg
                    else:
                        vpd__lbtmm = op(hfxg__pqyye, ndvw__vgoj)
                    clu__eocg[skne__jfjdw
                        ] = bodo.hiframes.pd_timestamp_ext.integer_to_timedelta64(
                        vpd__lbtmm)
                return bodo.hiframes.pd_series_ext.init_series(clu__eocg,
                    ljb__yzswj, gcp__crwv)
            return impl
        if (bodo.hiframes.pd_series_ext.is_dt64_series_typ(rhs) and lhs ==
            bodo.hiframes.datetime_datetime_ext.datetime_datetime_type):
            ioak__ojueu = rhs.dtype('NaT')

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                cupg__fozk = bodo.hiframes.pd_series_ext.get_series_data(rhs)
                ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(rhs)
                gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(rhs)
                ahhzf__rwrhu = len(cupg__fozk)
                clu__eocg = np.empty(ahhzf__rwrhu, timedelta64_dtype)
                kffy__lkmg = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                    ioak__ojueu)
                owzj__znf = (bodo.hiframes.pd_timestamp_ext.
                    datetime_datetime_to_dt64(lhs))
                ndvw__vgoj = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                    owzj__znf)
                for skne__jfjdw in numba.parfors.parfor.internal_prange(
                    ahhzf__rwrhu):
                    hfxg__pqyye = (bodo.hiframes.pd_timestamp_ext.
                        dt64_to_integer(cupg__fozk[skne__jfjdw]))
                    if ndvw__vgoj == kffy__lkmg or hfxg__pqyye == kffy__lkmg:
                        vpd__lbtmm = kffy__lkmg
                    else:
                        vpd__lbtmm = op(ndvw__vgoj, hfxg__pqyye)
                    clu__eocg[skne__jfjdw
                        ] = bodo.hiframes.pd_timestamp_ext.integer_to_timedelta64(
                        vpd__lbtmm)
                return bodo.hiframes.pd_series_ext.init_series(clu__eocg,
                    ljb__yzswj, gcp__crwv)
            return impl
        if (bodo.hiframes.pd_series_ext.is_timedelta64_series_typ(lhs) and 
            rhs == bodo.hiframes.datetime_timedelta_ext.datetime_timedelta_type
            ):
            ioak__ojueu = lhs.dtype('NaT')

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                cupg__fozk = bodo.hiframes.pd_series_ext.get_series_data(lhs)
                ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(lhs)
                gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(lhs)
                ahhzf__rwrhu = len(cupg__fozk)
                clu__eocg = np.empty(ahhzf__rwrhu, timedelta64_dtype)
                kffy__lkmg = (bodo.hiframes.pd_timestamp_ext.
                    timedelta64_to_integer(ioak__ojueu))
                tfqy__aph = (bodo.hiframes.pd_timestamp_ext.
                    datetime_timedelta_to_timedelta64(rhs))
                mmesb__keih = (bodo.hiframes.pd_timestamp_ext.
                    timedelta64_to_integer(tfqy__aph))
                for skne__jfjdw in numba.parfors.parfor.internal_prange(
                    ahhzf__rwrhu):
                    uoyl__csi = (bodo.hiframes.pd_timestamp_ext.
                        timedelta64_to_integer(cupg__fozk[skne__jfjdw]))
                    if mmesb__keih == kffy__lkmg or uoyl__csi == kffy__lkmg:
                        vpd__lbtmm = kffy__lkmg
                    else:
                        vpd__lbtmm = op(uoyl__csi, mmesb__keih)
                    clu__eocg[skne__jfjdw
                        ] = bodo.hiframes.pd_timestamp_ext.integer_to_timedelta64(
                        vpd__lbtmm)
                return bodo.hiframes.pd_series_ext.init_series(clu__eocg,
                    ljb__yzswj, gcp__crwv)
            return impl
        if (bodo.hiframes.pd_series_ext.is_timedelta64_series_typ(rhs) and 
            lhs == bodo.hiframes.datetime_timedelta_ext.datetime_timedelta_type
            ):
            ioak__ojueu = rhs.dtype('NaT')

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                cupg__fozk = bodo.hiframes.pd_series_ext.get_series_data(rhs)
                ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(rhs)
                gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(rhs)
                ahhzf__rwrhu = len(cupg__fozk)
                clu__eocg = np.empty(ahhzf__rwrhu, timedelta64_dtype)
                kffy__lkmg = (bodo.hiframes.pd_timestamp_ext.
                    timedelta64_to_integer(ioak__ojueu))
                tfqy__aph = (bodo.hiframes.pd_timestamp_ext.
                    datetime_timedelta_to_timedelta64(lhs))
                mmesb__keih = (bodo.hiframes.pd_timestamp_ext.
                    timedelta64_to_integer(tfqy__aph))
                for skne__jfjdw in numba.parfors.parfor.internal_prange(
                    ahhzf__rwrhu):
                    uoyl__csi = (bodo.hiframes.pd_timestamp_ext.
                        timedelta64_to_integer(cupg__fozk[skne__jfjdw]))
                    if mmesb__keih == kffy__lkmg or uoyl__csi == kffy__lkmg:
                        vpd__lbtmm = kffy__lkmg
                    else:
                        vpd__lbtmm = op(mmesb__keih, uoyl__csi)
                    clu__eocg[skne__jfjdw
                        ] = bodo.hiframes.pd_timestamp_ext.integer_to_timedelta64(
                        vpd__lbtmm)
                return bodo.hiframes.pd_series_ext.init_series(clu__eocg,
                    ljb__yzswj, gcp__crwv)
            return impl
        raise BodoError(f'{op} not supported for data types {lhs} and {rhs}.')
    return overload_series_dt_binop


def create_cmp_op_overload(op):

    def overload_series_dt64_cmp(lhs, rhs):
        if op == operator.ne:
            fanbc__nxuan = True
        else:
            fanbc__nxuan = False
        if (bodo.hiframes.pd_series_ext.is_timedelta64_series_typ(lhs) and 
            rhs == bodo.hiframes.datetime_timedelta_ext.datetime_timedelta_type
            ):
            ioak__ojueu = lhs.dtype('NaT')

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                cupg__fozk = bodo.hiframes.pd_series_ext.get_series_data(lhs)
                ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(lhs)
                gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(lhs)
                ahhzf__rwrhu = len(cupg__fozk)
                lgwae__mka = bodo.libs.bool_arr_ext.alloc_bool_array(
                    ahhzf__rwrhu)
                kffy__lkmg = (bodo.hiframes.pd_timestamp_ext.
                    timedelta64_to_integer(ioak__ojueu))
                gthzf__jhhq = (bodo.hiframes.pd_timestamp_ext.
                    datetime_timedelta_to_timedelta64(rhs))
                cabd__crf = (bodo.hiframes.pd_timestamp_ext.
                    timedelta64_to_integer(gthzf__jhhq))
                for skne__jfjdw in numba.parfors.parfor.internal_prange(
                    ahhzf__rwrhu):
                    rgz__modyi = (bodo.hiframes.pd_timestamp_ext.
                        timedelta64_to_integer(cupg__fozk[skne__jfjdw]))
                    if rgz__modyi == kffy__lkmg or cabd__crf == kffy__lkmg:
                        vpd__lbtmm = fanbc__nxuan
                    else:
                        vpd__lbtmm = op(rgz__modyi, cabd__crf)
                    lgwae__mka[skne__jfjdw] = vpd__lbtmm
                return bodo.hiframes.pd_series_ext.init_series(lgwae__mka,
                    ljb__yzswj, gcp__crwv)
            return impl
        if (bodo.hiframes.pd_series_ext.is_timedelta64_series_typ(rhs) and 
            lhs == bodo.hiframes.datetime_timedelta_ext.datetime_timedelta_type
            ):
            ioak__ojueu = rhs.dtype('NaT')

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                cupg__fozk = bodo.hiframes.pd_series_ext.get_series_data(rhs)
                ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(rhs)
                gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(rhs)
                ahhzf__rwrhu = len(cupg__fozk)
                lgwae__mka = bodo.libs.bool_arr_ext.alloc_bool_array(
                    ahhzf__rwrhu)
                kffy__lkmg = (bodo.hiframes.pd_timestamp_ext.
                    timedelta64_to_integer(ioak__ojueu))
                kjuiw__qipi = (bodo.hiframes.pd_timestamp_ext.
                    datetime_timedelta_to_timedelta64(lhs))
                rgz__modyi = (bodo.hiframes.pd_timestamp_ext.
                    timedelta64_to_integer(kjuiw__qipi))
                for skne__jfjdw in numba.parfors.parfor.internal_prange(
                    ahhzf__rwrhu):
                    cabd__crf = (bodo.hiframes.pd_timestamp_ext.
                        timedelta64_to_integer(cupg__fozk[skne__jfjdw]))
                    if rgz__modyi == kffy__lkmg or cabd__crf == kffy__lkmg:
                        vpd__lbtmm = fanbc__nxuan
                    else:
                        vpd__lbtmm = op(rgz__modyi, cabd__crf)
                    lgwae__mka[skne__jfjdw] = vpd__lbtmm
                return bodo.hiframes.pd_series_ext.init_series(lgwae__mka,
                    ljb__yzswj, gcp__crwv)
            return impl
        if bodo.hiframes.pd_series_ext.is_dt64_series_typ(lhs
            ) and rhs == bodo.hiframes.pd_timestamp_ext.pd_timestamp_type:
            ioak__ojueu = lhs.dtype('NaT')

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                cupg__fozk = bodo.hiframes.pd_series_ext.get_series_data(lhs)
                ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(lhs)
                gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(lhs)
                ahhzf__rwrhu = len(cupg__fozk)
                lgwae__mka = bodo.libs.bool_arr_ext.alloc_bool_array(
                    ahhzf__rwrhu)
                kffy__lkmg = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                    ioak__ojueu)
                for skne__jfjdw in numba.parfors.parfor.internal_prange(
                    ahhzf__rwrhu):
                    rgz__modyi = (bodo.hiframes.pd_timestamp_ext.
                        dt64_to_integer(cupg__fozk[skne__jfjdw]))
                    if rgz__modyi == kffy__lkmg or rhs.value == kffy__lkmg:
                        vpd__lbtmm = fanbc__nxuan
                    else:
                        vpd__lbtmm = op(rgz__modyi, rhs.value)
                    lgwae__mka[skne__jfjdw] = vpd__lbtmm
                return bodo.hiframes.pd_series_ext.init_series(lgwae__mka,
                    ljb__yzswj, gcp__crwv)
            return impl
        if (lhs == bodo.hiframes.pd_timestamp_ext.pd_timestamp_type and
            bodo.hiframes.pd_series_ext.is_dt64_series_typ(rhs)):
            ioak__ojueu = rhs.dtype('NaT')

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                cupg__fozk = bodo.hiframes.pd_series_ext.get_series_data(rhs)
                ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(rhs)
                gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(rhs)
                ahhzf__rwrhu = len(cupg__fozk)
                lgwae__mka = bodo.libs.bool_arr_ext.alloc_bool_array(
                    ahhzf__rwrhu)
                kffy__lkmg = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                    ioak__ojueu)
                for skne__jfjdw in numba.parfors.parfor.internal_prange(
                    ahhzf__rwrhu):
                    cabd__crf = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                        cupg__fozk[skne__jfjdw])
                    if cabd__crf == kffy__lkmg or lhs.value == kffy__lkmg:
                        vpd__lbtmm = fanbc__nxuan
                    else:
                        vpd__lbtmm = op(lhs.value, cabd__crf)
                    lgwae__mka[skne__jfjdw] = vpd__lbtmm
                return bodo.hiframes.pd_series_ext.init_series(lgwae__mka,
                    ljb__yzswj, gcp__crwv)
            return impl
        if bodo.hiframes.pd_series_ext.is_dt64_series_typ(lhs) and (rhs ==
            bodo.libs.str_ext.string_type or bodo.utils.typing.
            is_overload_constant_str(rhs)):
            ioak__ojueu = lhs.dtype('NaT')

            def impl(lhs, rhs):
                cupg__fozk = bodo.hiframes.pd_series_ext.get_series_data(lhs)
                ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(lhs)
                gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(lhs)
                numba.parfors.parfor.init_prange()
                ahhzf__rwrhu = len(cupg__fozk)
                lgwae__mka = bodo.libs.bool_arr_ext.alloc_bool_array(
                    ahhzf__rwrhu)
                kffy__lkmg = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                    ioak__ojueu)
                sep__qzip = bodo.hiframes.pd_timestamp_ext.parse_datetime_str(
                    rhs)
                pikyw__ceax = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                    sep__qzip)
                for skne__jfjdw in numba.parfors.parfor.internal_prange(
                    ahhzf__rwrhu):
                    rgz__modyi = (bodo.hiframes.pd_timestamp_ext.
                        dt64_to_integer(cupg__fozk[skne__jfjdw]))
                    if rgz__modyi == kffy__lkmg or pikyw__ceax == kffy__lkmg:
                        vpd__lbtmm = fanbc__nxuan
                    else:
                        vpd__lbtmm = op(rgz__modyi, pikyw__ceax)
                    lgwae__mka[skne__jfjdw] = vpd__lbtmm
                return bodo.hiframes.pd_series_ext.init_series(lgwae__mka,
                    ljb__yzswj, gcp__crwv)
            return impl
        if bodo.hiframes.pd_series_ext.is_dt64_series_typ(rhs) and (lhs ==
            bodo.libs.str_ext.string_type or bodo.utils.typing.
            is_overload_constant_str(lhs)):
            ioak__ojueu = rhs.dtype('NaT')

            def impl(lhs, rhs):
                cupg__fozk = bodo.hiframes.pd_series_ext.get_series_data(rhs)
                ljb__yzswj = bodo.hiframes.pd_series_ext.get_series_index(rhs)
                gcp__crwv = bodo.hiframes.pd_series_ext.get_series_name(rhs)
                numba.parfors.parfor.init_prange()
                ahhzf__rwrhu = len(cupg__fozk)
                lgwae__mka = bodo.libs.bool_arr_ext.alloc_bool_array(
                    ahhzf__rwrhu)
                kffy__lkmg = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                    ioak__ojueu)
                sep__qzip = bodo.hiframes.pd_timestamp_ext.parse_datetime_str(
                    lhs)
                pikyw__ceax = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                    sep__qzip)
                for skne__jfjdw in numba.parfors.parfor.internal_prange(
                    ahhzf__rwrhu):
                    owzj__znf = bodo.hiframes.pd_timestamp_ext.dt64_to_integer(
                        cupg__fozk[skne__jfjdw])
                    if owzj__znf == kffy__lkmg or pikyw__ceax == kffy__lkmg:
                        vpd__lbtmm = fanbc__nxuan
                    else:
                        vpd__lbtmm = op(pikyw__ceax, owzj__znf)
                    lgwae__mka[skne__jfjdw] = vpd__lbtmm
                return bodo.hiframes.pd_series_ext.init_series(lgwae__mka,
                    ljb__yzswj, gcp__crwv)
            return impl
        raise BodoError(
            f'{op} operator not supported for data types {lhs} and {rhs}.')
    return overload_series_dt64_cmp


series_dt_unsupported_methods = {'to_period', 'to_pydatetime',
    'tz_localize', 'tz_convert', 'asfreq', 'to_timestamp'}
series_dt_unsupported_attrs = {'time', 'timetz', 'tz', 'freq', 'qyear',
    'start_time', 'end_time'}


def _install_series_dt_unsupported():
    for xby__gbu in series_dt_unsupported_attrs:
        jfok__rrx = 'Series.dt.' + xby__gbu
        overload_attribute(SeriesDatetimePropertiesType, xby__gbu)(
            create_unsupported_overload(jfok__rrx))
    for pdzdq__fet in series_dt_unsupported_methods:
        jfok__rrx = 'Series.dt.' + pdzdq__fet
        overload_method(SeriesDatetimePropertiesType, pdzdq__fet,
            no_unliteral=True)(create_unsupported_overload(jfok__rrx))


_install_series_dt_unsupported()
