import operator
import llvmlite.binding as ll
import numba
import numba.core.typing.typeof
import numpy as np
from llvmlite import ir as lir
from llvmlite.llvmpy.core import Type as LLType
from numba.core import cgutils, types
from numba.core.imputils import impl_ret_borrowed, impl_ret_new_ref
from numba.extending import box, intrinsic, make_attribute_wrapper, models, overload, overload_attribute, register_model
import bodo
from bodo.libs import hstr_ext
from bodo.libs.array_item_arr_ext import offset_type
from bodo.libs.str_arr_ext import _get_str_binary_arr_payload, _memcpy, char_arr_type, get_data_ptr, null_bitmap_arr_type, offset_arr_type, string_array_type
ll.add_symbol('array_setitem', hstr_ext.array_setitem)
ll.add_symbol('array_getptr1', hstr_ext.array_getptr1)
ll.add_symbol('dtor_str_arr_split_view', hstr_ext.dtor_str_arr_split_view)
ll.add_symbol('str_arr_split_view_impl', hstr_ext.str_arr_split_view_impl)
ll.add_symbol('str_arr_split_view_alloc', hstr_ext.str_arr_split_view_alloc)
char_typ = types.uint8
data_ctypes_type = types.ArrayCTypes(types.Array(char_typ, 1, 'C'))
offset_ctypes_type = types.ArrayCTypes(types.Array(offset_type, 1, 'C'))


class StringArraySplitViewType(types.ArrayCompatible):

    def __init__(self):
        super(StringArraySplitViewType, self).__init__(name=
            'StringArraySplitViewType()')

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    @property
    def dtype(self):
        return string_array_type

    def copy(self):
        return StringArraySplitViewType()


string_array_split_view_type = StringArraySplitViewType()


class StringArraySplitViewPayloadType(types.Type):

    def __init__(self):
        super(StringArraySplitViewPayloadType, self).__init__(name=
            'StringArraySplitViewPayloadType()')


str_arr_split_view_payload_type = StringArraySplitViewPayloadType()


@register_model(StringArraySplitViewPayloadType)
class StringArrayPayloadModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        pwxyh__utj = [('index_offsets', types.CPointer(offset_type)), (
            'data_offsets', types.CPointer(offset_type)), ('null_bitmap',
            types.CPointer(char_typ))]
        models.StructModel.__init__(self, dmm, fe_type, pwxyh__utj)


str_arr_model_members = [('num_items', types.uint64), ('index_offsets',
    types.CPointer(offset_type)), ('data_offsets', types.CPointer(
    offset_type)), ('data', data_ctypes_type), ('null_bitmap', types.
    CPointer(char_typ)), ('meminfo', types.MemInfoPointer(
    str_arr_split_view_payload_type))]


@register_model(StringArraySplitViewType)
class StringArrayModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        models.StructModel.__init__(self, dmm, fe_type, str_arr_model_members)


make_attribute_wrapper(StringArraySplitViewType, 'num_items', '_num_items')
make_attribute_wrapper(StringArraySplitViewType, 'index_offsets',
    '_index_offsets')
make_attribute_wrapper(StringArraySplitViewType, 'data_offsets',
    '_data_offsets')
make_attribute_wrapper(StringArraySplitViewType, 'data', '_data')
make_attribute_wrapper(StringArraySplitViewType, 'null_bitmap', '_null_bitmap')


def construct_str_arr_split_view(context, builder):
    fvlvt__uwlk = context.get_value_type(str_arr_split_view_payload_type)
    mwuv__nsvo = context.get_abi_sizeof(fvlvt__uwlk)
    kzeu__dzcl = context.get_value_type(types.voidptr)
    aoh__ugsu = context.get_value_type(types.uintp)
    hzenh__xqe = lir.FunctionType(lir.VoidType(), [kzeu__dzcl, aoh__ugsu,
        kzeu__dzcl])
    nhid__ddfga = cgutils.get_or_insert_function(builder.module, hzenh__xqe,
        name='dtor_str_arr_split_view')
    isqjr__qcx = context.nrt.meminfo_alloc_dtor(builder, context.
        get_constant(types.uintp, mwuv__nsvo), nhid__ddfga)
    dkx__wqsc = context.nrt.meminfo_data(builder, isqjr__qcx)
    qkx__gxfgm = builder.bitcast(dkx__wqsc, fvlvt__uwlk.as_pointer())
    return isqjr__qcx, qkx__gxfgm


@intrinsic
def compute_split_view(typingctx, str_arr_typ, sep_typ=None):
    assert str_arr_typ == string_array_type and isinstance(sep_typ, types.
        StringLiteral)

    def codegen(context, builder, sig, args):
        xyul__hpu, lugay__cyr = args
        isqjr__qcx, qkx__gxfgm = construct_str_arr_split_view(context, builder)
        yac__mouq = _get_str_binary_arr_payload(context, builder, xyul__hpu,
            string_array_type)
        bmu__nbe = lir.FunctionType(lir.VoidType(), [qkx__gxfgm.type, lir.
            IntType(64), lir.IntType(offset_type.bitwidth).as_pointer(),
            lir.IntType(8).as_pointer(), lir.IntType(8).as_pointer(), lir.
            IntType(8)])
        vwf__syd = cgutils.get_or_insert_function(builder.module, bmu__nbe,
            name='str_arr_split_view_impl')
        dpoio__zmub = context.make_helper(builder, offset_arr_type,
            yac__mouq.offsets).data
        moc__fzle = context.make_helper(builder, char_arr_type, yac__mouq.data
            ).data
        rzaj__bqc = context.make_helper(builder, null_bitmap_arr_type,
            yac__mouq.null_bitmap).data
        djc__nib = context.get_constant(types.int8, ord(sep_typ.literal_value))
        builder.call(vwf__syd, [qkx__gxfgm, yac__mouq.n_arrays, dpoio__zmub,
            moc__fzle, rzaj__bqc, djc__nib])
        sases__khoby = cgutils.create_struct_proxy(
            str_arr_split_view_payload_type)(context, builder, value=
            builder.load(qkx__gxfgm))
        yjqg__bgsu = context.make_helper(builder, string_array_split_view_type)
        yjqg__bgsu.num_items = yac__mouq.n_arrays
        yjqg__bgsu.index_offsets = sases__khoby.index_offsets
        yjqg__bgsu.data_offsets = sases__khoby.data_offsets
        yjqg__bgsu.data = context.compile_internal(builder, lambda S:
            get_data_ptr(S), data_ctypes_type(string_array_type), [xyul__hpu])
        yjqg__bgsu.null_bitmap = sases__khoby.null_bitmap
        yjqg__bgsu.meminfo = isqjr__qcx
        jtmw__zyof = yjqg__bgsu._getvalue()
        return impl_ret_new_ref(context, builder,
            string_array_split_view_type, jtmw__zyof)
    return string_array_split_view_type(string_array_type, sep_typ), codegen


@box(StringArraySplitViewType)
def box_str_arr_split_view(typ, val, c):
    context = c.context
    builder = c.builder
    fqlq__ekl = context.make_helper(builder, string_array_split_view_type, val)
    qcci__yyp = context.insert_const_string(builder.module, 'numpy')
    hihzl__xem = c.pyapi.import_module_noblock(qcci__yyp)
    dtype = c.pyapi.object_getattr_string(hihzl__xem, 'object_')
    vpw__dstt = builder.sext(fqlq__ekl.num_items, c.pyapi.longlong)
    bopm__hgn = c.pyapi.long_from_longlong(vpw__dstt)
    zsvvu__ess = c.pyapi.call_method(hihzl__xem, 'ndarray', (bopm__hgn, dtype))
    cyvzp__igkp = LLType.function(lir.IntType(8).as_pointer(), [c.pyapi.
        pyobj, c.pyapi.py_ssize_t])
    vtht__wme = c.pyapi._get_function(cyvzp__igkp, name='array_getptr1')
    tdvnc__phz = LLType.function(lir.VoidType(), [c.pyapi.pyobj, lir.
        IntType(8).as_pointer(), c.pyapi.pyobj])
    anxny__smjtx = c.pyapi._get_function(tdvnc__phz, name='array_setitem')
    ypj__dbwm = c.pyapi.object_getattr_string(hihzl__xem, 'nan')
    with cgutils.for_range(builder, fqlq__ekl.num_items) as loop:
        str_ind = loop.index
        ytb__ohpap = builder.sext(builder.load(builder.gep(fqlq__ekl.
            index_offsets, [str_ind])), lir.IntType(64))
        mie__vqy = builder.sext(builder.load(builder.gep(fqlq__ekl.
            index_offsets, [builder.add(str_ind, str_ind.type(1))])), lir.
            IntType(64))
        azhj__vua = builder.lshr(str_ind, lir.Constant(lir.IntType(64), 3))
        bcidp__yiyk = builder.gep(fqlq__ekl.null_bitmap, [azhj__vua])
        ukamt__equbw = builder.load(bcidp__yiyk)
        esdum__eneq = builder.trunc(builder.and_(str_ind, lir.Constant(lir.
            IntType(64), 7)), lir.IntType(8))
        val = builder.and_(builder.lshr(ukamt__equbw, esdum__eneq), lir.
            Constant(lir.IntType(8), 1))
        ptqq__caxq = builder.sub(mie__vqy, ytb__ohpap)
        ptqq__caxq = builder.sub(ptqq__caxq, ptqq__caxq.type(1))
        hie__thfnj = builder.call(vtht__wme, [zsvvu__ess, str_ind])
        qpm__ykw = c.builder.icmp_unsigned('!=', val, val.type(0))
        with c.builder.if_else(qpm__ykw) as (then, otherwise):
            with then:
                jav__hwuil = c.pyapi.list_new(ptqq__caxq)
                with c.builder.if_then(cgutils.is_not_null(c.builder,
                    jav__hwuil), likely=True):
                    with cgutils.for_range(c.builder, ptqq__caxq) as loop:
                        fokz__mak = builder.add(ytb__ohpap, loop.index)
                        data_start = builder.load(builder.gep(fqlq__ekl.
                            data_offsets, [fokz__mak]))
                        data_start = builder.add(data_start, data_start.type(1)
                            )
                        ogj__omg = builder.load(builder.gep(fqlq__ekl.
                            data_offsets, [builder.add(fokz__mak, fokz__mak
                            .type(1))]))
                        yzyjf__vxc = builder.gep(builder.extract_value(
                            fqlq__ekl.data, 0), [data_start])
                        pkbo__cuxg = builder.sext(builder.sub(ogj__omg,
                            data_start), lir.IntType(64))
                        qladk__lfgef = c.pyapi.string_from_string_and_size(
                            yzyjf__vxc, pkbo__cuxg)
                        c.pyapi.list_setitem(jav__hwuil, loop.index,
                            qladk__lfgef)
                builder.call(anxny__smjtx, [zsvvu__ess, hie__thfnj, jav__hwuil]
                    )
            with otherwise:
                builder.call(anxny__smjtx, [zsvvu__ess, hie__thfnj, ypj__dbwm])
    c.pyapi.decref(hihzl__xem)
    c.pyapi.decref(dtype)
    c.pyapi.decref(ypj__dbwm)
    return zsvvu__ess


@intrinsic
def pre_alloc_str_arr_view(typingctx, num_items_t, num_offsets_t, data_t=None):
    assert num_items_t == types.intp and num_offsets_t == types.intp

    def codegen(context, builder, sig, args):
        ajm__wmnl, hgfo__zezeo, yzyjf__vxc = args
        isqjr__qcx, qkx__gxfgm = construct_str_arr_split_view(context, builder)
        bmu__nbe = lir.FunctionType(lir.VoidType(), [qkx__gxfgm.type, lir.
            IntType(64), lir.IntType(64)])
        vwf__syd = cgutils.get_or_insert_function(builder.module, bmu__nbe,
            name='str_arr_split_view_alloc')
        builder.call(vwf__syd, [qkx__gxfgm, ajm__wmnl, hgfo__zezeo])
        sases__khoby = cgutils.create_struct_proxy(
            str_arr_split_view_payload_type)(context, builder, value=
            builder.load(qkx__gxfgm))
        yjqg__bgsu = context.make_helper(builder, string_array_split_view_type)
        yjqg__bgsu.num_items = ajm__wmnl
        yjqg__bgsu.index_offsets = sases__khoby.index_offsets
        yjqg__bgsu.data_offsets = sases__khoby.data_offsets
        yjqg__bgsu.data = yzyjf__vxc
        yjqg__bgsu.null_bitmap = sases__khoby.null_bitmap
        context.nrt.incref(builder, data_t, yzyjf__vxc)
        yjqg__bgsu.meminfo = isqjr__qcx
        jtmw__zyof = yjqg__bgsu._getvalue()
        return impl_ret_new_ref(context, builder,
            string_array_split_view_type, jtmw__zyof)
    return string_array_split_view_type(types.intp, types.intp, data_t
        ), codegen


@intrinsic
def get_c_arr_ptr(typingctx, c_arr, ind_t=None):
    assert isinstance(c_arr, (types.CPointer, types.ArrayCTypes))

    def codegen(context, builder, sig, args):
        vsz__qwys, ind = args
        if isinstance(sig.args[0], types.ArrayCTypes):
            vsz__qwys = builder.extract_value(vsz__qwys, 0)
        return builder.bitcast(builder.gep(vsz__qwys, [ind]), lir.IntType(8
            ).as_pointer())
    return types.voidptr(c_arr, ind_t), codegen


@intrinsic
def getitem_c_arr(typingctx, c_arr, ind_t=None):

    def codegen(context, builder, sig, args):
        vsz__qwys, ind = args
        return builder.load(builder.gep(vsz__qwys, [ind]))
    return c_arr.dtype(c_arr, ind_t), codegen


@intrinsic
def setitem_c_arr(typingctx, c_arr, ind_t, item_t=None):

    def codegen(context, builder, sig, args):
        vsz__qwys, ind, uhk__gan = args
        axvp__lmmfr = builder.gep(vsz__qwys, [ind])
        builder.store(uhk__gan, axvp__lmmfr)
    return types.void(c_arr, ind_t, c_arr.dtype), codegen


@intrinsic
def get_array_ctypes_ptr(typingctx, arr_ctypes_t, ind_t=None):

    def codegen(context, builder, sig, args):
        cie__fxk, ind = args
        xlva__mqx = context.make_helper(builder, arr_ctypes_t, cie__fxk)
        ttjd__oacw = context.make_helper(builder, arr_ctypes_t)
        ttjd__oacw.data = builder.gep(xlva__mqx.data, [ind])
        ttjd__oacw.meminfo = xlva__mqx.meminfo
        kafg__ryk = ttjd__oacw._getvalue()
        return impl_ret_borrowed(context, builder, arr_ctypes_t, kafg__ryk)
    return arr_ctypes_t(arr_ctypes_t, ind_t), codegen


@numba.njit(no_cpython_wrapper=True)
def get_split_view_index(arr, item_ind, str_ind):
    psk__ctvqu = bodo.libs.int_arr_ext.get_bit_bitmap_arr(arr._null_bitmap,
        item_ind)
    if not psk__ctvqu:
        return 0, 0, 0
    fokz__mak = getitem_c_arr(arr._index_offsets, item_ind)
    cfe__oba = getitem_c_arr(arr._index_offsets, item_ind + 1) - 1
    ruh__ndukw = cfe__oba - fokz__mak
    if str_ind >= ruh__ndukw:
        return 0, 0, 0
    data_start = getitem_c_arr(arr._data_offsets, fokz__mak + str_ind)
    data_start += 1
    if fokz__mak + str_ind == 0:
        data_start = 0
    ogj__omg = getitem_c_arr(arr._data_offsets, fokz__mak + str_ind + 1)
    scq__hjjt = ogj__omg - data_start
    return 1, data_start, scq__hjjt


@numba.njit(no_cpython_wrapper=True)
def get_split_view_data_ptr(arr, data_start):
    return get_array_ctypes_ptr(arr._data, data_start)


@overload(len, no_unliteral=True)
def str_arr_split_view_len_overload(arr):
    if arr == string_array_split_view_type:
        return lambda arr: np.int64(arr._num_items)


@overload_attribute(StringArraySplitViewType, 'shape')
def overload_split_view_arr_shape(A):
    return lambda A: (np.int64(A._num_items),)


@overload(operator.getitem, no_unliteral=True)
def str_arr_split_view_getitem_overload(A, ind):
    if A != string_array_split_view_type:
        return
    if A == string_array_split_view_type and isinstance(ind, types.Integer):
        sher__fbofi = numba.cpython.unicode.PY_UNICODE_1BYTE_KIND

        def _impl(A, ind):
            fokz__mak = getitem_c_arr(A._index_offsets, ind)
            cfe__oba = getitem_c_arr(A._index_offsets, ind + 1)
            bjkpn__wtg = cfe__oba - fokz__mak - 1
            xyul__hpu = bodo.libs.str_arr_ext.pre_alloc_string_array(bjkpn__wtg
                , -1)
            for doyh__isp in range(bjkpn__wtg):
                data_start = getitem_c_arr(A._data_offsets, fokz__mak +
                    doyh__isp)
                data_start += 1
                if fokz__mak + doyh__isp == 0:
                    data_start = 0
                ogj__omg = getitem_c_arr(A._data_offsets, fokz__mak +
                    doyh__isp + 1)
                scq__hjjt = ogj__omg - data_start
                axvp__lmmfr = get_array_ctypes_ptr(A._data, data_start)
                pkzk__htyav = bodo.libs.str_arr_ext.decode_utf8(axvp__lmmfr,
                    scq__hjjt)
                xyul__hpu[doyh__isp] = pkzk__htyav
            return xyul__hpu
        return _impl
    if A == string_array_split_view_type and ind == types.Array(types.bool_,
        1, 'C'):
        ewd__epqbb = offset_type.bitwidth // 8

        def _impl(A, ind):
            bjkpn__wtg = len(A)
            if bjkpn__wtg != len(ind):
                raise IndexError(
                    'boolean index did not match indexed array along dimension 0'
                    )
            ajm__wmnl = 0
            hgfo__zezeo = 0
            for doyh__isp in range(bjkpn__wtg):
                if ind[doyh__isp]:
                    ajm__wmnl += 1
                    fokz__mak = getitem_c_arr(A._index_offsets, doyh__isp)
                    cfe__oba = getitem_c_arr(A._index_offsets, doyh__isp + 1)
                    hgfo__zezeo += cfe__oba - fokz__mak
            zsvvu__ess = pre_alloc_str_arr_view(ajm__wmnl, hgfo__zezeo, A._data
                )
            item_ind = 0
            gqer__fmhzx = 0
            for doyh__isp in range(bjkpn__wtg):
                if ind[doyh__isp]:
                    fokz__mak = getitem_c_arr(A._index_offsets, doyh__isp)
                    cfe__oba = getitem_c_arr(A._index_offsets, doyh__isp + 1)
                    cwa__dhe = cfe__oba - fokz__mak
                    setitem_c_arr(zsvvu__ess._index_offsets, item_ind,
                        gqer__fmhzx)
                    axvp__lmmfr = get_c_arr_ptr(A._data_offsets, fokz__mak)
                    vicq__cnyl = get_c_arr_ptr(zsvvu__ess._data_offsets,
                        gqer__fmhzx)
                    _memcpy(vicq__cnyl, axvp__lmmfr, cwa__dhe, ewd__epqbb)
                    psk__ctvqu = bodo.libs.int_arr_ext.get_bit_bitmap_arr(A
                        ._null_bitmap, doyh__isp)
                    bodo.libs.int_arr_ext.set_bit_to_arr(zsvvu__ess.
                        _null_bitmap, item_ind, psk__ctvqu)
                    item_ind += 1
                    gqer__fmhzx += cwa__dhe
            setitem_c_arr(zsvvu__ess._index_offsets, item_ind, gqer__fmhzx)
            return zsvvu__ess
        return _impl
