"""Table data type for storing dataframe column arrays. Supports storing many columns
(e.g. >10k) efficiently.
"""
import operator
from collections import defaultdict
import numba
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.core.imputils import impl_ret_borrowed, lower_constant
from numba.cpython.listobj import ListInstance
from numba.extending import NativeValue, box, infer_getattr, intrinsic, lower_builtin, lower_getattr, make_attribute_wrapper, models, overload, register_model, typeof_impl, unbox
from numba.np.arrayobj import _getitem_array_single_int
from numba.parfors.array_analysis import ArrayAnalysis
from bodo.utils.cg_helpers import is_ll_eq
from bodo.utils.templates import OverloadedKeyAttributeTemplate
from bodo.utils.typing import BodoError, get_overload_const_int, is_list_like_index_type, is_overload_constant_int


class Table:

    def __init__(self, arrs, usecols=None, num_arrs=-1):
        if usecols is not None:
            assert num_arrs != -1, 'num_arrs must be provided if usecols is not None'
            hkx__cxkg = 0
            kefv__snqt = []
            for i in range(usecols[-1] + 1):
                if i == usecols[hkx__cxkg]:
                    kefv__snqt.append(arrs[hkx__cxkg])
                    hkx__cxkg += 1
                else:
                    kefv__snqt.append(None)
            for pnmq__nosr in range(usecols[-1] + 1, num_arrs):
                kefv__snqt.append(None)
            self.arrays = kefv__snqt
        else:
            self.arrays = arrs
        self.block_0 = arrs

    def __eq__(self, other):
        return isinstance(other, Table) and len(self.arrays) == len(other.
            arrays) and all((ceng__eeby == cqeq__femws).all() for 
            ceng__eeby, cqeq__femws in zip(self.arrays, other.arrays))

    def __str__(self) ->str:
        return str(self.arrays)

    def to_pandas(self, index=None):
        ucjh__ctv = len(self.arrays)
        lnb__sorza = dict(zip(range(ucjh__ctv), self.arrays))
        df = pd.DataFrame(lnb__sorza, index)
        return df


class TableType(types.ArrayCompatible):

    def __init__(self, arr_types, has_runtime_cols=False):
        self.arr_types = arr_types
        self.has_runtime_cols = has_runtime_cols
        ffpzw__ukz = []
        rpxsw__qyhzx = []
        pop__rdgpn = {}
        ooo__tglvf = defaultdict(int)
        tgabm__ckjd = defaultdict(list)
        if not has_runtime_cols:
            for i, wtwv__xklmq in enumerate(arr_types):
                if wtwv__xklmq not in pop__rdgpn:
                    pop__rdgpn[wtwv__xklmq] = len(pop__rdgpn)
                jmnnj__bljf = pop__rdgpn[wtwv__xklmq]
                ffpzw__ukz.append(jmnnj__bljf)
                rpxsw__qyhzx.append(ooo__tglvf[jmnnj__bljf])
                ooo__tglvf[jmnnj__bljf] += 1
                tgabm__ckjd[jmnnj__bljf].append(i)
        self.block_nums = ffpzw__ukz
        self.block_offsets = rpxsw__qyhzx
        self.type_to_blk = pop__rdgpn
        self.block_to_arr_ind = tgabm__ckjd
        super(TableType, self).__init__(name=
            f'TableType({arr_types}, {has_runtime_cols})')

    @property
    def as_array(self):
        return types.Array(types.undefined, 2, 'C')

    @property
    def key(self):
        return self.arr_types, self.has_runtime_cols

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


@typeof_impl.register(Table)
def typeof_table(val, c):
    return TableType(tuple(numba.typeof(whp__bmf) for whp__bmf in val.arrays))


@register_model(TableType)
class TableTypeModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        if fe_type.has_runtime_cols:
            dxox__ndxwp = [(f'block_{i}', types.List(wtwv__xklmq)) for i,
                wtwv__xklmq in enumerate(fe_type.arr_types)]
        else:
            dxox__ndxwp = [(f'block_{jmnnj__bljf}', types.List(wtwv__xklmq)
                ) for wtwv__xklmq, jmnnj__bljf in fe_type.type_to_blk.items()]
        dxox__ndxwp.append(('parent', types.pyobject))
        dxox__ndxwp.append(('len', types.int64))
        super(TableTypeModel, self).__init__(dmm, fe_type, dxox__ndxwp)


make_attribute_wrapper(TableType, 'block_0', 'block_0')
make_attribute_wrapper(TableType, 'len', '_len')


@infer_getattr
class TableTypeAttribute(OverloadedKeyAttributeTemplate):
    key = TableType

    def resolve_shape(self, df):
        return types.Tuple([types.int64, types.int64])


@unbox(TableType)
def unbox_table(typ, val, c):
    dqfwv__plf = c.pyapi.object_getattr_string(val, 'arrays')
    qbfpr__aqeos = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    qbfpr__aqeos.parent = cgutils.get_null_value(qbfpr__aqeos.parent.type)
    cosdu__qgb = c.pyapi.make_none()
    ibi__jxr = c.context.get_constant(types.int64, 0)
    eqej__tgij = cgutils.alloca_once_value(c.builder, ibi__jxr)
    for wtwv__xklmq, jmnnj__bljf in typ.type_to_blk.items():
        pid__dlaz = c.context.get_constant(types.int64, len(typ.
            block_to_arr_ind[jmnnj__bljf]))
        pnmq__nosr, obvvm__xfqj = ListInstance.allocate_ex(c.context, c.
            builder, types.List(wtwv__xklmq), pid__dlaz)
        obvvm__xfqj.size = pid__dlaz
        qfaal__nnzzy = c.context.make_constant_array(c.builder, types.Array
            (types.int64, 1, 'C'), np.array(typ.block_to_arr_ind[
            jmnnj__bljf], dtype=np.int64))
        zgic__xhvf = c.context.make_array(types.Array(types.int64, 1, 'C'))(c
            .context, c.builder, qfaal__nnzzy)
        with cgutils.for_range(c.builder, pid__dlaz) as loop:
            i = loop.index
            zlag__ncoca = _getitem_array_single_int(c.context, c.builder,
                types.int64, types.Array(types.int64, 1, 'C'), zgic__xhvf, i)
            zfnsi__hziq = c.pyapi.long_from_longlong(zlag__ncoca)
            akf__lhqlc = c.pyapi.object_getitem(dqfwv__plf, zfnsi__hziq)
            fkjc__sgfv = c.builder.icmp_unsigned('==', akf__lhqlc, cosdu__qgb)
            with c.builder.if_else(fkjc__sgfv) as (then, orelse):
                with then:
                    ekc__cruj = c.context.get_constant_null(wtwv__xklmq)
                    obvvm__xfqj.inititem(i, ekc__cruj, incref=False)
                with orelse:
                    hpaj__xoa = c.pyapi.call_method(akf__lhqlc, '__len__', ())
                    nujgp__mwfu = c.pyapi.long_as_longlong(hpaj__xoa)
                    c.builder.store(nujgp__mwfu, eqej__tgij)
                    c.pyapi.decref(hpaj__xoa)
                    whp__bmf = c.pyapi.to_native_value(wtwv__xklmq, akf__lhqlc
                        ).value
                    obvvm__xfqj.inititem(i, whp__bmf, incref=False)
            c.pyapi.decref(akf__lhqlc)
            c.pyapi.decref(zfnsi__hziq)
        setattr(qbfpr__aqeos, f'block_{jmnnj__bljf}', obvvm__xfqj.value)
    qbfpr__aqeos.len = c.builder.load(eqej__tgij)
    c.pyapi.decref(dqfwv__plf)
    c.pyapi.decref(cosdu__qgb)
    rrpp__joyks = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(qbfpr__aqeos._getvalue(), is_error=rrpp__joyks)


@box(TableType)
def box_table(typ, val, c, ensure_unboxed=None):
    from bodo.hiframes.boxing import get_df_obj_column_codegen
    qbfpr__aqeos = cgutils.create_struct_proxy(typ)(c.context, c.builder, val)
    if typ.has_runtime_cols:
        mohi__cyuqs = c.context.get_constant(types.int64, 0)
        for i, wtwv__xklmq in enumerate(typ.arr_types):
            kefv__snqt = getattr(qbfpr__aqeos, f'block_{i}')
            hzetu__ckc = ListInstance(c.context, c.builder, types.List(
                wtwv__xklmq), kefv__snqt)
            mohi__cyuqs = c.builder.add(mohi__cyuqs, hzetu__ckc.size)
        ezqx__algum = c.pyapi.list_new(mohi__cyuqs)
        xihaw__smva = c.context.get_constant(types.int64, 0)
        for i, wtwv__xklmq in enumerate(typ.arr_types):
            kefv__snqt = getattr(qbfpr__aqeos, f'block_{i}')
            hzetu__ckc = ListInstance(c.context, c.builder, types.List(
                wtwv__xklmq), kefv__snqt)
            with cgutils.for_range(c.builder, hzetu__ckc.size) as loop:
                i = loop.index
                whp__bmf = hzetu__ckc.getitem(i)
                c.context.nrt.incref(c.builder, wtwv__xklmq, whp__bmf)
                idx = c.builder.add(xihaw__smva, i)
                c.pyapi.list_setitem(ezqx__algum, idx, c.pyapi.
                    from_native_value(wtwv__xklmq, whp__bmf, c.env_manager))
            xihaw__smva = c.builder.add(xihaw__smva, hzetu__ckc.size)
        hfdq__ych = c.pyapi.unserialize(c.pyapi.serialize_object(Table))
        jgbjw__zis = c.pyapi.call_function_objargs(hfdq__ych, (ezqx__algum,))
        c.pyapi.decref(hfdq__ych)
        c.pyapi.decref(ezqx__algum)
        c.context.nrt.decref(c.builder, typ, val)
        return jgbjw__zis
    ezqx__algum = c.pyapi.list_new(c.context.get_constant(types.int64, len(
        typ.arr_types)))
    cvd__owdc = cgutils.is_not_null(c.builder, qbfpr__aqeos.parent)
    if ensure_unboxed is None:
        ensure_unboxed = c.context.get_constant(types.bool_, False)
    for wtwv__xklmq, jmnnj__bljf in typ.type_to_blk.items():
        kefv__snqt = getattr(qbfpr__aqeos, f'block_{jmnnj__bljf}')
        hzetu__ckc = ListInstance(c.context, c.builder, types.List(
            wtwv__xklmq), kefv__snqt)
        qfaal__nnzzy = c.context.make_constant_array(c.builder, types.Array
            (types.int64, 1, 'C'), np.array(typ.block_to_arr_ind[
            jmnnj__bljf], dtype=np.int64))
        zgic__xhvf = c.context.make_array(types.Array(types.int64, 1, 'C'))(c
            .context, c.builder, qfaal__nnzzy)
        with cgutils.for_range(c.builder, hzetu__ckc.size) as loop:
            i = loop.index
            zlag__ncoca = _getitem_array_single_int(c.context, c.builder,
                types.int64, types.Array(types.int64, 1, 'C'), zgic__xhvf, i)
            whp__bmf = hzetu__ckc.getitem(i)
            cca__sizfe = cgutils.alloca_once_value(c.builder, whp__bmf)
            vht__lykbz = cgutils.alloca_once_value(c.builder, c.context.
                get_constant_null(wtwv__xklmq))
            ytmx__yekoh = is_ll_eq(c.builder, cca__sizfe, vht__lykbz)
            with c.builder.if_else(c.builder.and_(ytmx__yekoh, c.builder.
                not_(ensure_unboxed))) as (then, orelse):
                with then:
                    cosdu__qgb = c.pyapi.make_none()
                    c.pyapi.list_setitem(ezqx__algum, zlag__ncoca, cosdu__qgb)
                with orelse:
                    akf__lhqlc = cgutils.alloca_once(c.builder, c.context.
                        get_value_type(types.pyobject))
                    with c.builder.if_else(c.builder.and_(ytmx__yekoh,
                        cvd__owdc)) as (arr_then, arr_orelse):
                        with arr_then:
                            wnlq__epjui = get_df_obj_column_codegen(c.
                                context, c.builder, c.pyapi, qbfpr__aqeos.
                                parent, zlag__ncoca, wtwv__xklmq)
                            c.builder.store(wnlq__epjui, akf__lhqlc)
                        with arr_orelse:
                            c.context.nrt.incref(c.builder, wtwv__xklmq,
                                whp__bmf)
                            c.builder.store(c.pyapi.from_native_value(
                                wtwv__xklmq, whp__bmf, c.env_manager),
                                akf__lhqlc)
                    c.pyapi.list_setitem(ezqx__algum, zlag__ncoca, c.
                        builder.load(akf__lhqlc))
    hfdq__ych = c.pyapi.unserialize(c.pyapi.serialize_object(Table))
    jgbjw__zis = c.pyapi.call_function_objargs(hfdq__ych, (ezqx__algum,))
    c.pyapi.decref(hfdq__ych)
    c.pyapi.decref(ezqx__algum)
    c.context.nrt.decref(c.builder, typ, val)
    return jgbjw__zis


@lower_builtin(len, TableType)
def table_len_lower(context, builder, sig, args):
    impl = table_len_overload(*sig.args)
    return context.compile_internal(builder, impl, sig, args)


def table_len_overload(T):
    if not isinstance(T, TableType):
        return

    def impl(T):
        return T._len
    return impl


@lower_getattr(TableType, 'shape')
def lower_table_shape(context, builder, typ, val):
    impl = table_shape_overload(typ)
    return context.compile_internal(builder, impl, types.Tuple([types.int64,
        types.int64])(typ), (val,))


def table_shape_overload(T):
    if T.has_runtime_cols:

        def impl(T):
            return T._len, compute_num_runtime_columns(T)
        return impl
    ncols = len(T.arr_types)
    return lambda T: (T._len, types.int64(ncols))


@intrinsic
def compute_num_runtime_columns(typingctx, table_type):
    assert isinstance(table_type, TableType)

    def codegen(context, builder, sig, args):
        table_arg, = args
        qbfpr__aqeos = cgutils.create_struct_proxy(table_type)(context,
            builder, table_arg)
        nwd__emso = context.get_constant(types.int64, 0)
        for i, wtwv__xklmq in enumerate(table_type.arr_types):
            kefv__snqt = getattr(qbfpr__aqeos, f'block_{i}')
            hzetu__ckc = ListInstance(context, builder, types.List(
                wtwv__xklmq), kefv__snqt)
            nwd__emso = builder.add(nwd__emso, hzetu__ckc.size)
        return nwd__emso
    sig = types.int64(table_type)
    return sig, codegen


def get_table_data_codegen(context, builder, table_arg, col_ind, table_type):
    arr_type = table_type.arr_types[col_ind]
    qbfpr__aqeos = cgutils.create_struct_proxy(table_type)(context, builder,
        table_arg)
    jmnnj__bljf = table_type.block_nums[col_ind]
    yypzg__gzmun = table_type.block_offsets[col_ind]
    kefv__snqt = getattr(qbfpr__aqeos, f'block_{jmnnj__bljf}')
    hzetu__ckc = ListInstance(context, builder, types.List(arr_type),
        kefv__snqt)
    whp__bmf = hzetu__ckc.getitem(yypzg__gzmun)
    return whp__bmf


@intrinsic
def get_table_data(typingctx, table_type, ind_typ=None):
    assert isinstance(table_type, TableType)
    assert is_overload_constant_int(ind_typ)
    col_ind = get_overload_const_int(ind_typ)
    arr_type = table_type.arr_types[col_ind]

    def codegen(context, builder, sig, args):
        table_arg, pnmq__nosr = args
        whp__bmf = get_table_data_codegen(context, builder, table_arg,
            col_ind, table_type)
        return impl_ret_borrowed(context, builder, arr_type, whp__bmf)
    sig = arr_type(table_type, ind_typ)
    return sig, codegen


@intrinsic
def del_column(typingctx, table_type, ind_typ=None):
    assert isinstance(table_type, TableType)
    assert is_overload_constant_int(ind_typ)
    col_ind = get_overload_const_int(ind_typ)
    arr_type = table_type.arr_types[col_ind]

    def codegen(context, builder, sig, args):
        table_arg, pnmq__nosr = args
        qbfpr__aqeos = cgutils.create_struct_proxy(table_type)(context,
            builder, table_arg)
        jmnnj__bljf = table_type.block_nums[col_ind]
        yypzg__gzmun = table_type.block_offsets[col_ind]
        kefv__snqt = getattr(qbfpr__aqeos, f'block_{jmnnj__bljf}')
        hzetu__ckc = ListInstance(context, builder, types.List(arr_type),
            kefv__snqt)
        whp__bmf = hzetu__ckc.getitem(yypzg__gzmun)
        context.nrt.decref(builder, arr_type, whp__bmf)
        ekc__cruj = context.get_constant_null(arr_type)
        hzetu__ckc.inititem(yypzg__gzmun, ekc__cruj, incref=False)
    sig = types.void(table_type, ind_typ)
    return sig, codegen


def set_table_data_codegen(context, builder, in_table_type, in_table,
    out_table_type, arr_type, arr_arg, col_ind, is_new_col):
    in_table = cgutils.create_struct_proxy(in_table_type)(context, builder,
        in_table)
    out_table = cgutils.create_struct_proxy(out_table_type)(context, builder)
    out_table.len = in_table.len
    out_table.parent = in_table.parent
    ibi__jxr = context.get_constant(types.int64, 0)
    nsfhs__dsg = context.get_constant(types.int64, 1)
    kxvsz__urhld = arr_type not in in_table_type.type_to_blk
    for wtwv__xklmq, jmnnj__bljf in out_table_type.type_to_blk.items():
        if wtwv__xklmq in in_table_type.type_to_blk:
            havel__orczd = in_table_type.type_to_blk[wtwv__xklmq]
            obvvm__xfqj = ListInstance(context, builder, types.List(
                wtwv__xklmq), getattr(in_table, f'block_{havel__orczd}'))
            context.nrt.incref(builder, types.List(wtwv__xklmq),
                obvvm__xfqj.value)
            setattr(out_table, f'block_{jmnnj__bljf}', obvvm__xfqj.value)
    if kxvsz__urhld:
        pnmq__nosr, obvvm__xfqj = ListInstance.allocate_ex(context, builder,
            types.List(arr_type), nsfhs__dsg)
        obvvm__xfqj.size = nsfhs__dsg
        obvvm__xfqj.inititem(ibi__jxr, arr_arg, incref=True)
        jmnnj__bljf = out_table_type.type_to_blk[arr_type]
        setattr(out_table, f'block_{jmnnj__bljf}', obvvm__xfqj.value)
        if not is_new_col:
            _rm_old_array(col_ind, out_table_type, out_table, in_table_type,
                context, builder)
    else:
        jmnnj__bljf = out_table_type.type_to_blk[arr_type]
        obvvm__xfqj = ListInstance(context, builder, types.List(arr_type),
            getattr(out_table, f'block_{jmnnj__bljf}'))
        if is_new_col:
            n = obvvm__xfqj.size
            yve__yxj = builder.add(n, nsfhs__dsg)
            obvvm__xfqj.resize(yve__yxj)
            obvvm__xfqj.inititem(n, arr_arg, incref=True)
        elif arr_type == in_table_type.arr_types[col_ind]:
            kwtjd__vyqkz = context.get_constant(types.int64, out_table_type
                .block_offsets[col_ind])
            obvvm__xfqj.setitem(kwtjd__vyqkz, arr_arg, True)
        else:
            _rm_old_array(col_ind, out_table_type, out_table, in_table_type,
                context, builder)
            kwtjd__vyqkz = context.get_constant(types.int64, out_table_type
                .block_offsets[col_ind])
            n = obvvm__xfqj.size
            yve__yxj = builder.add(n, nsfhs__dsg)
            obvvm__xfqj.resize(yve__yxj)
            context.nrt.incref(builder, arr_type, obvvm__xfqj.getitem(
                kwtjd__vyqkz))
            obvvm__xfqj.move(builder.add(kwtjd__vyqkz, nsfhs__dsg),
                kwtjd__vyqkz, builder.sub(n, kwtjd__vyqkz))
            obvvm__xfqj.setitem(kwtjd__vyqkz, arr_arg, incref=True)
    return out_table._getvalue()


def _rm_old_array(col_ind, out_table_type, out_table, in_table_type,
    context, builder):
    xdua__yod = in_table_type.arr_types[col_ind]
    if xdua__yod in out_table_type.type_to_blk:
        jmnnj__bljf = out_table_type.type_to_blk[xdua__yod]
        sdx__eys = getattr(out_table, f'block_{jmnnj__bljf}')
        ttq__wfe = types.List(xdua__yod)
        kwtjd__vyqkz = context.get_constant(types.int64, in_table_type.
            block_offsets[col_ind])
        wriu__ygo = ttq__wfe.dtype(ttq__wfe, types.intp)
        otn__kihdt = context.compile_internal(builder, lambda lst, i: lst.
            pop(i), wriu__ygo, (sdx__eys, kwtjd__vyqkz))
        context.nrt.decref(builder, xdua__yod, otn__kihdt)


@intrinsic
def set_table_data(typingctx, table_type, ind_type, arr_type=None):
    assert isinstance(table_type, TableType), 'invalid input to set_table_data'
    assert is_overload_constant_int(ind_type
        ), 'set_table_data expects const index'
    col_ind = get_overload_const_int(ind_type)
    is_new_col = col_ind == len(table_type.arr_types)
    hib__emi = list(table_type.arr_types)
    if is_new_col:
        hib__emi.append(arr_type)
    else:
        hib__emi[col_ind] = arr_type
    out_table_type = TableType(tuple(hib__emi))

    def codegen(context, builder, sig, args):
        table_arg, pnmq__nosr, gxoeo__nzx = args
        out_table = set_table_data_codegen(context, builder, table_type,
            table_arg, out_table_type, arr_type, gxoeo__nzx, col_ind,
            is_new_col)
        return out_table
    return out_table_type(table_type, ind_type, arr_type), codegen


def alias_ext_dummy_func(lhs_name, args, alias_map, arg_aliases):
    assert len(args) >= 1
    numba.core.ir_utils._add_alias(lhs_name, args[0].name, alias_map,
        arg_aliases)


numba.core.ir_utils.alias_func_extensions['get_table_data',
    'bodo.hiframes.table'] = alias_ext_dummy_func


def get_table_data_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 2 and not kws
    pukzx__gal = args[0]
    if equiv_set.has_shape(pukzx__gal):
        return ArrayAnalysis.AnalyzeResult(shape=equiv_set.get_shape(
            pukzx__gal)[0], pre=[])


ArrayAnalysis._analyze_op_call_bodo_hiframes_table_get_table_data = (
    get_table_data_equiv)


@lower_constant(TableType)
def lower_constant_table(context, builder, table_type, pyval):
    faqv__uzch = []
    for wtwv__xklmq, jmnnj__bljf in table_type.type_to_blk.items():
        epvjh__vhdh = len(table_type.block_to_arr_ind[jmnnj__bljf])
        fsvf__zfuqt = []
        for i in range(epvjh__vhdh):
            zlag__ncoca = table_type.block_to_arr_ind[jmnnj__bljf][i]
            fsvf__zfuqt.append(pyval.arrays[zlag__ncoca])
        faqv__uzch.append(context.get_constant_generic(builder, types.List(
            wtwv__xklmq), fsvf__zfuqt))
    cloz__rzdx = context.get_constant_null(types.pyobject)
    caeq__zel = context.get_constant(types.int64, 0 if len(pyval.arrays) ==
        0 else len(pyval.arrays[0]))
    return lir.Constant.literal_struct(faqv__uzch + [cloz__rzdx, caeq__zel])


@intrinsic
def init_table(typingctx, table_type=None):
    assert isinstance(table_type, TableType), 'table type expected'

    def codegen(context, builder, sig, args):
        qbfpr__aqeos = cgutils.create_struct_proxy(table_type)(context, builder
            )
        for wtwv__xklmq, jmnnj__bljf in table_type.type_to_blk.items():
            ldh__mui = context.get_constant_null(types.List(wtwv__xklmq))
            setattr(qbfpr__aqeos, f'block_{jmnnj__bljf}', ldh__mui)
        return qbfpr__aqeos._getvalue()
    sig = table_type(table_type)
    return sig, codegen


@intrinsic
def get_table_block(typingctx, table_type, blk_type=None):
    assert isinstance(table_type, TableType), 'table type expected'
    assert is_overload_constant_int(blk_type)
    jmnnj__bljf = get_overload_const_int(blk_type)
    arr_type = None
    for wtwv__xklmq, cqeq__femws in table_type.type_to_blk.items():
        if cqeq__femws == jmnnj__bljf:
            arr_type = wtwv__xklmq
            break
    assert arr_type is not None, 'invalid table type block'
    jyyu__hlu = types.List(arr_type)

    def codegen(context, builder, sig, args):
        qbfpr__aqeos = cgutils.create_struct_proxy(table_type)(context,
            builder, args[0])
        kefv__snqt = getattr(qbfpr__aqeos, f'block_{jmnnj__bljf}')
        return impl_ret_borrowed(context, builder, jyyu__hlu, kefv__snqt)
    sig = jyyu__hlu(table_type, blk_type)
    return sig, codegen


@intrinsic
def ensure_column_unboxed(typingctx, table_type, arr_list_t, ind_t,
    arr_ind_t=None):
    assert isinstance(table_type, TableType), 'table type expected'
    sig = types.none(table_type, arr_list_t, ind_t, arr_ind_t)
    return sig, ensure_column_unboxed_codegen


def ensure_column_unboxed_codegen(context, builder, sig, args):
    from bodo.hiframes.boxing import get_df_obj_column_codegen
    table_arg, qsgac__vahm, wjknp__krmi, qrpt__vasi = args
    cxmq__phh = context.get_python_api(builder)
    qbfpr__aqeos = cgutils.create_struct_proxy(sig.args[0])(context,
        builder, table_arg)
    cvd__owdc = cgutils.is_not_null(builder, qbfpr__aqeos.parent)
    hzetu__ckc = ListInstance(context, builder, sig.args[1], qsgac__vahm)
    qmyhs__oob = hzetu__ckc.getitem(wjknp__krmi)
    cca__sizfe = cgutils.alloca_once_value(builder, qmyhs__oob)
    vht__lykbz = cgutils.alloca_once_value(builder, context.
        get_constant_null(sig.args[1].dtype))
    ytmx__yekoh = is_ll_eq(builder, cca__sizfe, vht__lykbz)
    with builder.if_then(ytmx__yekoh):
        with builder.if_else(cvd__owdc) as (then, orelse):
            with then:
                akf__lhqlc = get_df_obj_column_codegen(context, builder,
                    cxmq__phh, qbfpr__aqeos.parent, qrpt__vasi, sig.args[1]
                    .dtype)
                whp__bmf = cxmq__phh.to_native_value(sig.args[1].dtype,
                    akf__lhqlc).value
                hzetu__ckc.inititem(wjknp__krmi, whp__bmf, incref=False)
                cxmq__phh.decref(akf__lhqlc)
            with orelse:
                context.call_conv.return_user_exc(builder, BodoError, (
                    'unexpected null table column',))


@intrinsic
def set_table_block(typingctx, table_type, arr_list_type, blk_type=None):
    assert isinstance(table_type, TableType), 'table type expected'
    assert isinstance(arr_list_type, types.List), 'list type expected'
    assert is_overload_constant_int(blk_type), 'blk should be const int'
    jmnnj__bljf = get_overload_const_int(blk_type)

    def codegen(context, builder, sig, args):
        table_arg, wkoq__fzm, pnmq__nosr = args
        in_table = cgutils.create_struct_proxy(table_type)(context, builder,
            table_arg)
        setattr(in_table, f'block_{jmnnj__bljf}', wkoq__fzm)
        return impl_ret_borrowed(context, builder, table_type, in_table.
            _getvalue())
    sig = table_type(table_type, arr_list_type, blk_type)
    return sig, codegen


@intrinsic
def set_table_len(typingctx, table_type, l_type=None):
    assert isinstance(table_type, TableType), 'table type expected'

    def codegen(context, builder, sig, args):
        table_arg, eyww__ouw = args
        in_table = cgutils.create_struct_proxy(table_type)(context, builder,
            table_arg)
        in_table.len = eyww__ouw
        return impl_ret_borrowed(context, builder, table_type, in_table.
            _getvalue())
    sig = table_type(table_type, l_type)
    return sig, codegen


@intrinsic
def alloc_list_like(typingctx, list_type=None):
    assert isinstance(list_type, types.List), 'list type expected'

    def codegen(context, builder, sig, args):
        meob__ndcys = ListInstance(context, builder, list_type, args[0])
        nhlzt__glvl = meob__ndcys.size
        pnmq__nosr, obvvm__xfqj = ListInstance.allocate_ex(context, builder,
            list_type, nhlzt__glvl)
        obvvm__xfqj.size = nhlzt__glvl
        return obvvm__xfqj.value
    sig = list_type(list_type)
    return sig, codegen


def _get_idx_length(idx):
    pass


@overload(_get_idx_length)
def overload_get_idx_length(idx, n):
    if is_list_like_index_type(idx) and idx.dtype == types.bool_:
        return lambda idx, n: idx.sum()
    assert isinstance(idx, types.SliceType), 'slice index expected'

    def impl(idx, n):
        dqnqf__kely = numba.cpython.unicode._normalize_slice(idx, n)
        return numba.cpython.unicode._slice_span(dqnqf__kely)
    return impl


def gen_table_filter(T, used_cols=None):
    from bodo.utils.conversion import ensure_contig_if_np
    msuul__tlv = {'init_table': init_table, 'get_table_block':
        get_table_block, 'ensure_column_unboxed': ensure_column_unboxed,
        'set_table_block': set_table_block, 'set_table_len': set_table_len,
        'alloc_list_like': alloc_list_like, '_get_idx_length':
        _get_idx_length, 'ensure_contig_if_np': ensure_contig_if_np}
    if used_cols is not None:
        msuul__tlv['used_cols'] = np.array(used_cols, dtype=np.int64)
    gadc__usy = 'def impl(T, idx):\n'
    gadc__usy += f'  T2 = init_table(T)\n'
    gadc__usy += f'  l = 0\n'
    if used_cols is not None and len(used_cols) == 0:
        gadc__usy += f'  l = _get_idx_length(idx, len(T))\n'
        gadc__usy += f'  T2 = set_table_len(T2, l)\n'
        gadc__usy += f'  return T2\n'
        swx__rjdv = {}
        exec(gadc__usy, msuul__tlv, swx__rjdv)
        return swx__rjdv['impl']
    if used_cols is not None:
        gadc__usy += f'  used_set = set(used_cols)\n'
    for jmnnj__bljf in T.type_to_blk.values():
        msuul__tlv[f'arr_inds_{jmnnj__bljf}'] = np.array(T.block_to_arr_ind
            [jmnnj__bljf], dtype=np.int64)
        gadc__usy += (
            f'  arr_list_{jmnnj__bljf} = get_table_block(T, {jmnnj__bljf})\n')
        gadc__usy += (
            f'  out_arr_list_{jmnnj__bljf} = alloc_list_like(arr_list_{jmnnj__bljf})\n'
            )
        gadc__usy += f'  for i in range(len(arr_list_{jmnnj__bljf})):\n'
        gadc__usy += f'    arr_ind_{jmnnj__bljf} = arr_inds_{jmnnj__bljf}[i]\n'
        if used_cols is not None:
            gadc__usy += (
                f'    if arr_ind_{jmnnj__bljf} not in used_set: continue\n')
        gadc__usy += f"""    ensure_column_unboxed(T, arr_list_{jmnnj__bljf}, i, arr_ind_{jmnnj__bljf})
"""
        gadc__usy += f"""    out_arr_{jmnnj__bljf} = ensure_contig_if_np(arr_list_{jmnnj__bljf}[i][idx])
"""
        gadc__usy += f'    l = len(out_arr_{jmnnj__bljf})\n'
        gadc__usy += (
            f'    out_arr_list_{jmnnj__bljf}[i] = out_arr_{jmnnj__bljf}\n')
        gadc__usy += (
            f'  T2 = set_table_block(T2, out_arr_list_{jmnnj__bljf}, {jmnnj__bljf})\n'
            )
    gadc__usy += f'  T2 = set_table_len(T2, l)\n'
    gadc__usy += f'  return T2\n'
    swx__rjdv = {}
    exec(gadc__usy, msuul__tlv, swx__rjdv)
    return swx__rjdv['impl']


@overload(operator.getitem, no_unliteral=True)
def table_getitem(T, idx):
    if not isinstance(T, TableType):
        return
    return gen_table_filter(T)


@intrinsic
def init_runtime_table_from_lists(typingctx, arr_list_tup_typ, nrows_typ=None):
    assert isinstance(arr_list_tup_typ, types.BaseTuple
        ), 'init_runtime_table_from_lists requires a tuple of list of arrays'
    if isinstance(arr_list_tup_typ, types.UniTuple):
        if arr_list_tup_typ.dtype.dtype == types.undefined:
            return
        wdv__wnf = [arr_list_tup_typ.dtype.dtype] * len(arr_list_tup_typ)
    else:
        wdv__wnf = []
        for typ in arr_list_tup_typ:
            if typ.dtype == types.undefined:
                return
            wdv__wnf.append(typ.dtype)
    assert isinstance(nrows_typ, types.Integer
        ), 'init_runtime_table_from_lists requires an integer length'

    def codegen(context, builder, sig, args):
        obyrg__ggacv, lkbj__qih = args
        qbfpr__aqeos = cgutils.create_struct_proxy(table_type)(context, builder
            )
        qbfpr__aqeos.len = lkbj__qih
        faqv__uzch = cgutils.unpack_tuple(builder, obyrg__ggacv)
        for i, kefv__snqt in enumerate(faqv__uzch):
            setattr(qbfpr__aqeos, f'block_{i}', kefv__snqt)
            context.nrt.incref(builder, types.List(wdv__wnf[i]), kefv__snqt)
        return qbfpr__aqeos._getvalue()
    table_type = TableType(tuple(wdv__wnf), True)
    sig = table_type(arr_list_tup_typ, nrows_typ)
    return sig, codegen
