"""
Analysis and transformation for HDF5 support.
"""
import types as pytypes
import numba
from numba.core import ir, types
from numba.core.ir_utils import compile_to_numba_ir, find_callname, find_const, get_definition, guard, replace_arg_nodes, require
import bodo
import bodo.io
from bodo.utils.transform import get_const_value_inner


class H5_IO:

    def __init__(self, func_ir, _locals, flags, arg_types):
        self.func_ir = func_ir
        self.locals = _locals
        self.flags = flags
        self.arg_types = arg_types

    def handle_possible_h5_read(self, assign, lhs, rhs):
        kwce__iwh = self._get_h5_type(lhs, rhs)
        if kwce__iwh is not None:
            tlft__uhcc = str(kwce__iwh.dtype)
            wcwg__uymz = 'def _h5_read_impl(dset, index):\n'
            wcwg__uymz += (
                "  arr = bodo.io.h5_api.h5_read_dummy(dset, {}, '{}', index)\n"
                .format(kwce__iwh.ndim, tlft__uhcc))
            ohryu__pqj = {}
            exec(wcwg__uymz, {}, ohryu__pqj)
            bppvt__bygt = ohryu__pqj['_h5_read_impl']
            achg__wge = compile_to_numba_ir(bppvt__bygt, {'bodo': bodo}
                ).blocks.popitem()[1]
            rto__fohae = rhs.index if rhs.op == 'getitem' else rhs.index_var
            replace_arg_nodes(achg__wge, [rhs.value, rto__fohae])
            hjzzq__gscoo = achg__wge.body[:-3]
            hjzzq__gscoo[-1].target = assign.target
            return hjzzq__gscoo
        return None

    def _get_h5_type(self, lhs, rhs):
        kwce__iwh = self._get_h5_type_locals(lhs)
        if kwce__iwh is not None:
            return kwce__iwh
        return guard(self._infer_h5_typ, rhs)

    def _infer_h5_typ(self, rhs):
        require(rhs.op in ('getitem', 'static_getitem'))
        rto__fohae = rhs.index if rhs.op == 'getitem' else rhs.index_var
        ekn__obkgq = guard(find_const, self.func_ir, rto__fohae)
        require(not isinstance(ekn__obkgq, str))
        val_def = rhs
        obj_name_list = []
        while True:
            val_def = get_definition(self.func_ir, val_def.value)
            require(isinstance(val_def, ir.Expr))
            if val_def.op == 'call':
                return self._get_h5_type_file(val_def, obj_name_list)
            require(val_def.op in ('getitem', 'static_getitem'))
            ksr__sjkh = (val_def.index if val_def.op == 'getitem' else
                val_def.index_var)
            geo__slm = get_const_value_inner(self.func_ir, ksr__sjkh,
                arg_types=self.arg_types)
            obj_name_list.append(geo__slm)

    def _get_h5_type_file(self, val_def, obj_name_list):
        require(len(obj_name_list) > 0)
        require(find_callname(self.func_ir, val_def) == ('File', 'h5py'))
        require(len(val_def.args) > 0)
        mmsux__grnkl = get_const_value_inner(self.func_ir, val_def.args[0],
            arg_types=self.arg_types)
        obj_name_list.reverse()
        import h5py
        hkf__eplma = h5py.File(mmsux__grnkl, 'r')
        oig__pozn = hkf__eplma
        for geo__slm in obj_name_list:
            oig__pozn = oig__pozn[geo__slm]
        require(isinstance(oig__pozn, h5py.Dataset))
        mbzum__hsgu = len(oig__pozn.shape)
        jch__fshnv = numba.np.numpy_support.from_dtype(oig__pozn.dtype)
        hkf__eplma.close()
        return types.Array(jch__fshnv, mbzum__hsgu, 'C')

    def _get_h5_type_locals(self, varname):
        skpkm__jcbuc = self.locals.pop(varname, None)
        if skpkm__jcbuc is None and varname is not None:
            skpkm__jcbuc = self.flags.h5_types.get(varname, None)
        return skpkm__jcbuc
