import asyncio
import os
import threading
from collections import defaultdict
from concurrent import futures
import pyarrow.parquet as pq
from bodo.io.fs_io import get_s3_bucket_region_njit


def get_parquet_filesnames_from_deltalake(delta_lake_path):
    try:
        from deltalake import DeltaTable
    except Exception as azytv__myftv:
        raise ImportError(
            "Bodo Error: please pip install the 'deltalake' package to read parquet from delta lake"
            )
    yepv__gff = None
    nqw__pyz = delta_lake_path.rstrip('/')
    tpr__nztce = 'AWS_DEFAULT_REGION' in os.environ
    zkje__cjbhl = os.environ.get('AWS_DEFAULT_REGION', '')
    wbun__cse = False
    if delta_lake_path.startswith('s3://'):
        jbd__ytyl = get_s3_bucket_region_njit(delta_lake_path, parallel=False)
        if jbd__ytyl != '':
            os.environ['AWS_DEFAULT_REGION'] = jbd__ytyl
            wbun__cse = True
    ifnjw__jizmb = DeltaTable(delta_lake_path)
    yepv__gff = ifnjw__jizmb.files()
    yepv__gff = [(nqw__pyz + '/' + lupk__rbjpf) for lupk__rbjpf in sorted(
        yepv__gff)]
    if wbun__cse:
        if tpr__nztce:
            os.environ['AWS_DEFAULT_REGION'] = zkje__cjbhl
        else:
            del os.environ['AWS_DEFAULT_REGION']
    return yepv__gff


def get_dataset_schema(dataset):
    if dataset.metadata is None and dataset.schema is None:
        if dataset.common_metadata is not None:
            dataset.schema = dataset.common_metadata.schema
        else:
            dataset.schema = dataset.pieces[0].get_metadata().schema
    elif dataset.schema is None:
        dataset.schema = dataset.metadata.schema
    nmbs__cbnci = dataset.schema.to_arrow_schema()
    if dataset.partitions is not None:
        for wwm__xdy in dataset.partitions.partition_names:
            if nmbs__cbnci.get_field_index(wwm__xdy) != -1:
                ddn__wfv = nmbs__cbnci.get_field_index(wwm__xdy)
                nmbs__cbnci = nmbs__cbnci.remove(ddn__wfv)
    return nmbs__cbnci


class VisitLevelThread(threading.Thread):

    def __init__(self, manifest):
        threading.Thread.__init__(self)
        self.manifest = manifest
        self.exc = None

    def run(self):
        try:
            manifest = self.manifest
            manifest.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(manifest.loop)
            manifest.loop.run_until_complete(manifest._visit_level(0,
                manifest.dirpath, []))
        except Exception as azytv__myftv:
            self.exc = azytv__myftv
        finally:
            if hasattr(manifest, 'loop') and not manifest.loop.is_closed():
                manifest.loop.close()

    def join(self):
        super(VisitLevelThread, self).join()
        if self.exc:
            raise self.exc


class ParquetManifest:

    def __init__(self, dirpath, open_file_func=None, filesystem=None,
        pathsep='/', partition_scheme='hive', metadata_nthreads=1):
        filesystem, dirpath = pq._get_filesystem_and_path(filesystem, dirpath)
        self.filesystem = filesystem
        self.open_file_func = open_file_func
        self.pathsep = pathsep
        self.dirpath = pq._stringify_path(dirpath)
        self.partition_scheme = partition_scheme
        self.partitions = pq.ParquetPartitions()
        self.pieces = []
        self._metadata_nthreads = metadata_nthreads
        self._thread_pool = futures.ThreadPoolExecutor(max_workers=
            metadata_nthreads)
        self.common_metadata_path = None
        self.metadata_path = None
        self.delta_lake_filter = set()
        self.partition_vals = defaultdict(set)
        pmlr__rvd = VisitLevelThread(self)
        pmlr__rvd.start()
        pmlr__rvd.join()
        for yneqk__iygme in self.partition_vals.keys():
            self.partition_vals[yneqk__iygme] = sorted(self.partition_vals[
                yneqk__iygme])
        for gmtyu__ezj in self.partitions.levels:
            gmtyu__ezj.keys = sorted(gmtyu__ezj.keys)
        for axlan__dxpr in self.pieces:
            if axlan__dxpr.partition_keys is not None:
                axlan__dxpr.partition_keys = [(gtvgo__iuuu, self.
                    partition_vals[gtvgo__iuuu].index(pqk__ywxbj)) for 
                    gtvgo__iuuu, pqk__ywxbj in axlan__dxpr.partition_keys]
        self.pieces.sort(key=lambda piece: piece.path)
        if self.common_metadata_path is None:
            self.common_metadata_path = self.metadata_path
        self._thread_pool.shutdown()

    async def _visit_level(self, yge__fudv, base_path, ple__zmt):
        fs = self.filesystem
        lmu__gofi, gxa__aayn, cxp__gfnit = await self.loop.run_in_executor(self
            ._thread_pool, lambda fs, base_bath: next(fs.walk(base_path)),
            fs, base_path)
        if yge__fudv == 0 and '_delta_log' in gxa__aayn:
            self.delta_lake_filter = set(get_parquet_filesnames_from_deltalake
                (base_path))
        xujpg__rbxux = []
        for nqw__pyz in cxp__gfnit:
            if nqw__pyz == '':
                continue
            edcip__qmv = self.pathsep.join((base_path, nqw__pyz))
            if nqw__pyz.endswith('_common_metadata'):
                self.common_metadata_path = edcip__qmv
            elif nqw__pyz.endswith('_metadata'):
                self.metadata_path = edcip__qmv
            elif self._should_silently_exclude(nqw__pyz):
                continue
            elif self.delta_lake_filter and edcip__qmv not in self.delta_lake_filter:
                continue
            else:
                xujpg__rbxux.append(edcip__qmv)
        rxl__rfo = [self.pathsep.join((base_path, hcji__luccx)) for
            hcji__luccx in gxa__aayn if not pq._is_private_directory(
            hcji__luccx)]
        xujpg__rbxux.sort()
        rxl__rfo.sort()
        if len(xujpg__rbxux) > 0 and len(rxl__rfo) > 0:
            raise ValueError('Found files in an intermediate directory: {}'
                .format(base_path))
        elif len(rxl__rfo) > 0:
            await self._visit_directories(yge__fudv, rxl__rfo, ple__zmt)
        else:
            self._push_pieces(xujpg__rbxux, ple__zmt)

    async def _visit_directories(self, yge__fudv, gxa__aayn, ple__zmt):
        dse__fpbkg = []
        for nqw__pyz in gxa__aayn:
            meb__dywkh, omg__ygf = pq._path_split(nqw__pyz, self.pathsep)
            gtvgo__iuuu, gtv__idxi = pq._parse_hive_partition(omg__ygf)
            nlj__kzko = self.partitions.get_index(yge__fudv, gtvgo__iuuu,
                gtv__idxi)
            self.partition_vals[gtvgo__iuuu].add(gtv__idxi)
            cet__qtwpt = ple__zmt + [(gtvgo__iuuu, gtv__idxi)]
            dse__fpbkg.append(self._visit_level(yge__fudv + 1, nqw__pyz,
                cet__qtwpt))
        await asyncio.wait(dse__fpbkg)


ParquetManifest._should_silently_exclude = (pq.ParquetManifest.
    _should_silently_exclude)
ParquetManifest._parse_partition = pq.ParquetManifest._parse_partition
ParquetManifest._push_pieces = pq.ParquetManifest._push_pieces
pq.ParquetManifest = ParquetManifest


def pieces(self):
    return self._pieces


pq.ParquetDataset.pieces = property(pieces)


def partitions(self):
    return self._partitions


pq.ParquetDataset.partitions = property(partitions)
