import os
import warnings
from collections import defaultdict
import llvmlite.binding as ll
import numba
import numpy as np
import pandas as pd
import pyarrow
import pyarrow.dataset as ds
from numba.core import ir, types
from numba.core.ir_utils import compile_to_numba_ir, get_definition, guard, mk_unique_var, next_label, replace_arg_nodes
from numba.extending import NativeValue, intrinsic, models, overload, register_model, unbox
from pyarrow import null
import bodo
import bodo.ir.parquet_ext
import bodo.utils.tracing as tracing
from bodo.hiframes.datetime_date_ext import datetime_date_array_type, datetime_date_type
from bodo.hiframes.pd_categorical_ext import CategoricalArrayType, PDCategoricalDtype
from bodo.hiframes.table import TableType
from bodo.io.fs_io import get_hdfs_fs, get_s3_fs_from_path, get_s3_subtree_fs
from bodo.libs.array import cpp_table_to_py_table, delete_table, info_from_table, info_to_array, table_type
from bodo.libs.array_item_arr_ext import ArrayItemArrayType
from bodo.libs.binary_arr_ext import binary_array_type, bytes_type
from bodo.libs.bool_arr_ext import boolean_array
from bodo.libs.decimal_arr_ext import DecimalArrayType
from bodo.libs.distributed_api import get_end, get_start
from bodo.libs.int_arr_ext import IntegerArrayType
from bodo.libs.str_arr_ext import string_array_type
from bodo.libs.str_ext import string_type, unicode_to_utf8
from bodo.libs.struct_arr_ext import StructArrayType
from bodo.transforms import distributed_pass
from bodo.utils.transform import get_const_value
from bodo.utils.typing import BodoError, BodoWarning, FileInfo, get_overload_const_str, get_overload_constant_dict
from bodo.utils.utils import check_and_propagate_cpp_exception, numba_to_c_type, sanitize_varname
use_nullable_int_arr = True
from urllib.parse import urlparse
import bodo.io.pa_parquet
REMOTE_FILESYSTEMS = {'s3', 'gcs', 'gs', 'http', 'hdfs', 'abfs', 'abfss'}


class ParquetPredicateType(types.Type):

    def __init__(self):
        super(ParquetPredicateType, self).__init__(name=
            'ParquetPredicateType()')


parquet_predicate_type = ParquetPredicateType()
types.parquet_predicate_type = parquet_predicate_type
register_model(ParquetPredicateType)(models.OpaqueModel)


@unbox(ParquetPredicateType)
def unbox_parquet_predicate_type(typ, val, c):
    c.pyapi.incref(val)
    return NativeValue(val)


class ReadParquetFilepathType(types.Opaque):

    def __init__(self):
        super(ReadParquetFilepathType, self).__init__(name=
            'ReadParquetFilepathType')


read_parquet_fpath_type = ReadParquetFilepathType()
types.read_parquet_fpath_type = read_parquet_fpath_type
register_model(ReadParquetFilepathType)(models.OpaqueModel)


@unbox(ReadParquetFilepathType)
def unbox_read_parquet_fpath_type(typ, val, c):
    c.pyapi.incref(val)
    return NativeValue(val)


class StorageOptionsDictType(types.Opaque):

    def __init__(self):
        super(StorageOptionsDictType, self).__init__(name=
            'StorageOptionsDictType')


storage_options_dict_type = StorageOptionsDictType()
types.storage_options_dict_type = storage_options_dict_type
register_model(StorageOptionsDictType)(models.OpaqueModel)


@unbox(StorageOptionsDictType)
def unbox_storage_options_dict_type(typ, val, c):
    c.pyapi.incref(val)
    return NativeValue(val)


class ParquetFileInfo(FileInfo):

    def __init__(self, columns, storage_options=None):
        self.columns = columns
        self.storage_options = storage_options
        super().__init__()

    def _get_schema(self, fname):
        try:
            return parquet_file_schema(fname, selected_columns=self.columns,
                storage_options=self.storage_options)
        except OSError as mrrx__bcyks:
            if 'non-file path' in str(mrrx__bcyks):
                raise FileNotFoundError(str(mrrx__bcyks))
            raise


class ParquetHandler:

    def __init__(self, func_ir, typingctx, args, _locals):
        self.func_ir = func_ir
        self.typingctx = typingctx
        self.args = args
        self.locals = _locals

    def gen_parquet_read(self, file_name, lhs, columns, storage_options=None):
        huiq__tfgdd = lhs.scope
        jaa__rxdbz = lhs.loc
        fctzl__vjn = None
        if lhs.name in self.locals:
            fctzl__vjn = self.locals[lhs.name]
            self.locals.pop(lhs.name)
        goedq__uth = {}
        if lhs.name + ':convert' in self.locals:
            goedq__uth = self.locals[lhs.name + ':convert']
            self.locals.pop(lhs.name + ':convert')
        if fctzl__vjn is None:
            rfb__hkhtq = (
                'Parquet schema not available. Either path argument should be constant for Bodo to look at the file at compile time or schema should be provided. For more information, see: https://docs.bodo.ai/latest/source/programming_with_bodo/file_io.html#non-constant-filepaths'
                )
            jnu__jksmv = get_const_value(file_name, self.func_ir,
                rfb__hkhtq, arg_types=self.args, file_info=ParquetFileInfo(
                columns, storage_options=storage_options))
            ckvrn__guiw = False
            wpi__cpn = guard(get_definition, self.func_ir, file_name)
            if isinstance(wpi__cpn, ir.Arg):
                typ = self.args[wpi__cpn.index]
                if isinstance(typ, types.FilenameType):
                    (col_names, ceh__vmn, vvmhv__dllbp, col_indices,
                        partition_names) = typ.schema
                    ckvrn__guiw = True
            if not ckvrn__guiw:
                (col_names, ceh__vmn, vvmhv__dllbp, col_indices,
                    partition_names) = (parquet_file_schema(jnu__jksmv,
                    columns, storage_options=storage_options))
        else:
            mzg__piqpu = list(fctzl__vjn.keys())
            vreb__vep = {c: rkyu__vbhn for rkyu__vbhn, c in enumerate(
                mzg__piqpu)}
            axf__cndj = [ylkx__xtxxu for ylkx__xtxxu in fctzl__vjn.values()]
            vvmhv__dllbp = 'index' if 'index' in vreb__vep else None
            if columns is None:
                selected_columns = mzg__piqpu
            else:
                selected_columns = columns
            col_indices = [vreb__vep[c] for c in selected_columns]
            ceh__vmn = [axf__cndj[vreb__vep[c]] for c in selected_columns]
            col_names = selected_columns
            vvmhv__dllbp = vvmhv__dllbp if vvmhv__dllbp in col_names else None
            partition_names = []
        ivmmy__zfvg = None if isinstance(vvmhv__dllbp, dict
            ) or vvmhv__dllbp is None else vvmhv__dllbp
        index_column_index = None
        index_column_type = types.none
        if ivmmy__zfvg:
            jbeon__zspz = col_names.index(ivmmy__zfvg)
            col_indices = col_indices.copy()
            ceh__vmn = ceh__vmn.copy()
            index_column_index = col_indices.pop(jbeon__zspz)
            index_column_type = ceh__vmn.pop(jbeon__zspz)
        for rkyu__vbhn, c in enumerate(col_names):
            if c in goedq__uth:
                ceh__vmn[rkyu__vbhn] = goedq__uth[c]
        qpl__ivu = [ir.Var(huiq__tfgdd, mk_unique_var('pq_table'),
            jaa__rxdbz), ir.Var(huiq__tfgdd, mk_unique_var('pq_index'),
            jaa__rxdbz)]
        hinj__xdxre = [bodo.ir.parquet_ext.ParquetReader(file_name, lhs.
            name, col_names, col_indices, ceh__vmn, qpl__ivu, jaa__rxdbz,
            partition_names, storage_options, index_column_index,
            index_column_type)]
        return (col_names, qpl__ivu, vvmhv__dllbp, hinj__xdxre, ceh__vmn,
            index_column_type)


def determine_filter_cast(pq_node, typemap, filter_val, orig_colname_map):
    brn__gqli = filter_val[0]
    gzx__nrrzs = pq_node.original_out_types[orig_colname_map[brn__gqli]]
    weh__onzxg = bodo.utils.typing.element_type(gzx__nrrzs)
    if brn__gqli in pq_node.partition_names:
        if weh__onzxg == types.unicode_type:
            dlvkh__zmrus = '.cast(pyarrow.string(), safe=False)'
        elif isinstance(weh__onzxg, types.Integer):
            dlvkh__zmrus = f'.cast(pyarrow.{weh__onzxg.name}(), safe=False)'
        else:
            dlvkh__zmrus = ''
    else:
        dlvkh__zmrus = ''
    xbbm__rbx = typemap[filter_val[2].name]
    if not bodo.utils.typing.is_common_scalar_dtype([weh__onzxg, xbbm__rbx]):
        if not bodo.utils.typing.is_safe_arrow_cast(weh__onzxg, xbbm__rbx):
            raise BodoError(
                f'Unsupport Arrow cast from {weh__onzxg} to {xbbm__rbx} in filter pushdown. Please try a comparison that avoids casting the column.'
                )
        if weh__onzxg == types.unicode_type:
            return ".cast(pyarrow.timestamp('ns'), safe=False)", ''
        elif weh__onzxg in (bodo.datetime64ns, bodo.pd_timestamp_type):
            return dlvkh__zmrus, ".cast(pyarrow.timestamp('ns'), safe=False)"
    return dlvkh__zmrus, ''


def pq_distributed_run(pq_node, array_dists, typemap, calltypes, typingctx,
    targetctx, meta_head_only_info=None):
    xbbbk__ukmq = len(pq_node.out_vars)
    extra_args = ''
    dnf_filter_str = 'None'
    expr_filter_str = 'None'
    noev__wdrk, qgfno__ibcqm = bodo.ir.connector.generate_filter_map(pq_node
        .filters)
    if pq_node.filters:
        ugc__cbgv = []
        ytbe__qopjp = []
        qdz__wwo = False
        tupvr__tsds = None
        orig_colname_map = {c: rkyu__vbhn for rkyu__vbhn, c in enumerate(
            pq_node.original_df_colnames)}
        for gyrtf__uwo in pq_node.filters:
            tcwu__xms = []
            rbpks__xgfe = []
            dylzv__emc = set()
            for mgbx__sbbcp in gyrtf__uwo:
                if isinstance(mgbx__sbbcp[2], ir.Var):
                    smu__nfa, dyly__heqo = determine_filter_cast(pq_node,
                        typemap, mgbx__sbbcp, orig_colname_map)
                    rbpks__xgfe.append(
                        f"(ds.field('{mgbx__sbbcp[0]}'){smu__nfa} {mgbx__sbbcp[1]} ds.scalar({noev__wdrk[mgbx__sbbcp[2].name]}){dyly__heqo})"
                        )
                else:
                    assert mgbx__sbbcp[2
                        ] == 'NULL', 'unsupport constant used in filter pushdown'
                    if mgbx__sbbcp[1] == 'is not':
                        prefix = '~'
                    else:
                        prefix = ''
                    rbpks__xgfe.append(
                        f"({prefix}ds.field('{mgbx__sbbcp[0]}').is_null())")
                if mgbx__sbbcp[0] in pq_node.partition_names and isinstance(
                    mgbx__sbbcp[2], ir.Var):
                    vmii__vdv = (
                        f"('{mgbx__sbbcp[0]}', '{mgbx__sbbcp[1]}', {noev__wdrk[mgbx__sbbcp[2].name]})"
                        )
                    tcwu__xms.append(vmii__vdv)
                    dylzv__emc.add(vmii__vdv)
                else:
                    qdz__wwo = True
            if tupvr__tsds is None:
                tupvr__tsds = dylzv__emc
            else:
                tupvr__tsds.intersection_update(dylzv__emc)
            bprn__xapew = ', '.join(tcwu__xms)
            agp__thg = ' & '.join(rbpks__xgfe)
            if bprn__xapew:
                ugc__cbgv.append(f'[{bprn__xapew}]')
            ytbe__qopjp.append(f'({agp__thg})')
        jkmmp__tuiow = ', '.join(ugc__cbgv)
        cpjsv__lvkk = ' | '.join(ytbe__qopjp)
        if qdz__wwo:
            if tupvr__tsds:
                vzfac__cszpg = sorted(tupvr__tsds)
                dnf_filter_str = f"[[{', '.join(vzfac__cszpg)}]]"
        elif jkmmp__tuiow:
            dnf_filter_str = f'[{jkmmp__tuiow}]'
        expr_filter_str = f'({cpjsv__lvkk})'
        extra_args = ', '.join(noev__wdrk.values())
    legw__prd = ', '.join(f'out{rkyu__vbhn}' for rkyu__vbhn in range(
        xbbbk__ukmq))
    dcn__wrm = f'def pq_impl(fname, {extra_args}):\n'
    dcn__wrm += (
        f'    (total_rows, {legw__prd},) = _pq_reader_py(fname, {extra_args})\n'
        )
    uodbi__cpgy = {}
    exec(dcn__wrm, {}, uodbi__cpgy)
    krvbm__ftix = uodbi__cpgy['pq_impl']
    parallel = False
    if array_dists is not None:
        tlvc__aniny = pq_node.out_vars[0].name
        parallel = array_dists[tlvc__aniny] in (distributed_pass.
            Distribution.OneD, distributed_pass.Distribution.OneD_Var)
        lwj__iuk = pq_node.out_vars[1].name
        assert typemap[lwj__iuk] == types.none or not parallel or array_dists[
            lwj__iuk] in (distributed_pass.Distribution.OneD,
            distributed_pass.Distribution.OneD_Var
            ), 'pq data/index parallelization does not match'
    jhqh__saq = _gen_pq_reader_py(pq_node.df_colnames, pq_node.col_indices,
        pq_node.type_usecol_offset, pq_node.out_types, pq_node.
        storage_options, pq_node.partition_names, dnf_filter_str,
        expr_filter_str, extra_args, parallel, meta_head_only_info, pq_node
        .index_column_index, pq_node.index_column_type)
    vxav__van = typemap[pq_node.file_name.name]
    xyj__tlmda = (vxav__van,) + tuple(typemap[mgbx__sbbcp.name] for
        mgbx__sbbcp in qgfno__ibcqm)
    box__wnl = compile_to_numba_ir(krvbm__ftix, {'_pq_reader_py': jhqh__saq
        }, typingctx=typingctx, targetctx=targetctx, arg_typs=xyj__tlmda,
        typemap=typemap, calltypes=calltypes).blocks.popitem()[1]
    replace_arg_nodes(box__wnl, [pq_node.file_name] + qgfno__ibcqm)
    hinj__xdxre = box__wnl.body[:-3]
    if meta_head_only_info:
        hinj__xdxre[-1 - xbbbk__ukmq].target = meta_head_only_info[1]
    hinj__xdxre[-2].target = pq_node.out_vars[0]
    hinj__xdxre[-1].target = pq_node.out_vars[1]
    return hinj__xdxre


distributed_pass.distributed_run_extensions[bodo.ir.parquet_ext.ParquetReader
    ] = pq_distributed_run


def get_filters_pyobject(dnf_filter_str, expr_filter_str, vars):
    pass


@overload(get_filters_pyobject, no_unliteral=True)
def overload_get_filters_pyobject(dnf_filter_str, expr_filter_str, var_tup):
    pvb__mzgd = get_overload_const_str(dnf_filter_str)
    sgfo__kfqb = get_overload_const_str(expr_filter_str)
    qwny__wxatk = ', '.join(f'f{rkyu__vbhn}' for rkyu__vbhn in range(len(
        var_tup)))
    dcn__wrm = 'def impl(dnf_filter_str, expr_filter_str, var_tup):\n'
    if len(var_tup):
        dcn__wrm += f'  {qwny__wxatk}, = var_tup\n'
    dcn__wrm += """  with numba.objmode(dnf_filters_py='parquet_predicate_type', expr_filters_py='parquet_predicate_type'):
"""
    dcn__wrm += f'    dnf_filters_py = {pvb__mzgd}\n'
    dcn__wrm += f'    expr_filters_py = {sgfo__kfqb}\n'
    dcn__wrm += '  return (dnf_filters_py, expr_filters_py)\n'
    uodbi__cpgy = {}
    exec(dcn__wrm, globals(), uodbi__cpgy)
    return uodbi__cpgy['impl']


@numba.njit
def get_fname_pyobject(fname):
    with numba.objmode(fname_py='read_parquet_fpath_type'):
        fname_py = fname
    return fname_py


def get_storage_options_pyobject(storage_options):
    pass


@overload(get_storage_options_pyobject, no_unliteral=True)
def overload_get_storage_options_pyobject(storage_options):
    akc__ulupb = get_overload_constant_dict(storage_options)
    dcn__wrm = 'def impl(storage_options):\n'
    dcn__wrm += (
        "  with numba.objmode(storage_options_py='storage_options_dict_type'):\n"
        )
    dcn__wrm += f'    storage_options_py = {str(akc__ulupb)}\n'
    dcn__wrm += '  return storage_options_py\n'
    uodbi__cpgy = {}
    exec(dcn__wrm, globals(), uodbi__cpgy)
    return uodbi__cpgy['impl']


def _gen_pq_reader_py(col_names, col_indices, type_usecol_offset, out_types,
    storage_options, partition_names, dnf_filter_str, expr_filter_str,
    extra_args, is_parallel, meta_head_only_info, index_column_index,
    index_column_type):
    nwbqr__kkj = next_label()
    yuvc__sda = ',' if extra_args else ''
    dcn__wrm = f'def pq_reader_py(fname,{extra_args}):\n'
    dcn__wrm += (
        f"    ev = bodo.utils.tracing.Event('read_parquet', {is_parallel})\n")
    dcn__wrm += "    ev.add_attribute('fname', fname)\n"
    dcn__wrm += f"""    bucket_region = bodo.io.fs_io.get_s3_bucket_region_njit(fname, parallel={is_parallel})
"""
    dcn__wrm += f"""    dnf_filters, expr_filters = get_filters_pyobject("{dnf_filter_str}", "{expr_filter_str}", ({extra_args}{yuvc__sda}))
"""
    dcn__wrm += '    fname_py = get_fname_pyobject(fname)\n'
    storage_options['bodo_dummy'] = 'dummy'
    dcn__wrm += (
        f'    storage_options_py = get_storage_options_pyobject({str(storage_options)})\n'
        )
    tot_rows_to_read = -1
    if meta_head_only_info and meta_head_only_info[0] is not None:
        tot_rows_to_read = meta_head_only_info[0]
    eln__lpo = not type_usecol_offset
    ixl__dtqrz = [sanitize_varname(c) for c in col_names]
    partition_names = [sanitize_varname(c) for c in partition_names]
    isr__ymbn = {c: rkyu__vbhn for rkyu__vbhn, c in enumerate(col_indices)}
    rlw__fpgod = {c: rkyu__vbhn for rkyu__vbhn, c in enumerate(ixl__dtqrz)}
    wqua__efdb = []
    iael__amks = set()
    for rkyu__vbhn in type_usecol_offset:
        if ixl__dtqrz[rkyu__vbhn] not in partition_names:
            wqua__efdb.append(col_indices[rkyu__vbhn])
        else:
            iael__amks.add(col_indices[rkyu__vbhn])
    if index_column_index is not None:
        wqua__efdb.append(index_column_index)
    wqua__efdb = sorted(wqua__efdb)
    idjlm__cxku = {c: rkyu__vbhn for rkyu__vbhn, c in enumerate(wqua__efdb)}

    def is_nullable(typ):
        return bodo.utils.utils.is_array_typ(typ, False) and not isinstance(typ
            , types.Array)
    vyrr__sfj = [(int(is_nullable(out_types[isr__ymbn[pvu__ycfjm]])) if 
        pvu__ycfjm != index_column_index else int(is_nullable(
        index_column_type))) for pvu__ycfjm in wqua__efdb]
    dvjn__imwsr = []
    gdtw__tszd = {}
    dxnyv__mbiqc = []
    zouo__bnhzr = []
    for rkyu__vbhn, wlhy__gdd in enumerate(partition_names):
        try:
            unlic__nnk = rlw__fpgod[wlhy__gdd]
            if col_indices[unlic__nnk] not in iael__amks:
                continue
        except (KeyError, ValueError) as duysw__cptl:
            continue
        gdtw__tszd[wlhy__gdd] = len(dvjn__imwsr)
        dvjn__imwsr.append(wlhy__gdd)
        dxnyv__mbiqc.append(rkyu__vbhn)
        xyxw__tiyl = out_types[unlic__nnk].dtype
        sfx__gdpj = bodo.hiframes.pd_categorical_ext.get_categories_int_type(
            xyxw__tiyl)
        zouo__bnhzr.append(numba_to_c_type(sfx__gdpj))
    dcn__wrm += f'    total_rows_np = np.array([0], dtype=np.int64)\n'
    if len(dxnyv__mbiqc) > 0:
        dcn__wrm += f"""    out_table = pq_read(fname_py, {is_parallel}, unicode_to_utf8(bucket_region), dnf_filters, expr_filters, storage_options_py, {tot_rows_to_read}, selected_cols_arr_{nwbqr__kkj}.ctypes, {len(wqua__efdb)}, nullable_cols_arr_{nwbqr__kkj}.ctypes, np.array({dxnyv__mbiqc}, dtype=np.int32).ctypes, np.array({zouo__bnhzr}, dtype=np.int32).ctypes, {len(dxnyv__mbiqc)}, total_rows_np.ctypes)
"""
    else:
        dcn__wrm += f"""    out_table = pq_read(fname_py, {is_parallel}, unicode_to_utf8(bucket_region), dnf_filters, expr_filters, storage_options_py, {tot_rows_to_read}, selected_cols_arr_{nwbqr__kkj}.ctypes, {len(wqua__efdb)}, nullable_cols_arr_{nwbqr__kkj}.ctypes, 0, 0, 0, total_rows_np.ctypes)
"""
    dcn__wrm += '    check_and_propagate_cpp_exception()\n'
    hofw__rzwzs = 'None'
    dvuhr__qht = index_column_type
    sqi__qayr = TableType(tuple(out_types))
    if eln__lpo:
        sqi__qayr = types.none
    if index_column_index is not None:
        ztz__wzoti = idjlm__cxku[index_column_index]
        hofw__rzwzs = (
            f'info_to_array(info_from_table(out_table, {ztz__wzoti}), index_arr_type)'
            )
    dcn__wrm += f'    index_arr = {hofw__rzwzs}\n'
    if eln__lpo:
        dys__ebjv = None
    else:
        dys__ebjv = []
        hzx__wmhls = 0
        for rkyu__vbhn, mbvu__xucgg in enumerate(col_indices):
            if hzx__wmhls < len(type_usecol_offset
                ) and rkyu__vbhn == type_usecol_offset[hzx__wmhls]:
                nzf__zlt = col_indices[rkyu__vbhn]
                if nzf__zlt in iael__amks:
                    qlc__xyxoh = ixl__dtqrz[rkyu__vbhn]
                    dys__ebjv.append(len(wqua__efdb) + gdtw__tszd[qlc__xyxoh])
                else:
                    dys__ebjv.append(idjlm__cxku[mbvu__xucgg])
                hzx__wmhls += 1
            else:
                dys__ebjv.append(-1)
        dys__ebjv = np.array(dys__ebjv, dtype=np.int64)
    if eln__lpo:
        dcn__wrm += '    T = None\n'
    else:
        dcn__wrm += f"""    T = cpp_table_to_py_table(out_table, table_idx_{nwbqr__kkj}, py_table_type_{nwbqr__kkj})
"""
    dcn__wrm += '    delete_table(out_table)\n'
    dcn__wrm += f'    total_rows = total_rows_np[0]\n'
    dcn__wrm += f'    ev.finalize()\n'
    dcn__wrm += '    return (total_rows, T, index_arr)\n'
    uodbi__cpgy = {}
    fymoe__zcov = {f'py_table_type_{nwbqr__kkj}': sqi__qayr,
        f'table_idx_{nwbqr__kkj}': dys__ebjv,
        f'selected_cols_arr_{nwbqr__kkj}': np.array(wqua__efdb, np.int32),
        f'nullable_cols_arr_{nwbqr__kkj}': np.array(vyrr__sfj, np.int32),
        'index_arr_type': dvuhr__qht, 'cpp_table_to_py_table':
        cpp_table_to_py_table, 'info_to_array': info_to_array,
        'info_from_table': info_from_table, 'delete_table': delete_table,
        'check_and_propagate_cpp_exception':
        check_and_propagate_cpp_exception, 'pq_read': _pq_read,
        'unicode_to_utf8': unicode_to_utf8, 'get_filters_pyobject':
        get_filters_pyobject, 'get_storage_options_pyobject':
        get_storage_options_pyobject, 'get_fname_pyobject':
        get_fname_pyobject, 'np': np, 'pd': pd, 'bodo': bodo}
    exec(dcn__wrm, fymoe__zcov, uodbi__cpgy)
    jhqh__saq = uodbi__cpgy['pq_reader_py']
    cnyv__ash = numba.njit(jhqh__saq, no_cpython_wrapper=True)
    return cnyv__ash


import pyarrow as pa
_pa_numba_typ_map = {pa.bool_(): types.bool_, pa.int8(): types.int8, pa.
    int16(): types.int16, pa.int32(): types.int32, pa.int64(): types.int64,
    pa.uint8(): types.uint8, pa.uint16(): types.uint16, pa.uint32(): types.
    uint32, pa.uint64(): types.uint64, pa.float32(): types.float32, pa.
    float64(): types.float64, pa.string(): string_type, pa.binary():
    bytes_type, pa.date32(): datetime_date_type, pa.date64(): types.
    NPDatetime('ns'), pa.timestamp('ns'): types.NPDatetime('ns'), pa.
    timestamp('us'): types.NPDatetime('ns'), pa.timestamp('ms'): types.
    NPDatetime('ns'), pa.timestamp('s'): types.NPDatetime('ns'), null():
    string_type}


def _get_numba_typ_from_pa_typ(pa_typ, is_index, nullable_from_metadata,
    category_info):
    if isinstance(pa_typ.type, pa.ListType):
        return ArrayItemArrayType(_get_numba_typ_from_pa_typ(pa_typ.type.
            value_field, is_index, nullable_from_metadata, category_info))
    if isinstance(pa_typ.type, pa.StructType):
        qwxbo__pdjo = []
        mjp__czoy = []
        for kee__pxakz in pa_typ.flatten():
            mjp__czoy.append(kee__pxakz.name.split('.')[-1])
            qwxbo__pdjo.append(_get_numba_typ_from_pa_typ(kee__pxakz,
                is_index, nullable_from_metadata, category_info))
        return StructArrayType(tuple(qwxbo__pdjo), tuple(mjp__czoy))
    if isinstance(pa_typ.type, pa.Decimal128Type):
        return DecimalArrayType(pa_typ.type.precision, pa_typ.type.scale)
    if isinstance(pa_typ.type, pa.DictionaryType):
        if pa_typ.type.value_type != pa.string():
            raise BodoError(
                f'Parquet Categorical data type should be string, not {pa_typ.type.value_type}'
                )
        nvjks__eyo = _pa_numba_typ_map[pa_typ.type.index_type]
        vqb__lychc = PDCategoricalDtype(category_info[pa_typ.name], bodo.
            string_type, pa_typ.type.ordered, int_type=nvjks__eyo)
        return CategoricalArrayType(vqb__lychc)
    if pa_typ.type not in _pa_numba_typ_map:
        raise BodoError('Arrow data type {} not supported yet'.format(
            pa_typ.type))
    fdwoq__vxbxg = _pa_numba_typ_map[pa_typ.type]
    if fdwoq__vxbxg == datetime_date_type:
        return datetime_date_array_type
    if fdwoq__vxbxg == bytes_type:
        return binary_array_type
    argl__uwp = (string_array_type if fdwoq__vxbxg == string_type else
        types.Array(fdwoq__vxbxg, 1, 'C'))
    if fdwoq__vxbxg == types.bool_:
        argl__uwp = boolean_array
    if nullable_from_metadata is not None:
        ffkrr__obn = nullable_from_metadata
    else:
        ffkrr__obn = use_nullable_int_arr
    if ffkrr__obn and not is_index and isinstance(fdwoq__vxbxg, types.Integer
        ) and pa_typ.nullable:
        argl__uwp = IntegerArrayType(fdwoq__vxbxg)
    return argl__uwp


def get_parquet_dataset(fpath, get_row_counts=True, dnf_filters=None,
    expr_filters=None, storage_options=None, read_categories=False,
    is_parallel=False, tot_rows_to_read=None):
    if get_row_counts:
        bmy__yajb = tracing.Event('get_parquet_dataset')
    import time
    import pyarrow as pa
    import pyarrow.parquet as pq
    from mpi4py import MPI
    xwxf__xkpvr = MPI.COMM_WORLD
    if isinstance(fpath, list):
        dws__ptva = urlparse(fpath[0])
        protocol = dws__ptva.scheme
        pbd__rln = dws__ptva.netloc
        for rkyu__vbhn in range(len(fpath)):
            ark__wszri = fpath[rkyu__vbhn]
            mhwua__lmx = urlparse(ark__wszri)
            if mhwua__lmx.scheme != protocol:
                raise BodoError(
                    'All parquet files must use the same filesystem protocol')
            if mhwua__lmx.netloc != pbd__rln:
                raise BodoError(
                    'All parquet files must be in the same S3 bucket')
            fpath[rkyu__vbhn] = ark__wszri.rstrip('/')
    else:
        dws__ptva = urlparse(fpath)
        protocol = dws__ptva.scheme
        fpath = fpath.rstrip('/')
    if protocol in {'gcs', 'gs'}:
        try:
            import gcsfs
        except ImportError as duysw__cptl:
            xvcp__uxn = """Couldn't import gcsfs, which is required for Google cloud access. gcsfs can be installed by calling 'conda install -c conda-forge gcsfs'.
"""
            raise BodoError(xvcp__uxn)
    if protocol == 'http':
        try:
            import fsspec
        except ImportError as duysw__cptl:
            xvcp__uxn = """Couldn't import fsspec, which is required for http access. fsspec can be installed by calling 'conda install -c conda-forge fsspec'.
"""
    fs = []

    def getfs(parallel=False):
        if len(fs) == 1:
            return fs[0]
        if protocol == 's3':
            fs.append(get_s3_fs_from_path(fpath, parallel=parallel,
                storage_options=storage_options) if not isinstance(fpath,
                list) else get_s3_fs_from_path(fpath[0], parallel=parallel,
                storage_options=storage_options))
        elif protocol in {'gcs', 'gs'}:
            bluz__pjysd = gcsfs.GCSFileSystem(token=None)
            fs.append(bluz__pjysd)
        elif protocol == 'http':
            fs.append(fsspec.filesystem('http'))
        elif protocol in {'hdfs', 'abfs', 'abfss'}:
            fs.append(get_hdfs_fs(fpath) if not isinstance(fpath, list) else
                get_hdfs_fs(fpath[0]))
        else:
            fs.append(None)
        return fs[0]

    def get_legacy_fs():
        if protocol in {'s3', 'hdfs', 'abfs', 'abfss'}:
            from fsspec.implementations.arrow import ArrowFSWrapper
            return ArrowFSWrapper(getfs())
        else:
            return getfs()

    def glob(protocol, fs, path):
        if not protocol and fs is None:
            from fsspec.implementations.local import LocalFileSystem
            fs = LocalFileSystem()
        if isinstance(fs, pyarrow.fs.FileSystem):
            from fsspec.implementations.arrow import ArrowFSWrapper
            fs = ArrowFSWrapper(fs)
        try:
            if protocol in {'hdfs', 'abfs', 'abfss'}:
                prefix = f'{protocol}://{dws__ptva.netloc}'
                path = path[len(prefix):]
            nfeo__dffig = fs.glob(path)
            if protocol == 's3':
                nfeo__dffig = [('s3://' + ark__wszri) for ark__wszri in
                    nfeo__dffig if not ark__wszri.startswith('s3://')]
            elif protocol in {'hdfs', 'abfs', 'abfss'}:
                nfeo__dffig = [(prefix + ark__wszri) for ark__wszri in
                    nfeo__dffig]
        except:
            raise BodoError(
                f'glob pattern expansion not supported for {protocol}')
        if len(nfeo__dffig) == 0:
            raise BodoError('No files found matching glob pattern')
        return nfeo__dffig
    hwgm__qudyn = False
    if get_row_counts:
        bphkh__ycy = getfs(parallel=True)
        hwgm__qudyn = bodo.parquet_validate_schema
    if bodo.get_rank() == 0:
        rondd__kkof = 1
        vjp__hxzjc = os.cpu_count()
        if vjp__hxzjc is not None and vjp__hxzjc > 1:
            rondd__kkof = vjp__hxzjc // 2
        try:
            if get_row_counts:
                fnk__umgr = tracing.Event('pq.ParquetDataset', is_parallel=
                    False)
                if tracing.is_tracing():
                    fnk__umgr.add_attribute('dnf_filter', str(dnf_filters))
            oap__dyh = pa.io_thread_count()
            pa.set_io_thread_count(rondd__kkof)
            if '*' in fpath:
                fpath = glob(protocol, getfs(), fpath)
            if protocol == 's3':
                if isinstance(fpath, list):
                    get_legacy_fs().info(fpath[0])
                else:
                    get_legacy_fs().info(fpath)
            if protocol in {'hdfs', 'abfs', 'abfss'}:
                prefix = f'{protocol}://{dws__ptva.netloc}'
                if isinstance(fpath, list):
                    kens__uatc = [ark__wszri[len(prefix):] for ark__wszri in
                        fpath]
                else:
                    kens__uatc = fpath[len(prefix):]
            else:
                kens__uatc = fpath
            tot__ihw = pq.ParquetDataset(kens__uatc, filesystem=
                get_legacy_fs(), filters=None, use_legacy_dataset=True,
                validate_schema=False, metadata_nthreads=rondd__kkof)
            pa.set_io_thread_count(oap__dyh)
            wpfmo__ppc = bodo.io.pa_parquet.get_dataset_schema(tot__ihw)
            if dnf_filters:
                if get_row_counts:
                    fnk__umgr.add_attribute('num_pieces_before_filter', len
                        (tot__ihw.pieces))
                nflkf__cmq = time.time()
                tot__ihw._filter(dnf_filters)
                if get_row_counts:
                    fnk__umgr.add_attribute('dnf_filter_time', time.time() -
                        nflkf__cmq)
                    fnk__umgr.add_attribute('num_pieces_after_filter', len(
                        tot__ihw.pieces))
            if get_row_counts:
                fnk__umgr.finalize()
            tot__ihw._metadata.fs = None
        except Exception as mrrx__bcyks:
            xwxf__xkpvr.bcast(mrrx__bcyks)
            raise BodoError(
                f"""error from pyarrow: {type(mrrx__bcyks).__name__}: {str(mrrx__bcyks)}
"""
                )
        if get_row_counts:
            iyr__xer = tracing.Event('bcast dataset')
        xwxf__xkpvr.bcast(tot__ihw)
        xwxf__xkpvr.bcast(wpfmo__ppc)
    else:
        if get_row_counts:
            iyr__xer = tracing.Event('bcast dataset')
        tot__ihw = xwxf__xkpvr.bcast(None)
        if isinstance(tot__ihw, Exception):
            rwxl__uuwk = tot__ihw
            raise BodoError(
                f'error from pyarrow: {type(rwxl__uuwk).__name__}: {str(rwxl__uuwk)}\n'
                )
        wpfmo__ppc = xwxf__xkpvr.bcast(None)
    if get_row_counts:
        iyr__xer.finalize()
    tot__ihw._bodo_total_rows = 0
    if get_row_counts and tot_rows_to_read == 0:
        get_row_counts = hwgm__qudyn = False
        for aydhi__stcl in tot__ihw.pieces:
            aydhi__stcl._bodo_num_rows = 0
    if get_row_counts or hwgm__qudyn:
        if get_row_counts and tracing.is_tracing():
            ktkbl__beew = tracing.Event('get_row_counts')
            ktkbl__beew.add_attribute('g_num_pieces', len(tot__ihw.pieces))
            ktkbl__beew.add_attribute('g_expr_filters', str(expr_filters))
        bbnk__byhg = 0.0
        num_pieces = len(tot__ihw.pieces)
        start = get_start(num_pieces, bodo.get_size(), bodo.get_rank())
        wym__kcbr = get_end(num_pieces, bodo.get_size(), bodo.get_rank())
        gwn__pisze = 0
        qofv__pgzr = 0
        rxma__ctgk = 0
        igsnf__rso = True
        tot__ihw._metadata.fs = getfs()
        if expr_filters is not None:
            import random
            random.seed(37)
            qqr__yjc = random.sample(tot__ihw.pieces, k=len(tot__ihw.pieces))
        else:
            qqr__yjc = tot__ihw.pieces
        for aydhi__stcl in qqr__yjc:
            aydhi__stcl._bodo_num_rows = 0
        fpaths = [aydhi__stcl.path for aydhi__stcl in qqr__yjc[start:wym__kcbr]
            ]
        if protocol == 's3':
            pbd__rln = dws__ptva.netloc
            prefix = 's3://' + pbd__rln + '/'
            fpaths = [ark__wszri[len(prefix):] for ark__wszri in fpaths]
            sajm__ocd = get_s3_subtree_fs(pbd__rln, region=getfs().region,
                storage_options=storage_options)
        else:
            sajm__ocd = getfs()
        pa.set_io_thread_count(4)
        pa.set_cpu_count(4)
        wrdw__vtj = ds.dataset(fpaths, filesystem=sajm__ocd, partitioning=
            ds.partitioning(flavor='hive'))
        for idng__stlm, evy__zzis in zip(qqr__yjc[start:wym__kcbr],
            wrdw__vtj.get_fragments()):
            nflkf__cmq = time.time()
            odw__itcat = evy__zzis.scanner(schema=wrdw__vtj.schema, filter=
                expr_filters, use_threads=True).count_rows()
            bbnk__byhg += time.time() - nflkf__cmq
            idng__stlm._bodo_num_rows = odw__itcat
            gwn__pisze += odw__itcat
            qofv__pgzr += evy__zzis.num_row_groups
            rxma__ctgk += sum(mus__cby.total_byte_size for mus__cby in
                evy__zzis.row_groups)
            if hwgm__qudyn:
                agj__byra = evy__zzis.metadata.schema.to_arrow_schema()
                if wpfmo__ppc != agj__byra:
                    print('Schema in {!s} was different. \n{!s}\n\nvs\n\n{!s}'
                        .format(idng__stlm, agj__byra, wpfmo__ppc))
                    igsnf__rso = False
                    break
        if hwgm__qudyn:
            igsnf__rso = xwxf__xkpvr.allreduce(igsnf__rso, op=MPI.LAND)
            if not igsnf__rso:
                raise BodoError("Schema in parquet files don't match")
        if get_row_counts:
            tot__ihw._bodo_total_rows = xwxf__xkpvr.allreduce(gwn__pisze,
                op=MPI.SUM)
            uscr__jbkkf = xwxf__xkpvr.allreduce(qofv__pgzr, op=MPI.SUM)
            sdqfx__xul = xwxf__xkpvr.allreduce(rxma__ctgk, op=MPI.SUM)
            ujzv__ljrd = np.array([aydhi__stcl._bodo_num_rows for
                aydhi__stcl in tot__ihw.pieces])
            ujzv__ljrd = xwxf__xkpvr.allreduce(ujzv__ljrd, op=MPI.SUM)
            for aydhi__stcl, icle__yzhmh in zip(tot__ihw.pieces, ujzv__ljrd):
                aydhi__stcl._bodo_num_rows = icle__yzhmh
            if is_parallel and bodo.get_rank(
                ) == 0 and uscr__jbkkf < bodo.get_size() and uscr__jbkkf != 0:
                warnings.warn(BodoWarning(
                    f"""Total number of row groups in parquet dataset {fpath} ({uscr__jbkkf}) is too small for effective IO parallelization.
For best performance the number of row groups should be greater than the number of workers ({bodo.get_size()})
"""
                    ))
            if uscr__jbkkf == 0:
                dro__fvzhg = 0
            else:
                dro__fvzhg = sdqfx__xul // uscr__jbkkf
            if (bodo.get_rank() == 0 and sdqfx__xul >= 20 * 1048576 and 
                dro__fvzhg < 1048576 and protocol in REMOTE_FILESYSTEMS):
                warnings.warn(BodoWarning(
                    f'Parquet average row group size is small ({dro__fvzhg} bytes) and can have negative impact on performance when reading from remote sources'
                    ))
            if tracing.is_tracing():
                ktkbl__beew.add_attribute('g_total_num_row_groups', uscr__jbkkf
                    )
                if expr_filters is not None:
                    ktkbl__beew.add_attribute('total_scan_time', bbnk__byhg)
                yod__lat = np.array([aydhi__stcl._bodo_num_rows for
                    aydhi__stcl in tot__ihw.pieces])
                gzhc__sdvw = np.percentile(yod__lat, [25, 50, 75])
                ktkbl__beew.add_attribute('g_row_counts_min', yod__lat.min())
                ktkbl__beew.add_attribute('g_row_counts_Q1', gzhc__sdvw[0])
                ktkbl__beew.add_attribute('g_row_counts_median', gzhc__sdvw[1])
                ktkbl__beew.add_attribute('g_row_counts_Q3', gzhc__sdvw[2])
                ktkbl__beew.add_attribute('g_row_counts_max', yod__lat.max())
                ktkbl__beew.add_attribute('g_row_counts_mean', yod__lat.mean())
                ktkbl__beew.add_attribute('g_row_counts_std', yod__lat.std())
                ktkbl__beew.add_attribute('g_row_counts_sum', yod__lat.sum())
                ktkbl__beew.finalize()
    tot__ihw._prefix = ''
    if protocol in {'hdfs', 'abfs', 'abfss'}:
        prefix = f'{protocol}://{dws__ptva.netloc}'
        if len(tot__ihw.pieces) > 0:
            idng__stlm = tot__ihw.pieces[0]
            if not idng__stlm.path.startswith(prefix):
                tot__ihw._prefix = prefix
    if read_categories:
        _add_categories_to_pq_dataset(tot__ihw)
    if get_row_counts:
        bmy__yajb.finalize()
    return tot__ihw


def get_scanner_batches(fpaths, expr_filters, selected_fields,
    avg_num_pieces, is_parallel, storage_options, region, prefix):
    import pyarrow as pa
    vjp__hxzjc = os.cpu_count()
    if vjp__hxzjc is None or vjp__hxzjc == 0:
        vjp__hxzjc = 2
    ysvbs__qhp = min(4, vjp__hxzjc)
    bfc__pfvwl = min(16, vjp__hxzjc)
    if is_parallel and len(fpaths) > bfc__pfvwl and len(fpaths
        ) / avg_num_pieces >= 2.0:
        pa.set_io_thread_count(bfc__pfvwl)
        pa.set_cpu_count(bfc__pfvwl)
    else:
        pa.set_io_thread_count(ysvbs__qhp)
        pa.set_cpu_count(ysvbs__qhp)
    if fpaths[0].startswith('s3://'):
        pbd__rln = urlparse(fpaths[0]).netloc
        prefix = 's3://' + pbd__rln + '/'
        fpaths = [ark__wszri[len(prefix):] for ark__wszri in fpaths]
        sajm__ocd = get_s3_subtree_fs(pbd__rln, region=region,
            storage_options=storage_options)
    elif prefix and prefix.startswith(('hdfs', 'abfs', 'abfss')):
        sajm__ocd = get_hdfs_fs(prefix + fpaths[0])
    elif fpaths[0].startswith(('gcs', 'gs')):
        import gcsfs
        sajm__ocd = gcsfs.GCSFileSystem(token=None)
    else:
        sajm__ocd = None
    tot__ihw = ds.dataset(fpaths, filesystem=sajm__ocd, partitioning=ds.
        partitioning(flavor='hive'))
    col_names = tot__ihw.schema.names
    ihar__ekqz = [col_names[wrctw__tvyj] for wrctw__tvyj in selected_fields]
    masw__yxjro = tot__ihw.scanner(columns=ihar__ekqz, filter=expr_filters,
        use_threads=True).to_reader()
    return tot__ihw, masw__yxjro


def _add_categories_to_pq_dataset(pq_dataset):
    import pyarrow as pa
    from mpi4py import MPI
    if len(pq_dataset.pieces) < 1:
        raise BodoError(
            'No pieces found in Parquet dataset. Cannot get read categorical values'
            )
    xzd__vze = pq_dataset.schema.to_arrow_schema()
    dqyt__cypgs = [c for c in xzd__vze.names if isinstance(xzd__vze.field(c
        ).type, pa.DictionaryType)]
    if len(dqyt__cypgs) == 0:
        pq_dataset._category_info = {}
        return
    xwxf__xkpvr = MPI.COMM_WORLD
    if bodo.get_rank() == 0:
        try:
            briua__jbdan = pq_dataset.pieces[0].open()
            mus__cby = briua__jbdan.read_row_group(0, dqyt__cypgs)
            category_info = {c: tuple(mus__cby.column(c).chunk(0).
                dictionary.to_pylist()) for c in dqyt__cypgs}
            del briua__jbdan, mus__cby
        except Exception as mrrx__bcyks:
            xwxf__xkpvr.bcast(mrrx__bcyks)
            raise mrrx__bcyks
        xwxf__xkpvr.bcast(category_info)
    else:
        category_info = xwxf__xkpvr.bcast(None)
        if isinstance(category_info, Exception):
            rwxl__uuwk = category_info
            raise rwxl__uuwk
    pq_dataset._category_info = category_info


def get_pandas_metadata(schema, num_pieces):
    vvmhv__dllbp = None
    nullable_from_metadata = defaultdict(lambda : None)
    haop__pjav = b'pandas'
    if schema.metadata is not None and haop__pjav in schema.metadata:
        import json
        mztc__wrf = json.loads(schema.metadata[haop__pjav].decode('utf8'))
        vihe__jzy = len(mztc__wrf['index_columns'])
        if vihe__jzy > 1:
            raise BodoError('read_parquet: MultiIndex not supported yet')
        vvmhv__dllbp = mztc__wrf['index_columns'][0] if vihe__jzy else None
        if not isinstance(vvmhv__dllbp, str) and (not isinstance(
            vvmhv__dllbp, dict) or num_pieces != 1):
            vvmhv__dllbp = None
        for ugd__jklf in mztc__wrf['columns']:
            iogs__agq = ugd__jklf['name']
            if ugd__jklf['pandas_type'].startswith('int'
                ) and iogs__agq is not None:
                if ugd__jklf['numpy_type'].startswith('Int'):
                    nullable_from_metadata[iogs__agq] = True
                else:
                    nullable_from_metadata[iogs__agq] = False
    return vvmhv__dllbp, nullable_from_metadata


def parquet_file_schema(file_name, selected_columns, storage_options=None):
    col_names = []
    ceh__vmn = []
    pq_dataset = get_parquet_dataset(file_name, get_row_counts=False,
        storage_options=storage_options, read_categories=True)
    partition_names = [] if pq_dataset.partitions is None else [pq_dataset.
        partitions.levels[rkyu__vbhn].name for rkyu__vbhn in range(len(
        pq_dataset.partitions.partition_names))]
    xzd__vze = pq_dataset.schema.to_arrow_schema()
    num_pieces = len(pq_dataset.pieces)
    col_names = xzd__vze.names
    vvmhv__dllbp, nullable_from_metadata = get_pandas_metadata(xzd__vze,
        num_pieces)
    axf__cndj = [_get_numba_typ_from_pa_typ(xzd__vze.field(c), c ==
        vvmhv__dllbp, nullable_from_metadata[c], pq_dataset._category_info) for
        c in col_names]
    if partition_names:
        col_names += partition_names
        axf__cndj += [_get_partition_cat_dtype(pq_dataset.partitions.levels
            [rkyu__vbhn]) for rkyu__vbhn in range(len(partition_names))]
    dpwo__ftgq = {c: rkyu__vbhn for rkyu__vbhn, c in enumerate(col_names)}
    if selected_columns is None:
        selected_columns = col_names
    for c in selected_columns:
        if c not in dpwo__ftgq:
            raise BodoError(f'Selected column {c} not in Parquet file schema')
    if vvmhv__dllbp and not isinstance(vvmhv__dllbp, dict
        ) and vvmhv__dllbp not in selected_columns:
        selected_columns.append(vvmhv__dllbp)
    col_indices = [dpwo__ftgq[c] for c in selected_columns]
    ceh__vmn = [axf__cndj[dpwo__ftgq[c]] for c in selected_columns]
    col_names = selected_columns
    return col_names, ceh__vmn, vvmhv__dllbp, col_indices, partition_names


def _get_partition_cat_dtype(part_set):
    jin__gmenw = part_set.dictionary.to_pandas()
    rpd__eksu = bodo.typeof(jin__gmenw).dtype
    vqb__lychc = PDCategoricalDtype(tuple(jin__gmenw), rpd__eksu, False)
    return CategoricalArrayType(vqb__lychc)


_pq_read = types.ExternalFunction('pq_read', table_type(
    read_parquet_fpath_type, types.boolean, types.voidptr,
    parquet_predicate_type, parquet_predicate_type,
    storage_options_dict_type, types.int64, types.voidptr, types.int32,
    types.voidptr, types.voidptr, types.voidptr, types.int32, types.voidptr))
from llvmlite import ir as lir
from numba.core import cgutils
if bodo.utils.utils.has_pyarrow():
    from bodo.io import arrow_cpp
    ll.add_symbol('pq_read', arrow_cpp.pq_read)
    ll.add_symbol('pq_write', arrow_cpp.pq_write)
    ll.add_symbol('pq_write_partitioned', arrow_cpp.pq_write_partitioned)


@intrinsic
def parquet_write_table_cpp(typingctx, filename_t, table_t, col_names_t,
    index_t, write_index, metadata_t, compression_t, is_parallel_t,
    write_range_index, start, stop, step, name, bucket_region):

    def codegen(context, builder, sig, args):
        wsq__yyjm = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(1), lir.
            IntType(8).as_pointer(), lir.IntType(8).as_pointer(), lir.
            IntType(1), lir.IntType(1), lir.IntType(32), lir.IntType(32),
            lir.IntType(32), lir.IntType(8).as_pointer(), lir.IntType(8).
            as_pointer()])
        qlx__osbf = cgutils.get_or_insert_function(builder.module,
            wsq__yyjm, name='pq_write')
        builder.call(qlx__osbf, args)
        bodo.utils.utils.inlined_check_and_propagate_cpp_exception(context,
            builder)
    return types.void(types.voidptr, table_t, col_names_t, index_t, types.
        boolean, types.voidptr, types.voidptr, types.boolean, types.boolean,
        types.int32, types.int32, types.int32, types.voidptr, types.voidptr
        ), codegen


@intrinsic
def parquet_write_table_partitioned_cpp(typingctx, filename_t, data_table_t,
    col_names_t, col_names_no_partitions_t, cat_table_t, part_col_idxs_t,
    num_part_col_t, compression_t, is_parallel_t, bucket_region):

    def codegen(context, builder, sig, args):
        wsq__yyjm = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(32), lir
            .IntType(8).as_pointer(), lir.IntType(1), lir.IntType(8).
            as_pointer()])
        qlx__osbf = cgutils.get_or_insert_function(builder.module,
            wsq__yyjm, name='pq_write_partitioned')
        builder.call(qlx__osbf, args)
        bodo.utils.utils.inlined_check_and_propagate_cpp_exception(context,
            builder)
    return types.void(types.voidptr, data_table_t, col_names_t,
        col_names_no_partitions_t, cat_table_t, types.voidptr, types.int32,
        types.voidptr, types.boolean, types.voidptr), codegen
