"""
Implementation of pd.read_sql in BODO.
We piggyback on the pandas implementation. Future plan is to have a faster
version for this task.
"""
import numba
import numpy as np
import pandas as pd
from numba.core import ir, ir_utils, typeinfer, types
from numba.core.ir_utils import compile_to_numba_ir, replace_arg_nodes
import bodo
import bodo.ir.connector
from bodo import objmode
from bodo.ir.csv_ext import _get_dtype_str
from bodo.libs.array import delete_table, info_from_table, info_to_array, table_type
from bodo.libs.distributed_api import bcast, bcast_scalar
from bodo.libs.str_ext import string_type, unicode_to_utf8
from bodo.transforms import distributed_analysis, distributed_pass
from bodo.utils.typing import BodoError
from bodo.utils.utils import check_and_propagate_cpp_exception, sanitize_varname
MPI_ROOT = 0


class SqlReader(ir.Stmt):

    def __init__(self, sql_request, connection, df_out, df_colnames,
        out_vars, out_types, converted_colnames, db_type, loc):
        self.connector_typ = 'sql'
        self.sql_request = sql_request
        self.connection = connection
        self.df_out = df_out
        self.df_colnames = df_colnames
        self.out_vars = out_vars
        self.out_types = out_types
        self.converted_colnames = converted_colnames
        self.loc = loc
        self.limit = req_limit(sql_request)
        self.db_type = db_type
        self.filters = None

    def __repr__(self):
        return (
            '{} = ReadSql(sql_request={}, connection={}, col_names={}, types={}, vars={}, limit={})'
            .format(self.df_out, self.sql_request, self.connection, self.
            df_colnames, self.out_types, self.out_vars, self.limit))


def remove_dead_sql(sql_node, lives_no_aliases, lives, arg_aliases,
    alias_map, func_ir, typemap):
    hmgz__bswy = []
    kxu__okxqu = []
    awtbf__pha = []
    for aiu__uupaa, oci__jempx in enumerate(sql_node.out_vars):
        if oci__jempx.name in lives:
            hmgz__bswy.append(sql_node.df_colnames[aiu__uupaa])
            kxu__okxqu.append(sql_node.out_vars[aiu__uupaa])
            awtbf__pha.append(sql_node.out_types[aiu__uupaa])
    sql_node.df_colnames = hmgz__bswy
    sql_node.out_vars = kxu__okxqu
    sql_node.out_types = awtbf__pha
    if len(sql_node.out_vars) == 0:
        return None
    return sql_node


def sql_distributed_run(sql_node, array_dists, typemap, calltypes,
    typingctx, targetctx):
    parallel = False
    if array_dists is not None:
        parallel = True
        for dofoc__vop in sql_node.out_vars:
            if array_dists[dofoc__vop.name
                ] != distributed_pass.Distribution.OneD and array_dists[
                dofoc__vop.name] != distributed_pass.Distribution.OneD_Var:
                parallel = False
    vvqi__bwg = len(sql_node.out_vars)
    sqzv__kfal = ', '.join('arr' + str(aiu__uupaa) for aiu__uupaa in range(
        vvqi__bwg))
    twpd__okurq, cewee__qqqso = bodo.ir.connector.generate_filter_map(sql_node
        .filters)
    crnpi__yro = ', '.join(twpd__okurq.values())
    cuy__yuy = f'def sql_impl(sql_request, conn, {crnpi__yro}):\n'
    if sql_node.filters:
        kpfjq__lmglz = []
        for nnn__amaqy in sql_node.filters:
            rjuxz__mhahy = [' '.join(['(', ettq__nvjs[0], ettq__nvjs[1], 
                '{' + twpd__okurq[ettq__nvjs[2].name] + '}' if isinstance(
                ettq__nvjs[2], ir.Var) else ettq__nvjs[2], ')']) for
                ettq__nvjs in nnn__amaqy]
            kpfjq__lmglz.append(' ( ' + ' AND '.join(rjuxz__mhahy) + ' ) ')
        hprb__wex = ' WHERE ' + ' OR '.join(kpfjq__lmglz)
        for aiu__uupaa, xxpmp__oujn in enumerate(twpd__okurq.values()):
            cuy__yuy += f'    {xxpmp__oujn} = get_sql_literal({xxpmp__oujn})\n'
        cuy__yuy += f'    sql_request = f"{{sql_request}} {hprb__wex}"\n'
    cuy__yuy += '    ({},) = _sql_reader_py(sql_request, conn)\n'.format(
        sqzv__kfal)
    qar__kippd = {}
    exec(cuy__yuy, {}, qar__kippd)
    rxiw__wbo = qar__kippd['sql_impl']
    gdrei__xnsmg = _gen_sql_reader_py(sql_node.df_colnames, sql_node.
        out_types, typingctx, targetctx, sql_node.db_type, sql_node.limit,
        sql_node.converted_colnames, parallel)
    wrhhc__rysre = compile_to_numba_ir(rxiw__wbo, {'_sql_reader_py':
        gdrei__xnsmg, 'bcast_scalar': bcast_scalar, 'bcast': bcast,
        'get_sql_literal': _get_snowflake_sql_literal}, typingctx=typingctx,
        targetctx=targetctx, arg_typs=(string_type, string_type) + tuple(
        typemap[dofoc__vop.name] for dofoc__vop in cewee__qqqso), typemap=
        typemap, calltypes=calltypes).blocks.popitem()[1]
    hjh__bwt = escape_column_names(sql_node.df_colnames, sql_node.db_type,
        sql_node.converted_colnames)
    if sql_node.db_type == 'oracle':
        ehey__ral = ('SELECT ' + hjh__bwt + ' FROM (' + sql_node.
            sql_request + ') TEMP')
    else:
        ehey__ral = ('SELECT ' + hjh__bwt + ' FROM (' + sql_node.
            sql_request + ') as TEMP')
    replace_arg_nodes(wrhhc__rysre, [ir.Const(ehey__ral, sql_node.loc), ir.
        Const(sql_node.connection, sql_node.loc)] + cewee__qqqso)
    jbncr__smukh = wrhhc__rysre.body[:-3]
    for aiu__uupaa in range(len(sql_node.out_vars)):
        jbncr__smukh[-len(sql_node.out_vars) + aiu__uupaa
            ].target = sql_node.out_vars[aiu__uupaa]
    return jbncr__smukh


def escape_column_names(col_names, db_type, converted_colnames):
    if db_type in ('snowflake', 'oracle'):
        cusx__cex = [(zkg__qfq.upper() if zkg__qfq in converted_colnames else
            zkg__qfq) for zkg__qfq in col_names]
        hjh__bwt = ', '.join([f'"{zkg__qfq}"' for zkg__qfq in cusx__cex])
    elif db_type == 'mysql' or db_type == 'mysql+pymysql':
        hjh__bwt = ', '.join([f'`{zkg__qfq}`' for zkg__qfq in col_names])
    else:
        hjh__bwt = ', '.join([f'"{zkg__qfq}"' for zkg__qfq in col_names])
    return hjh__bwt


@numba.generated_jit(nopython=True, no_cpython_wrapper=True)
def _get_snowflake_sql_literal(filter_value):
    ysxj__uwumc = types.unliteral(filter_value)
    if ysxj__uwumc == types.unicode_type:
        return lambda filter_value: f'$${filter_value}$$'
    elif isinstance(ysxj__uwumc, (types.Integer, types.Float)
        ) or filter_value == types.bool_:
        return lambda filter_value: str(filter_value)
    elif ysxj__uwumc == bodo.pd_timestamp_type:

        def impl(filter_value):
            xmpzy__yesu = filter_value.nanosecond
            fbrxh__zwy = ''
            if xmpzy__yesu < 10:
                fbrxh__zwy = '00'
            elif xmpzy__yesu < 100:
                fbrxh__zwy = '0'
            return (
                f"timestamp '{filter_value.strftime('%Y-%m-%d %H:%M:%S.%f')}{fbrxh__zwy}{xmpzy__yesu}'"
                )
        return impl
    elif ysxj__uwumc == bodo.datetime_date_type:
        return (lambda filter_value:
            f"date '{filter_value.strftime('%Y-%m-%d')}'")
    else:
        raise BodoError(
            f'pd.read_sql(): Internal error, unsupported type {ysxj__uwumc} used in filter pushdown.'
            )


numba.parfors.array_analysis.array_analysis_extensions[SqlReader
    ] = bodo.ir.connector.connector_array_analysis
distributed_analysis.distributed_analysis_extensions[SqlReader
    ] = bodo.ir.connector.connector_distributed_analysis
typeinfer.typeinfer_extensions[SqlReader
    ] = bodo.ir.connector.connector_typeinfer
ir_utils.visit_vars_extensions[SqlReader
    ] = bodo.ir.connector.visit_vars_connector
ir_utils.remove_dead_extensions[SqlReader] = remove_dead_sql
numba.core.analysis.ir_extension_usedefs[SqlReader
    ] = bodo.ir.connector.connector_usedefs
ir_utils.copy_propagate_extensions[SqlReader
    ] = bodo.ir.connector.get_copies_connector
ir_utils.apply_copy_propagate_extensions[SqlReader
    ] = bodo.ir.connector.apply_copies_connector
ir_utils.build_defs_extensions[SqlReader
    ] = bodo.ir.connector.build_connector_definitions
distributed_pass.distributed_run_extensions[SqlReader] = sql_distributed_run
compiled_funcs = []


@numba.njit
def sqlalchemy_check():
    with numba.objmode():
        sqlalchemy_check_()


def sqlalchemy_check_():
    try:
        import sqlalchemy
    except ImportError as pwex__bnu:
        upvm__lpr = (
            "Using URI string without sqlalchemy installed. sqlalchemy can be installed by calling 'conda install -c conda-forge sqlalchemy'."
            )
        raise BodoError(upvm__lpr)


def req_limit(sql_request):
    import re
    jfeg__ufnrw = re.compile('LIMIT\\s+(\\d+)\\s*$', re.IGNORECASE)
    wwvc__knbh = jfeg__ufnrw.search(sql_request)
    if wwvc__knbh:
        return int(wwvc__knbh.group(1))
    else:
        return None


def _gen_sql_reader_py(col_names, col_typs, typingctx, targetctx, db_type,
    limit, converted_colnames, parallel):
    vgv__xuac = [sanitize_varname(yquvs__zpsos) for yquvs__zpsos in col_names]
    btzhi__ogbt = ["{}='{}'".format(hxyd__rjfz, _get_dtype_str(ock__uha)) for
        hxyd__rjfz, ock__uha in zip(vgv__xuac, col_typs)]
    if bodo.sql_access_method == 'multiple_access_by_block':
        cuy__yuy = 'def sql_reader_py(sql_request,conn):\n'
        cuy__yuy += '  sqlalchemy_check()\n'
        cuy__yuy += '  rank = bodo.libs.distributed_api.get_rank()\n'
        cuy__yuy += '  n_pes = bodo.libs.distributed_api.get_size()\n'
        cuy__yuy += '  with objmode({}):\n'.format(', '.join(btzhi__ogbt))
        cuy__yuy += '    list_df_block = []\n'
        cuy__yuy += '    block_size = 50000\n'
        cuy__yuy += '    iter = 0\n'
        cuy__yuy += '    while(True):\n'
        cuy__yuy += '      offset = (iter * n_pes + rank) * block_size\n'
        if db_type == 'oracle':
            cuy__yuy += """      sql_cons = 'select * from (' + sql_request + ') OFFSET ' + str(offset) + ' ROWS FETCH NEXT ' + str(block_size) + ' ROWS ONLY'
"""
        else:
            cuy__yuy += """      sql_cons = 'select * from (' + sql_request + ') x LIMIT ' + str(block_size) + ' OFFSET ' + str(offset)
"""
        cuy__yuy += '      df_block = pd.read_sql(sql_cons, conn)\n'
        cuy__yuy += '      if df_block.size == 0:\n'
        cuy__yuy += '        break\n'
        cuy__yuy += '      list_df_block.append(df_block)\n'
        cuy__yuy += '      iter += 1\n'
        cuy__yuy += '    df_ret = pd.concat(list_df_block)\n'
        for hxyd__rjfz, zze__mijxi in zip(vgv__xuac, col_names):
            cuy__yuy += "    {} = df_ret['{}'].values\n".format(hxyd__rjfz,
                zze__mijxi)
        cuy__yuy += '  return ({},)\n'.format(', '.join(sfyyr__aodh for
            sfyyr__aodh in vgv__xuac))
    if bodo.sql_access_method == 'multiple_access_nb_row_first':
        cuy__yuy = 'def sql_reader_py(sql_request, conn):\n'
        if db_type == 'snowflake':
            qiw__aejp = {}
            for aiu__uupaa, ossm__ski in enumerate(col_typs):
                qiw__aejp[f'col_{aiu__uupaa}_type'] = ossm__ski
            cuy__yuy += (
                f"  ev = bodo.utils.tracing.Event('read_snowflake', {parallel})\n"
                )

            def is_nullable(typ):
                return bodo.utils.utils.is_array_typ(typ, False
                    ) and not isinstance(typ, types.Array)
            urlgx__awhz = [int(is_nullable(ossm__ski)) for ossm__ski in
                col_typs]
            cuy__yuy += f"""  out_table = snowflake_read(unicode_to_utf8(sql_request), unicode_to_utf8(conn), {parallel}, {len(col_names)}, np.array({urlgx__awhz}, dtype=np.int32).ctypes)
"""
            cuy__yuy += '  check_and_propagate_cpp_exception()\n'
            for aiu__uupaa, wspz__vumcd in enumerate(vgv__xuac):
                cuy__yuy += f"""  {wspz__vumcd} = info_to_array(info_from_table(out_table, {aiu__uupaa}), col_{aiu__uupaa}_type)
"""
            cuy__yuy += '  delete_table(out_table)\n'
            cuy__yuy += f'  ev.finalize()\n'
        else:
            cuy__yuy += '  sqlalchemy_check()\n'
            if parallel:
                cuy__yuy += '  rank = bodo.libs.distributed_api.get_rank()\n'
                if limit is not None:
                    cuy__yuy += f'  nb_row = {limit}\n'
                else:
                    cuy__yuy += '  with objmode(nb_row="int64"):\n'
                    cuy__yuy += f'     if rank == {MPI_ROOT}:\n'
                    cuy__yuy += """         sql_cons = 'select count(*) from (' + sql_request + ') x'
"""
                    cuy__yuy += (
                        '         frame = pd.read_sql(sql_cons, conn)\n')
                    cuy__yuy += '         nb_row = frame.iat[0,0]\n'
                    cuy__yuy += '     else:\n'
                    cuy__yuy += '         nb_row = 0\n'
                    cuy__yuy += '  nb_row = bcast_scalar(nb_row)\n'
                cuy__yuy += '  with objmode({}):\n'.format(', '.join(
                    btzhi__ogbt))
                cuy__yuy += """    offset, limit = bodo.libs.distributed_api.get_start_count(nb_row)
"""
                hjh__bwt = escape_column_names(col_names, db_type,
                    converted_colnames)
                if db_type == 'oracle':
                    cuy__yuy += f"""    sql_cons = 'select {hjh__bwt} from (' + sql_request + ') OFFSET ' + str(offset) + ' ROWS FETCH NEXT ' + str(limit) + ' ROWS ONLY'
"""
                else:
                    cuy__yuy += f"""    sql_cons = 'select {hjh__bwt} from (' + sql_request + ') x LIMIT ' + str(limit) + ' OFFSET ' + str(offset)
"""
                cuy__yuy += '    df_ret = pd.read_sql(sql_cons, conn)\n'
            else:
                cuy__yuy += '  with objmode({}):\n'.format(', '.join(
                    btzhi__ogbt))
                cuy__yuy += '    df_ret = pd.read_sql(sql_request, conn)\n'
            for hxyd__rjfz, zze__mijxi in zip(vgv__xuac, col_names):
                cuy__yuy += "    {} = df_ret['{}'].values\n".format(hxyd__rjfz,
                    zze__mijxi)
        cuy__yuy += '  return ({},)\n'.format(', '.join(sfyyr__aodh for
            sfyyr__aodh in vgv__xuac))
    ufadx__nwloa = {'bodo': bodo}
    if db_type == 'snowflake':
        ufadx__nwloa.update(qiw__aejp)
        ufadx__nwloa.update({'np': np, 'unicode_to_utf8': unicode_to_utf8,
            'check_and_propagate_cpp_exception':
            check_and_propagate_cpp_exception, 'snowflake_read':
            _snowflake_read, 'info_to_array': info_to_array,
            'info_from_table': info_from_table, 'delete_table': delete_table})
    else:
        ufadx__nwloa.update({'sqlalchemy_check': sqlalchemy_check, 'pd': pd,
            'objmode': objmode, 'bcast_scalar': bcast_scalar})
    qar__kippd = {}
    exec(cuy__yuy, ufadx__nwloa, qar__kippd)
    gdrei__xnsmg = qar__kippd['sql_reader_py']
    svgs__pmn = numba.njit(gdrei__xnsmg)
    compiled_funcs.append(svgs__pmn)
    return svgs__pmn


_snowflake_read = types.ExternalFunction('snowflake_read', table_type(types
    .voidptr, types.voidptr, types.boolean, types.int64, types.voidptr))
import llvmlite.binding as ll
from bodo.io import arrow_cpp
ll.add_symbol('snowflake_read', arrow_cpp.snowflake_read)
