"""
Implements array operations for usage by DataFrames and Series
such as count and max.
"""
import numba
import numpy as np
import pandas as pd
from numba import generated_jit
from numba.core import types
from numba.extending import overload
import bodo
from bodo.hiframes.datetime_date_ext import datetime_date_array_type
from bodo.hiframes.pd_categorical_ext import CategoricalArrayType
from bodo.utils.typing import element_type, is_hashable_type, is_iterable_type, is_overload_true, is_overload_zero


@numba.njit
def array_op_median(arr, skipna=True, parallel=False):
    njcr__huc = np.empty(1, types.float64)
    bodo.libs.array_kernels.median_series_computation(njcr__huc.ctypes, arr,
        parallel, skipna)
    return njcr__huc[0]


def array_op_isna(arr):
    pass


@overload(array_op_isna)
def overload_array_op_isna(arr):

    def impl(arr):
        numba.parfors.parfor.init_prange()
        xljck__boksh = len(arr)
        nuz__yed = np.empty(xljck__boksh, np.bool_)
        for lpon__rvoql in numba.parfors.parfor.internal_prange(xljck__boksh):
            nuz__yed[lpon__rvoql] = bodo.libs.array_kernels.isna(arr,
                lpon__rvoql)
        return nuz__yed
    return impl


def array_op_count(arr):
    pass


@overload(array_op_count)
def overload_array_op_count(arr):

    def impl(arr):
        numba.parfors.parfor.init_prange()
        favd__jntwg = 0
        for lpon__rvoql in numba.parfors.parfor.internal_prange(len(arr)):
            fpc__iqvy = 0
            if not bodo.libs.array_kernels.isna(arr, lpon__rvoql):
                fpc__iqvy = 1
            favd__jntwg += fpc__iqvy
        njcr__huc = favd__jntwg
        return njcr__huc
    return impl


def array_op_describe(arr):
    pass


def array_op_describe_impl(arr):
    fmla__ryww = array_op_count(arr)
    kfe__cym = array_op_min(arr)
    kky__yhit = array_op_max(arr)
    eypz__nba = array_op_mean(arr)
    ymhup__akz = array_op_std(arr)
    prxy__kox = array_op_quantile(arr, 0.25)
    wxjvj__cuat = array_op_quantile(arr, 0.5)
    ear__uvdto = array_op_quantile(arr, 0.75)
    return (fmla__ryww, eypz__nba, ymhup__akz, kfe__cym, prxy__kox,
        wxjvj__cuat, ear__uvdto, kky__yhit)


def array_op_describe_dt_impl(arr):
    fmla__ryww = array_op_count(arr)
    kfe__cym = array_op_min(arr)
    kky__yhit = array_op_max(arr)
    eypz__nba = array_op_mean(arr)
    prxy__kox = array_op_quantile(arr, 0.25)
    wxjvj__cuat = array_op_quantile(arr, 0.5)
    ear__uvdto = array_op_quantile(arr, 0.75)
    return (fmla__ryww, eypz__nba, kfe__cym, prxy__kox, wxjvj__cuat,
        ear__uvdto, kky__yhit)


@overload(array_op_describe)
def overload_array_op_describe(arr):
    if arr.dtype == bodo.datetime64ns:
        return array_op_describe_dt_impl
    return array_op_describe_impl


def array_op_min(arr):
    pass


@overload(array_op_min)
def overload_array_op_min(arr):
    if arr.dtype == bodo.timedelta64ns:

        def impl_td64(arr):
            numba.parfors.parfor.init_prange()
            lqlzh__axeu = numba.cpython.builtins.get_type_max_value(np.int64)
            favd__jntwg = 0
            for lpon__rvoql in numba.parfors.parfor.internal_prange(len(arr)):
                xpcng__mjs = lqlzh__axeu
                fpc__iqvy = 0
                if not bodo.libs.array_kernels.isna(arr, lpon__rvoql):
                    xpcng__mjs = (bodo.hiframes.pd_timestamp_ext.
                        timedelta64_to_integer(arr[lpon__rvoql]))
                    fpc__iqvy = 1
                lqlzh__axeu = min(lqlzh__axeu, xpcng__mjs)
                favd__jntwg += fpc__iqvy
            return bodo.hiframes.pd_index_ext._tdi_val_finalize(lqlzh__axeu,
                favd__jntwg)
        return impl_td64
    if arr.dtype == bodo.datetime64ns:

        def impl_dt64(arr):
            numba.parfors.parfor.init_prange()
            lqlzh__axeu = numba.cpython.builtins.get_type_max_value(np.int64)
            favd__jntwg = 0
            for lpon__rvoql in numba.parfors.parfor.internal_prange(len(arr)):
                xpcng__mjs = lqlzh__axeu
                fpc__iqvy = 0
                if not bodo.libs.array_kernels.isna(arr, lpon__rvoql):
                    xpcng__mjs = (bodo.hiframes.pd_timestamp_ext.
                        dt64_to_integer(arr[lpon__rvoql]))
                    fpc__iqvy = 1
                lqlzh__axeu = min(lqlzh__axeu, xpcng__mjs)
                favd__jntwg += fpc__iqvy
            return bodo.hiframes.pd_index_ext._dti_val_finalize(lqlzh__axeu,
                favd__jntwg)
        return impl_dt64
    if isinstance(arr, CategoricalArrayType):

        def impl_cat(arr):
            txdv__nsj = (bodo.hiframes.pd_categorical_ext.
                get_categorical_arr_codes(arr))
            numba.parfors.parfor.init_prange()
            lqlzh__axeu = numba.cpython.builtins.get_type_max_value(np.int64)
            favd__jntwg = 0
            for lpon__rvoql in numba.parfors.parfor.internal_prange(len(
                txdv__nsj)):
                gdoxa__gncyk = txdv__nsj[lpon__rvoql]
                if gdoxa__gncyk == -1:
                    continue
                lqlzh__axeu = min(lqlzh__axeu, gdoxa__gncyk)
                favd__jntwg += 1
            njcr__huc = bodo.hiframes.series_kernels._box_cat_val(lqlzh__axeu,
                arr.dtype, favd__jntwg)
            return njcr__huc
        return impl_cat
    if arr == datetime_date_array_type:

        def impl_date(arr):
            numba.parfors.parfor.init_prange()
            lqlzh__axeu = bodo.hiframes.series_kernels._get_date_max_value()
            favd__jntwg = 0
            for lpon__rvoql in numba.parfors.parfor.internal_prange(len(arr)):
                xpcng__mjs = lqlzh__axeu
                fpc__iqvy = 0
                if not bodo.libs.array_kernels.isna(arr, lpon__rvoql):
                    xpcng__mjs = arr[lpon__rvoql]
                    fpc__iqvy = 1
                lqlzh__axeu = min(lqlzh__axeu, xpcng__mjs)
                favd__jntwg += fpc__iqvy
            njcr__huc = bodo.hiframes.series_kernels._sum_handle_nan(
                lqlzh__axeu, favd__jntwg)
            return njcr__huc
        return impl_date

    def impl(arr):
        numba.parfors.parfor.init_prange()
        lqlzh__axeu = bodo.hiframes.series_kernels._get_type_max_value(arr.
            dtype)
        favd__jntwg = 0
        for lpon__rvoql in numba.parfors.parfor.internal_prange(len(arr)):
            xpcng__mjs = lqlzh__axeu
            fpc__iqvy = 0
            if not bodo.libs.array_kernels.isna(arr, lpon__rvoql):
                xpcng__mjs = arr[lpon__rvoql]
                fpc__iqvy = 1
            lqlzh__axeu = min(lqlzh__axeu, xpcng__mjs)
            favd__jntwg += fpc__iqvy
        njcr__huc = bodo.hiframes.series_kernels._sum_handle_nan(lqlzh__axeu,
            favd__jntwg)
        return njcr__huc
    return impl


def array_op_max(arr):
    pass


@overload(array_op_max)
def overload_array_op_max(arr):
    if arr.dtype == bodo.timedelta64ns:

        def impl_td64(arr):
            numba.parfors.parfor.init_prange()
            lqlzh__axeu = numba.cpython.builtins.get_type_min_value(np.int64)
            favd__jntwg = 0
            for lpon__rvoql in numba.parfors.parfor.internal_prange(len(arr)):
                xpcng__mjs = lqlzh__axeu
                fpc__iqvy = 0
                if not bodo.libs.array_kernels.isna(arr, lpon__rvoql):
                    xpcng__mjs = (bodo.hiframes.pd_timestamp_ext.
                        timedelta64_to_integer(arr[lpon__rvoql]))
                    fpc__iqvy = 1
                lqlzh__axeu = max(lqlzh__axeu, xpcng__mjs)
                favd__jntwg += fpc__iqvy
            return bodo.hiframes.pd_index_ext._tdi_val_finalize(lqlzh__axeu,
                favd__jntwg)
        return impl_td64
    if arr.dtype == bodo.datetime64ns:

        def impl_dt64(arr):
            numba.parfors.parfor.init_prange()
            lqlzh__axeu = numba.cpython.builtins.get_type_min_value(np.int64)
            favd__jntwg = 0
            for lpon__rvoql in numba.parfors.parfor.internal_prange(len(arr)):
                xpcng__mjs = lqlzh__axeu
                fpc__iqvy = 0
                if not bodo.libs.array_kernels.isna(arr, lpon__rvoql):
                    xpcng__mjs = (bodo.hiframes.pd_timestamp_ext.
                        dt64_to_integer(arr[lpon__rvoql]))
                    fpc__iqvy = 1
                lqlzh__axeu = max(lqlzh__axeu, xpcng__mjs)
                favd__jntwg += fpc__iqvy
            return bodo.hiframes.pd_index_ext._dti_val_finalize(lqlzh__axeu,
                favd__jntwg)
        return impl_dt64
    if isinstance(arr, CategoricalArrayType):

        def impl_cat(arr):
            txdv__nsj = (bodo.hiframes.pd_categorical_ext.
                get_categorical_arr_codes(arr))
            numba.parfors.parfor.init_prange()
            lqlzh__axeu = -1
            for lpon__rvoql in numba.parfors.parfor.internal_prange(len(
                txdv__nsj)):
                lqlzh__axeu = max(lqlzh__axeu, txdv__nsj[lpon__rvoql])
            njcr__huc = bodo.hiframes.series_kernels._box_cat_val(lqlzh__axeu,
                arr.dtype, 1)
            return njcr__huc
        return impl_cat
    if arr == datetime_date_array_type:

        def impl_date(arr):
            numba.parfors.parfor.init_prange()
            lqlzh__axeu = bodo.hiframes.series_kernels._get_date_min_value()
            favd__jntwg = 0
            for lpon__rvoql in numba.parfors.parfor.internal_prange(len(arr)):
                xpcng__mjs = lqlzh__axeu
                fpc__iqvy = 0
                if not bodo.libs.array_kernels.isna(arr, lpon__rvoql):
                    xpcng__mjs = arr[lpon__rvoql]
                    fpc__iqvy = 1
                lqlzh__axeu = max(lqlzh__axeu, xpcng__mjs)
                favd__jntwg += fpc__iqvy
            njcr__huc = bodo.hiframes.series_kernels._sum_handle_nan(
                lqlzh__axeu, favd__jntwg)
            return njcr__huc
        return impl_date

    def impl(arr):
        numba.parfors.parfor.init_prange()
        lqlzh__axeu = bodo.hiframes.series_kernels._get_type_min_value(arr.
            dtype)
        favd__jntwg = 0
        for lpon__rvoql in numba.parfors.parfor.internal_prange(len(arr)):
            xpcng__mjs = lqlzh__axeu
            fpc__iqvy = 0
            if not bodo.libs.array_kernels.isna(arr, lpon__rvoql):
                xpcng__mjs = arr[lpon__rvoql]
                fpc__iqvy = 1
            lqlzh__axeu = max(lqlzh__axeu, xpcng__mjs)
            favd__jntwg += fpc__iqvy
        njcr__huc = bodo.hiframes.series_kernels._sum_handle_nan(lqlzh__axeu,
            favd__jntwg)
        return njcr__huc
    return impl


def array_op_mean(arr):
    pass


@overload(array_op_mean)
def overload_array_op_mean(arr):
    if arr.dtype == bodo.datetime64ns:

        def impl(arr):
            return pd.Timestamp(types.int64(bodo.libs.array_ops.
                array_op_mean(arr.view(np.int64))))
        return impl
    ixhz__aond = types.float64
    pzmps__hiknt = types.float64
    if isinstance(arr, types.Array) and arr.dtype == types.float32:
        ixhz__aond = types.float32
        pzmps__hiknt = types.float32
    bbigb__dvk = ixhz__aond(0)
    ajvkf__whbp = pzmps__hiknt(0)
    wsx__cojl = pzmps__hiknt(1)

    def impl(arr):
        numba.parfors.parfor.init_prange()
        lqlzh__axeu = bbigb__dvk
        favd__jntwg = ajvkf__whbp
        for lpon__rvoql in numba.parfors.parfor.internal_prange(len(arr)):
            xpcng__mjs = bbigb__dvk
            fpc__iqvy = ajvkf__whbp
            if not bodo.libs.array_kernels.isna(arr, lpon__rvoql):
                xpcng__mjs = arr[lpon__rvoql]
                fpc__iqvy = wsx__cojl
            lqlzh__axeu += xpcng__mjs
            favd__jntwg += fpc__iqvy
        njcr__huc = bodo.hiframes.series_kernels._mean_handle_nan(lqlzh__axeu,
            favd__jntwg)
        return njcr__huc
    return impl


def array_op_var(arr, skipna, ddof):
    pass


@overload(array_op_var)
def overload_array_op_var(arr, skipna, ddof):

    def impl(arr, skipna, ddof):
        numba.parfors.parfor.init_prange()
        urth__udn = 0.0
        ajss__zyzoi = 0.0
        favd__jntwg = 0
        for lpon__rvoql in numba.parfors.parfor.internal_prange(len(arr)):
            xpcng__mjs = 0.0
            fpc__iqvy = 0
            if not bodo.libs.array_kernels.isna(arr, lpon__rvoql
                ) or not skipna:
                xpcng__mjs = arr[lpon__rvoql]
                fpc__iqvy = 1
            urth__udn += xpcng__mjs
            ajss__zyzoi += xpcng__mjs * xpcng__mjs
            favd__jntwg += fpc__iqvy
        njcr__huc = bodo.hiframes.series_kernels._compute_var_nan_count_ddof(
            urth__udn, ajss__zyzoi, favd__jntwg, ddof)
        return njcr__huc
    return impl


def array_op_std(arr, skipna=True, ddof=1):
    pass


@overload(array_op_std)
def overload_array_op_std(arr, skipna=True, ddof=1):
    if arr.dtype == bodo.datetime64ns:

        def impl_dt64(arr, skipna=True, ddof=1):
            return pd.Timedelta(types.int64(array_op_var(arr.view(np.int64),
                skipna, ddof) ** 0.5))
        return impl_dt64
    return lambda arr, skipna=True, ddof=1: array_op_var(arr, skipna, ddof
        ) ** 0.5


def array_op_quantile(arr, q):
    pass


@overload(array_op_quantile)
def overload_array_op_quantile(arr, q):
    if is_iterable_type(q):
        if arr.dtype == bodo.datetime64ns:

            def _impl_list_dt(arr, q):
                nuz__yed = np.empty(len(q), np.int64)
                for lpon__rvoql in range(len(q)):
                    afj__wik = np.float64(q[lpon__rvoql])
                    nuz__yed[lpon__rvoql] = bodo.libs.array_kernels.quantile(
                        arr.view(np.int64), afj__wik)
                return nuz__yed.view(np.dtype('datetime64[ns]'))
            return _impl_list_dt

        def impl_list(arr, q):
            nuz__yed = np.empty(len(q), np.float64)
            for lpon__rvoql in range(len(q)):
                afj__wik = np.float64(q[lpon__rvoql])
                nuz__yed[lpon__rvoql] = bodo.libs.array_kernels.quantile(arr,
                    afj__wik)
            return nuz__yed
        return impl_list
    if arr.dtype == bodo.datetime64ns:

        def _impl_dt(arr, q):
            return pd.Timestamp(bodo.libs.array_kernels.quantile(arr.view(
                np.int64), np.float64(q)))
        return _impl_dt

    def impl(arr, q):
        return bodo.libs.array_kernels.quantile(arr, np.float64(q))
    return impl


def array_op_sum(arr, skipna, min_count):
    pass


@overload(array_op_sum, no_unliteral=True)
def overload_array_op_sum(arr, skipna, min_count):
    if isinstance(arr.dtype, types.Integer):
        isf__gsshi = types.intp
    elif arr.dtype == types.bool_:
        isf__gsshi = np.int64
    else:
        isf__gsshi = arr.dtype
    rajj__gbvr = isf__gsshi(0)
    if isinstance(arr.dtype, types.Float) and (not is_overload_true(skipna) or
        not is_overload_zero(min_count)):

        def impl(arr, skipna, min_count):
            numba.parfors.parfor.init_prange()
            lqlzh__axeu = rajj__gbvr
            xljck__boksh = len(arr)
            favd__jntwg = 0
            for lpon__rvoql in numba.parfors.parfor.internal_prange(
                xljck__boksh):
                xpcng__mjs = rajj__gbvr
                fpc__iqvy = 0
                if not bodo.libs.array_kernels.isna(arr, lpon__rvoql
                    ) or not skipna:
                    xpcng__mjs = arr[lpon__rvoql]
                    fpc__iqvy = 1
                lqlzh__axeu += xpcng__mjs
                favd__jntwg += fpc__iqvy
            njcr__huc = bodo.hiframes.series_kernels._var_handle_mincount(
                lqlzh__axeu, favd__jntwg, min_count)
            return njcr__huc
    else:

        def impl(arr, skipna, min_count):
            numba.parfors.parfor.init_prange()
            lqlzh__axeu = rajj__gbvr
            xljck__boksh = len(arr)
            for lpon__rvoql in numba.parfors.parfor.internal_prange(
                xljck__boksh):
                xpcng__mjs = rajj__gbvr
                if not bodo.libs.array_kernels.isna(arr, lpon__rvoql):
                    xpcng__mjs = arr[lpon__rvoql]
                lqlzh__axeu += xpcng__mjs
            return lqlzh__axeu
    return impl


def array_op_prod(arr, skipna, min_count):
    pass


@overload(array_op_prod)
def overload_array_op_prod(arr, skipna, min_count):
    mpbpw__frvu = arr.dtype(1)
    if arr.dtype == types.bool_:
        mpbpw__frvu = 1
    if isinstance(arr.dtype, types.Float):

        def impl(arr, skipna, min_count):
            numba.parfors.parfor.init_prange()
            lqlzh__axeu = mpbpw__frvu
            favd__jntwg = 0
            for lpon__rvoql in numba.parfors.parfor.internal_prange(len(arr)):
                xpcng__mjs = mpbpw__frvu
                fpc__iqvy = 0
                if not bodo.libs.array_kernels.isna(arr, lpon__rvoql
                    ) or not skipna:
                    xpcng__mjs = arr[lpon__rvoql]
                    fpc__iqvy = 1
                favd__jntwg += fpc__iqvy
                lqlzh__axeu *= xpcng__mjs
            njcr__huc = bodo.hiframes.series_kernels._var_handle_mincount(
                lqlzh__axeu, favd__jntwg, min_count)
            return njcr__huc
    else:

        def impl(arr, skipna, min_count):
            numba.parfors.parfor.init_prange()
            lqlzh__axeu = mpbpw__frvu
            for lpon__rvoql in numba.parfors.parfor.internal_prange(len(arr)):
                xpcng__mjs = mpbpw__frvu
                if not bodo.libs.array_kernels.isna(arr, lpon__rvoql):
                    xpcng__mjs = arr[lpon__rvoql]
                lqlzh__axeu *= xpcng__mjs
            return lqlzh__axeu
    return impl


def array_op_idxmax(arr, index):
    pass


@overload(array_op_idxmax, inline='always')
def overload_array_op_idxmax(arr, index):

    def impl(arr, index):
        lpon__rvoql = bodo.libs.array_kernels._nan_argmax(arr)
        return index[lpon__rvoql]
    return impl


def array_op_idxmin(arr, index):
    pass


@overload(array_op_idxmin, inline='always')
def overload_array_op_idxmin(arr, index):

    def impl(arr, index):
        lpon__rvoql = bodo.libs.array_kernels._nan_argmin(arr)
        return index[lpon__rvoql]
    return impl


def _convert_isin_values(values, use_hash_impl):
    pass


@overload(_convert_isin_values, no_unliteral=True)
def overload_convert_isin_values(values, use_hash_impl):
    if is_overload_true(use_hash_impl):

        def impl(values, use_hash_impl):
            cqjt__hyh = {}
            for kyv__vfl in values:
                cqjt__hyh[bodo.utils.conversion.box_if_dt64(kyv__vfl)] = 0
            return cqjt__hyh
        return impl
    else:

        def impl(values, use_hash_impl):
            return values
        return impl


def array_op_isin(arr, values):
    pass


@overload(array_op_isin, inline='always')
def overload_array_op_isin(arr, values):
    use_hash_impl = element_type(values) == element_type(arr
        ) and is_hashable_type(element_type(values))

    def impl(arr, values):
        values = bodo.libs.array_ops._convert_isin_values(values, use_hash_impl
            )
        numba.parfors.parfor.init_prange()
        xljck__boksh = len(arr)
        nuz__yed = np.empty(xljck__boksh, np.bool_)
        for lpon__rvoql in numba.parfors.parfor.internal_prange(xljck__boksh):
            nuz__yed[lpon__rvoql] = bodo.utils.conversion.box_if_dt64(arr[
                lpon__rvoql]) in values
        return nuz__yed
    return impl


@generated_jit(nopython=True)
def array_unique_vector_map(in_arr_tup):
    znwq__nosb = len(in_arr_tup) != 1
    ekb__jjhl = list(in_arr_tup.types)
    zhfe__hqvzh = 'def impl(in_arr_tup):\n'
    zhfe__hqvzh += '  n = len(in_arr_tup[0])\n'
    if znwq__nosb:
        ona__tkou = ', '.join([f'in_arr_tup[{lpon__rvoql}][unused]' for
            lpon__rvoql in range(len(in_arr_tup))])
        hucp__arkxt = ', '.join(['False' for tptt__jvbsq in range(len(
            in_arr_tup))])
        zhfe__hqvzh += f"""  arr_map = {{bodo.libs.nullable_tuple_ext.build_nullable_tuple(({ona__tkou},), ({hucp__arkxt},)): 0 for unused in range(0)}}
"""
        zhfe__hqvzh += '  map_vector = np.empty(n, np.int64)\n'
        for lpon__rvoql, bwt__qivmq in enumerate(ekb__jjhl):
            zhfe__hqvzh += f'  in_lst_{lpon__rvoql} = []\n'
            if bwt__qivmq == bodo.string_array_type:
                zhfe__hqvzh += f'  total_len_{lpon__rvoql} = 0\n'
            zhfe__hqvzh += f'  null_in_lst_{lpon__rvoql} = []\n'
        zhfe__hqvzh += '  for i in range(n):\n'
        jdv__xqd = ', '.join([f'in_arr_tup[{lpon__rvoql}][i]' for
            lpon__rvoql in range(len(ekb__jjhl))])
        zamvl__ocsi = ', '.join([
            f'bodo.libs.array_kernels.isna(in_arr_tup[{lpon__rvoql}], i)' for
            lpon__rvoql in range(len(ekb__jjhl))])
        zhfe__hqvzh += f"""    data_val = bodo.libs.nullable_tuple_ext.build_nullable_tuple(({jdv__xqd},), ({zamvl__ocsi},))
"""
        zhfe__hqvzh += '    if data_val not in arr_map:\n'
        zhfe__hqvzh += '      set_val = len(arr_map)\n'
        zhfe__hqvzh += '      values_tup = data_val._data\n'
        zhfe__hqvzh += '      nulls_tup = data_val._null_values\n'
        for lpon__rvoql, bwt__qivmq in enumerate(ekb__jjhl):
            zhfe__hqvzh += (
                f'      in_lst_{lpon__rvoql}.append(values_tup[{lpon__rvoql}])\n'
                )
            zhfe__hqvzh += (
                f'      null_in_lst_{lpon__rvoql}.append(nulls_tup[{lpon__rvoql}])\n'
                )
            if bwt__qivmq == bodo.string_array_type:
                zhfe__hqvzh += f"""      total_len_{lpon__rvoql}  += nulls_tup[{lpon__rvoql}] * len(values_tup[{lpon__rvoql}])
"""
        zhfe__hqvzh += '      arr_map[data_val] = len(arr_map)\n'
        zhfe__hqvzh += '    else:\n'
        zhfe__hqvzh += '      set_val = arr_map[data_val]\n'
        zhfe__hqvzh += '    map_vector[i] = set_val\n'
        zhfe__hqvzh += '  n_rows = len(arr_map)\n'
        for lpon__rvoql, bwt__qivmq in enumerate(ekb__jjhl):
            if bwt__qivmq == bodo.string_array_type:
                zhfe__hqvzh += f"""  out_arr_{lpon__rvoql} = bodo.libs.str_arr_ext.pre_alloc_string_array(n_rows, total_len_{lpon__rvoql})
"""
            else:
                zhfe__hqvzh += f"""  out_arr_{lpon__rvoql} = bodo.utils.utils.alloc_type(n_rows, in_arr_tup[{lpon__rvoql}], (-1,))
"""
        zhfe__hqvzh += '  for j in range(len(arr_map)):\n'
        for lpon__rvoql in range(len(ekb__jjhl)):
            zhfe__hqvzh += f'    if null_in_lst_{lpon__rvoql}[j]:\n'
            zhfe__hqvzh += (
                f'      bodo.libs.array_kernels.setna(out_arr_{lpon__rvoql}, j)\n'
                )
            zhfe__hqvzh += '    else:\n'
            zhfe__hqvzh += (
                f'      out_arr_{lpon__rvoql}[j] = in_lst_{lpon__rvoql}[j]\n')
        uwv__jgkhr = ', '.join([f'out_arr_{lpon__rvoql}' for lpon__rvoql in
            range(len(ekb__jjhl))])
        zhfe__hqvzh += f'  return ({uwv__jgkhr},), map_vector\n'
    else:
        zhfe__hqvzh += '  in_arr = in_arr_tup[0]\n'
        zhfe__hqvzh += (
            f'  arr_map = {{in_arr[unused]: 0 for unused in range(0)}}\n')
        zhfe__hqvzh += '  map_vector = np.empty(n, np.int64)\n'
        zhfe__hqvzh += '  is_na = 0\n'
        zhfe__hqvzh += '  in_lst = []\n'
        if ekb__jjhl[0] == bodo.string_array_type:
            zhfe__hqvzh += '  total_len = 0\n'
        zhfe__hqvzh += '  for i in range(n):\n'
        zhfe__hqvzh += '    if bodo.libs.array_kernels.isna(in_arr, i):\n'
        zhfe__hqvzh += '      is_na = 1\n'
        zhfe__hqvzh += (
            '      # Always put NA in the last location. We can safely use\n')
        zhfe__hqvzh += (
            '      # -1 because in_arr[-1] == in_arr[len(in_arr) - 1]\n')
        zhfe__hqvzh += '      set_val = -1\n'
        zhfe__hqvzh += '    else:\n'
        zhfe__hqvzh += '      data_val = in_arr[i]\n'
        zhfe__hqvzh += '      if data_val not in arr_map:\n'
        zhfe__hqvzh += '        set_val = len(arr_map)\n'
        zhfe__hqvzh += '        in_lst.append(data_val)\n'
        if ekb__jjhl[0] == bodo.string_array_type:
            zhfe__hqvzh += '        total_len += len(data_val)\n'
        zhfe__hqvzh += '        arr_map[data_val] = len(arr_map)\n'
        zhfe__hqvzh += '      else:\n'
        zhfe__hqvzh += '        set_val = arr_map[data_val]\n'
        zhfe__hqvzh += '    map_vector[i] = set_val\n'
        zhfe__hqvzh += '  n_rows = len(arr_map) + is_na\n'
        if ekb__jjhl[0] == bodo.string_array_type:
            zhfe__hqvzh += """  out_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(n_rows, total_len)
"""
        else:
            zhfe__hqvzh += (
                '  out_arr = bodo.utils.utils.alloc_type(n_rows, in_arr, (-1,))\n'
                )
        zhfe__hqvzh += '  for j in range(len(arr_map)):\n'
        zhfe__hqvzh += '    out_arr[j] = in_lst[j]\n'
        zhfe__hqvzh += '  if is_na:\n'
        zhfe__hqvzh += (
            '    bodo.libs.array_kernels.setna(out_arr, n_rows - 1)\n')
        zhfe__hqvzh += f'  return (out_arr,), map_vector\n'
    holkp__dnr = {}
    exec(zhfe__hqvzh, {'bodo': bodo, 'np': np}, holkp__dnr)
    impl = holkp__dnr['impl']
    return impl
