import atexit
import datetime
import operator
import sys
import time
import warnings
from collections import defaultdict
from decimal import Decimal
from enum import Enum
import llvmlite.binding as ll
import numba
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from mpi4py import MPI
from numba.core import cgutils, ir_utils, types
from numba.core.typing import signature
from numba.core.typing.builtins import IndexValueType
from numba.core.typing.templates import AbstractTemplate, infer_global
from numba.extending import intrinsic, models, overload, register_jitable, register_model
from numba.parfors.array_analysis import ArrayAnalysis
import bodo
from bodo.hiframes.datetime_date_ext import datetime_date_array_type
from bodo.hiframes.datetime_timedelta_ext import datetime_timedelta_array_type
from bodo.hiframes.pd_categorical_ext import CategoricalArrayType
from bodo.libs import hdist
from bodo.libs.array_item_arr_ext import ArrayItemArrayType, np_offset_type, offset_type
from bodo.libs.binary_arr_ext import binary_array_type
from bodo.libs.bool_arr_ext import boolean_array
from bodo.libs.decimal_arr_ext import Decimal128Type, DecimalArrayType
from bodo.libs.int_arr_ext import IntegerArrayType, set_bit_to_arr
from bodo.libs.interval_arr_ext import IntervalArrayType
from bodo.libs.map_arr_ext import MapArrayType
from bodo.libs.str_arr_ext import convert_len_arr_to_offset, get_bit_bitmap, get_data_ptr, get_null_bitmap_ptr, get_offset_ptr, num_total_chars, pre_alloc_string_array, set_bit_to, string_array_type
from bodo.libs.struct_arr_ext import StructArrayType
from bodo.libs.tuple_arr_ext import TupleArrayType
from bodo.utils.typing import BodoError, BodoWarning, is_overload_false, is_overload_none, raise_bodo_error
from bodo.utils.utils import CTypeEnum, check_and_propagate_cpp_exception, empty_like_type, numba_to_c_type, tuple_to_scalar
ll.add_symbol('dist_get_time', hdist.dist_get_time)
ll.add_symbol('get_time', hdist.get_time)
ll.add_symbol('dist_reduce', hdist.dist_reduce)
ll.add_symbol('dist_arr_reduce', hdist.dist_arr_reduce)
ll.add_symbol('dist_exscan', hdist.dist_exscan)
ll.add_symbol('dist_irecv', hdist.dist_irecv)
ll.add_symbol('dist_isend', hdist.dist_isend)
ll.add_symbol('dist_wait', hdist.dist_wait)
ll.add_symbol('dist_get_item_pointer', hdist.dist_get_item_pointer)
ll.add_symbol('get_dummy_ptr', hdist.get_dummy_ptr)
ll.add_symbol('allgather', hdist.allgather)
ll.add_symbol('comm_req_alloc', hdist.comm_req_alloc)
ll.add_symbol('comm_req_dealloc', hdist.comm_req_dealloc)
ll.add_symbol('req_array_setitem', hdist.req_array_setitem)
ll.add_symbol('dist_waitall', hdist.dist_waitall)
ll.add_symbol('oneD_reshape_shuffle', hdist.oneD_reshape_shuffle)
ll.add_symbol('permutation_int', hdist.permutation_int)
ll.add_symbol('permutation_array_index', hdist.permutation_array_index)
ll.add_symbol('c_get_rank', hdist.dist_get_rank)
ll.add_symbol('c_get_size', hdist.dist_get_size)
ll.add_symbol('c_barrier', hdist.barrier)
ll.add_symbol('c_alltoall', hdist.c_alltoall)
ll.add_symbol('c_gather_scalar', hdist.c_gather_scalar)
ll.add_symbol('c_gatherv', hdist.c_gatherv)
ll.add_symbol('c_scatterv', hdist.c_scatterv)
ll.add_symbol('c_allgatherv', hdist.c_allgatherv)
ll.add_symbol('c_bcast', hdist.c_bcast)
ll.add_symbol('c_recv', hdist.dist_recv)
ll.add_symbol('c_send', hdist.dist_send)
mpi_req_numba_type = getattr(types, 'int' + str(8 * hdist.mpi_req_num_bytes))
MPI_ROOT = 0
ANY_SOURCE = np.int32(hdist.ANY_SOURCE)


class Reduce_Type(Enum):
    Sum = 0
    Prod = 1
    Min = 2
    Max = 3
    Argmin = 4
    Argmax = 5
    Or = 6
    Concat = 7
    No_Op = 8


_get_rank = types.ExternalFunction('c_get_rank', types.int32())
_get_size = types.ExternalFunction('c_get_size', types.int32())
_barrier = types.ExternalFunction('c_barrier', types.int32())


@numba.njit
def get_rank():
    return _get_rank()


@numba.njit
def get_size():
    return _get_size()


@numba.njit
def barrier():
    _barrier()


_get_time = types.ExternalFunction('get_time', types.float64())
dist_time = types.ExternalFunction('dist_get_time', types.float64())


@overload(time.time, no_unliteral=True)
def overload_time_time():
    return lambda : _get_time()


@numba.generated_jit(nopython=True)
def get_type_enum(arr):
    arr = arr.instance_type if isinstance(arr, types.TypeRef) else arr
    dtype = arr.dtype
    if isinstance(dtype, bodo.hiframes.pd_categorical_ext.PDCategoricalDtype):
        dtype = bodo.hiframes.pd_categorical_ext.get_categories_int_type(dtype)
    typ_val = numba_to_c_type(dtype)
    return lambda arr: np.int32(typ_val)


INT_MAX = np.iinfo(np.int32).max
_send = types.ExternalFunction('c_send', types.void(types.voidptr, types.
    int32, types.int32, types.int32, types.int32))


@numba.njit
def send(val, rank, tag):
    send_arr = np.full(1, val)
    sch__occf = get_type_enum(send_arr)
    _send(send_arr.ctypes, 1, sch__occf, rank, tag)


_recv = types.ExternalFunction('c_recv', types.void(types.voidptr, types.
    int32, types.int32, types.int32, types.int32))


@numba.njit
def recv(dtype, rank, tag):
    recv_arr = np.empty(1, dtype)
    sch__occf = get_type_enum(recv_arr)
    _recv(recv_arr.ctypes, 1, sch__occf, rank, tag)
    return recv_arr[0]


_isend = types.ExternalFunction('dist_isend', mpi_req_numba_type(types.
    voidptr, types.int32, types.int32, types.int32, types.int32, types.bool_))


@numba.generated_jit(nopython=True)
def isend(arr, size, pe, tag, cond=True):
    if isinstance(arr, types.Array):

        def impl(arr, size, pe, tag, cond=True):
            sch__occf = get_type_enum(arr)
            return _isend(arr.ctypes, size, sch__occf, pe, tag, cond)
        return impl
    if isinstance(arr, (IntegerArrayType, DecimalArrayType)) or arr in (
        boolean_array, datetime_date_array_type):
        sch__occf = np.int32(numba_to_c_type(arr.dtype))
        mgfar__mzdui = np.int32(numba_to_c_type(types.uint8))

        def impl_nullable(arr, size, pe, tag, cond=True):
            eskss__qqbz = size + 7 >> 3
            rnbd__ktgi = _isend(arr._data.ctypes, size, sch__occf, pe, tag,
                cond)
            hpoca__rcpj = _isend(arr._null_bitmap.ctypes, eskss__qqbz,
                mgfar__mzdui, pe, tag, cond)
            return rnbd__ktgi, hpoca__rcpj
        return impl_nullable
    if arr in [binary_array_type, string_array_type]:
        hfp__jtfqs = np.int32(numba_to_c_type(offset_type))
        mgfar__mzdui = np.int32(numba_to_c_type(types.uint8))

        def impl_str_arr(arr, size, pe, tag, cond=True):
            yhlr__hwylu = np.int64(bodo.libs.str_arr_ext.num_total_chars(arr))
            send(yhlr__hwylu, pe, tag - 1)
            eskss__qqbz = size + 7 >> 3
            _send(bodo.libs.str_arr_ext.get_offset_ptr(arr), size + 1,
                hfp__jtfqs, pe, tag)
            _send(bodo.libs.str_arr_ext.get_data_ptr(arr), yhlr__hwylu,
                mgfar__mzdui, pe, tag)
            _send(bodo.libs.str_arr_ext.get_null_bitmap_ptr(arr),
                eskss__qqbz, mgfar__mzdui, pe, tag)
            return None
        return impl_str_arr
    typ_enum = numba_to_c_type(types.uint8)

    def impl_voidptr(arr, size, pe, tag, cond=True):
        return _isend(arr, size, typ_enum, pe, tag, cond)
    return impl_voidptr


_irecv = types.ExternalFunction('dist_irecv', mpi_req_numba_type(types.
    voidptr, types.int32, types.int32, types.int32, types.int32, types.bool_))


@numba.generated_jit(nopython=True)
def irecv(arr, size, pe, tag, cond=True):
    if isinstance(arr, types.Array):

        def impl(arr, size, pe, tag, cond=True):
            sch__occf = get_type_enum(arr)
            return _irecv(arr.ctypes, size, sch__occf, pe, tag, cond)
        return impl
    if isinstance(arr, (IntegerArrayType, DecimalArrayType)) or arr in (
        boolean_array, datetime_date_array_type):
        sch__occf = np.int32(numba_to_c_type(arr.dtype))
        mgfar__mzdui = np.int32(numba_to_c_type(types.uint8))

        def impl_nullable(arr, size, pe, tag, cond=True):
            eskss__qqbz = size + 7 >> 3
            rnbd__ktgi = _irecv(arr._data.ctypes, size, sch__occf, pe, tag,
                cond)
            hpoca__rcpj = _irecv(arr._null_bitmap.ctypes, eskss__qqbz,
                mgfar__mzdui, pe, tag, cond)
            return rnbd__ktgi, hpoca__rcpj
        return impl_nullable
    if arr in [binary_array_type, string_array_type]:
        hfp__jtfqs = np.int32(numba_to_c_type(offset_type))
        mgfar__mzdui = np.int32(numba_to_c_type(types.uint8))
        if arr == binary_array_type:
            mtb__qvt = 'bodo.libs.binary_arr_ext.pre_alloc_binary_array'
        else:
            mtb__qvt = 'bodo.libs.str_arr_ext.pre_alloc_string_array'
        inakp__umpn = f"""def impl(arr, size, pe, tag, cond=True):
            # recv the number of string characters and resize buffer to proper size
            n_chars = bodo.libs.distributed_api.recv(np.int64, pe, tag - 1)
            new_arr = {mtb__qvt}(size, n_chars)
            bodo.libs.str_arr_ext.move_str_binary_arr_payload(arr, new_arr)

            n_bytes = (size + 7) >> 3
            bodo.libs.distributed_api._recv(
                bodo.libs.str_arr_ext.get_offset_ptr(arr),
                size + 1,
                offset_typ_enum,
                pe,
                tag,
            )
            bodo.libs.distributed_api._recv(
                bodo.libs.str_arr_ext.get_data_ptr(arr), n_chars, char_typ_enum, pe, tag
            )
            bodo.libs.distributed_api._recv(
                bodo.libs.str_arr_ext.get_null_bitmap_ptr(arr),
                n_bytes,
                char_typ_enum,
                pe,
                tag,
            )
            return None"""
        lpgl__akc = dict()
        exec(inakp__umpn, {'bodo': bodo, 'np': np, 'offset_typ_enum':
            hfp__jtfqs, 'char_typ_enum': mgfar__mzdui}, lpgl__akc)
        impl = lpgl__akc['impl']
        return impl
    raise BodoError(f'irecv(): array type {arr} not supported yet')


_alltoall = types.ExternalFunction('c_alltoall', types.void(types.voidptr,
    types.voidptr, types.int32, types.int32))


@numba.njit
def alltoall(send_arr, recv_arr, count):
    assert count < INT_MAX
    sch__occf = get_type_enum(send_arr)
    _alltoall(send_arr.ctypes, recv_arr.ctypes, np.int32(count), sch__occf)


@numba.generated_jit(nopython=True)
def gather_scalar(data, allgather=False, warn_if_rep=True, root=MPI_ROOT):
    data = types.unliteral(data)
    typ_val = numba_to_c_type(data)
    dtype = data

    def gather_scalar_impl(data, allgather=False, warn_if_rep=True, root=
        MPI_ROOT):
        n_pes = bodo.libs.distributed_api.get_size()
        rank = bodo.libs.distributed_api.get_rank()
        send = np.full(1, data, dtype)
        mwkgk__dqtpy = n_pes if rank == root or allgather else 0
        wzpx__tvf = np.empty(mwkgk__dqtpy, dtype)
        c_gather_scalar(send.ctypes, wzpx__tvf.ctypes, np.int32(typ_val),
            allgather, np.int32(root))
        return wzpx__tvf
    return gather_scalar_impl


c_gather_scalar = types.ExternalFunction('c_gather_scalar', types.void(
    types.voidptr, types.voidptr, types.int32, types.bool_, types.int32))
c_gatherv = types.ExternalFunction('c_gatherv', types.void(types.voidptr,
    types.int32, types.voidptr, types.voidptr, types.voidptr, types.int32,
    types.bool_, types.int32))
c_scatterv = types.ExternalFunction('c_scatterv', types.void(types.voidptr,
    types.voidptr, types.voidptr, types.voidptr, types.int32, types.int32))


@intrinsic
def value_to_ptr(typingctx, val_tp=None):

    def codegen(context, builder, sig, args):
        denop__yeb = cgutils.alloca_once(builder, args[0].type)
        builder.store(args[0], denop__yeb)
        return builder.bitcast(denop__yeb, lir.IntType(8).as_pointer())
    return types.voidptr(val_tp), codegen


@intrinsic
def load_val_ptr(typingctx, ptr_tp, val_tp=None):

    def codegen(context, builder, sig, args):
        denop__yeb = builder.bitcast(args[0], args[1].type.as_pointer())
        return builder.load(denop__yeb)
    return val_tp(ptr_tp, val_tp), codegen


_dist_reduce = types.ExternalFunction('dist_reduce', types.void(types.
    voidptr, types.voidptr, types.int32, types.int32))
_dist_arr_reduce = types.ExternalFunction('dist_arr_reduce', types.void(
    types.voidptr, types.int64, types.int32, types.int32))


@numba.generated_jit(nopython=True)
def dist_reduce(value, reduce_op):
    if isinstance(value, types.Array):
        typ_enum = np.int32(numba_to_c_type(value.dtype))

        def impl_arr(value, reduce_op):
            A = np.ascontiguousarray(value)
            _dist_arr_reduce(A.ctypes, A.size, reduce_op, typ_enum)
            return A
        return impl_arr
    dldlc__koffo = types.unliteral(value)
    if isinstance(dldlc__koffo, IndexValueType):
        dldlc__koffo = dldlc__koffo.val_typ
        vjn__zjn = [types.bool_, types.uint8, types.int8, types.uint16,
            types.int16, types.uint32, types.int32, types.float32, types.
            float64]
        if not sys.platform.startswith('win'):
            vjn__zjn.append(types.int64)
            vjn__zjn.append(bodo.datetime64ns)
            vjn__zjn.append(bodo.timedelta64ns)
            vjn__zjn.append(bodo.datetime_date_type)
        if dldlc__koffo not in vjn__zjn:
            raise BodoError('argmin/argmax not supported for type {}'.
                format(dldlc__koffo))
    typ_enum = np.int32(numba_to_c_type(dldlc__koffo))

    def impl(value, reduce_op):
        hhqj__zdhe = value_to_ptr(value)
        tbwh__frn = value_to_ptr(value)
        _dist_reduce(hhqj__zdhe, tbwh__frn, reduce_op, typ_enum)
        return load_val_ptr(tbwh__frn, value)
    return impl


_dist_exscan = types.ExternalFunction('dist_exscan', types.void(types.
    voidptr, types.voidptr, types.int32, types.int32))


@numba.generated_jit(nopython=True)
def dist_exscan(value, reduce_op):
    dldlc__koffo = types.unliteral(value)
    typ_enum = np.int32(numba_to_c_type(dldlc__koffo))
    opb__ovaav = dldlc__koffo(0)

    def impl(value, reduce_op):
        hhqj__zdhe = value_to_ptr(value)
        tbwh__frn = value_to_ptr(opb__ovaav)
        _dist_exscan(hhqj__zdhe, tbwh__frn, reduce_op, typ_enum)
        return load_val_ptr(tbwh__frn, value)
    return impl


@numba.njit
def get_bit(bits, i):
    return bits[i >> 3] >> (i & 7) & 1


@numba.njit
def copy_gathered_null_bytes(null_bitmap_ptr, tmp_null_bytes,
    recv_counts_nulls, recv_counts):
    foqat__cgotx = 0
    embm__psc = 0
    for i in range(len(recv_counts)):
        svyn__xolv = recv_counts[i]
        eskss__qqbz = recv_counts_nulls[i]
        tpbfk__wagy = tmp_null_bytes[foqat__cgotx:foqat__cgotx + eskss__qqbz]
        for yiw__qxcq in range(svyn__xolv):
            set_bit_to(null_bitmap_ptr, embm__psc, get_bit(tpbfk__wagy,
                yiw__qxcq))
            embm__psc += 1
        foqat__cgotx += eskss__qqbz


@numba.generated_jit(nopython=True)
def gatherv(data, allgather=False, warn_if_rep=True, root=MPI_ROOT):
    from bodo.libs.csr_matrix_ext import CSRMatrixType
    bodo.hiframes.pd_dataframe_ext.check_runtime_cols_unsupported(data,
        'bodo.gatherv()')
    if isinstance(data, CategoricalArrayType):

        def impl_cat(data, allgather=False, warn_if_rep=True, root=MPI_ROOT):
            kkn__zkzgp = bodo.gatherv(data.codes, allgather, root=root)
            return bodo.hiframes.pd_categorical_ext.init_categorical_array(
                kkn__zkzgp, data.dtype)
        return impl_cat
    if isinstance(data, types.Array):
        typ_val = numba_to_c_type(data.dtype)

        def gatherv_impl(data, allgather=False, warn_if_rep=True, root=MPI_ROOT
            ):
            data = np.ascontiguousarray(data)
            rank = bodo.libs.distributed_api.get_rank()
            daqx__maivf = data.size
            recv_counts = gather_scalar(np.int32(daqx__maivf), allgather,
                root=root)
            ipb__udqey = recv_counts.sum()
            xvui__nqwib = empty_like_type(ipb__udqey, data)
            jpd__alai = np.empty(1, np.int32)
            if rank == root or allgather:
                jpd__alai = bodo.ir.join.calc_disp(recv_counts)
            c_gatherv(data.ctypes, np.int32(daqx__maivf), xvui__nqwib.
                ctypes, recv_counts.ctypes, jpd__alai.ctypes, np.int32(
                typ_val), allgather, np.int32(root))
            return xvui__nqwib.reshape((-1,) + data.shape[1:])
        return gatherv_impl
    if data == string_array_type:

        def gatherv_str_arr_impl(data, allgather=False, warn_if_rep=True,
            root=MPI_ROOT):
            xvui__nqwib = bodo.gatherv(data._data, allgather, warn_if_rep, root
                )
            return bodo.libs.str_arr_ext.init_str_arr(xvui__nqwib)
        return gatherv_str_arr_impl
    if data == binary_array_type:

        def gatherv_binary_arr_impl(data, allgather=False, warn_if_rep=True,
            root=MPI_ROOT):
            xvui__nqwib = bodo.gatherv(data._data, allgather, warn_if_rep, root
                )
            return bodo.libs.binary_arr_ext.init_binary_arr(xvui__nqwib)
        return gatherv_binary_arr_impl
    if data == datetime_timedelta_array_type:
        typ_val = numba_to_c_type(types.int64)
        mgfar__mzdui = np.int32(numba_to_c_type(types.uint8))

        def gatherv_impl_int_arr(data, allgather=False, warn_if_rep=True,
            root=MPI_ROOT):
            rank = bodo.libs.distributed_api.get_rank()
            daqx__maivf = len(data)
            eskss__qqbz = daqx__maivf + 7 >> 3
            recv_counts = gather_scalar(np.int32(daqx__maivf), allgather,
                root=root)
            ipb__udqey = recv_counts.sum()
            xvui__nqwib = empty_like_type(ipb__udqey, data)
            jpd__alai = np.empty(1, np.int32)
            recv_counts_nulls = np.empty(1, np.int32)
            mcu__cirkw = np.empty(1, np.int32)
            tmp_null_bytes = np.empty(1, np.uint8)
            if rank == root or allgather:
                jpd__alai = bodo.ir.join.calc_disp(recv_counts)
                recv_counts_nulls = np.empty(len(recv_counts), np.int32)
                for i in range(len(recv_counts)):
                    recv_counts_nulls[i] = recv_counts[i] + 7 >> 3
                mcu__cirkw = bodo.ir.join.calc_disp(recv_counts_nulls)
                tmp_null_bytes = np.empty(recv_counts_nulls.sum(), np.uint8)
            c_gatherv(data._days_data.ctypes, np.int32(daqx__maivf),
                xvui__nqwib._days_data.ctypes, recv_counts.ctypes,
                jpd__alai.ctypes, np.int32(typ_val), allgather, np.int32(root))
            c_gatherv(data._seconds_data.ctypes, np.int32(daqx__maivf),
                xvui__nqwib._seconds_data.ctypes, recv_counts.ctypes,
                jpd__alai.ctypes, np.int32(typ_val), allgather, np.int32(root))
            c_gatherv(data._microseconds_data.ctypes, np.int32(daqx__maivf),
                xvui__nqwib._microseconds_data.ctypes, recv_counts.ctypes,
                jpd__alai.ctypes, np.int32(typ_val), allgather, np.int32(root))
            c_gatherv(data._null_bitmap.ctypes, np.int32(eskss__qqbz),
                tmp_null_bytes.ctypes, recv_counts_nulls.ctypes, mcu__cirkw
                .ctypes, mgfar__mzdui, allgather, np.int32(root))
            copy_gathered_null_bytes(xvui__nqwib._null_bitmap.ctypes,
                tmp_null_bytes, recv_counts_nulls, recv_counts)
            return xvui__nqwib
        return gatherv_impl_int_arr
    if isinstance(data, (IntegerArrayType, DecimalArrayType)) or data in (
        boolean_array, datetime_date_array_type):
        typ_val = numba_to_c_type(data.dtype)
        mgfar__mzdui = np.int32(numba_to_c_type(types.uint8))

        def gatherv_impl_int_arr(data, allgather=False, warn_if_rep=True,
            root=MPI_ROOT):
            rank = bodo.libs.distributed_api.get_rank()
            daqx__maivf = len(data)
            eskss__qqbz = daqx__maivf + 7 >> 3
            recv_counts = gather_scalar(np.int32(daqx__maivf), allgather,
                root=root)
            ipb__udqey = recv_counts.sum()
            xvui__nqwib = empty_like_type(ipb__udqey, data)
            jpd__alai = np.empty(1, np.int32)
            recv_counts_nulls = np.empty(1, np.int32)
            mcu__cirkw = np.empty(1, np.int32)
            tmp_null_bytes = np.empty(1, np.uint8)
            if rank == root or allgather:
                jpd__alai = bodo.ir.join.calc_disp(recv_counts)
                recv_counts_nulls = np.empty(len(recv_counts), np.int32)
                for i in range(len(recv_counts)):
                    recv_counts_nulls[i] = recv_counts[i] + 7 >> 3
                mcu__cirkw = bodo.ir.join.calc_disp(recv_counts_nulls)
                tmp_null_bytes = np.empty(recv_counts_nulls.sum(), np.uint8)
            c_gatherv(data._data.ctypes, np.int32(daqx__maivf), xvui__nqwib
                ._data.ctypes, recv_counts.ctypes, jpd__alai.ctypes, np.
                int32(typ_val), allgather, np.int32(root))
            c_gatherv(data._null_bitmap.ctypes, np.int32(eskss__qqbz),
                tmp_null_bytes.ctypes, recv_counts_nulls.ctypes, mcu__cirkw
                .ctypes, mgfar__mzdui, allgather, np.int32(root))
            copy_gathered_null_bytes(xvui__nqwib._null_bitmap.ctypes,
                tmp_null_bytes, recv_counts_nulls, recv_counts)
            return xvui__nqwib
        return gatherv_impl_int_arr
    if isinstance(data, IntervalArrayType):

        def impl_interval_arr(data, allgather=False, warn_if_rep=True, root
            =MPI_ROOT):
            zhmj__wnbi = bodo.gatherv(data._left, allgather, warn_if_rep, root)
            assb__ixk = bodo.gatherv(data._right, allgather, warn_if_rep, root)
            return bodo.libs.interval_arr_ext.init_interval_array(zhmj__wnbi,
                assb__ixk)
        return impl_interval_arr
    if isinstance(data, bodo.hiframes.pd_series_ext.SeriesType):

        def impl(data, allgather=False, warn_if_rep=True, root=MPI_ROOT):
            arr = bodo.hiframes.pd_series_ext.get_series_data(data)
            index = bodo.hiframes.pd_series_ext.get_series_index(data)
            mqsgk__rsekc = bodo.hiframes.pd_series_ext.get_series_name(data)
            out_arr = bodo.libs.distributed_api.gatherv(arr, allgather,
                warn_if_rep, root)
            euu__jsi = bodo.gatherv(index, allgather, warn_if_rep, root)
            return bodo.hiframes.pd_series_ext.init_series(out_arr,
                euu__jsi, mqsgk__rsekc)
        return impl
    if isinstance(data, bodo.hiframes.pd_index_ext.RangeIndexType):
        izh__gbfkj = np.iinfo(np.int64).max
        tpk__mpe = np.iinfo(np.int64).min

        def impl_range_index(data, allgather=False, warn_if_rep=True, root=
            MPI_ROOT):
            start = data._start
            stop = data._stop
            if len(data) == 0:
                start = izh__gbfkj
                stop = tpk__mpe
            start = bodo.libs.distributed_api.dist_reduce(start, np.int32(
                Reduce_Type.Min.value))
            stop = bodo.libs.distributed_api.dist_reduce(stop, np.int32(
                Reduce_Type.Max.value))
            total_len = bodo.libs.distributed_api.dist_reduce(len(data), np
                .int32(Reduce_Type.Sum.value))
            if start == izh__gbfkj and stop == tpk__mpe:
                start = 0
                stop = 0
            ljt__owdmb = max(0, -(-(stop - start) // data._step))
            if ljt__owdmb < total_len:
                stop = start + data._step * total_len
            if bodo.get_rank() != root and not allgather:
                start = 0
                stop = 0
            return bodo.hiframes.pd_index_ext.init_range_index(start, stop,
                data._step, data._name)
        return impl_range_index
    if bodo.hiframes.pd_index_ext.is_pd_index_type(data):
        from bodo.hiframes.pd_index_ext import PeriodIndexType
        if isinstance(data, PeriodIndexType):
            zcjh__ejv = data.freq

            def impl_pd_index(data, allgather=False, warn_if_rep=True, root
                =MPI_ROOT):
                arr = bodo.libs.distributed_api.gatherv(data._data,
                    allgather, root=root)
                return bodo.hiframes.pd_index_ext.init_period_index(arr,
                    data._name, zcjh__ejv)
        else:

            def impl_pd_index(data, allgather=False, warn_if_rep=True, root
                =MPI_ROOT):
                arr = bodo.libs.distributed_api.gatherv(data._data,
                    allgather, root=root)
                return bodo.utils.conversion.index_from_array(arr, data._name)
        return impl_pd_index
    if isinstance(data, bodo.hiframes.pd_multi_index_ext.MultiIndexType):

        def impl_multi_index(data, allgather=False, warn_if_rep=True, root=
            MPI_ROOT):
            xvui__nqwib = bodo.gatherv(data._data, allgather, root=root)
            return bodo.hiframes.pd_multi_index_ext.init_multi_index(
                xvui__nqwib, data._names, data._name)
        return impl_multi_index
    if isinstance(data, bodo.hiframes.table.TableType):
        zhgu__ubbme = {'bodo': bodo, 'get_table_block': bodo.hiframes.table
            .get_table_block, 'ensure_column_unboxed': bodo.hiframes.table.
            ensure_column_unboxed, 'set_table_block': bodo.hiframes.table.
            set_table_block, 'set_table_len': bodo.hiframes.table.
            set_table_len, 'alloc_list_like': bodo.hiframes.table.
            alloc_list_like, 'init_table': bodo.hiframes.table.init_table}
        inakp__umpn = (
            f'def impl_table(data, allgather=False, warn_if_rep=True, root={MPI_ROOT}):\n'
            )
        inakp__umpn += '  T = data\n'
        inakp__umpn += '  T2 = init_table(T)\n'
        for fnim__yjdxp in data.type_to_blk.values():
            zhgu__ubbme[f'arr_inds_{fnim__yjdxp}'] = np.array(data.
                block_to_arr_ind[fnim__yjdxp], dtype=np.int64)
            inakp__umpn += (
                f'  arr_list_{fnim__yjdxp} = get_table_block(T, {fnim__yjdxp})\n'
                )
            inakp__umpn += f"""  out_arr_list_{fnim__yjdxp} = alloc_list_like(arr_list_{fnim__yjdxp})
"""
            inakp__umpn += f'  for i in range(len(arr_list_{fnim__yjdxp})):\n'
            inakp__umpn += (
                f'    arr_ind_{fnim__yjdxp} = arr_inds_{fnim__yjdxp}[i]\n')
            inakp__umpn += f"""    ensure_column_unboxed(T, arr_list_{fnim__yjdxp}, i, arr_ind_{fnim__yjdxp})
"""
            inakp__umpn += f"""    out_arr_{fnim__yjdxp} = bodo.gatherv(arr_list_{fnim__yjdxp}[i], allgather, warn_if_rep, root)
"""
            inakp__umpn += (
                f'    out_arr_list_{fnim__yjdxp}[i] = out_arr_{fnim__yjdxp}\n')
            inakp__umpn += f"""  T2 = set_table_block(T2, out_arr_list_{fnim__yjdxp}, {fnim__yjdxp})
"""
        inakp__umpn += (
            f'  length = T._len if bodo.get_rank() == root or allgather else 0\n'
            )
        inakp__umpn += f'  T2 = set_table_len(T2, length)\n'
        inakp__umpn += f'  return T2\n'
        lpgl__akc = {}
        exec(inakp__umpn, zhgu__ubbme, lpgl__akc)
        kmp__cfc = lpgl__akc['impl_table']
        return kmp__cfc
    if isinstance(data, bodo.hiframes.pd_dataframe_ext.DataFrameType):
        abcew__rrdu = len(data.columns)
        if abcew__rrdu == 0:
            return (lambda data, allgather=False, warn_if_rep=True, root=
                MPI_ROOT: bodo.hiframes.pd_dataframe_ext.init_dataframe((),
                bodo.hiframes.pd_dataframe_ext.get_dataframe_index(data), ()))
        hqjpi__yzmde = ', '.join(f'g_data_{i}' for i in range(abcew__rrdu))
        syt__fvrz = bodo.utils.transform.gen_const_tup(data.columns)
        inakp__umpn = (
            'def impl_df(data, allgather=False, warn_if_rep=True, root={}):\n'
            .format(MPI_ROOT))
        if data.is_table_format:
            from bodo.transforms.distributed_analysis import Distribution
            sev__kps = bodo.hiframes.pd_dataframe_ext.DataFrameType(data.
                data, data.index, data.columns, Distribution.REP, True)
            zhgu__ubbme = {'bodo': bodo, 'df_type': sev__kps}
            hqjpi__yzmde = 'T2'
            syt__fvrz = 'df_type'
            inakp__umpn += (
                '  T = bodo.hiframes.pd_dataframe_ext.get_dataframe_table(data)\n'
                )
            inakp__umpn += (
                '  T2 = bodo.gatherv(T, allgather, warn_if_rep, root)\n')
        else:
            zhgu__ubbme = {'bodo': bodo}
            for i in range(abcew__rrdu):
                inakp__umpn += (
                    """  data_{} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(data, {})
"""
                    .format(i, i))
                inakp__umpn += (
                    """  g_data_{} = bodo.gatherv(data_{}, allgather, warn_if_rep, root)
"""
                    .format(i, i))
        inakp__umpn += (
            '  index = bodo.hiframes.pd_dataframe_ext.get_dataframe_index(data)\n'
            )
        inakp__umpn += (
            '  g_index = bodo.gatherv(index, allgather, warn_if_rep, root)\n')
        inakp__umpn += (
            """  return bodo.hiframes.pd_dataframe_ext.init_dataframe(({},), g_index, {})
"""
            .format(hqjpi__yzmde, syt__fvrz))
        lpgl__akc = {}
        exec(inakp__umpn, zhgu__ubbme, lpgl__akc)
        dgt__qkc = lpgl__akc['impl_df']
        return dgt__qkc
    if isinstance(data, ArrayItemArrayType):
        hysi__kjlz = np.int32(numba_to_c_type(types.int32))
        mgfar__mzdui = np.int32(numba_to_c_type(types.uint8))

        def gatherv_array_item_arr_impl(data, allgather=False, warn_if_rep=
            True, root=MPI_ROOT):
            rank = bodo.libs.distributed_api.get_rank()
            uhe__nuo = bodo.libs.array_item_arr_ext.get_offsets(data)
            zoet__any = bodo.libs.array_item_arr_ext.get_data(data)
            qdh__nzwcg = bodo.libs.array_item_arr_ext.get_null_bitmap(data)
            daqx__maivf = len(data)
            qsfpo__tlc = np.empty(daqx__maivf, np.uint32)
            eskss__qqbz = daqx__maivf + 7 >> 3
            for i in range(daqx__maivf):
                qsfpo__tlc[i] = uhe__nuo[i + 1] - uhe__nuo[i]
            recv_counts = gather_scalar(np.int32(daqx__maivf), allgather,
                root=root)
            ipb__udqey = recv_counts.sum()
            jpd__alai = np.empty(1, np.int32)
            recv_counts_nulls = np.empty(1, np.int32)
            mcu__cirkw = np.empty(1, np.int32)
            tmp_null_bytes = np.empty(1, np.uint8)
            if rank == root or allgather:
                jpd__alai = bodo.ir.join.calc_disp(recv_counts)
                recv_counts_nulls = np.empty(len(recv_counts), np.int32)
                for nxd__iemn in range(len(recv_counts)):
                    recv_counts_nulls[nxd__iemn] = recv_counts[nxd__iemn
                        ] + 7 >> 3
                mcu__cirkw = bodo.ir.join.calc_disp(recv_counts_nulls)
                tmp_null_bytes = np.empty(recv_counts_nulls.sum(), np.uint8)
            azjmu__pixds = np.empty(ipb__udqey + 1, np.uint32)
            nzrd__otgia = bodo.gatherv(zoet__any, allgather, warn_if_rep, root)
            twl__tiqws = np.empty(ipb__udqey + 7 >> 3, np.uint8)
            c_gatherv(qsfpo__tlc.ctypes, np.int32(daqx__maivf),
                azjmu__pixds.ctypes, recv_counts.ctypes, jpd__alai.ctypes,
                hysi__kjlz, allgather, np.int32(root))
            c_gatherv(qdh__nzwcg.ctypes, np.int32(eskss__qqbz),
                tmp_null_bytes.ctypes, recv_counts_nulls.ctypes, mcu__cirkw
                .ctypes, mgfar__mzdui, allgather, np.int32(root))
            dummy_use(data)
            yuf__xetax = np.empty(ipb__udqey + 1, np.uint64)
            convert_len_arr_to_offset(azjmu__pixds.ctypes, yuf__xetax.
                ctypes, ipb__udqey)
            copy_gathered_null_bytes(twl__tiqws.ctypes, tmp_null_bytes,
                recv_counts_nulls, recv_counts)
            out_arr = bodo.libs.array_item_arr_ext.init_array_item_array(
                ipb__udqey, nzrd__otgia, yuf__xetax, twl__tiqws)
            return out_arr
        return gatherv_array_item_arr_impl
    if isinstance(data, StructArrayType):
        xzxym__exf = data.names
        mgfar__mzdui = np.int32(numba_to_c_type(types.uint8))

        def impl_struct_arr(data, allgather=False, warn_if_rep=True, root=
            MPI_ROOT):
            laepp__yumcq = bodo.libs.struct_arr_ext.get_data(data)
            cnv__vwod = bodo.libs.struct_arr_ext.get_null_bitmap(data)
            smyub__ajc = bodo.gatherv(laepp__yumcq, allgather=allgather,
                root=root)
            rank = bodo.libs.distributed_api.get_rank()
            daqx__maivf = len(data)
            eskss__qqbz = daqx__maivf + 7 >> 3
            recv_counts = gather_scalar(np.int32(daqx__maivf), allgather,
                root=root)
            ipb__udqey = recv_counts.sum()
            zbi__gvvl = np.empty(ipb__udqey + 7 >> 3, np.uint8)
            recv_counts_nulls = np.empty(1, np.int32)
            mcu__cirkw = np.empty(1, np.int32)
            tmp_null_bytes = np.empty(1, np.uint8)
            if rank == root or allgather:
                recv_counts_nulls = np.empty(len(recv_counts), np.int32)
                for i in range(len(recv_counts)):
                    recv_counts_nulls[i] = recv_counts[i] + 7 >> 3
                mcu__cirkw = bodo.ir.join.calc_disp(recv_counts_nulls)
                tmp_null_bytes = np.empty(recv_counts_nulls.sum(), np.uint8)
            c_gatherv(cnv__vwod.ctypes, np.int32(eskss__qqbz),
                tmp_null_bytes.ctypes, recv_counts_nulls.ctypes, mcu__cirkw
                .ctypes, mgfar__mzdui, allgather, np.int32(root))
            copy_gathered_null_bytes(zbi__gvvl.ctypes, tmp_null_bytes,
                recv_counts_nulls, recv_counts)
            return bodo.libs.struct_arr_ext.init_struct_arr(smyub__ajc,
                zbi__gvvl, xzxym__exf)
        return impl_struct_arr
    if data == binary_array_type:

        def impl_bin_arr(data, allgather=False, warn_if_rep=True, root=MPI_ROOT
            ):
            xvui__nqwib = bodo.gatherv(data._data, allgather, warn_if_rep, root
                )
            return bodo.libs.binary_arr_ext.init_binary_arr(xvui__nqwib)
        return impl_bin_arr
    if isinstance(data, TupleArrayType):

        def impl_tuple_arr(data, allgather=False, warn_if_rep=True, root=
            MPI_ROOT):
            xvui__nqwib = bodo.gatherv(data._data, allgather, warn_if_rep, root
                )
            return bodo.libs.tuple_arr_ext.init_tuple_arr(xvui__nqwib)
        return impl_tuple_arr
    if isinstance(data, MapArrayType):

        def impl_map_arr(data, allgather=False, warn_if_rep=True, root=MPI_ROOT
            ):
            xvui__nqwib = bodo.gatherv(data._data, allgather, warn_if_rep, root
                )
            return bodo.libs.map_arr_ext.init_map_arr(xvui__nqwib)
        return impl_map_arr
    if isinstance(data, CSRMatrixType):

        def impl_csr_matrix(data, allgather=False, warn_if_rep=True, root=
            MPI_ROOT):
            xvui__nqwib = bodo.gatherv(data.data, allgather, warn_if_rep, root)
            lhwcb__jxrzr = bodo.gatherv(data.indices, allgather,
                warn_if_rep, root)
            gsuid__lqacu = bodo.gatherv(data.indptr, allgather, warn_if_rep,
                root)
            dfgf__dmrig = gather_scalar(data.shape[0], allgather, root=root)
            fjhec__gbak = dfgf__dmrig.sum()
            abcew__rrdu = bodo.libs.distributed_api.dist_reduce(data.shape[
                1], np.int32(Reduce_Type.Max.value))
            ohu__cjjw = np.empty(fjhec__gbak + 1, np.int64)
            lhwcb__jxrzr = lhwcb__jxrzr.astype(np.int64)
            ohu__cjjw[0] = 0
            tsqj__whei = 1
            grqfy__ktkh = 0
            for tkki__puxpj in dfgf__dmrig:
                for pzlm__iwhc in range(tkki__puxpj):
                    kyd__vkeo = gsuid__lqacu[grqfy__ktkh + 1] - gsuid__lqacu[
                        grqfy__ktkh]
                    ohu__cjjw[tsqj__whei] = ohu__cjjw[tsqj__whei - 1
                        ] + kyd__vkeo
                    tsqj__whei += 1
                    grqfy__ktkh += 1
                grqfy__ktkh += 1
            return bodo.libs.csr_matrix_ext.init_csr_matrix(xvui__nqwib,
                lhwcb__jxrzr, ohu__cjjw, (fjhec__gbak, abcew__rrdu))
        return impl_csr_matrix
    if isinstance(data, types.BaseTuple):
        inakp__umpn = (
            'def impl_tuple(data, allgather=False, warn_if_rep=True, root={}):\n'
            .format(MPI_ROOT))
        inakp__umpn += '  return ({}{})\n'.format(', '.join(
            'bodo.gatherv(data[{}], allgather, warn_if_rep, root)'.format(i
            ) for i in range(len(data))), ',' if len(data) > 0 else '')
        lpgl__akc = {}
        exec(inakp__umpn, {'bodo': bodo}, lpgl__akc)
        cbmep__iqtb = lpgl__akc['impl_tuple']
        return cbmep__iqtb
    if data is types.none:
        return (lambda data, allgather=False, warn_if_rep=True, root=
            MPI_ROOT: None)
    raise BodoError('gatherv() not available for {}'.format(data))


@numba.generated_jit(nopython=True)
def rebalance(data, dests=None, random=False, random_seed=None, parallel=False
    ):
    bodo.hiframes.pd_dataframe_ext.check_runtime_cols_unsupported(data,
        'bodo.rebalance()')
    inakp__umpn = (
        'def impl(data, dests=None, random=False, random_seed=None, parallel=False):\n'
        )
    inakp__umpn += '    if random:\n'
    inakp__umpn += '        if random_seed is None:\n'
    inakp__umpn += '            random = 1\n'
    inakp__umpn += '        else:\n'
    inakp__umpn += '            random = 2\n'
    inakp__umpn += '    if random_seed is None:\n'
    inakp__umpn += '        random_seed = -1\n'
    if isinstance(data, bodo.hiframes.pd_dataframe_ext.DataFrameType):
        xwib__qwcgn = data
        abcew__rrdu = len(xwib__qwcgn.columns)
        for i in range(abcew__rrdu):
            inakp__umpn += f"""    data_{i} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(data, {i})
"""
        inakp__umpn += """    ind_arr = bodo.utils.conversion.index_to_array(bodo.hiframes.pd_dataframe_ext.get_dataframe_index(data))
"""
        hqjpi__yzmde = ', '.join(f'data_{i}' for i in range(abcew__rrdu))
        inakp__umpn += ('    info_list_total = [{}, array_to_info(ind_arr)]\n'
            .format(', '.join('array_to_info(data_{})'.format(tcz__ejld) for
            tcz__ejld in range(abcew__rrdu))))
        inakp__umpn += (
            '    table_total = arr_info_list_to_table(info_list_total)\n')
        inakp__umpn += '    if dests is None:\n'
        inakp__umpn += """        out_table = shuffle_renormalization(table_total, random, random_seed, parallel)
"""
        inakp__umpn += '    else:\n'
        inakp__umpn += """        out_table = shuffle_renormalization_group(table_total, random, random_seed, parallel, len(dests), np.array(dests, dtype=np.int32).ctypes)
"""
        for sey__ygk in range(abcew__rrdu):
            inakp__umpn += (
                """    out_arr_{0} = info_to_array(info_from_table(out_table, {0}), data_{0})
"""
                .format(sey__ygk))
        inakp__umpn += (
            """    out_arr_index = info_to_array(info_from_table(out_table, {}), ind_arr)
"""
            .format(abcew__rrdu))
        inakp__umpn += '    delete_table(out_table)\n'
        inakp__umpn += '    if parallel:\n'
        inakp__umpn += '        delete_table(table_total)\n'
        hqjpi__yzmde = ', '.join('out_arr_{}'.format(i) for i in range(
            abcew__rrdu))
        syt__fvrz = bodo.utils.transform.gen_const_tup(xwib__qwcgn.columns)
        index = 'bodo.utils.conversion.index_from_array(out_arr_index)'
        inakp__umpn += (
            '    return bodo.hiframes.pd_dataframe_ext.init_dataframe(({},), {}, {})\n'
            .format(hqjpi__yzmde, index, syt__fvrz))
    elif isinstance(data, bodo.hiframes.pd_series_ext.SeriesType):
        inakp__umpn += (
            '    data_0 = bodo.hiframes.pd_series_ext.get_series_data(data)\n')
        inakp__umpn += """    ind_arr = bodo.utils.conversion.index_to_array(bodo.hiframes.pd_series_ext.get_series_index(data))
"""
        inakp__umpn += (
            '    name = bodo.hiframes.pd_series_ext.get_series_name(data)\n')
        inakp__umpn += """    table_total = arr_info_list_to_table([array_to_info(data_0), array_to_info(ind_arr)])
"""
        inakp__umpn += '    if dests is None:\n'
        inakp__umpn += """        out_table = shuffle_renormalization(table_total, random, random_seed, parallel)
"""
        inakp__umpn += '    else:\n'
        inakp__umpn += """        out_table = shuffle_renormalization_group(table_total, random, random_seed, parallel, len(dests), np.array(dests, dtype=np.int32).ctypes)
"""
        inakp__umpn += (
            '    out_arr_0 = info_to_array(info_from_table(out_table, 0), data_0)\n'
            )
        inakp__umpn += """    out_arr_index = info_to_array(info_from_table(out_table, 1), ind_arr)
"""
        inakp__umpn += '    delete_table(out_table)\n'
        inakp__umpn += '    if parallel:\n'
        inakp__umpn += '        delete_table(table_total)\n'
        index = 'bodo.utils.conversion.index_from_array(out_arr_index)'
        inakp__umpn += f"""    return bodo.hiframes.pd_series_ext.init_series(out_arr_0, {index}, name)
"""
    elif isinstance(data, types.Array):
        assert is_overload_false(random
            ), 'Call random_shuffle instead of rebalance'
        inakp__umpn += '    if not parallel:\n'
        inakp__umpn += '        return data\n'
        inakp__umpn += """    dim0_global_size = bodo.libs.distributed_api.dist_reduce(data.shape[0], np.int32(bodo.libs.distributed_api.Reduce_Type.Sum.value))
"""
        inakp__umpn += '    if dests is None:\n'
        inakp__umpn += """        dim0_local_size = bodo.libs.distributed_api.get_node_portion(dim0_global_size, bodo.get_size(), bodo.get_rank())
"""
        inakp__umpn += '    elif bodo.get_rank() not in dests:\n'
        inakp__umpn += '        dim0_local_size = 0\n'
        inakp__umpn += '    else:\n'
        inakp__umpn += """        dim0_local_size = bodo.libs.distributed_api.get_node_portion(dim0_global_size, len(dests), dests.index(bodo.get_rank()))
"""
        inakp__umpn += """    out = np.empty((dim0_local_size,) + tuple(data.shape[1:]), dtype=data.dtype)
"""
        inakp__umpn += """    bodo.libs.distributed_api.dist_oneD_reshape_shuffle(out, data, dim0_global_size, dests)
"""
        inakp__umpn += '    return out\n'
    elif bodo.utils.utils.is_array_typ(data, False):
        inakp__umpn += (
            '    table_total = arr_info_list_to_table([array_to_info(data)])\n'
            )
        inakp__umpn += '    if dests is None:\n'
        inakp__umpn += """        out_table = shuffle_renormalization(table_total, random, random_seed, parallel)
"""
        inakp__umpn += '    else:\n'
        inakp__umpn += """        out_table = shuffle_renormalization_group(table_total, random, random_seed, parallel, len(dests), np.array(dests, dtype=np.int32).ctypes)
"""
        inakp__umpn += (
            '    out_arr = info_to_array(info_from_table(out_table, 0), data)\n'
            )
        inakp__umpn += '    delete_table(out_table)\n'
        inakp__umpn += '    if parallel:\n'
        inakp__umpn += '        delete_table(table_total)\n'
        inakp__umpn += '    return out_arr\n'
    else:
        raise BodoError(f'Type {data} not supported for bodo.rebalance')
    lpgl__akc = {}
    exec(inakp__umpn, {'np': np, 'bodo': bodo, 'array_to_info': bodo.libs.
        array.array_to_info, 'shuffle_renormalization': bodo.libs.array.
        shuffle_renormalization, 'shuffle_renormalization_group': bodo.libs
        .array.shuffle_renormalization_group, 'arr_info_list_to_table':
        bodo.libs.array.arr_info_list_to_table, 'info_from_table': bodo.
        libs.array.info_from_table, 'info_to_array': bodo.libs.array.
        info_to_array, 'delete_table': bodo.libs.array.delete_table}, lpgl__akc
        )
    impl = lpgl__akc['impl']
    return impl


@numba.generated_jit(nopython=True)
def random_shuffle(data, seed=None, dests=None, parallel=False):
    inakp__umpn = 'def impl(data, seed=None, dests=None, parallel=False):\n'
    if isinstance(data, types.Array):
        if not is_overload_none(dests):
            raise BodoError('not supported')
        inakp__umpn += '    if seed is None:\n'
        inakp__umpn += """        seed = bodo.libs.distributed_api.bcast_scalar(np.random.randint(0, 2**31))
"""
        inakp__umpn += '    np.random.seed(seed)\n'
        inakp__umpn += '    if not parallel:\n'
        inakp__umpn += '        data = data.copy()\n'
        inakp__umpn += '        np.random.shuffle(data)\n'
        inakp__umpn += '        return data\n'
        inakp__umpn += '    else:\n'
        inakp__umpn += """        dim0_global_size = bodo.libs.distributed_api.dist_reduce(data.shape[0], np.int32(bodo.libs.distributed_api.Reduce_Type.Sum.value))
"""
        inakp__umpn += '        permutation = np.arange(dim0_global_size)\n'
        inakp__umpn += '        np.random.shuffle(permutation)\n'
        inakp__umpn += """        dim0_local_size = bodo.libs.distributed_api.get_node_portion(dim0_global_size, bodo.get_size(), bodo.get_rank())
"""
        inakp__umpn += """        output = np.empty((dim0_local_size,) + tuple(data.shape[1:]), dtype=data.dtype)
"""
        inakp__umpn += (
            '        dtype_size = bodo.io.np_io.get_dtype_size(data.dtype)\n')
        inakp__umpn += """        bodo.libs.distributed_api.dist_permutation_array_index(output, dim0_global_size, dtype_size, data, permutation, len(permutation))
"""
        inakp__umpn += '        return output\n'
    else:
        inakp__umpn += """    return bodo.libs.distributed_api.rebalance(data, dests=dests, random=True, random_seed=seed, parallel=parallel)
"""
    lpgl__akc = {}
    exec(inakp__umpn, {'np': np, 'bodo': bodo}, lpgl__akc)
    impl = lpgl__akc['impl']
    return impl


@numba.generated_jit(nopython=True)
def allgatherv(data, warn_if_rep=True, root=MPI_ROOT):
    return lambda data, warn_if_rep=True, root=MPI_ROOT: gatherv(data, True,
        warn_if_rep, root)


@numba.njit
def get_scatter_null_bytes_buff(null_bitmap_ptr, sendcounts, sendcounts_nulls):
    if bodo.get_rank() != MPI_ROOT:
        return np.empty(1, np.uint8)
    tkt__tkw = np.empty(sendcounts_nulls.sum(), np.uint8)
    foqat__cgotx = 0
    embm__psc = 0
    for mcxa__cizk in range(len(sendcounts)):
        svyn__xolv = sendcounts[mcxa__cizk]
        eskss__qqbz = sendcounts_nulls[mcxa__cizk]
        tpbfk__wagy = tkt__tkw[foqat__cgotx:foqat__cgotx + eskss__qqbz]
        for yiw__qxcq in range(svyn__xolv):
            set_bit_to_arr(tpbfk__wagy, yiw__qxcq, get_bit_bitmap(
                null_bitmap_ptr, embm__psc))
            embm__psc += 1
        foqat__cgotx += eskss__qqbz
    return tkt__tkw


def _bcast_dtype(data):
    try:
        from mpi4py import MPI
    except:
        raise BodoError('mpi4py is required for scatterv')
    pfkp__hzgkm = MPI.COMM_WORLD
    data = pfkp__hzgkm.bcast(data)
    return data


@numba.generated_jit(nopython=True, no_cpython_wrapper=True)
def _get_scatterv_send_counts(send_counts, n_pes, n):
    if not is_overload_none(send_counts):
        return lambda send_counts, n_pes, n: send_counts

    def impl(send_counts, n_pes, n):
        send_counts = np.empty(n_pes, np.int32)
        for i in range(n_pes):
            send_counts[i] = get_node_portion(n, n_pes, i)
        return send_counts
    return impl


@numba.generated_jit(nopython=True, no_cpython_wrapper=True)
def _scatterv_np(data, send_counts=None, warn_if_dist=True):
    typ_val = numba_to_c_type(data.dtype)
    zdz__lhaem = data.ndim
    dtype = data.dtype
    if dtype == types.NPDatetime('ns'):
        dtype = np.dtype('datetime64[ns]')
    elif dtype == types.NPTimedelta('ns'):
        dtype = np.dtype('timedelta64[ns]')
    ujxek__mnic = (0,) * zdz__lhaem

    def scatterv_arr_impl(data, send_counts=None, warn_if_dist=True):
        rank = bodo.libs.distributed_api.get_rank()
        n_pes = bodo.libs.distributed_api.get_size()
        pwqw__ayk = np.ascontiguousarray(data)
        uxcma__qveft = data.ctypes
        jivnm__gxmnd = ujxek__mnic
        if rank == MPI_ROOT:
            jivnm__gxmnd = pwqw__ayk.shape
        jivnm__gxmnd = bcast_tuple(jivnm__gxmnd)
        opg__cabz = get_tuple_prod(jivnm__gxmnd[1:])
        send_counts = _get_scatterv_send_counts(send_counts, n_pes,
            jivnm__gxmnd[0])
        send_counts *= opg__cabz
        daqx__maivf = send_counts[rank]
        quc__rhz = np.empty(daqx__maivf, dtype)
        jpd__alai = bodo.ir.join.calc_disp(send_counts)
        c_scatterv(uxcma__qveft, send_counts.ctypes, jpd__alai.ctypes,
            quc__rhz.ctypes, np.int32(daqx__maivf), np.int32(typ_val))
        return quc__rhz.reshape((-1,) + jivnm__gxmnd[1:])
    return scatterv_arr_impl


def _get_name_value_for_type(name_typ):
    assert isinstance(name_typ, (types.UnicodeType, types.StringLiteral)
        ) or name_typ == types.none
    return None if name_typ == types.none else '_' + str(ir_utils.next_label())


def get_value_for_type(dtype):
    if isinstance(dtype, types.Array):
        return np.zeros((1,) * dtype.ndim, numba.np.numpy_support.as_dtype(
            dtype.dtype))
    if dtype == string_array_type:
        return pd.array(['A'], 'string')
    if dtype == binary_array_type:
        return np.array([b'A'], dtype=object)
    if isinstance(dtype, IntegerArrayType):
        vrfhm__ibc = '{}Int{}'.format('' if dtype.dtype.signed else 'U',
            dtype.dtype.bitwidth)
        return pd.array([3], vrfhm__ibc)
    if dtype == boolean_array:
        return pd.array([True], 'boolean')
    if isinstance(dtype, DecimalArrayType):
        return np.array([Decimal('32.1')])
    if dtype == datetime_date_array_type:
        return np.array([datetime.date(2011, 8, 9)])
    if dtype == datetime_timedelta_array_type:
        return np.array([datetime.timedelta(33)])
    if bodo.hiframes.pd_index_ext.is_pd_index_type(dtype):
        mqsgk__rsekc = _get_name_value_for_type(dtype.name_typ)
        if isinstance(dtype, bodo.hiframes.pd_index_ext.RangeIndexType):
            return pd.RangeIndex(1, name=mqsgk__rsekc)
        rmjw__vmzm = bodo.utils.typing.get_index_data_arr_types(dtype)[0]
        arr = get_value_for_type(rmjw__vmzm)
        return pd.Index(arr, name=mqsgk__rsekc)
    if isinstance(dtype, bodo.hiframes.pd_multi_index_ext.MultiIndexType):
        mqsgk__rsekc = _get_name_value_for_type(dtype.name_typ)
        xzxym__exf = tuple(_get_name_value_for_type(t) for t in dtype.names_typ
            )
        bnouc__bdaq = tuple(get_value_for_type(t) for t in dtype.array_types)
        val = pd.MultiIndex.from_arrays(bnouc__bdaq, names=xzxym__exf)
        val.name = mqsgk__rsekc
        return val
    if isinstance(dtype, bodo.hiframes.pd_series_ext.SeriesType):
        mqsgk__rsekc = _get_name_value_for_type(dtype.name_typ)
        arr = get_value_for_type(dtype.data)
        index = get_value_for_type(dtype.index)
        return pd.Series(arr, index, name=mqsgk__rsekc)
    if isinstance(dtype, bodo.hiframes.pd_dataframe_ext.DataFrameType):
        bnouc__bdaq = tuple(get_value_for_type(t) for t in dtype.data)
        index = get_value_for_type(dtype.index)
        return pd.DataFrame({mqsgk__rsekc: arr for mqsgk__rsekc, arr in zip
            (dtype.columns, bnouc__bdaq)}, index)
    if isinstance(dtype, CategoricalArrayType):
        return pd.Categorical.from_codes([0], dtype.dtype.categories)
    if isinstance(dtype, types.BaseTuple):
        return tuple(get_value_for_type(t) for t in dtype.types)
    if isinstance(dtype, ArrayItemArrayType):
        return pd.Series([get_value_for_type(dtype.dtype),
            get_value_for_type(dtype.dtype)]).values
    if isinstance(dtype, IntervalArrayType):
        rmjw__vmzm = get_value_for_type(dtype.arr_type)
        return pd.arrays.IntervalArray([pd.Interval(rmjw__vmzm[0],
            rmjw__vmzm[0])])
    raise BodoError(f'get_value_for_type(dtype): Missing data type {dtype}')


def scatterv(data, send_counts=None, warn_if_dist=True):
    rank = bodo.libs.distributed_api.get_rank()
    if rank != MPI_ROOT and data is not None:
        warnings.warn(BodoWarning(
            "bodo.scatterv(): A non-None value for 'data' was found on a rank other than the root. This data won't be sent to any other ranks and will be overwritten with data from rank 0."
            ))
    dtype = bodo.typeof(data)
    dtype = _bcast_dtype(dtype)
    if rank != MPI_ROOT:
        data = get_value_for_type(dtype)
    return scatterv_impl(data, send_counts)


@overload(scatterv)
def scatterv_overload(data, send_counts=None, warn_if_dist=True):
    bodo.hiframes.pd_dataframe_ext.check_runtime_cols_unsupported(data,
        'bodo.scatterv()')
    return lambda data, send_counts=None, warn_if_dist=True: scatterv_impl(data
        , send_counts)


@numba.generated_jit(nopython=True)
def scatterv_impl(data, send_counts=None, warn_if_dist=True):
    if isinstance(data, types.Array):
        return lambda data, send_counts=None, warn_if_dist=True: _scatterv_np(
            data, send_counts)
    if data in [binary_array_type, string_array_type]:
        hysi__kjlz = np.int32(numba_to_c_type(types.int32))
        mgfar__mzdui = np.int32(numba_to_c_type(types.uint8))
        if data == binary_array_type:
            mtb__qvt = 'bodo.libs.binary_arr_ext.pre_alloc_binary_array'
        else:
            mtb__qvt = 'bodo.libs.str_arr_ext.pre_alloc_string_array'
        inakp__umpn = f"""def impl(
            data, send_counts=None, warn_if_dist=True
        ):  # pragma: no cover
            rank = bodo.libs.distributed_api.get_rank()
            n_pes = bodo.libs.distributed_api.get_size()
            n_all = bodo.libs.distributed_api.bcast_scalar(len(data))

            # convert offsets to lengths of strings
            send_arr_lens = np.empty(
                len(data), np.uint32
            )  # XXX offset type is offset_type, lengths for comm are uint32
            for i in range(len(data)):
                send_arr_lens[i] = bodo.libs.str_arr_ext.get_str_arr_item_length(
                    data, i
                )

            # ------- calculate buffer counts -------

            send_counts = bodo.libs.distributed_api._get_scatterv_send_counts(send_counts, n_pes, n_all)

            # displacements
            displs = bodo.ir.join.calc_disp(send_counts)

            # compute send counts for characters
            send_counts_char = np.empty(n_pes, np.int32)
            if rank == 0:
                curr_str = 0
                for i in range(n_pes):
                    c = 0
                    for _ in range(send_counts[i]):
                        c += send_arr_lens[curr_str]
                        curr_str += 1
                    send_counts_char[i] = c

            bodo.libs.distributed_api.bcast(send_counts_char)

            # displacements for characters
            displs_char = bodo.ir.join.calc_disp(send_counts_char)

            # compute send counts for nulls
            send_counts_nulls = np.empty(n_pes, np.int32)
            for i in range(n_pes):
                send_counts_nulls[i] = (send_counts[i] + 7) >> 3

            # displacements for nulls
            displs_nulls = bodo.ir.join.calc_disp(send_counts_nulls)

            # alloc output array
            n_loc = send_counts[rank]  # total number of elements on this PE
            n_loc_char = send_counts_char[rank]
            recv_arr = {mtb__qvt}(n_loc, n_loc_char)

            # ----- string lengths -----------

            recv_lens = np.empty(n_loc, np.uint32)
            bodo.libs.distributed_api.c_scatterv(
                send_arr_lens.ctypes,
                send_counts.ctypes,
                displs.ctypes,
                recv_lens.ctypes,
                np.int32(n_loc),
                int32_typ_enum,
            )

            # TODO: don't hardcode offset type. Also, if offset is 32 bit we can
            # use the same buffer
            bodo.libs.str_arr_ext.convert_len_arr_to_offset(recv_lens.ctypes, bodo.libs.str_arr_ext.get_offset_ptr(recv_arr), n_loc)

            # ----- string characters -----------

            bodo.libs.distributed_api.c_scatterv(
                bodo.libs.str_arr_ext.get_data_ptr(data),
                send_counts_char.ctypes,
                displs_char.ctypes,
                bodo.libs.str_arr_ext.get_data_ptr(recv_arr),
                np.int32(n_loc_char),
                char_typ_enum,
            )

            # ----------- null bitmap -------------

            n_recv_bytes = (n_loc + 7) >> 3

            send_null_bitmap = bodo.libs.distributed_api.get_scatter_null_bytes_buff(
                bodo.libs.str_arr_ext.get_null_bitmap_ptr(data), send_counts, send_counts_nulls
            )

            bodo.libs.distributed_api.c_scatterv(
                send_null_bitmap.ctypes,
                send_counts_nulls.ctypes,
                displs_nulls.ctypes,
                bodo.libs.str_arr_ext.get_null_bitmap_ptr(recv_arr),
                np.int32(n_recv_bytes),
                char_typ_enum,
            )

            return recv_arr"""
        lpgl__akc = dict()
        exec(inakp__umpn, {'bodo': bodo, 'np': np, 'int32_typ_enum':
            hysi__kjlz, 'char_typ_enum': mgfar__mzdui}, lpgl__akc)
        impl = lpgl__akc['impl']
        return impl
    if isinstance(data, ArrayItemArrayType):
        hysi__kjlz = np.int32(numba_to_c_type(types.int32))
        mgfar__mzdui = np.int32(numba_to_c_type(types.uint8))

        def scatterv_array_item_impl(data, send_counts=None, warn_if_dist=True
            ):
            isyst__nsm = bodo.libs.array_item_arr_ext.get_offsets(data)
            qzieo__dvq = bodo.libs.array_item_arr_ext.get_data(data)
            gpka__amjy = bodo.libs.array_item_arr_ext.get_null_bitmap(data)
            rank = bodo.libs.distributed_api.get_rank()
            n_pes = bodo.libs.distributed_api.get_size()
            rjtu__vuy = bcast_scalar(len(data))
            hztnl__wetii = np.empty(len(data), np.uint32)
            for i in range(len(data)):
                hztnl__wetii[i] = isyst__nsm[i + 1] - isyst__nsm[i]
            send_counts = _get_scatterv_send_counts(send_counts, n_pes,
                rjtu__vuy)
            jpd__alai = bodo.ir.join.calc_disp(send_counts)
            fwmfp__vqjz = np.empty(n_pes, np.int32)
            if rank == 0:
                idwl__szinf = 0
                for i in range(n_pes):
                    yyep__npsy = 0
                    for pzlm__iwhc in range(send_counts[i]):
                        yyep__npsy += hztnl__wetii[idwl__szinf]
                        idwl__szinf += 1
                    fwmfp__vqjz[i] = yyep__npsy
            bcast(fwmfp__vqjz)
            hnof__dsibs = np.empty(n_pes, np.int32)
            for i in range(n_pes):
                hnof__dsibs[i] = send_counts[i] + 7 >> 3
            mcu__cirkw = bodo.ir.join.calc_disp(hnof__dsibs)
            daqx__maivf = send_counts[rank]
            ippz__xxzp = np.empty(daqx__maivf + 1, np_offset_type)
            jnxyb__cfbo = bodo.libs.distributed_api.scatterv_impl(qzieo__dvq,
                fwmfp__vqjz)
            oke__xyg = daqx__maivf + 7 >> 3
            vqc__mkspv = np.empty(oke__xyg, np.uint8)
            kssv__qjy = np.empty(daqx__maivf, np.uint32)
            c_scatterv(hztnl__wetii.ctypes, send_counts.ctypes, jpd__alai.
                ctypes, kssv__qjy.ctypes, np.int32(daqx__maivf), hysi__kjlz)
            convert_len_arr_to_offset(kssv__qjy.ctypes, ippz__xxzp.ctypes,
                daqx__maivf)
            ldjyz__jdx = get_scatter_null_bytes_buff(gpka__amjy.ctypes,
                send_counts, hnof__dsibs)
            c_scatterv(ldjyz__jdx.ctypes, hnof__dsibs.ctypes, mcu__cirkw.
                ctypes, vqc__mkspv.ctypes, np.int32(oke__xyg), mgfar__mzdui)
            return bodo.libs.array_item_arr_ext.init_array_item_array(
                daqx__maivf, jnxyb__cfbo, ippz__xxzp, vqc__mkspv)
        return scatterv_array_item_impl
    if isinstance(data, (IntegerArrayType, DecimalArrayType)) or data in (
        boolean_array, datetime_date_array_type):
        mgfar__mzdui = np.int32(numba_to_c_type(types.uint8))
        if isinstance(data, IntegerArrayType):
            jfbpn__flt = bodo.libs.int_arr_ext.init_integer_array
        if isinstance(data, DecimalArrayType):
            precision = data.precision
            scale = data.scale
            jfbpn__flt = numba.njit(no_cpython_wrapper=True)(lambda d, b:
                bodo.libs.decimal_arr_ext.init_decimal_array(d, b,
                precision, scale))
        if data == boolean_array:
            jfbpn__flt = bodo.libs.bool_arr_ext.init_bool_array
        if data == datetime_date_array_type:
            jfbpn__flt = (bodo.hiframes.datetime_date_ext.
                init_datetime_date_array)

        def scatterv_impl_int_arr(data, send_counts=None, warn_if_dist=True):
            n_pes = bodo.libs.distributed_api.get_size()
            pwqw__ayk = data._data
            cnv__vwod = data._null_bitmap
            cgsnz__qlq = len(pwqw__ayk)
            gmny__jxf = _scatterv_np(pwqw__ayk, send_counts)
            rjtu__vuy = bcast_scalar(cgsnz__qlq)
            wfuw__igpx = len(gmny__jxf) + 7 >> 3
            jcx__azzl = np.empty(wfuw__igpx, np.uint8)
            send_counts = _get_scatterv_send_counts(send_counts, n_pes,
                rjtu__vuy)
            hnof__dsibs = np.empty(n_pes, np.int32)
            for i in range(n_pes):
                hnof__dsibs[i] = send_counts[i] + 7 >> 3
            mcu__cirkw = bodo.ir.join.calc_disp(hnof__dsibs)
            ldjyz__jdx = get_scatter_null_bytes_buff(cnv__vwod.ctypes,
                send_counts, hnof__dsibs)
            c_scatterv(ldjyz__jdx.ctypes, hnof__dsibs.ctypes, mcu__cirkw.
                ctypes, jcx__azzl.ctypes, np.int32(wfuw__igpx), mgfar__mzdui)
            return jfbpn__flt(gmny__jxf, jcx__azzl)
        return scatterv_impl_int_arr
    if isinstance(data, IntervalArrayType):

        def impl_interval_arr(data, send_counts=None, warn_if_dist=True):
            rblyq__xrfui = bodo.libs.distributed_api.scatterv_impl(data.
                _left, send_counts)
            vox__zbvxr = bodo.libs.distributed_api.scatterv_impl(data.
                _right, send_counts)
            return bodo.libs.interval_arr_ext.init_interval_array(rblyq__xrfui,
                vox__zbvxr)
        return impl_interval_arr
    if isinstance(data, bodo.hiframes.pd_index_ext.RangeIndexType):

        def impl_range_index(data, send_counts=None, warn_if_dist=True):
            rank = bodo.libs.distributed_api.get_rank()
            n_pes = bodo.libs.distributed_api.get_size()
            start = data._start
            stop = data._stop
            fdyy__rjv = data._step
            mqsgk__rsekc = data._name
            mqsgk__rsekc = bcast_scalar(mqsgk__rsekc)
            start = bcast_scalar(start)
            stop = bcast_scalar(stop)
            fdyy__rjv = bcast_scalar(fdyy__rjv)
            dciq__bhdv = bodo.libs.array_kernels.calc_nitems(start, stop,
                fdyy__rjv)
            chunk_start = bodo.libs.distributed_api.get_start(dciq__bhdv,
                n_pes, rank)
            chunk_count = bodo.libs.distributed_api.get_node_portion(dciq__bhdv
                , n_pes, rank)
            ybqv__nxm = start + fdyy__rjv * chunk_start
            tuj__pnl = start + fdyy__rjv * (chunk_start + chunk_count)
            tuj__pnl = min(tuj__pnl, stop)
            return bodo.hiframes.pd_index_ext.init_range_index(ybqv__nxm,
                tuj__pnl, fdyy__rjv, mqsgk__rsekc)
        return impl_range_index
    if isinstance(data, bodo.hiframes.pd_index_ext.PeriodIndexType):
        zcjh__ejv = data.freq

        def impl_period_index(data, send_counts=None, warn_if_dist=True):
            pwqw__ayk = data._data
            mqsgk__rsekc = data._name
            mqsgk__rsekc = bcast_scalar(mqsgk__rsekc)
            arr = bodo.libs.distributed_api.scatterv_impl(pwqw__ayk,
                send_counts)
            return bodo.hiframes.pd_index_ext.init_period_index(arr,
                mqsgk__rsekc, zcjh__ejv)
        return impl_period_index
    if bodo.hiframes.pd_index_ext.is_pd_index_type(data):

        def impl_pd_index(data, send_counts=None, warn_if_dist=True):
            pwqw__ayk = data._data
            mqsgk__rsekc = data._name
            mqsgk__rsekc = bcast_scalar(mqsgk__rsekc)
            arr = bodo.libs.distributed_api.scatterv_impl(pwqw__ayk,
                send_counts)
            return bodo.utils.conversion.index_from_array(arr, mqsgk__rsekc)
        return impl_pd_index
    if isinstance(data, bodo.hiframes.pd_multi_index_ext.MultiIndexType):

        def impl_multi_index(data, send_counts=None, warn_if_dist=True):
            xvui__nqwib = bodo.libs.distributed_api.scatterv_impl(data.
                _data, send_counts)
            mqsgk__rsekc = bcast_scalar(data._name)
            xzxym__exf = bcast_tuple(data._names)
            return bodo.hiframes.pd_multi_index_ext.init_multi_index(
                xvui__nqwib, xzxym__exf, mqsgk__rsekc)
        return impl_multi_index
    if isinstance(data, bodo.hiframes.pd_series_ext.SeriesType):

        def impl_series(data, send_counts=None, warn_if_dist=True):
            arr = bodo.hiframes.pd_series_ext.get_series_data(data)
            index = bodo.hiframes.pd_series_ext.get_series_index(data)
            mqsgk__rsekc = bodo.hiframes.pd_series_ext.get_series_name(data)
            jrg__uwv = bcast_scalar(mqsgk__rsekc)
            out_arr = bodo.libs.distributed_api.scatterv_impl(arr, send_counts)
            euu__jsi = bodo.libs.distributed_api.scatterv_impl(index,
                send_counts)
            return bodo.hiframes.pd_series_ext.init_series(out_arr,
                euu__jsi, jrg__uwv)
        return impl_series
    if isinstance(data, bodo.hiframes.pd_dataframe_ext.DataFrameType):
        abcew__rrdu = len(data.columns)
        hqjpi__yzmde = ', '.join('g_data_{}'.format(i) for i in range(
            abcew__rrdu))
        syt__fvrz = bodo.utils.transform.gen_const_tup(data.columns)
        inakp__umpn = (
            'def impl_df(data, send_counts=None, warn_if_dist=True):\n')
        for i in range(abcew__rrdu):
            inakp__umpn += (
                """  data_{} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(data, {})
"""
                .format(i, i))
            inakp__umpn += (
                """  g_data_{} = bodo.libs.distributed_api.scatterv_impl(data_{}, send_counts)
"""
                .format(i, i))
        inakp__umpn += (
            '  index = bodo.hiframes.pd_dataframe_ext.get_dataframe_index(data)\n'
            )
        inakp__umpn += (
            '  g_index = bodo.libs.distributed_api.scatterv_impl(index, send_counts)\n'
            )
        inakp__umpn += (
            """  return bodo.hiframes.pd_dataframe_ext.init_dataframe(({},), g_index, {})
"""
            .format(hqjpi__yzmde, syt__fvrz))
        lpgl__akc = {}
        exec(inakp__umpn, {'bodo': bodo}, lpgl__akc)
        dgt__qkc = lpgl__akc['impl_df']
        return dgt__qkc
    if isinstance(data, CategoricalArrayType):

        def impl_cat(data, send_counts=None, warn_if_dist=True):
            kkn__zkzgp = bodo.libs.distributed_api.scatterv_impl(data.codes,
                send_counts)
            return bodo.hiframes.pd_categorical_ext.init_categorical_array(
                kkn__zkzgp, data.dtype)
        return impl_cat
    if isinstance(data, types.BaseTuple):
        inakp__umpn = (
            'def impl_tuple(data, send_counts=None, warn_if_dist=True):\n')
        inakp__umpn += '  return ({}{})\n'.format(', '.join(
            'bodo.libs.distributed_api.scatterv_impl(data[{}], send_counts)'
            .format(i) for i in range(len(data))), ',' if len(data) > 0 else ''
            )
        lpgl__akc = {}
        exec(inakp__umpn, {'bodo': bodo}, lpgl__akc)
        cbmep__iqtb = lpgl__akc['impl_tuple']
        return cbmep__iqtb
    if data is types.none:
        return lambda data, send_counts=None, warn_if_dist=True: None
    raise BodoError('scatterv() not available for {}'.format(data))


@intrinsic
def cptr_to_voidptr(typingctx, cptr_tp=None):

    def codegen(context, builder, sig, args):
        return builder.bitcast(args[0], lir.IntType(8).as_pointer())
    return types.voidptr(cptr_tp), codegen


def bcast(data):
    return


@overload(bcast, no_unliteral=True)
def bcast_overload(data):
    if isinstance(data, types.Array):

        def bcast_impl(data):
            typ_enum = get_type_enum(data)
            count = data.size
            assert count < INT_MAX
            c_bcast(data.ctypes, np.int32(count), typ_enum, np.array([-1]).
                ctypes, 0)
            return
        return bcast_impl
    if isinstance(data, DecimalArrayType):

        def bcast_decimal_arr(data):
            count = data._data.size
            assert count < INT_MAX
            c_bcast(data._data.ctypes, np.int32(count), CTypeEnum.Int128.
                value, np.array([-1]).ctypes, 0)
            bcast(data._null_bitmap)
            return
        return bcast_decimal_arr
    if isinstance(data, IntegerArrayType) or data in (boolean_array,
        datetime_date_array_type):

        def bcast_impl_int_arr(data):
            bcast(data._data)
            bcast(data._null_bitmap)
            return
        return bcast_impl_int_arr
    if data in [binary_array_type, string_array_type]:
        hfp__jtfqs = np.int32(numba_to_c_type(offset_type))
        mgfar__mzdui = np.int32(numba_to_c_type(types.uint8))

        def bcast_str_impl(data):
            daqx__maivf = len(data)
            oyd__iejfg = num_total_chars(data)
            assert daqx__maivf < INT_MAX
            assert oyd__iejfg < INT_MAX
            bdhi__pehqz = get_offset_ptr(data)
            uxcma__qveft = get_data_ptr(data)
            null_bitmap_ptr = get_null_bitmap_ptr(data)
            eskss__qqbz = daqx__maivf + 7 >> 3
            c_bcast(bdhi__pehqz, np.int32(daqx__maivf + 1), hfp__jtfqs, np.
                array([-1]).ctypes, 0)
            c_bcast(uxcma__qveft, np.int32(oyd__iejfg), mgfar__mzdui, np.
                array([-1]).ctypes, 0)
            c_bcast(null_bitmap_ptr, np.int32(eskss__qqbz), mgfar__mzdui,
                np.array([-1]).ctypes, 0)
        return bcast_str_impl


c_bcast = types.ExternalFunction('c_bcast', types.void(types.voidptr, types
    .int32, types.int32, types.voidptr, types.int32))


def bcast_scalar(val):
    return val


@overload(bcast_scalar, no_unliteral=True)
def bcast_scalar_overload(val):
    val = types.unliteral(val)
    if not (isinstance(val, (types.Integer, types.Float)) or val in [bodo.
        datetime64ns, bodo.timedelta64ns, bodo.string_type, types.none,
        types.bool_]):
        raise_bodo_error(
            f'bcast_scalar requires an argument of type Integer, Float, datetime64ns, timedelta64ns, string, None, or Bool. Found type {val}'
            )
    if val == types.none:
        return lambda val: None
    if val == bodo.string_type:
        mgfar__mzdui = np.int32(numba_to_c_type(types.uint8))

        def impl_str(val):
            rank = bodo.libs.distributed_api.get_rank()
            if rank != MPI_ROOT:
                hkeok__mxiw = 0
                pvvug__safj = np.empty(0, np.uint8).ctypes
            else:
                pvvug__safj, hkeok__mxiw = (bodo.libs.str_ext.
                    unicode_to_utf8_and_len(val))
            hkeok__mxiw = bodo.libs.distributed_api.bcast_scalar(hkeok__mxiw)
            if rank != MPI_ROOT:
                ltxh__scd = np.empty(hkeok__mxiw + 1, np.uint8)
                ltxh__scd[hkeok__mxiw] = 0
                pvvug__safj = ltxh__scd.ctypes
            c_bcast(pvvug__safj, np.int32(hkeok__mxiw), mgfar__mzdui, np.
                array([-1]).ctypes, 0)
            return bodo.libs.str_arr_ext.decode_utf8(pvvug__safj, hkeok__mxiw)
        return impl_str
    typ_val = numba_to_c_type(val)
    inakp__umpn = (
        """def bcast_scalar_impl(val):
  send = np.empty(1, dtype)
  send[0] = val
  c_bcast(send.ctypes, np.int32(1), np.int32({}), np.array([-1]).ctypes, 0)
  return send[0]
"""
        .format(typ_val))
    dtype = numba.np.numpy_support.as_dtype(val)
    lpgl__akc = {}
    exec(inakp__umpn, {'bodo': bodo, 'np': np, 'c_bcast': c_bcast, 'dtype':
        dtype}, lpgl__akc)
    jqxau__eroj = lpgl__akc['bcast_scalar_impl']
    return jqxau__eroj


def bcast_tuple(val):
    return val


@overload(bcast_tuple, no_unliteral=True)
def overload_bcast_tuple(val):
    assert isinstance(val, types.BaseTuple)
    phzq__nwvz = len(val)
    inakp__umpn = 'def bcast_tuple_impl(val):\n'
    inakp__umpn += '  return ({}{})'.format(','.join(
        'bcast_scalar(val[{}])'.format(i) for i in range(phzq__nwvz)), ',' if
        phzq__nwvz else '')
    lpgl__akc = {}
    exec(inakp__umpn, {'bcast_scalar': bcast_scalar}, lpgl__akc)
    ymrf__wxfq = lpgl__akc['bcast_tuple_impl']
    return ymrf__wxfq


def prealloc_str_for_bcast(arr):
    return arr


@overload(prealloc_str_for_bcast, no_unliteral=True)
def prealloc_str_for_bcast_overload(arr):
    if arr == string_array_type:

        def prealloc_impl(arr):
            rank = bodo.libs.distributed_api.get_rank()
            daqx__maivf = bcast_scalar(len(arr))
            cihzj__yit = bcast_scalar(np.int64(num_total_chars(arr)))
            if rank != MPI_ROOT:
                arr = pre_alloc_string_array(daqx__maivf, cihzj__yit)
            return arr
        return prealloc_impl
    return lambda arr: arr


def get_local_slice(idx, arr_start, total_len):
    return idx


@overload(get_local_slice, no_unliteral=True, jit_options={'cache': True,
    'no_cpython_wrapper': True})
def get_local_slice_overload(idx, arr_start, total_len):

    def impl(idx, arr_start, total_len):
        slice_index = numba.cpython.unicode._normalize_slice(idx, total_len)
        start = slice_index.start
        fdyy__rjv = slice_index.step
        uqhtg__zdh = 0 if fdyy__rjv == 1 or start > arr_start else abs(
            fdyy__rjv - arr_start % fdyy__rjv) % fdyy__rjv
        ybqv__nxm = max(arr_start, slice_index.start) - arr_start + uqhtg__zdh
        tuj__pnl = max(slice_index.stop - arr_start, 0)
        return slice(ybqv__nxm, tuj__pnl, fdyy__rjv)
    return impl


def slice_getitem(arr, slice_index, arr_start, total_len):
    return arr[slice_index]


@overload(slice_getitem, no_unliteral=True, jit_options={'cache': True})
def slice_getitem_overload(arr, slice_index, arr_start, total_len):

    def getitem_impl(arr, slice_index, arr_start, total_len):
        mfpe__sdbc = get_local_slice(slice_index, arr_start, total_len)
        return bodo.utils.conversion.ensure_contig_if_np(arr[mfpe__sdbc])
    return getitem_impl


def slice_getitem_from_start(arr, slice_index):
    return arr[slice_index]


@overload(slice_getitem_from_start, no_unliteral=True)
def slice_getitem_from_start_overload(arr, slice_index):
    if arr == bodo.hiframes.datetime_date_ext.datetime_date_array_type:

        def getitem_datetime_date_impl(arr, slice_index):
            rank = bodo.libs.distributed_api.get_rank()
            nxd__iemn = slice_index.stop
            A = bodo.hiframes.datetime_date_ext.alloc_datetime_date_array(
                nxd__iemn)
            if rank == 0:
                A = arr[:nxd__iemn]
            bodo.libs.distributed_api.bcast(A)
            return A
        return getitem_datetime_date_impl
    if (arr == bodo.hiframes.datetime_timedelta_ext.
        datetime_timedelta_array_type):

        def getitem_datetime_timedelta_impl(arr, slice_index):
            rank = bodo.libs.distributed_api.get_rank()
            nxd__iemn = slice_index.stop
            A = (bodo.hiframes.datetime_timedelta_ext.
                alloc_datetime_timedelta_array(nxd__iemn))
            if rank == 0:
                A = arr[:nxd__iemn]
            bodo.libs.distributed_api.bcast(A)
            return A
        return getitem_datetime_timedelta_impl
    if isinstance(arr.dtype, Decimal128Type):
        precision = arr.dtype.precision
        scale = arr.dtype.scale

        def getitem_decimal_impl(arr, slice_index):
            rank = bodo.libs.distributed_api.get_rank()
            nxd__iemn = slice_index.stop
            A = bodo.libs.decimal_arr_ext.alloc_decimal_array(nxd__iemn,
                precision, scale)
            if rank == 0:
                for i in range(nxd__iemn):
                    A._data[i] = arr._data[i]
                    xxc__njqn = bodo.libs.int_arr_ext.get_bit_bitmap_arr(arr
                        ._null_bitmap, i)
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap, i,
                        xxc__njqn)
            bodo.libs.distributed_api.bcast(A)
            return A
        return getitem_decimal_impl
    if arr == string_array_type:

        def getitem_str_impl(arr, slice_index):
            rank = bodo.libs.distributed_api.get_rank()
            nxd__iemn = slice_index.stop
            yhlr__hwylu = np.uint64(0)
            if rank == 0:
                out_arr = arr[:nxd__iemn]
                yhlr__hwylu = num_total_chars(out_arr)
            yhlr__hwylu = bcast_scalar(yhlr__hwylu)
            if rank != 0:
                out_arr = pre_alloc_string_array(nxd__iemn, yhlr__hwylu)
            bodo.libs.distributed_api.bcast(out_arr)
            return out_arr
        return getitem_str_impl
    rmjw__vmzm = arr

    def getitem_impl(arr, slice_index):
        rank = bodo.libs.distributed_api.get_rank()
        nxd__iemn = slice_index.stop
        out_arr = bodo.utils.utils.alloc_type(tuple_to_scalar((nxd__iemn,) +
            arr.shape[1:]), rmjw__vmzm)
        if rank == 0:
            out_arr = arr[:nxd__iemn]
        bodo.libs.distributed_api.bcast(out_arr)
        return out_arr
    return getitem_impl


dummy_use = numba.njit(lambda a: None)


def int_getitem(arr, ind, arr_start, total_len, is_1D):
    return arr[ind]


def transform_str_getitem_output(data, length):
    pass


@overload(transform_str_getitem_output)
def overload_transform_str_getitem_output(data, length):
    if data == bodo.string_type:
        return lambda data, length: bodo.libs.str_arr_ext.decode_utf8(data.
            _data, length)
    if data == types.Array(types.uint8, 1, 'C'):
        return lambda data, length: bodo.libs.binary_arr_ext.init_bytes_type(
            data, length)
    raise BodoError(
        f'Internal Error: Expected String or Uint8 Array, found {data}')


@overload(int_getitem, no_unliteral=True)
def int_getitem_overload(arr, ind, arr_start, total_len, is_1D):
    if arr in [bodo.binary_array_type, string_array_type]:
        sqax__ssag = numba.cpython.unicode.PY_UNICODE_1BYTE_KIND
        mgfar__mzdui = np.int32(numba_to_c_type(types.uint8))
        lsir__trudw = arr.dtype

        def str_getitem_impl(arr, ind, arr_start, total_len, is_1D):
            if ind >= total_len:
                raise IndexError('index out of bounds')
            ind = ind % total_len
            root = np.int32(0)
            pugus__nid = np.int32(10)
            tag = np.int32(11)
            cpdm__gxje = np.zeros(1, np.int64)
            if arr_start <= ind < arr_start + len(arr):
                ind = ind - arr_start
                zoet__any = arr._data
                mchcz__hkfm = bodo.libs.array_item_arr_ext.get_offsets_ind(
                    zoet__any, ind)
                dfu__zewws = bodo.libs.array_item_arr_ext.get_offsets_ind(
                    zoet__any, ind + 1)
                length = dfu__zewws - mchcz__hkfm
                denop__yeb = zoet__any[ind]
                cpdm__gxje[0] = length
                isend(cpdm__gxje, np.int32(1), root, pugus__nid, True)
                isend(denop__yeb, np.int32(length), root, tag, True)
            rank = bodo.libs.distributed_api.get_rank()
            val = bodo.libs.str_ext.alloc_empty_bytes_or_string_data(
                lsir__trudw, sqax__ssag, 0, 1)
            ljt__owdmb = 0
            if rank == root:
                ljt__owdmb = recv(np.int64, ANY_SOURCE, pugus__nid)
                val = bodo.libs.str_ext.alloc_empty_bytes_or_string_data(
                    lsir__trudw, sqax__ssag, ljt__owdmb, 1)
                uxcma__qveft = bodo.libs.str_ext.get_unicode_or_numpy_data(val)
                _recv(uxcma__qveft, np.int32(ljt__owdmb), mgfar__mzdui,
                    ANY_SOURCE, tag)
            dummy_use(cpdm__gxje)
            ljt__owdmb = bcast_scalar(ljt__owdmb)
            if rank != root:
                val = bodo.libs.str_ext.alloc_empty_bytes_or_string_data(
                    lsir__trudw, sqax__ssag, ljt__owdmb, 1)
            uxcma__qveft = bodo.libs.str_ext.get_unicode_or_numpy_data(val)
            c_bcast(uxcma__qveft, np.int32(ljt__owdmb), mgfar__mzdui, np.
                array([-1]).ctypes, 0)
            val = transform_str_getitem_output(val, ljt__owdmb)
            return val
        return str_getitem_impl
    if isinstance(arr, bodo.CategoricalArrayType):
        iigl__xzs = bodo.hiframes.pd_categorical_ext.get_categories_int_type(
            arr.dtype)

        def cat_getitem_impl(arr, ind, arr_start, total_len, is_1D):
            if ind >= total_len:
                raise IndexError('index out of bounds')
            ind = ind % total_len
            root = np.int32(0)
            tag = np.int32(11)
            send_arr = np.zeros(1, iigl__xzs)
            if arr_start <= ind < arr_start + len(arr):
                kkn__zkzgp = (bodo.hiframes.pd_categorical_ext.
                    get_categorical_arr_codes(arr))
                data = kkn__zkzgp[ind - arr_start]
                send_arr = np.full(1, data, iigl__xzs)
                isend(send_arr, np.int32(1), root, tag, True)
            rank = bodo.libs.distributed_api.get_rank()
            val = iigl__xzs(-1)
            if rank == root:
                val = recv(iigl__xzs, ANY_SOURCE, tag)
            dummy_use(send_arr)
            val = bcast_scalar(val)
            ynosm__vwp = arr.dtype.categories[max(val, 0)]
            return ynosm__vwp
        return cat_getitem_impl
    tbcw__fnm = arr.dtype

    def getitem_impl(arr, ind, arr_start, total_len, is_1D):
        if ind >= total_len:
            raise IndexError('index out of bounds')
        ind = ind % total_len
        root = np.int32(0)
        tag = np.int32(11)
        send_arr = np.zeros(1, tbcw__fnm)
        if arr_start <= ind < arr_start + len(arr):
            data = arr[ind - arr_start]
            send_arr = np.full(1, data)
            isend(send_arr, np.int32(1), root, tag, True)
        rank = bodo.libs.distributed_api.get_rank()
        val = np.zeros(1, tbcw__fnm)[0]
        if rank == root:
            val = recv(tbcw__fnm, ANY_SOURCE, tag)
        dummy_use(send_arr)
        val = bcast_scalar(val)
        return val
    return getitem_impl


c_alltoallv = types.ExternalFunction('c_alltoallv', types.void(types.
    voidptr, types.voidptr, types.voidptr, types.voidptr, types.voidptr,
    types.voidptr, types.int32))


@numba.generated_jit(nopython=True, no_cpython_wrapper=True)
def alltoallv(send_data, out_data, send_counts, recv_counts, send_disp,
    recv_disp):
    typ_enum = get_type_enum(send_data)
    ega__yxxts = get_type_enum(out_data)
    assert typ_enum == ega__yxxts
    if isinstance(send_data, (IntegerArrayType, DecimalArrayType)
        ) or send_data in (boolean_array, datetime_date_array_type):
        return (lambda send_data, out_data, send_counts, recv_counts,
            send_disp, recv_disp: c_alltoallv(send_data._data.ctypes,
            out_data._data.ctypes, send_counts.ctypes, recv_counts.ctypes,
            send_disp.ctypes, recv_disp.ctypes, typ_enum))
    if isinstance(send_data, bodo.CategoricalArrayType):
        return (lambda send_data, out_data, send_counts, recv_counts,
            send_disp, recv_disp: c_alltoallv(send_data.codes.ctypes,
            out_data.codes.ctypes, send_counts.ctypes, recv_counts.ctypes,
            send_disp.ctypes, recv_disp.ctypes, typ_enum))
    return (lambda send_data, out_data, send_counts, recv_counts, send_disp,
        recv_disp: c_alltoallv(send_data.ctypes, out_data.ctypes,
        send_counts.ctypes, recv_counts.ctypes, send_disp.ctypes, recv_disp
        .ctypes, typ_enum))


def alltoallv_tup(send_data, out_data, send_counts, recv_counts, send_disp,
    recv_disp):
    return


@overload(alltoallv_tup, no_unliteral=True)
def alltoallv_tup_overload(send_data, out_data, send_counts, recv_counts,
    send_disp, recv_disp):
    count = send_data.count
    assert out_data.count == count
    inakp__umpn = (
        'def f(send_data, out_data, send_counts, recv_counts, send_disp, recv_disp):\n'
        )
    for i in range(count):
        inakp__umpn += (
            """  alltoallv(send_data[{}], out_data[{}], send_counts, recv_counts, send_disp, recv_disp)
"""
            .format(i, i))
    inakp__umpn += '  return\n'
    lpgl__akc = {}
    exec(inakp__umpn, {'alltoallv': alltoallv}, lpgl__akc)
    qdjr__rcdx = lpgl__akc['f']
    return qdjr__rcdx


@numba.njit
def get_start_count(n):
    rank = bodo.libs.distributed_api.get_rank()
    n_pes = bodo.libs.distributed_api.get_size()
    start = bodo.libs.distributed_api.get_start(n, n_pes, rank)
    count = bodo.libs.distributed_api.get_node_portion(n, n_pes, rank)
    return start, count


@numba.njit
def get_start(total_size, pes, rank):
    wzpx__tvf = total_size % pes
    jtwh__zqqa = (total_size - wzpx__tvf) // pes
    return rank * jtwh__zqqa + min(rank, wzpx__tvf)


@numba.njit
def get_end(total_size, pes, rank):
    wzpx__tvf = total_size % pes
    jtwh__zqqa = (total_size - wzpx__tvf) // pes
    return (rank + 1) * jtwh__zqqa + min(rank + 1, wzpx__tvf)


@numba.njit
def get_node_portion(total_size, pes, rank):
    wzpx__tvf = total_size % pes
    jtwh__zqqa = (total_size - wzpx__tvf) // pes
    if rank < wzpx__tvf:
        return jtwh__zqqa + 1
    else:
        return jtwh__zqqa


@numba.generated_jit(nopython=True)
def dist_cumsum(in_arr, out_arr):
    opb__ovaav = in_arr.dtype(0)
    gmnla__jwm = np.int32(Reduce_Type.Sum.value)

    def cumsum_impl(in_arr, out_arr):
        yyep__npsy = opb__ovaav
        for rbfd__afhd in np.nditer(in_arr):
            yyep__npsy += rbfd__afhd.item()
        mpvy__ejnij = dist_exscan(yyep__npsy, gmnla__jwm)
        for i in range(in_arr.size):
            mpvy__ejnij += in_arr[i]
            out_arr[i] = mpvy__ejnij
        return 0
    return cumsum_impl


@numba.generated_jit(nopython=True)
def dist_cumprod(in_arr, out_arr):
    fuf__new = in_arr.dtype(1)
    gmnla__jwm = np.int32(Reduce_Type.Prod.value)

    def cumprod_impl(in_arr, out_arr):
        yyep__npsy = fuf__new
        for rbfd__afhd in np.nditer(in_arr):
            yyep__npsy *= rbfd__afhd.item()
        mpvy__ejnij = dist_exscan(yyep__npsy, gmnla__jwm)
        if get_rank() == 0:
            mpvy__ejnij = fuf__new
        for i in range(in_arr.size):
            mpvy__ejnij *= in_arr[i]
            out_arr[i] = mpvy__ejnij
        return 0
    return cumprod_impl


@numba.generated_jit(nopython=True)
def dist_cummin(in_arr, out_arr):
    if isinstance(in_arr.dtype, types.Float):
        fuf__new = np.finfo(in_arr.dtype(1).dtype).max
    else:
        fuf__new = np.iinfo(in_arr.dtype(1).dtype).max
    gmnla__jwm = np.int32(Reduce_Type.Min.value)

    def cummin_impl(in_arr, out_arr):
        yyep__npsy = fuf__new
        for rbfd__afhd in np.nditer(in_arr):
            yyep__npsy = min(yyep__npsy, rbfd__afhd.item())
        mpvy__ejnij = dist_exscan(yyep__npsy, gmnla__jwm)
        if get_rank() == 0:
            mpvy__ejnij = fuf__new
        for i in range(in_arr.size):
            mpvy__ejnij = min(mpvy__ejnij, in_arr[i])
            out_arr[i] = mpvy__ejnij
        return 0
    return cummin_impl


@numba.generated_jit(nopython=True)
def dist_cummax(in_arr, out_arr):
    if isinstance(in_arr.dtype, types.Float):
        fuf__new = np.finfo(in_arr.dtype(1).dtype).min
    else:
        fuf__new = np.iinfo(in_arr.dtype(1).dtype).min
    fuf__new = in_arr.dtype(1)
    gmnla__jwm = np.int32(Reduce_Type.Max.value)

    def cummax_impl(in_arr, out_arr):
        yyep__npsy = fuf__new
        for rbfd__afhd in np.nditer(in_arr):
            yyep__npsy = max(yyep__npsy, rbfd__afhd.item())
        mpvy__ejnij = dist_exscan(yyep__npsy, gmnla__jwm)
        if get_rank() == 0:
            mpvy__ejnij = fuf__new
        for i in range(in_arr.size):
            mpvy__ejnij = max(mpvy__ejnij, in_arr[i])
            out_arr[i] = mpvy__ejnij
        return 0
    return cummax_impl


_allgather = types.ExternalFunction('allgather', types.void(types.voidptr,
    types.int32, types.voidptr, types.int32))


@numba.njit
def allgather(arr, val):
    sch__occf = get_type_enum(arr)
    _allgather(arr.ctypes, 1, value_to_ptr(val), sch__occf)


def dist_return(A):
    return A


def dist_return_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 1 and not kws
    aolsh__ddcry = args[0]
    if equiv_set.has_shape(aolsh__ddcry):
        return ArrayAnalysis.AnalyzeResult(shape=aolsh__ddcry, pre=[])
    return None


ArrayAnalysis._analyze_op_call_bodo_libs_distributed_api_dist_return = (
    dist_return_equiv)


def threaded_return(A):
    return A


@numba.njit
def set_arr_local(arr, ind, val):
    arr[ind] = val


@numba.njit
def local_alloc_size(n, in_arr):
    return n


@infer_global(threaded_return)
@infer_global(dist_return)
class ThreadedRetTyper(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        assert len(args) == 1
        return signature(args[0], *args)


@numba.njit
def parallel_print(*args):
    print(*args)


@numba.njit
def single_print(*args):
    if bodo.libs.distributed_api.get_rank() == 0:
        print(*args)


@numba.njit(no_cpython_wrapper=True)
def print_if_not_empty(arg):
    if len(arg) != 0 or bodo.get_rank() == 0:
        print(arg)


_wait = types.ExternalFunction('dist_wait', types.void(mpi_req_numba_type,
    types.bool_))


@numba.generated_jit(nopython=True)
def wait(req, cond=True):
    if isinstance(req, types.BaseTuple):
        count = len(req.types)
        dxg__dcu = ','.join(f'_wait(req[{i}], cond)' for i in range(count))
        inakp__umpn = 'def f(req, cond=True):\n'
        inakp__umpn += f'  return {dxg__dcu}\n'
        lpgl__akc = {}
        exec(inakp__umpn, {'_wait': _wait}, lpgl__akc)
        impl = lpgl__akc['f']
        return impl
    if is_overload_none(req):
        return lambda req, cond=True: None
    return lambda req, cond=True: _wait(req, cond)


class ReqArrayType(types.Type):

    def __init__(self):
        super(ReqArrayType, self).__init__(name='ReqArrayType()')


req_array_type = ReqArrayType()
register_model(ReqArrayType)(models.OpaqueModel)
waitall = types.ExternalFunction('dist_waitall', types.void(types.int32,
    req_array_type))
comm_req_alloc = types.ExternalFunction('comm_req_alloc', req_array_type(
    types.int32))
comm_req_dealloc = types.ExternalFunction('comm_req_dealloc', types.void(
    req_array_type))
req_array_setitem = types.ExternalFunction('req_array_setitem', types.void(
    req_array_type, types.int64, mpi_req_numba_type))


@overload(operator.setitem, no_unliteral=True)
def overload_req_arr_setitem(A, idx, val):
    if A == req_array_type:
        assert val == mpi_req_numba_type
        return lambda A, idx, val: req_array_setitem(A, idx, val)


@numba.njit
def _get_local_range(start, stop, chunk_start, chunk_count):
    assert start >= 0 and stop > 0
    ybqv__nxm = max(start, chunk_start)
    tuj__pnl = min(stop, chunk_start + chunk_count)
    loss__irx = ybqv__nxm - chunk_start
    sopq__piuj = tuj__pnl - chunk_start
    if loss__irx < 0 or sopq__piuj < 0:
        loss__irx = 1
        sopq__piuj = 0
    return loss__irx, sopq__piuj


@register_jitable
def _set_if_in_range(A, val, index, chunk_start):
    if index >= chunk_start and index < chunk_start + len(A):
        A[index - chunk_start] = val


@register_jitable
def _root_rank_select(old_val, new_val):
    if get_rank() == 0:
        return old_val
    return new_val


def get_tuple_prod(t):
    return np.prod(t)


@overload(get_tuple_prod, no_unliteral=True)
def get_tuple_prod_overload(t):
    if t == numba.core.types.containers.Tuple(()):
        return lambda t: 1

    def get_tuple_prod_impl(t):
        wzpx__tvf = 1
        for a in t:
            wzpx__tvf *= a
        return wzpx__tvf
    return get_tuple_prod_impl


sig = types.void(types.voidptr, types.voidptr, types.intp, types.intp,
    types.intp, types.intp, types.int32, types.voidptr)
oneD_reshape_shuffle = types.ExternalFunction('oneD_reshape_shuffle', sig)


@numba.njit(no_cpython_wrapper=True, cache=True)
def dist_oneD_reshape_shuffle(lhs, in_arr, new_dim0_global_len, dest_ranks=None
    ):
    bguy__nygfu = np.ascontiguousarray(in_arr)
    cth__biaj = get_tuple_prod(bguy__nygfu.shape[1:])
    dndt__snjtg = get_tuple_prod(lhs.shape[1:])
    if dest_ranks is not None:
        bzw__hwggi = np.array(dest_ranks, dtype=np.int32)
    else:
        bzw__hwggi = np.empty(0, dtype=np.int32)
    dtype_size = bodo.io.np_io.get_dtype_size(in_arr.dtype)
    oneD_reshape_shuffle(lhs.ctypes, bguy__nygfu.ctypes,
        new_dim0_global_len, len(in_arr), dtype_size * dndt__snjtg, 
        dtype_size * cth__biaj, len(bzw__hwggi), bzw__hwggi.ctypes)
    check_and_propagate_cpp_exception()


permutation_int = types.ExternalFunction('permutation_int', types.void(
    types.voidptr, types.intp))


@numba.njit
def dist_permutation_int(lhs, n):
    permutation_int(lhs.ctypes, n)


permutation_array_index = types.ExternalFunction('permutation_array_index',
    types.void(types.voidptr, types.intp, types.intp, types.voidptr, types.
    int64, types.voidptr, types.intp))


@numba.njit
def dist_permutation_array_index(lhs, lhs_len, dtype_size, rhs, p, p_len):
    niabn__dcwd = np.ascontiguousarray(rhs)
    pvz__xoqpq = get_tuple_prod(niabn__dcwd.shape[1:])
    gmc__fzyrw = dtype_size * pvz__xoqpq
    permutation_array_index(lhs.ctypes, lhs_len, gmc__fzyrw, niabn__dcwd.
        ctypes, niabn__dcwd.shape[0], p.ctypes, p_len)
    check_and_propagate_cpp_exception()


from bodo.io import fsspec_reader, hdfs_reader, s3_reader
ll.add_symbol('finalize', hdist.finalize)
finalize = types.ExternalFunction('finalize', types.int32())
ll.add_symbol('finalize_s3', s3_reader.finalize_s3)
finalize_s3 = types.ExternalFunction('finalize_s3', types.int32())
ll.add_symbol('finalize_fsspec', fsspec_reader.finalize_fsspec)
finalize_fsspec = types.ExternalFunction('finalize_fsspec', types.int32())
ll.add_symbol('disconnect_hdfs', hdfs_reader.disconnect_hdfs)
disconnect_hdfs = types.ExternalFunction('disconnect_hdfs', types.int32())


def _check_for_cpp_errors():
    pass


@overload(_check_for_cpp_errors)
def overload_check_for_cpp_errors():
    return lambda : check_and_propagate_cpp_exception()


@numba.njit
def call_finalize():
    finalize()
    finalize_s3()
    finalize_fsspec()
    _check_for_cpp_errors()
    disconnect_hdfs()


def flush_stdout():
    if not sys.stdout.closed:
        sys.stdout.flush()


atexit.register(call_finalize)
atexit.register(flush_stdout)


def bcast_comm(data, comm_ranks, nranks):
    rank = bodo.libs.distributed_api.get_rank()
    dtype = bodo.typeof(data)
    dtype = _bcast_dtype(dtype)
    if rank != MPI_ROOT:
        data = get_value_for_type(dtype)
    return bcast_comm_impl(data, comm_ranks, nranks)


@overload(bcast_comm)
def bcast_comm_overload(data, comm_ranks, nranks):
    return lambda data, comm_ranks, nranks: bcast_comm_impl(data,
        comm_ranks, nranks)


@numba.generated_jit(nopython=True)
def bcast_comm_impl(data, comm_ranks, nranks):
    bodo.hiframes.pd_dataframe_ext.check_runtime_cols_unsupported(data,
        'bodo.bcast_comm()')
    if isinstance(data, (types.Integer, types.Float)):
        typ_val = numba_to_c_type(data)
        inakp__umpn = (
            """def bcast_scalar_impl(data, comm_ranks, nranks):
  send = np.empty(1, dtype)
  send[0] = data
  c_bcast(send.ctypes, np.int32(1), np.int32({}), comm_ranks,ctypes, np.int32({}))
  return send[0]
"""
            .format(typ_val, nranks))
        dtype = numba.np.numpy_support.as_dtype(data)
        lpgl__akc = {}
        exec(inakp__umpn, {'bodo': bodo, 'np': np, 'c_bcast': c_bcast,
            'dtype': dtype}, lpgl__akc)
        jqxau__eroj = lpgl__akc['bcast_scalar_impl']
        return jqxau__eroj
    if isinstance(data, types.Array):
        return lambda data, comm_ranks, nranks: _bcast_np(data, comm_ranks,
            nranks)
    if isinstance(data, bodo.hiframes.pd_dataframe_ext.DataFrameType):
        abcew__rrdu = len(data.columns)
        hqjpi__yzmde = ', '.join('g_data_{}'.format(i) for i in range(
            abcew__rrdu))
        syt__fvrz = bodo.utils.transform.gen_const_tup(data.columns)
        inakp__umpn = 'def impl_df(data, comm_ranks, nranks):\n'
        for i in range(abcew__rrdu):
            inakp__umpn += (
                """  data_{} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(data, {})
"""
                .format(i, i))
            inakp__umpn += (
                """  g_data_{} = bodo.libs.distributed_api.bcast_comm_impl(data_{}, comm_ranks, nranks)
"""
                .format(i, i))
        inakp__umpn += (
            '  index = bodo.hiframes.pd_dataframe_ext.get_dataframe_index(data)\n'
            )
        inakp__umpn += """  g_index = bodo.libs.distributed_api.bcast_comm_impl(index, comm_ranks, nranks)
"""
        inakp__umpn += (
            """  return bodo.hiframes.pd_dataframe_ext.init_dataframe(({},), g_index, {})
"""
            .format(hqjpi__yzmde, syt__fvrz))
        lpgl__akc = {}
        exec(inakp__umpn, {'bodo': bodo}, lpgl__akc)
        dgt__qkc = lpgl__akc['impl_df']
        return dgt__qkc
    if isinstance(data, bodo.hiframes.pd_index_ext.RangeIndexType):

        def impl_range_index(data, comm_ranks, nranks):
            rank = bodo.libs.distributed_api.get_rank()
            n_pes = bodo.libs.distributed_api.get_size()
            start = data._start
            stop = data._stop
            fdyy__rjv = data._step
            mqsgk__rsekc = data._name
            mqsgk__rsekc = bcast_scalar(mqsgk__rsekc)
            start = bcast_scalar(start)
            stop = bcast_scalar(stop)
            fdyy__rjv = bcast_scalar(fdyy__rjv)
            dciq__bhdv = bodo.libs.array_kernels.calc_nitems(start, stop,
                fdyy__rjv)
            chunk_start = bodo.libs.distributed_api.get_start(dciq__bhdv,
                n_pes, rank)
            chunk_count = bodo.libs.distributed_api.get_node_portion(dciq__bhdv
                , n_pes, rank)
            ybqv__nxm = start + fdyy__rjv * chunk_start
            tuj__pnl = start + fdyy__rjv * (chunk_start + chunk_count)
            tuj__pnl = min(tuj__pnl, stop)
            return bodo.hiframes.pd_index_ext.init_range_index(ybqv__nxm,
                tuj__pnl, fdyy__rjv, mqsgk__rsekc)
        return impl_range_index
    if bodo.hiframes.pd_index_ext.is_pd_index_type(data):

        def impl_pd_index(data, comm_ranks, nranks):
            pwqw__ayk = data._data
            mqsgk__rsekc = data._name
            arr = bodo.libs.distributed_api.bcast_comm_impl(pwqw__ayk,
                comm_ranks, nranks)
            return bodo.utils.conversion.index_from_array(arr, mqsgk__rsekc)
        return impl_pd_index
    if isinstance(data, bodo.hiframes.pd_series_ext.SeriesType):

        def impl_series(data, comm_ranks, nranks):
            arr = bodo.hiframes.pd_series_ext.get_series_data(data)
            index = bodo.hiframes.pd_series_ext.get_series_index(data)
            mqsgk__rsekc = bodo.hiframes.pd_series_ext.get_series_name(data)
            jrg__uwv = bodo.libs.distributed_api.bcast_comm_impl(mqsgk__rsekc,
                comm_ranks, nranks)
            out_arr = bodo.libs.distributed_api.bcast_comm_impl(arr,
                comm_ranks, nranks)
            euu__jsi = bodo.libs.distributed_api.bcast_comm_impl(index,
                comm_ranks, nranks)
            return bodo.hiframes.pd_series_ext.init_series(out_arr,
                euu__jsi, jrg__uwv)
        return impl_series
    if isinstance(data, types.BaseTuple):
        inakp__umpn = 'def impl_tuple(data, comm_ranks, nranks):\n'
        inakp__umpn += '  return ({}{})\n'.format(', '.join(
            'bcast_comm_impl(data[{}], comm_ranks, nranks)'.format(i) for i in
            range(len(data))), ',' if len(data) > 0 else '')
        lpgl__akc = {}
        exec(inakp__umpn, {'bcast_comm_impl': bcast_comm_impl}, lpgl__akc)
        cbmep__iqtb = lpgl__akc['impl_tuple']
        return cbmep__iqtb
    if data is types.none:
        return lambda data, comm_ranks, nranks: None


@numba.generated_jit(nopython=True, no_cpython_wrapper=True)
def _bcast_np(data, comm_ranks, nranks):
    typ_val = numba_to_c_type(data.dtype)
    zdz__lhaem = data.ndim
    dtype = data.dtype
    if dtype == types.NPDatetime('ns'):
        dtype = np.dtype('datetime64[ns]')
    elif dtype == types.NPTimedelta('ns'):
        dtype = np.dtype('timedelta64[ns]')
    ujxek__mnic = (0,) * zdz__lhaem

    def bcast_arr_impl(data, comm_ranks, nranks):
        rank = bodo.libs.distributed_api.get_rank()
        pwqw__ayk = np.ascontiguousarray(data)
        uxcma__qveft = data.ctypes
        jivnm__gxmnd = ujxek__mnic
        if rank == MPI_ROOT:
            jivnm__gxmnd = pwqw__ayk.shape
        jivnm__gxmnd = bcast_tuple(jivnm__gxmnd)
        opg__cabz = get_tuple_prod(jivnm__gxmnd[1:])
        send_counts = jivnm__gxmnd[0] * opg__cabz
        quc__rhz = np.empty(send_counts, dtype)
        if rank == MPI_ROOT:
            c_bcast(uxcma__qveft, np.int32(send_counts), np.int32(typ_val),
                comm_ranks.ctypes, np.int32(nranks))
            return data
        else:
            c_bcast(quc__rhz.ctypes, np.int32(send_counts), np.int32(
                typ_val), comm_ranks.ctypes, np.int32(nranks))
            return quc__rhz.reshape((-1,) + jivnm__gxmnd[1:])
    return bcast_arr_impl


node_ranks = None


def get_host_ranks():
    global node_ranks
    if node_ranks is None:
        pfkp__hzgkm = MPI.COMM_WORLD
        pize__tujs = MPI.Get_processor_name()
        nkw__uyxea = pfkp__hzgkm.allgather(pize__tujs)
        node_ranks = defaultdict(list)
        for i, tjbz__gacjh in enumerate(nkw__uyxea):
            node_ranks[tjbz__gacjh].append(i)
    return node_ranks


def create_subcomm_mpi4py(comm_ranks):
    pfkp__hzgkm = MPI.COMM_WORLD
    zqjh__tqh = pfkp__hzgkm.Get_group()
    sifr__ooiz = zqjh__tqh.Incl(comm_ranks)
    jcyj__zvw = pfkp__hzgkm.Create_group(sifr__ooiz)
    return jcyj__zvw


def get_nodes_first_ranks():
    dpznw__msa = get_host_ranks()
    return np.array([ynl__cgpw[0] for ynl__cgpw in dpznw__msa.values()],
        dtype='int32')


def get_num_nodes():
    return len(get_host_ranks())
