import operator
import re
import llvmlite.binding as ll
import numba
import numpy as np
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.core.typing.templates import AbstractTemplate, AttributeTemplate, bound_function, infer_getattr, infer_global, signature
from numba.extending import intrinsic, lower_cast, make_attribute_wrapper, models, overload, overload_attribute, register_jitable, register_model
from numba.parfors.array_analysis import ArrayAnalysis
import bodo
from bodo.libs import hstr_ext
from bodo.utils.typing import get_overload_const_int, get_overload_const_str, is_overload_constant_int, is_overload_constant_str


def unliteral_all(args):
    return tuple(types.unliteral(a) for a in args)


ll.add_symbol('del_str', hstr_ext.del_str)
ll.add_symbol('unicode_to_utf8', hstr_ext.unicode_to_utf8)
ll.add_symbol('memcmp', hstr_ext.memcmp)
ll.add_symbol('int_to_hex', hstr_ext.int_to_hex)
string_type = types.unicode_type


@numba.njit
def contains_regex(e, in_str):
    with numba.objmode(res='bool_'):
        res = bool(e.search(in_str))
    return res


@numba.generated_jit
def str_findall_count(regex, in_str):

    def _str_findall_count_impl(regex, in_str):
        with numba.objmode(res='int64'):
            res = len(regex.findall(in_str))
        return res
    return _str_findall_count_impl


utf8_str_type = types.ArrayCTypes(types.Array(types.uint8, 1, 'C'))


@intrinsic
def unicode_to_utf8_and_len(typingctx, str_typ=None):
    assert str_typ in (string_type, types.Optional(string_type)) or isinstance(
        str_typ, types.StringLiteral)
    rthfb__pymq = types.Tuple([utf8_str_type, types.int64])

    def codegen(context, builder, sig, args):
        mxcg__qqvo, = args
        jkxtf__umkr = cgutils.create_struct_proxy(string_type)(context,
            builder, value=mxcg__qqvo)
        yhuvx__cikeu = cgutils.create_struct_proxy(utf8_str_type)(context,
            builder)
        dvn__glucn = cgutils.create_struct_proxy(rthfb__pymq)(context, builder)
        is_ascii = builder.icmp_unsigned('==', jkxtf__umkr.is_ascii, lir.
            Constant(jkxtf__umkr.is_ascii.type, 1))
        with builder.if_else(is_ascii) as (then, orelse):
            with then:
                context.nrt.incref(builder, string_type, mxcg__qqvo)
                yhuvx__cikeu.data = jkxtf__umkr.data
                yhuvx__cikeu.meminfo = jkxtf__umkr.meminfo
                dvn__glucn.f1 = jkxtf__umkr.length
            with orelse:
                tzg__aozf = lir.FunctionType(lir.IntType(64), [lir.IntType(
                    8).as_pointer(), lir.IntType(8).as_pointer(), lir.
                    IntType(64), lir.IntType(32)])
                zzdrk__ntalx = cgutils.get_or_insert_function(builder.
                    module, tzg__aozf, name='unicode_to_utf8')
                zpfgn__esrgz = context.get_constant_null(types.voidptr)
                ffopv__xqcdz = builder.call(zzdrk__ntalx, [zpfgn__esrgz,
                    jkxtf__umkr.data, jkxtf__umkr.length, jkxtf__umkr.kind])
                dvn__glucn.f1 = ffopv__xqcdz
                clra__rdbrr = builder.add(ffopv__xqcdz, lir.Constant(lir.
                    IntType(64), 1))
                yhuvx__cikeu.meminfo = context.nrt.meminfo_alloc_aligned(
                    builder, size=clra__rdbrr, align=32)
                yhuvx__cikeu.data = context.nrt.meminfo_data(builder,
                    yhuvx__cikeu.meminfo)
                builder.call(zzdrk__ntalx, [yhuvx__cikeu.data, jkxtf__umkr.
                    data, jkxtf__umkr.length, jkxtf__umkr.kind])
                builder.store(lir.Constant(lir.IntType(8), 0), builder.gep(
                    yhuvx__cikeu.data, [ffopv__xqcdz]))
        dvn__glucn.f0 = yhuvx__cikeu._getvalue()
        return dvn__glucn._getvalue()
    return rthfb__pymq(string_type), codegen


def unicode_to_utf8(s):
    return s


@overload(unicode_to_utf8)
def overload_unicode_to_utf8(s):
    return lambda s: unicode_to_utf8_and_len(s)[0]


@overload(max)
def overload_builtin_max(lhs, rhs):
    if lhs == types.unicode_type and rhs == types.unicode_type:

        def impl(lhs, rhs):
            return lhs if lhs > rhs else rhs
        return impl


@overload(min)
def overload_builtin_min(lhs, rhs):
    if lhs == types.unicode_type and rhs == types.unicode_type:

        def impl(lhs, rhs):
            return lhs if lhs < rhs else rhs
        return impl


@intrinsic
def memcmp(typingctx, dest_t, src_t, count_t=None):

    def codegen(context, builder, sig, args):
        tzg__aozf = lir.FunctionType(lir.IntType(32), [lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(64)])
        efvyx__wgrri = cgutils.get_or_insert_function(builder.module,
            tzg__aozf, name='memcmp')
        return builder.call(efvyx__wgrri, args)
    return types.int32(types.voidptr, types.voidptr, types.intp), codegen


def int_to_str_len(n):
    return len(str(n))


@overload(int_to_str_len)
def overload_int_to_str_len(n):
    peqvh__ikd = n(10)

    def impl(n):
        if n == 0:
            return 1
        qfk__pnt = 0
        if n < 0:
            n = -n
            qfk__pnt += 1
        while n > 0:
            n = n // peqvh__ikd
            qfk__pnt += 1
        return qfk__pnt
    return impl


class StdStringType(types.Opaque):

    def __init__(self):
        super(StdStringType, self).__init__(name='StdStringType')


std_str_type = StdStringType()
register_model(StdStringType)(models.OpaqueModel)
del_str = types.ExternalFunction('del_str', types.void(std_str_type))
get_c_str = types.ExternalFunction('get_c_str', types.voidptr(std_str_type))
dummy_use = numba.njit(lambda a: None)


@overload(int)
def int_str_overload(in_str, base=10):
    if in_str == string_type:
        if is_overload_constant_int(base) and get_overload_const_int(base
            ) == 10:

            def _str_to_int_impl(in_str, base=10):
                val = _str_to_int64(in_str._data, in_str._length)
                dummy_use(in_str)
                return val
            return _str_to_int_impl

        def _str_to_int_base_impl(in_str, base=10):
            val = _str_to_int64_base(in_str._data, in_str._length, base)
            dummy_use(in_str)
            return val
        return _str_to_int_base_impl


@infer_global(float)
class StrToFloat(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        [vwos__yga] = args
        if isinstance(vwos__yga, StdStringType):
            return signature(types.float64, vwos__yga)
        if vwos__yga == string_type:
            return signature(types.float64, vwos__yga)


ll.add_symbol('init_string_const', hstr_ext.init_string_const)
ll.add_symbol('get_c_str', hstr_ext.get_c_str)
ll.add_symbol('str_to_int64', hstr_ext.str_to_int64)
ll.add_symbol('str_to_uint64', hstr_ext.str_to_uint64)
ll.add_symbol('str_to_int64_base', hstr_ext.str_to_int64_base)
ll.add_symbol('str_to_float64', hstr_ext.str_to_float64)
ll.add_symbol('str_to_float32', hstr_ext.str_to_float32)
ll.add_symbol('get_str_len', hstr_ext.get_str_len)
ll.add_symbol('str_from_float32', hstr_ext.str_from_float32)
ll.add_symbol('str_from_float64', hstr_ext.str_from_float64)
get_std_str_len = types.ExternalFunction('get_str_len', signature(types.
    intp, std_str_type))
init_string_from_chars = types.ExternalFunction('init_string_const',
    std_str_type(types.voidptr, types.intp))
_str_to_int64 = types.ExternalFunction('str_to_int64', signature(types.
    int64, types.voidptr, types.int64))
_str_to_uint64 = types.ExternalFunction('str_to_uint64', signature(types.
    uint64, types.voidptr, types.int64))
_str_to_int64_base = types.ExternalFunction('str_to_int64_base', signature(
    types.int64, types.voidptr, types.int64, types.int64))


def gen_unicode_to_std_str(context, builder, unicode_val):
    jkxtf__umkr = cgutils.create_struct_proxy(string_type)(context, builder,
        value=unicode_val)
    tzg__aozf = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.IntType(
        8).as_pointer(), lir.IntType(64)])
    rycj__qix = cgutils.get_or_insert_function(builder.module, tzg__aozf,
        name='init_string_const')
    return builder.call(rycj__qix, [jkxtf__umkr.data, jkxtf__umkr.length])


def gen_std_str_to_unicode(context, builder, std_str_val, del_str=False):
    kind = numba.cpython.unicode.PY_UNICODE_1BYTE_KIND

    def _std_str_to_unicode(std_str):
        length = bodo.libs.str_ext.get_std_str_len(std_str)
        argfl__lwb = numba.cpython.unicode._empty_string(kind, length, 1)
        bodo.libs.str_arr_ext._memcpy(argfl__lwb._data, bodo.libs.str_ext.
            get_c_str(std_str), length, 1)
        if del_str:
            bodo.libs.str_ext.del_str(std_str)
        return argfl__lwb
    val = context.compile_internal(builder, _std_str_to_unicode,
        string_type(bodo.libs.str_ext.std_str_type), [std_str_val])
    return val


def gen_get_unicode_chars(context, builder, unicode_val):
    jkxtf__umkr = cgutils.create_struct_proxy(string_type)(context, builder,
        value=unicode_val)
    return jkxtf__umkr.data


@intrinsic
def unicode_to_std_str(typingctx, unicode_t=None):

    def codegen(context, builder, sig, args):
        return gen_unicode_to_std_str(context, builder, args[0])
    return std_str_type(string_type), codegen


@intrinsic
def std_str_to_unicode(typingctx, unicode_t=None):

    def codegen(context, builder, sig, args):
        return gen_std_str_to_unicode(context, builder, args[0], True)
    return string_type(std_str_type), codegen


class RandomAccessStringArrayType(types.ArrayCompatible):

    def __init__(self):
        super(RandomAccessStringArrayType, self).__init__(name=
            'RandomAccessStringArrayType()')

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    @property
    def dtype(self):
        return string_type

    def copy(self):
        RandomAccessStringArrayType()


random_access_string_array = RandomAccessStringArrayType()


@register_model(RandomAccessStringArrayType)
class RandomAccessStringArrayModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        acnro__kql = [('data', types.List(string_type))]
        models.StructModel.__init__(self, dmm, fe_type, acnro__kql)


make_attribute_wrapper(RandomAccessStringArrayType, 'data', '_data')


@intrinsic
def alloc_random_access_string_array(typingctx, n_t=None):

    def codegen(context, builder, sig, args):
        isvob__pid, = args
        qskz__ileb = types.List(string_type)
        fsvj__zkq = numba.cpython.listobj.ListInstance.allocate(context,
            builder, qskz__ileb, isvob__pid)
        fsvj__zkq.size = isvob__pid
        rpscm__jwcpj = cgutils.create_struct_proxy(sig.return_type)(context,
            builder)
        rpscm__jwcpj.data = fsvj__zkq.value
        return rpscm__jwcpj._getvalue()
    return random_access_string_array(types.intp), codegen


@overload(operator.getitem, no_unliteral=True)
def random_access_str_arr_getitem(A, ind):
    if A != random_access_string_array:
        return
    if isinstance(ind, types.Integer):
        return lambda A, ind: A._data[ind]


@overload(operator.setitem)
def random_access_str_arr_setitem(A, idx, val):
    if A != random_access_string_array:
        return
    if isinstance(idx, types.Integer):
        assert val == string_type

        def impl_scalar(A, idx, val):
            A._data[idx] = val
        return impl_scalar


@overload(len, no_unliteral=True)
def overload_str_arr_len(A):
    if A == random_access_string_array:
        return lambda A: len(A._data)


@overload_attribute(RandomAccessStringArrayType, 'shape')
def overload_str_arr_shape(A):
    return lambda A: (len(A._data),)


def alloc_random_access_str_arr_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 1 and not kws
    return ArrayAnalysis.AnalyzeResult(shape=args[0], pre=[])


(ArrayAnalysis.
    _analyze_op_call_bodo_libs_str_ext_alloc_random_access_string_array
    ) = alloc_random_access_str_arr_equiv
str_from_float32 = types.ExternalFunction('str_from_float32', types.void(
    types.voidptr, types.float32))
str_from_float64 = types.ExternalFunction('str_from_float64', types.void(
    types.voidptr, types.float64))


def float_to_str(s, v):
    pass


@overload(float_to_str)
def float_to_str_overload(s, v):
    assert isinstance(v, types.Float)
    if v == types.float32:
        return lambda s, v: str_from_float32(s._data, v)
    return lambda s, v: str_from_float64(s._data, v)


@overload(str)
def float_str_overload(v):
    if isinstance(v, types.Float):
        kind = numba.cpython.unicode.PY_UNICODE_1BYTE_KIND

        def impl(v):
            if v == 0:
                return '0.0'
            sxut__rqana = 0
            okrc__hib = v
            if okrc__hib < 0:
                sxut__rqana = 1
                okrc__hib = -okrc__hib
            if okrc__hib < 1:
                vgsh__hjuqx = 1
            else:
                vgsh__hjuqx = 1 + int(np.floor(np.log10(okrc__hib)))
            length = sxut__rqana + vgsh__hjuqx + 1 + 6
            s = numba.cpython.unicode._malloc_string(kind, 1, length, True)
            float_to_str(s, v)
            return s
        return impl


@overload(format, no_unliteral=True)
def overload_format(value, format_spec=''):
    if is_overload_constant_str(format_spec) and get_overload_const_str(
        format_spec) == '':

        def impl_fast(value, format_spec=''):
            return str(value)
        return impl_fast

    def impl(value, format_spec=''):
        with numba.objmode(res='string'):
            res = format(value, format_spec)
        return res
    return impl


@lower_cast(StdStringType, types.float64)
def cast_str_to_float64(context, builder, fromty, toty, val):
    tzg__aozf = lir.FunctionType(lir.DoubleType(), [lir.IntType(8).
        as_pointer()])
    rycj__qix = cgutils.get_or_insert_function(builder.module, tzg__aozf,
        name='str_to_float64')
    res = builder.call(rycj__qix, (val,))
    bodo.utils.utils.inlined_check_and_propagate_cpp_exception(context, builder
        )
    return res


@lower_cast(StdStringType, types.float32)
def cast_str_to_float32(context, builder, fromty, toty, val):
    tzg__aozf = lir.FunctionType(lir.FloatType(), [lir.IntType(8).as_pointer()]
        )
    rycj__qix = cgutils.get_or_insert_function(builder.module, tzg__aozf,
        name='str_to_float32')
    res = builder.call(rycj__qix, (val,))
    bodo.utils.utils.inlined_check_and_propagate_cpp_exception(context, builder
        )
    return res


@lower_cast(string_type, types.float64)
def cast_unicode_str_to_float64(context, builder, fromty, toty, val):
    std_str = gen_unicode_to_std_str(context, builder, val)
    return cast_str_to_float64(context, builder, std_str_type, toty, std_str)


@lower_cast(string_type, types.float32)
def cast_unicode_str_to_float32(context, builder, fromty, toty, val):
    std_str = gen_unicode_to_std_str(context, builder, val)
    return cast_str_to_float32(context, builder, std_str_type, toty, std_str)


@lower_cast(string_type, types.int64)
@lower_cast(string_type, types.int32)
@lower_cast(string_type, types.int16)
@lower_cast(string_type, types.int8)
def cast_unicode_str_to_int64(context, builder, fromty, toty, val):
    jkxtf__umkr = cgutils.create_struct_proxy(string_type)(context, builder,
        value=val)
    tzg__aozf = lir.FunctionType(lir.IntType(toty.bitwidth), [lir.IntType(8
        ).as_pointer(), lir.IntType(64)])
    rycj__qix = cgutils.get_or_insert_function(builder.module, tzg__aozf,
        name='str_to_int64')
    res = builder.call(rycj__qix, (jkxtf__umkr.data, jkxtf__umkr.length))
    bodo.utils.utils.inlined_check_and_propagate_cpp_exception(context, builder
        )
    return res


@lower_cast(string_type, types.uint64)
@lower_cast(string_type, types.uint32)
@lower_cast(string_type, types.uint16)
@lower_cast(string_type, types.uint8)
def cast_unicode_str_to_uint64(context, builder, fromty, toty, val):
    jkxtf__umkr = cgutils.create_struct_proxy(string_type)(context, builder,
        value=val)
    tzg__aozf = lir.FunctionType(lir.IntType(toty.bitwidth), [lir.IntType(8
        ).as_pointer(), lir.IntType(64)])
    rycj__qix = cgutils.get_or_insert_function(builder.module, tzg__aozf,
        name='str_to_uint64')
    res = builder.call(rycj__qix, (jkxtf__umkr.data, jkxtf__umkr.length))
    bodo.utils.utils.inlined_check_and_propagate_cpp_exception(context, builder
        )
    return res


@infer_getattr
class StringAttribute(AttributeTemplate):
    key = types.UnicodeType

    @bound_function('str.format', no_unliteral=True)
    def resolve_format(self, string_typ, args, kws):
        kws = dict(kws)
        yooi__gqag = ', '.join('e{}'.format(ocz__rrbha) for ocz__rrbha in
            range(len(args)))
        if yooi__gqag:
            yooi__gqag += ', '
        ztdv__faf = ', '.join("{} = ''".format(a) for a in kws.keys())
        ldhhd__mqdr = f'def format_stub(string, {yooi__gqag} {ztdv__faf}):\n'
        ldhhd__mqdr += '    pass\n'
        dueba__psagf = {}
        exec(ldhhd__mqdr, {}, dueba__psagf)
        sei__ysysl = dueba__psagf['format_stub']
        hlity__fxwe = numba.core.utils.pysignature(sei__ysysl)
        nole__iosf = (string_typ,) + args + tuple(kws.values())
        return signature(string_typ, nole__iosf).replace(pysig=hlity__fxwe)


@numba.njit(cache=True)
def str_split(arr, pat, n):
    cas__siuyg = pat is not None and len(pat) > 1
    if cas__siuyg:
        pdxek__yww = re.compile(pat)
        if n == -1:
            n = 0
    elif n == 0:
        n = -1
    fsvj__zkq = len(arr)
    hfjm__mpl = 0
    eoiba__khnhk = 0
    for ocz__rrbha in numba.parfors.parfor.internal_prange(fsvj__zkq):
        if bodo.libs.array_kernels.isna(arr, ocz__rrbha):
            continue
        if cas__siuyg:
            xkq__jqcv = pdxek__yww.split(arr[ocz__rrbha], maxsplit=n)
        elif pat == '':
            xkq__jqcv = [''] + list(arr[ocz__rrbha]) + ['']
        else:
            xkq__jqcv = arr[ocz__rrbha].split(pat, n)
        hfjm__mpl += len(xkq__jqcv)
        for s in xkq__jqcv:
            eoiba__khnhk += bodo.libs.str_arr_ext.get_utf8_size(s)
    sea__ejw = bodo.libs.array_item_arr_ext.pre_alloc_array_item_array(
        fsvj__zkq, (hfjm__mpl, eoiba__khnhk), bodo.libs.str_arr_ext.
        string_array_type)
    vgrwv__ivsv = bodo.libs.array_item_arr_ext.get_offsets(sea__ejw)
    vvi__cpvsq = bodo.libs.array_item_arr_ext.get_null_bitmap(sea__ejw)
    ktq__ezxi = bodo.libs.array_item_arr_ext.get_data(sea__ejw)
    budl__ydrn = 0
    for vobs__rqte in numba.parfors.parfor.internal_prange(fsvj__zkq):
        vgrwv__ivsv[vobs__rqte] = budl__ydrn
        if bodo.libs.array_kernels.isna(arr, vobs__rqte):
            bodo.libs.int_arr_ext.set_bit_to_arr(vvi__cpvsq, vobs__rqte, 0)
            continue
        bodo.libs.int_arr_ext.set_bit_to_arr(vvi__cpvsq, vobs__rqte, 1)
        if cas__siuyg:
            xkq__jqcv = pdxek__yww.split(arr[vobs__rqte], maxsplit=n)
        elif pat == '':
            xkq__jqcv = [''] + list(arr[vobs__rqte]) + ['']
        else:
            xkq__jqcv = arr[vobs__rqte].split(pat, n)
        iawca__qisq = len(xkq__jqcv)
        for xmyi__zirav in range(iawca__qisq):
            s = xkq__jqcv[xmyi__zirav]
            ktq__ezxi[budl__ydrn] = s
            budl__ydrn += 1
    vgrwv__ivsv[fsvj__zkq] = budl__ydrn
    return sea__ejw


@overload(hex)
def overload_hex(x):
    if isinstance(x, types.Integer):
        kind = numba.cpython.unicode.PY_UNICODE_1BYTE_KIND

        def impl(x):
            x = np.int64(x)
            if x < 0:
                ynh__kasd = '-0x'
                x = x * -1
            else:
                ynh__kasd = '0x'
            x = np.uint64(x)
            if x == 0:
                aoow__hjbc = 1
            else:
                aoow__hjbc = fast_ceil_log2(x + 1)
                aoow__hjbc = (aoow__hjbc + 3) // 4
            length = len(ynh__kasd) + aoow__hjbc
            output = numba.cpython.unicode._empty_string(kind, length, 1)
            bodo.libs.str_arr_ext._memcpy(output._data, ynh__kasd._data,
                len(ynh__kasd), 1)
            int_to_hex(output, aoow__hjbc, len(ynh__kasd), x)
            return output
        return impl


@register_jitable
def fast_ceil_log2(x):
    coj__tnaxc = 0 if x & x - 1 == 0 else 1
    xonk__sob = [np.uint64(18446744069414584320), np.uint64(4294901760), np
        .uint64(65280), np.uint64(240), np.uint64(12), np.uint64(2)]
    lnpal__epzjf = 32
    for ocz__rrbha in range(len(xonk__sob)):
        paa__xvzyq = 0 if x & xonk__sob[ocz__rrbha] == 0 else lnpal__epzjf
        coj__tnaxc = coj__tnaxc + paa__xvzyq
        x = x >> paa__xvzyq
        lnpal__epzjf = lnpal__epzjf >> 1
    return coj__tnaxc


@intrinsic
def int_to_hex(typingctx, output, out_len, header_len, int_val):

    def codegen(context, builder, sig, args):
        output, out_len, header_len, int_val = args
        mvzru__ngq = cgutils.create_struct_proxy(sig.args[0])(context,
            builder, value=output)
        tzg__aozf = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer(), lir.IntType(64), lir.IntType(64)])
        hsu__ummw = cgutils.get_or_insert_function(builder.module,
            tzg__aozf, name='int_to_hex')
        cntsa__vdk = builder.inttoptr(builder.add(builder.ptrtoint(
            mvzru__ngq.data, lir.IntType(64)), header_len), lir.IntType(8).
            as_pointer())
        builder.call(hsu__ummw, (cntsa__vdk, out_len, int_val))
    return types.void(output, out_len, header_len, int_val), codegen


def alloc_empty_bytes_or_string_data(typ, kind, length, is_ascii=0):
    pass


@overload(alloc_empty_bytes_or_string_data)
def overload_alloc_empty_bytes_or_string_data(typ, kind, length, is_ascii=0):
    typ = typ.instance_type if isinstance(typ, types.TypeRef) else typ
    if typ == bodo.bytes_type:
        return lambda typ, kind, length, is_ascii=0: np.empty(length, np.uint8)
    if typ == string_type:
        return (lambda typ, kind, length, is_ascii=0: numba.cpython.unicode
            ._empty_string(kind, length, is_ascii))
    raise BodoError(
        f'Internal Error: Expected Bytes or String type, found {typ}')


def get_unicode_or_numpy_data(val):
    pass


@overload(get_unicode_or_numpy_data)
def overload_get_unicode_or_numpy_data(val):
    if val == string_type:
        return lambda val: val._data
    if isinstance(val, types.Array):
        return lambda val: val.ctypes
    raise BodoError(
        f'Internal Error: Expected String or Numpy Array, found {val}')
