import numpy as np
import pandas as pd
import numba
from numba.extending import overload
from bodo.utils.utils import alloc_arr_tup
MIN_MERGE = 32


@numba.njit(no_cpython_wrapper=True, cache=True)
def sort(key_arrs, lo, hi, data):
    zfy__spg = hi - lo
    if zfy__spg < 2:
        return
    if zfy__spg < MIN_MERGE:
        wiyve__eaie = countRunAndMakeAscending(key_arrs, lo, hi, data)
        binarySort(key_arrs, lo, hi, lo + wiyve__eaie, data)
        return
    stackSize, runBase, runLen, tmpLength, tmp, tmp_data, minGallop = (
        init_sort_start(key_arrs, data))
    ufk__pmoh = minRunLength(zfy__spg)
    while True:
        becww__ptmw = countRunAndMakeAscending(key_arrs, lo, hi, data)
        if becww__ptmw < ufk__pmoh:
            xmtls__jhj = zfy__spg if zfy__spg <= ufk__pmoh else ufk__pmoh
            binarySort(key_arrs, lo, lo + xmtls__jhj, lo + becww__ptmw, data)
            becww__ptmw = xmtls__jhj
        stackSize = pushRun(stackSize, runBase, runLen, lo, becww__ptmw)
        stackSize, tmpLength, tmp, tmp_data, minGallop = mergeCollapse(
            stackSize, runBase, runLen, key_arrs, data, tmpLength, tmp,
            tmp_data, minGallop)
        lo += becww__ptmw
        zfy__spg -= becww__ptmw
        if zfy__spg == 0:
            break
    assert lo == hi
    stackSize, tmpLength, tmp, tmp_data, minGallop = mergeForceCollapse(
        stackSize, runBase, runLen, key_arrs, data, tmpLength, tmp,
        tmp_data, minGallop)
    assert stackSize == 1


@numba.njit(no_cpython_wrapper=True, cache=True)
def binarySort(key_arrs, lo, hi, start, data):
    assert lo <= start and start <= hi
    if start == lo:
        start += 1
    while start < hi:
        vrana__kcn = getitem_arr_tup(key_arrs, start)
        byuhs__xiak = getitem_arr_tup(data, start)
        putru__xwjvs = lo
        whyz__yfeq = start
        assert putru__xwjvs <= whyz__yfeq
        while putru__xwjvs < whyz__yfeq:
            tzmq__yglk = putru__xwjvs + whyz__yfeq >> 1
            if vrana__kcn < getitem_arr_tup(key_arrs, tzmq__yglk):
                whyz__yfeq = tzmq__yglk
            else:
                putru__xwjvs = tzmq__yglk + 1
        assert putru__xwjvs == whyz__yfeq
        n = start - putru__xwjvs
        copyRange_tup(key_arrs, putru__xwjvs, key_arrs, putru__xwjvs + 1, n)
        copyRange_tup(data, putru__xwjvs, data, putru__xwjvs + 1, n)
        setitem_arr_tup(key_arrs, putru__xwjvs, vrana__kcn)
        setitem_arr_tup(data, putru__xwjvs, byuhs__xiak)
        start += 1


@numba.njit(no_cpython_wrapper=True, cache=True)
def countRunAndMakeAscending(key_arrs, lo, hi, data):
    assert lo < hi
    szht__gokz = lo + 1
    if szht__gokz == hi:
        return 1
    if getitem_arr_tup(key_arrs, szht__gokz) < getitem_arr_tup(key_arrs, lo):
        szht__gokz += 1
        while szht__gokz < hi and getitem_arr_tup(key_arrs, szht__gokz
            ) < getitem_arr_tup(key_arrs, szht__gokz - 1):
            szht__gokz += 1
        reverseRange(key_arrs, lo, szht__gokz, data)
    else:
        szht__gokz += 1
        while szht__gokz < hi and getitem_arr_tup(key_arrs, szht__gokz
            ) >= getitem_arr_tup(key_arrs, szht__gokz - 1):
            szht__gokz += 1
    return szht__gokz - lo


@numba.njit(no_cpython_wrapper=True, cache=True)
def reverseRange(key_arrs, lo, hi, data):
    hi -= 1
    while lo < hi:
        swap_arrs(key_arrs, lo, hi)
        swap_arrs(data, lo, hi)
        lo += 1
        hi -= 1


@numba.njit(no_cpython_wrapper=True, cache=True)
def minRunLength(n):
    assert n >= 0
    twgx__qrreh = 0
    while n >= MIN_MERGE:
        twgx__qrreh |= n & 1
        n >>= 1
    return n + twgx__qrreh


MIN_GALLOP = 7
INITIAL_TMP_STORAGE_LENGTH = 256


@numba.njit(no_cpython_wrapper=True, cache=True)
def init_sort_start(key_arrs, data):
    minGallop = MIN_GALLOP
    kwgj__tqfsz = len(key_arrs[0])
    tmpLength = (kwgj__tqfsz >> 1 if kwgj__tqfsz < 2 *
        INITIAL_TMP_STORAGE_LENGTH else INITIAL_TMP_STORAGE_LENGTH)
    tmp = alloc_arr_tup(tmpLength, key_arrs)
    tmp_data = alloc_arr_tup(tmpLength, data)
    stackSize = 0
    pvywz__uju = (5 if kwgj__tqfsz < 120 else 10 if kwgj__tqfsz < 1542 else
        19 if kwgj__tqfsz < 119151 else 40)
    runBase = np.empty(pvywz__uju, np.int64)
    runLen = np.empty(pvywz__uju, np.int64)
    return stackSize, runBase, runLen, tmpLength, tmp, tmp_data, minGallop


@numba.njit(no_cpython_wrapper=True, cache=True)
def pushRun(stackSize, runBase, runLen, runBase_val, runLen_val):
    runBase[stackSize] = runBase_val
    runLen[stackSize] = runLen_val
    stackSize += 1
    return stackSize


@numba.njit(no_cpython_wrapper=True, cache=True)
def mergeCollapse(stackSize, runBase, runLen, key_arrs, data, tmpLength,
    tmp, tmp_data, minGallop):
    while stackSize > 1:
        n = stackSize - 2
        if n >= 1 and runLen[n - 1] <= runLen[n] + runLen[n + 1
            ] or n >= 2 and runLen[n - 2] <= runLen[n] + runLen[n - 1]:
            if runLen[n - 1] < runLen[n + 1]:
                n -= 1
        elif runLen[n] > runLen[n + 1]:
            break
        stackSize, tmpLength, tmp, tmp_data, minGallop = mergeAt(stackSize,
            runBase, runLen, key_arrs, data, tmpLength, tmp, tmp_data,
            minGallop, n)
    return stackSize, tmpLength, tmp, tmp_data, minGallop


@numba.njit(no_cpython_wrapper=True, cache=True)
def mergeForceCollapse(stackSize, runBase, runLen, key_arrs, data,
    tmpLength, tmp, tmp_data, minGallop):
    while stackSize > 1:
        n = stackSize - 2
        if n > 0 and runLen[n - 1] < runLen[n + 1]:
            n -= 1
        stackSize, tmpLength, tmp, tmp_data, minGallop = mergeAt(stackSize,
            runBase, runLen, key_arrs, data, tmpLength, tmp, tmp_data,
            minGallop, n)
    return stackSize, tmpLength, tmp, tmp_data, minGallop


@numba.njit(no_cpython_wrapper=True, cache=True)
def mergeAt(stackSize, runBase, runLen, key_arrs, data, tmpLength, tmp,
    tmp_data, minGallop, i):
    assert stackSize >= 2
    assert i >= 0
    assert i == stackSize - 2 or i == stackSize - 3
    base1 = runBase[i]
    len1 = runLen[i]
    base2 = runBase[i + 1]
    len2 = runLen[i + 1]
    assert len1 > 0 and len2 > 0
    assert base1 + len1 == base2
    runLen[i] = len1 + len2
    if i == stackSize - 3:
        runBase[i + 1] = runBase[i + 2]
        runLen[i + 1] = runLen[i + 2]
    stackSize -= 1
    msw__sobws = gallopRight(getitem_arr_tup(key_arrs, base2), key_arrs,
        base1, len1, 0)
    assert msw__sobws >= 0
    base1 += msw__sobws
    len1 -= msw__sobws
    if len1 == 0:
        return stackSize, tmpLength, tmp, tmp_data, minGallop
    len2 = gallopLeft(getitem_arr_tup(key_arrs, base1 + len1 - 1), key_arrs,
        base2, len2, len2 - 1)
    assert len2 >= 0
    if len2 == 0:
        return stackSize, tmpLength, tmp, tmp_data, minGallop
    if len1 <= len2:
        tmpLength, tmp, tmp_data = ensureCapacity(tmpLength, tmp, tmp_data,
            key_arrs, data, len1)
        minGallop = mergeLo(key_arrs, data, tmp, tmp_data, minGallop, base1,
            len1, base2, len2)
    else:
        tmpLength, tmp, tmp_data = ensureCapacity(tmpLength, tmp, tmp_data,
            key_arrs, data, len2)
        minGallop = mergeHi(key_arrs, data, tmp, tmp_data, minGallop, base1,
            len1, base2, len2)
    return stackSize, tmpLength, tmp, tmp_data, minGallop


@numba.njit(no_cpython_wrapper=True, cache=True)
def gallopLeft(key, arr, base, _len, hint):
    assert _len > 0 and hint >= 0 and hint < _len
    tceig__tsmp = 0
    hijbk__xrfpn = 1
    if key > getitem_arr_tup(arr, base + hint):
        hnql__usyt = _len - hint
        while hijbk__xrfpn < hnql__usyt and key > getitem_arr_tup(arr, base +
            hint + hijbk__xrfpn):
            tceig__tsmp = hijbk__xrfpn
            hijbk__xrfpn = (hijbk__xrfpn << 1) + 1
            if hijbk__xrfpn <= 0:
                hijbk__xrfpn = hnql__usyt
        if hijbk__xrfpn > hnql__usyt:
            hijbk__xrfpn = hnql__usyt
        tceig__tsmp += hint
        hijbk__xrfpn += hint
    else:
        hnql__usyt = hint + 1
        while hijbk__xrfpn < hnql__usyt and key <= getitem_arr_tup(arr, 
            base + hint - hijbk__xrfpn):
            tceig__tsmp = hijbk__xrfpn
            hijbk__xrfpn = (hijbk__xrfpn << 1) + 1
            if hijbk__xrfpn <= 0:
                hijbk__xrfpn = hnql__usyt
        if hijbk__xrfpn > hnql__usyt:
            hijbk__xrfpn = hnql__usyt
        tmp = tceig__tsmp
        tceig__tsmp = hint - hijbk__xrfpn
        hijbk__xrfpn = hint - tmp
    assert -1 <= tceig__tsmp and tceig__tsmp < hijbk__xrfpn and hijbk__xrfpn <= _len
    tceig__tsmp += 1
    while tceig__tsmp < hijbk__xrfpn:
        nyuj__hdl = tceig__tsmp + (hijbk__xrfpn - tceig__tsmp >> 1)
        if key > getitem_arr_tup(arr, base + nyuj__hdl):
            tceig__tsmp = nyuj__hdl + 1
        else:
            hijbk__xrfpn = nyuj__hdl
    assert tceig__tsmp == hijbk__xrfpn
    return hijbk__xrfpn


@numba.njit(no_cpython_wrapper=True, cache=True)
def gallopRight(key, arr, base, _len, hint):
    assert _len > 0 and hint >= 0 and hint < _len
    hijbk__xrfpn = 1
    tceig__tsmp = 0
    if key < getitem_arr_tup(arr, base + hint):
        hnql__usyt = hint + 1
        while hijbk__xrfpn < hnql__usyt and key < getitem_arr_tup(arr, base +
            hint - hijbk__xrfpn):
            tceig__tsmp = hijbk__xrfpn
            hijbk__xrfpn = (hijbk__xrfpn << 1) + 1
            if hijbk__xrfpn <= 0:
                hijbk__xrfpn = hnql__usyt
        if hijbk__xrfpn > hnql__usyt:
            hijbk__xrfpn = hnql__usyt
        tmp = tceig__tsmp
        tceig__tsmp = hint - hijbk__xrfpn
        hijbk__xrfpn = hint - tmp
    else:
        hnql__usyt = _len - hint
        while hijbk__xrfpn < hnql__usyt and key >= getitem_arr_tup(arr, 
            base + hint + hijbk__xrfpn):
            tceig__tsmp = hijbk__xrfpn
            hijbk__xrfpn = (hijbk__xrfpn << 1) + 1
            if hijbk__xrfpn <= 0:
                hijbk__xrfpn = hnql__usyt
        if hijbk__xrfpn > hnql__usyt:
            hijbk__xrfpn = hnql__usyt
        tceig__tsmp += hint
        hijbk__xrfpn += hint
    assert -1 <= tceig__tsmp and tceig__tsmp < hijbk__xrfpn and hijbk__xrfpn <= _len
    tceig__tsmp += 1
    while tceig__tsmp < hijbk__xrfpn:
        nyuj__hdl = tceig__tsmp + (hijbk__xrfpn - tceig__tsmp >> 1)
        if key < getitem_arr_tup(arr, base + nyuj__hdl):
            hijbk__xrfpn = nyuj__hdl
        else:
            tceig__tsmp = nyuj__hdl + 1
    assert tceig__tsmp == hijbk__xrfpn
    return hijbk__xrfpn


@numba.njit(no_cpython_wrapper=True, cache=True)
def mergeLo(key_arrs, data, tmp, tmp_data, minGallop, base1, len1, base2, len2
    ):
    assert len1 > 0 and len2 > 0 and base1 + len1 == base2
    arr = key_arrs
    arr_data = data
    copyRange_tup(arr, base1, tmp, 0, len1)
    copyRange_tup(arr_data, base1, tmp_data, 0, len1)
    cursor1 = 0
    cursor2 = base2
    dest = base1
    setitem_arr_tup(arr, dest, getitem_arr_tup(arr, cursor2))
    copyElement_tup(arr_data, cursor2, arr_data, dest)
    cursor2 += 1
    dest += 1
    len2 -= 1
    if len2 == 0:
        copyRange_tup(tmp, cursor1, arr, dest, len1)
        copyRange_tup(tmp_data, cursor1, arr_data, dest, len1)
        return minGallop
    if len1 == 1:
        copyRange_tup(arr, cursor2, arr, dest, len2)
        copyRange_tup(arr_data, cursor2, arr_data, dest, len2)
        copyElement_tup(tmp, cursor1, arr, dest + len2)
        copyElement_tup(tmp_data, cursor1, arr_data, dest + len2)
        return minGallop
    len1, len2, cursor1, cursor2, dest, minGallop = mergeLo_inner(key_arrs,
        data, tmp_data, len1, len2, tmp, cursor1, cursor2, dest, minGallop)
    minGallop = 1 if minGallop < 1 else minGallop
    if len1 == 1:
        assert len2 > 0
        copyRange_tup(arr, cursor2, arr, dest, len2)
        copyRange_tup(arr_data, cursor2, arr_data, dest, len2)
        copyElement_tup(tmp, cursor1, arr, dest + len2)
        copyElement_tup(tmp_data, cursor1, arr_data, dest + len2)
    elif len1 == 0:
        raise ValueError('Comparison method violates its general contract!')
    else:
        assert len2 == 0
        assert len1 > 1
        copyRange_tup(tmp, cursor1, arr, dest, len1)
        copyRange_tup(tmp_data, cursor1, arr_data, dest, len1)
    return minGallop


@numba.njit(no_cpython_wrapper=True, cache=True)
def mergeLo_inner(arr, arr_data, tmp_data, len1, len2, tmp, cursor1,
    cursor2, dest, minGallop):
    while True:
        dli__yxy = 0
        dex__zny = 0
        while True:
            assert len1 > 1 and len2 > 0
            if getitem_arr_tup(arr, cursor2) < getitem_arr_tup(tmp, cursor1):
                copyElement_tup(arr, cursor2, arr, dest)
                copyElement_tup(arr_data, cursor2, arr_data, dest)
                cursor2 += 1
                dest += 1
                dex__zny += 1
                dli__yxy = 0
                len2 -= 1
                if len2 == 0:
                    return len1, len2, cursor1, cursor2, dest, minGallop
            else:
                copyElement_tup(tmp, cursor1, arr, dest)
                copyElement_tup(tmp_data, cursor1, arr_data, dest)
                cursor1 += 1
                dest += 1
                dli__yxy += 1
                dex__zny = 0
                len1 -= 1
                if len1 == 1:
                    return len1, len2, cursor1, cursor2, dest, minGallop
            if not dli__yxy | dex__zny < minGallop:
                break
        while True:
            assert len1 > 1 and len2 > 0
            dli__yxy = gallopRight(getitem_arr_tup(arr, cursor2), tmp,
                cursor1, len1, 0)
            if dli__yxy != 0:
                copyRange_tup(tmp, cursor1, arr, dest, dli__yxy)
                copyRange_tup(tmp_data, cursor1, arr_data, dest, dli__yxy)
                dest += dli__yxy
                cursor1 += dli__yxy
                len1 -= dli__yxy
                if len1 <= 1:
                    return len1, len2, cursor1, cursor2, dest, minGallop
            copyElement_tup(arr, cursor2, arr, dest)
            copyElement_tup(arr_data, cursor2, arr_data, dest)
            cursor2 += 1
            dest += 1
            len2 -= 1
            if len2 == 0:
                return len1, len2, cursor1, cursor2, dest, minGallop
            dex__zny = gallopLeft(getitem_arr_tup(tmp, cursor1), arr,
                cursor2, len2, 0)
            if dex__zny != 0:
                copyRange_tup(arr, cursor2, arr, dest, dex__zny)
                copyRange_tup(arr_data, cursor2, arr_data, dest, dex__zny)
                dest += dex__zny
                cursor2 += dex__zny
                len2 -= dex__zny
                if len2 == 0:
                    return len1, len2, cursor1, cursor2, dest, minGallop
            copyElement_tup(tmp, cursor1, arr, dest)
            copyElement_tup(tmp_data, cursor1, arr_data, dest)
            cursor1 += 1
            dest += 1
            len1 -= 1
            if len1 == 1:
                return len1, len2, cursor1, cursor2, dest, minGallop
            minGallop -= 1
            if not dli__yxy >= MIN_GALLOP | dex__zny >= MIN_GALLOP:
                break
        if minGallop < 0:
            minGallop = 0
        minGallop += 2
    return len1, len2, cursor1, cursor2, dest, minGallop


@numba.njit(no_cpython_wrapper=True, cache=True)
def mergeHi(key_arrs, data, tmp, tmp_data, minGallop, base1, len1, base2, len2
    ):
    assert len1 > 0 and len2 > 0 and base1 + len1 == base2
    arr = key_arrs
    arr_data = data
    copyRange_tup(arr, base2, tmp, 0, len2)
    copyRange_tup(arr_data, base2, tmp_data, 0, len2)
    cursor1 = base1 + len1 - 1
    cursor2 = len2 - 1
    dest = base2 + len2 - 1
    copyElement_tup(arr, cursor1, arr, dest)
    copyElement_tup(arr_data, cursor1, arr_data, dest)
    cursor1 -= 1
    dest -= 1
    len1 -= 1
    if len1 == 0:
        copyRange_tup(tmp, 0, arr, dest - (len2 - 1), len2)
        copyRange_tup(tmp_data, 0, arr_data, dest - (len2 - 1), len2)
        return minGallop
    if len2 == 1:
        dest -= len1
        cursor1 -= len1
        copyRange_tup(arr, cursor1 + 1, arr, dest + 1, len1)
        copyRange_tup(arr_data, cursor1 + 1, arr_data, dest + 1, len1)
        copyElement_tup(tmp, cursor2, arr, dest)
        copyElement_tup(tmp_data, cursor2, arr_data, dest)
        return minGallop
    len1, len2, tmp, cursor1, cursor2, dest, minGallop = mergeHi_inner(key_arrs
        , data, tmp_data, base1, len1, len2, tmp, cursor1, cursor2, dest,
        minGallop)
    minGallop = 1 if minGallop < 1 else minGallop
    if len2 == 1:
        assert len1 > 0
        dest -= len1
        cursor1 -= len1
        copyRange_tup(arr, cursor1 + 1, arr, dest + 1, len1)
        copyRange_tup(arr_data, cursor1 + 1, arr_data, dest + 1, len1)
        copyElement_tup(tmp, cursor2, arr, dest)
        copyElement_tup(tmp_data, cursor2, arr_data, dest)
    elif len2 == 0:
        raise ValueError('Comparison method violates its general contract!')
    else:
        assert len1 == 0
        assert len2 > 0
        copyRange_tup(tmp, 0, arr, dest - (len2 - 1), len2)
        copyRange_tup(tmp_data, 0, arr_data, dest - (len2 - 1), len2)
    return minGallop


@numba.njit(no_cpython_wrapper=True, cache=True)
def mergeHi_inner(arr, arr_data, tmp_data, base1, len1, len2, tmp, cursor1,
    cursor2, dest, minGallop):
    while True:
        dli__yxy = 0
        dex__zny = 0
        while True:
            assert len1 > 0 and len2 > 1
            if getitem_arr_tup(tmp, cursor2) < getitem_arr_tup(arr, cursor1):
                copyElement_tup(arr, cursor1, arr, dest)
                copyElement_tup(arr_data, cursor1, arr_data, dest)
                cursor1 -= 1
                dest -= 1
                dli__yxy += 1
                dex__zny = 0
                len1 -= 1
                if len1 == 0:
                    return len1, len2, tmp, cursor1, cursor2, dest, minGallop
            else:
                copyElement_tup(tmp, cursor2, arr, dest)
                copyElement_tup(tmp_data, cursor2, arr_data, dest)
                cursor2 -= 1
                dest -= 1
                dex__zny += 1
                dli__yxy = 0
                len2 -= 1
                if len2 == 1:
                    return len1, len2, tmp, cursor1, cursor2, dest, minGallop
            if not dli__yxy | dex__zny < minGallop:
                break
        while True:
            assert len1 > 0 and len2 > 1
            dli__yxy = len1 - gallopRight(getitem_arr_tup(tmp, cursor2),
                arr, base1, len1, len1 - 1)
            if dli__yxy != 0:
                dest -= dli__yxy
                cursor1 -= dli__yxy
                len1 -= dli__yxy
                copyRange_tup(arr, cursor1 + 1, arr, dest + 1, dli__yxy)
                copyRange_tup(arr_data, cursor1 + 1, arr_data, dest + 1,
                    dli__yxy)
                if len1 == 0:
                    return len1, len2, tmp, cursor1, cursor2, dest, minGallop
            copyElement_tup(tmp, cursor2, arr, dest)
            copyElement_tup(tmp_data, cursor2, arr_data, dest)
            cursor2 -= 1
            dest -= 1
            len2 -= 1
            if len2 == 1:
                return len1, len2, tmp, cursor1, cursor2, dest, minGallop
            dex__zny = len2 - gallopLeft(getitem_arr_tup(arr, cursor1), tmp,
                0, len2, len2 - 1)
            if dex__zny != 0:
                dest -= dex__zny
                cursor2 -= dex__zny
                len2 -= dex__zny
                copyRange_tup(tmp, cursor2 + 1, arr, dest + 1, dex__zny)
                copyRange_tup(tmp_data, cursor2 + 1, arr_data, dest + 1,
                    dex__zny)
                if len2 <= 1:
                    return len1, len2, tmp, cursor1, cursor2, dest, minGallop
            copyElement_tup(arr, cursor1, arr, dest)
            copyElement_tup(arr_data, cursor1, arr_data, dest)
            cursor1 -= 1
            dest -= 1
            len1 -= 1
            if len1 == 0:
                return len1, len2, tmp, cursor1, cursor2, dest, minGallop
            minGallop -= 1
            if not dli__yxy >= MIN_GALLOP | dex__zny >= MIN_GALLOP:
                break
        if minGallop < 0:
            minGallop = 0
        minGallop += 2
    return len1, len2, tmp, cursor1, cursor2, dest, minGallop


@numba.njit(no_cpython_wrapper=True, cache=True)
def ensureCapacity(tmpLength, tmp, tmp_data, key_arrs, data, minCapacity):
    skqi__guqx = len(key_arrs[0])
    if tmpLength < minCapacity:
        gzxy__kzp = minCapacity
        gzxy__kzp |= gzxy__kzp >> 1
        gzxy__kzp |= gzxy__kzp >> 2
        gzxy__kzp |= gzxy__kzp >> 4
        gzxy__kzp |= gzxy__kzp >> 8
        gzxy__kzp |= gzxy__kzp >> 16
        gzxy__kzp += 1
        if gzxy__kzp < 0:
            gzxy__kzp = minCapacity
        else:
            gzxy__kzp = min(gzxy__kzp, skqi__guqx >> 1)
        tmp = alloc_arr_tup(gzxy__kzp, key_arrs)
        tmp_data = alloc_arr_tup(gzxy__kzp, data)
        tmpLength = gzxy__kzp
    return tmpLength, tmp, tmp_data


def swap_arrs(data, lo, hi):
    for arr in data:
        qpd__aghf = arr[lo]
        arr[lo] = arr[hi]
        arr[hi] = qpd__aghf


@overload(swap_arrs, no_unliteral=True)
def swap_arrs_overload(arr_tup, lo, hi):
    brg__bzobx = arr_tup.count
    sgmcx__ejnl = 'def f(arr_tup, lo, hi):\n'
    for i in range(brg__bzobx):
        sgmcx__ejnl += '  tmp_v_{} = arr_tup[{}][lo]\n'.format(i, i)
        sgmcx__ejnl += '  arr_tup[{}][lo] = arr_tup[{}][hi]\n'.format(i, i)
        sgmcx__ejnl += '  arr_tup[{}][hi] = tmp_v_{}\n'.format(i, i)
    sgmcx__ejnl += '  return\n'
    jgxht__gvty = {}
    exec(sgmcx__ejnl, {}, jgxht__gvty)
    fyakn__cgev = jgxht__gvty['f']
    return fyakn__cgev


@numba.njit(no_cpython_wrapper=True, cache=True)
def copyRange(src_arr, src_pos, dst_arr, dst_pos, n):
    dst_arr[dst_pos:dst_pos + n] = src_arr[src_pos:src_pos + n]


def copyRange_tup(src_arr_tup, src_pos, dst_arr_tup, dst_pos, n):
    for src_arr, dst_arr in zip(src_arr_tup, dst_arr_tup):
        dst_arr[dst_pos:dst_pos + n] = src_arr[src_pos:src_pos + n]


@overload(copyRange_tup, no_unliteral=True)
def copyRange_tup_overload(src_arr_tup, src_pos, dst_arr_tup, dst_pos, n):
    brg__bzobx = src_arr_tup.count
    assert brg__bzobx == dst_arr_tup.count
    sgmcx__ejnl = 'def f(src_arr_tup, src_pos, dst_arr_tup, dst_pos, n):\n'
    for i in range(brg__bzobx):
        sgmcx__ejnl += (
            '  copyRange(src_arr_tup[{}], src_pos, dst_arr_tup[{}], dst_pos, n)\n'
            .format(i, i))
    sgmcx__ejnl += '  return\n'
    jgxht__gvty = {}
    exec(sgmcx__ejnl, {'copyRange': copyRange}, jgxht__gvty)
    yakv__dnnvs = jgxht__gvty['f']
    return yakv__dnnvs


@numba.njit(no_cpython_wrapper=True, cache=True)
def copyElement(src_arr, src_pos, dst_arr, dst_pos):
    dst_arr[dst_pos] = src_arr[src_pos]


def copyElement_tup(src_arr_tup, src_pos, dst_arr_tup, dst_pos):
    for src_arr, dst_arr in zip(src_arr_tup, dst_arr_tup):
        dst_arr[dst_pos] = src_arr[src_pos]


@overload(copyElement_tup, no_unliteral=True)
def copyElement_tup_overload(src_arr_tup, src_pos, dst_arr_tup, dst_pos):
    brg__bzobx = src_arr_tup.count
    assert brg__bzobx == dst_arr_tup.count
    sgmcx__ejnl = 'def f(src_arr_tup, src_pos, dst_arr_tup, dst_pos):\n'
    for i in range(brg__bzobx):
        sgmcx__ejnl += (
            '  copyElement(src_arr_tup[{}], src_pos, dst_arr_tup[{}], dst_pos)\n'
            .format(i, i))
    sgmcx__ejnl += '  return\n'
    jgxht__gvty = {}
    exec(sgmcx__ejnl, {'copyElement': copyElement}, jgxht__gvty)
    yakv__dnnvs = jgxht__gvty['f']
    return yakv__dnnvs


def getitem_arr_tup(arr_tup, ind):
    ybmq__frffl = [arr[ind] for arr in arr_tup]
    return tuple(ybmq__frffl)


@overload(getitem_arr_tup, no_unliteral=True)
def getitem_arr_tup_overload(arr_tup, ind):
    brg__bzobx = arr_tup.count
    sgmcx__ejnl = 'def f(arr_tup, ind):\n'
    sgmcx__ejnl += '  return ({}{})\n'.format(','.join(['arr_tup[{}][ind]'.
        format(i) for i in range(brg__bzobx)]), ',' if brg__bzobx == 1 else '')
    jgxht__gvty = {}
    exec(sgmcx__ejnl, {}, jgxht__gvty)
    geczx__accsx = jgxht__gvty['f']
    return geczx__accsx


def setitem_arr_tup(arr_tup, ind, val_tup):
    for arr, mpxoj__kml in zip(arr_tup, val_tup):
        arr[ind] = mpxoj__kml


@overload(setitem_arr_tup, no_unliteral=True)
def setitem_arr_tup_overload(arr_tup, ind, val_tup):
    brg__bzobx = arr_tup.count
    sgmcx__ejnl = 'def f(arr_tup, ind, val_tup):\n'
    for i in range(brg__bzobx):
        if isinstance(val_tup, numba.core.types.BaseTuple):
            sgmcx__ejnl += '  arr_tup[{}][ind] = val_tup[{}]\n'.format(i, i)
        else:
            assert arr_tup.count == 1
            sgmcx__ejnl += '  arr_tup[{}][ind] = val_tup\n'.format(i)
    sgmcx__ejnl += '  return\n'
    jgxht__gvty = {}
    exec(sgmcx__ejnl, {}, jgxht__gvty)
    geczx__accsx = jgxht__gvty['f']
    return geczx__accsx


def test():
    import time
    ealzm__mwqed = time.time()
    vjmn__psqo = np.ones(3)
    data = np.arange(3), np.ones(3)
    sort((vjmn__psqo,), 0, 3, data)
    print('compile time', time.time() - ealzm__mwqed)
    n = 210000
    np.random.seed(2)
    data = np.arange(n), np.random.ranf(n)
    ims__how = np.random.ranf(n)
    gqw__uisuf = pd.DataFrame({'A': ims__how, 'B': data[0], 'C': data[1]})
    ealzm__mwqed = time.time()
    yqgs__fxdu = gqw__uisuf.sort_values('A', inplace=False)
    mrcdg__yabg = time.time()
    sort((ims__how,), 0, n, data)
    print('Bodo', time.time() - mrcdg__yabg, 'Numpy', mrcdg__yabg -
        ealzm__mwqed)
    np.testing.assert_almost_equal(data[0], yqgs__fxdu.B.values)
    np.testing.assert_almost_equal(data[1], yqgs__fxdu.C.values)


if __name__ == '__main__':
    test()
