"""
Numba monkey patches to fix issues related to Bodo. Should be imported before any
other module in bodo package.
"""
import copy
import functools
import hashlib
import inspect
import itertools
import operator
import os
import re
import sys
import textwrap
import traceback
import types as pytypes
import warnings
from collections import OrderedDict
from collections.abc import Sequence
from contextlib import ExitStack
import numba
import numba.core.boxing
import numba.core.inline_closurecall
import numba.core.typing.listdecl
import numba.np.linalg
from numba.core import analysis, cgutils, errors, ir, ir_utils, types
from numba.core.compiler import Compiler
from numba.core.errors import ForceLiteralArg, LiteralTypingError, TypingError
from numba.core.ir_utils import GuardException, _create_function_from_code_obj, analysis, build_definitions, find_callname, guard, has_no_side_effect, mk_unique_var, remove_dead_extensions, replace_vars_inner, require, visit_vars_extensions, visit_vars_inner
from numba.core.types import literal
from numba.core.types.functions import _bt_as_lines, _ResolutionFailures, _termcolor, _unlit_non_poison
from numba.core.typing.templates import AbstractTemplate, Signature, _EmptyImplementationEntry, _inline_info, _OverloadAttributeTemplate, infer_global, signature
from numba.core.typing.typeof import Purpose, typeof
from numba.experimental.jitclass import base as jitclass_base
from numba.experimental.jitclass import decorators as jitclass_decorators
from numba.extending import NativeValue, lower_builtin, typeof_impl
from numba.parfors.parfor import get_expr_args
from bodo.utils.typing import BodoError
_check_numba_change = False
numba.core.typing.templates._IntrinsicTemplate.prefer_literal = True


def run_frontend(func, inline_closures=False, emit_dels=False):
    wbtb__soau = numba.core.bytecode.FunctionIdentity.from_function(func)
    aurw__dux = numba.core.interpreter.Interpreter(wbtb__soau)
    toqd__bzhvc = numba.core.bytecode.ByteCode(func_id=wbtb__soau)
    func_ir = aurw__dux.interpret(toqd__bzhvc)
    if inline_closures:
        from numba.core.inline_closurecall import InlineClosureCallPass


        class DummyPipeline:

            def __init__(self, f_ir):
                self.state = numba.core.compiler.StateDict()
                self.state.typingctx = None
                self.state.targetctx = None
                self.state.args = None
                self.state.func_ir = f_ir
                self.state.typemap = None
                self.state.return_type = None
                self.state.calltypes = None
        numba.core.rewrites.rewrite_registry.apply('before-inference',
            DummyPipeline(func_ir).state)
        lqt__jluxc = InlineClosureCallPass(func_ir, numba.core.cpu.
            ParallelOptions(False), {}, False)
        lqt__jluxc.run()
    dfh__wmh = numba.core.postproc.PostProcessor(func_ir)
    dfh__wmh.run(emit_dels)
    return func_ir


if _check_numba_change:
    lines = inspect.getsource(numba.core.compiler.run_frontend)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '8c2477a793b2c08d56430997880974ac12c5570e69c9e54d37d694b322ea18b6':
        warnings.warn('numba.core.compiler.run_frontend has changed')
numba.core.compiler.run_frontend = run_frontend


def visit_vars_stmt(stmt, callback, cbdata):
    for t, ovuka__mos in visit_vars_extensions.items():
        if isinstance(stmt, t):
            ovuka__mos(stmt, callback, cbdata)
            return
    if isinstance(stmt, ir.Assign):
        stmt.target = visit_vars_inner(stmt.target, callback, cbdata)
        stmt.value = visit_vars_inner(stmt.value, callback, cbdata)
    elif isinstance(stmt, ir.Arg):
        stmt.name = visit_vars_inner(stmt.name, callback, cbdata)
    elif isinstance(stmt, ir.Return):
        stmt.value = visit_vars_inner(stmt.value, callback, cbdata)
    elif isinstance(stmt, ir.Raise):
        stmt.exception = visit_vars_inner(stmt.exception, callback, cbdata)
    elif isinstance(stmt, ir.Branch):
        stmt.cond = visit_vars_inner(stmt.cond, callback, cbdata)
    elif isinstance(stmt, ir.Jump):
        stmt.target = visit_vars_inner(stmt.target, callback, cbdata)
    elif isinstance(stmt, ir.Del):
        var = ir.Var(None, stmt.value, stmt.loc)
        var = visit_vars_inner(var, callback, cbdata)
        stmt.value = var.name
    elif isinstance(stmt, ir.DelAttr):
        stmt.target = visit_vars_inner(stmt.target, callback, cbdata)
        stmt.attr = visit_vars_inner(stmt.attr, callback, cbdata)
    elif isinstance(stmt, ir.SetAttr):
        stmt.target = visit_vars_inner(stmt.target, callback, cbdata)
        stmt.attr = visit_vars_inner(stmt.attr, callback, cbdata)
        stmt.value = visit_vars_inner(stmt.value, callback, cbdata)
    elif isinstance(stmt, ir.DelItem):
        stmt.target = visit_vars_inner(stmt.target, callback, cbdata)
        stmt.index = visit_vars_inner(stmt.index, callback, cbdata)
    elif isinstance(stmt, ir.StaticSetItem):
        stmt.target = visit_vars_inner(stmt.target, callback, cbdata)
        stmt.index_var = visit_vars_inner(stmt.index_var, callback, cbdata)
        stmt.value = visit_vars_inner(stmt.value, callback, cbdata)
    elif isinstance(stmt, ir.SetItem):
        stmt.target = visit_vars_inner(stmt.target, callback, cbdata)
        stmt.index = visit_vars_inner(stmt.index, callback, cbdata)
        stmt.value = visit_vars_inner(stmt.value, callback, cbdata)
    elif isinstance(stmt, ir.Print):
        stmt.args = [visit_vars_inner(x, callback, cbdata) for x in stmt.args]
        stmt.vararg = visit_vars_inner(stmt.vararg, callback, cbdata)
    else:
        pass
    return


if _check_numba_change:
    lines = inspect.getsource(numba.core.ir_utils.visit_vars_stmt)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '52b7b645ba65c35f3cf564f936e113261db16a2dff1e80fbee2459af58844117':
        warnings.warn('numba.core.ir_utils.visit_vars_stmt has changed')
numba.core.ir_utils.visit_vars_stmt = visit_vars_stmt
old_run_pass = numba.core.typed_passes.InlineOverloads.run_pass


def InlineOverloads_run_pass(self, state):
    import bodo
    bodo.compiler.bodo_overload_inline_pass(state.func_ir, state.typingctx,
        state.targetctx, state.typemap, state.calltypes)
    return old_run_pass(self, state)


numba.core.typed_passes.InlineOverloads.run_pass = InlineOverloads_run_pass
from numba.core.ir_utils import _add_alias, alias_analysis_extensions, alias_func_extensions
_immutable_type_class = (types.Number, types.scalars._NPDatetimeBase, types
    .iterators.RangeType, types.UnicodeType)


def is_immutable_type(var, typemap):
    if typemap is None or var not in typemap:
        return False
    typ = typemap[var]
    if isinstance(typ, _immutable_type_class):
        return True
    if isinstance(typ, types.BaseTuple) and all(isinstance(t,
        _immutable_type_class) for t in typ.types):
        return True
    return False


def find_potential_aliases(blocks, args, typemap, func_ir, alias_map=None,
    arg_aliases=None):
    if alias_map is None:
        alias_map = {}
    if arg_aliases is None:
        arg_aliases = set(a for a in args if not is_immutable_type(a, typemap))
    func_ir._definitions = build_definitions(func_ir.blocks)
    tfs__vfxt = ['ravel', 'transpose', 'reshape']
    for yoz__msqpj in blocks.values():
        for ncc__qdj in yoz__msqpj.body:
            if type(ncc__qdj) in alias_analysis_extensions:
                ovuka__mos = alias_analysis_extensions[type(ncc__qdj)]
                ovuka__mos(ncc__qdj, args, typemap, func_ir, alias_map,
                    arg_aliases)
            if isinstance(ncc__qdj, ir.Assign):
                betea__ovk = ncc__qdj.value
                bqjwa__wqmxx = ncc__qdj.target.name
                if is_immutable_type(bqjwa__wqmxx, typemap):
                    continue
                if isinstance(betea__ovk, ir.Var
                    ) and bqjwa__wqmxx != betea__ovk.name:
                    _add_alias(bqjwa__wqmxx, betea__ovk.name, alias_map,
                        arg_aliases)
                if isinstance(betea__ovk, ir.Expr) and (betea__ovk.op ==
                    'cast' or betea__ovk.op in ['getitem', 'static_getitem']):
                    _add_alias(bqjwa__wqmxx, betea__ovk.value.name,
                        alias_map, arg_aliases)
                if isinstance(betea__ovk, ir.Expr
                    ) and betea__ovk.op == 'inplace_binop':
                    _add_alias(bqjwa__wqmxx, betea__ovk.lhs.name, alias_map,
                        arg_aliases)
                if isinstance(betea__ovk, ir.Expr
                    ) and betea__ovk.op == 'getattr' and betea__ovk.attr in [
                    'T', 'ctypes', 'flat']:
                    _add_alias(bqjwa__wqmxx, betea__ovk.value.name,
                        alias_map, arg_aliases)
                if isinstance(betea__ovk, ir.Expr
                    ) and betea__ovk.op == 'getattr' and betea__ovk.attr not in [
                    'shape'] and betea__ovk.value.name in arg_aliases:
                    _add_alias(bqjwa__wqmxx, betea__ovk.value.name,
                        alias_map, arg_aliases)
                if isinstance(betea__ovk, ir.Expr
                    ) and betea__ovk.op == 'getattr' and betea__ovk.attr in (
                    'loc', 'iloc', 'iat', '_obj', 'obj', 'codes', '_df'):
                    _add_alias(bqjwa__wqmxx, betea__ovk.value.name,
                        alias_map, arg_aliases)
                if isinstance(betea__ovk, ir.Expr) and betea__ovk.op in (
                    'build_tuple', 'build_list', 'build_set'
                    ) and not is_immutable_type(bqjwa__wqmxx, typemap):
                    for otcze__qfk in betea__ovk.items:
                        _add_alias(bqjwa__wqmxx, otcze__qfk.name, alias_map,
                            arg_aliases)
                if isinstance(betea__ovk, ir.Expr) and betea__ovk.op == 'call':
                    xuh__movn = guard(find_callname, func_ir, betea__ovk,
                        typemap)
                    if xuh__movn is None:
                        continue
                    arwdx__escdo, zzn__mmik = xuh__movn
                    if xuh__movn in alias_func_extensions:
                        adxsp__jsyo = alias_func_extensions[xuh__movn]
                        adxsp__jsyo(bqjwa__wqmxx, betea__ovk.args,
                            alias_map, arg_aliases)
                    if zzn__mmik == 'numpy' and arwdx__escdo in tfs__vfxt:
                        _add_alias(bqjwa__wqmxx, betea__ovk.args[0].name,
                            alias_map, arg_aliases)
                    if isinstance(zzn__mmik, ir.Var
                        ) and arwdx__escdo in tfs__vfxt:
                        _add_alias(bqjwa__wqmxx, zzn__mmik.name, alias_map,
                            arg_aliases)
    oeo__zct = copy.deepcopy(alias_map)
    for otcze__qfk in oeo__zct:
        for vusy__nezu in oeo__zct[otcze__qfk]:
            alias_map[otcze__qfk] |= alias_map[vusy__nezu]
        for vusy__nezu in oeo__zct[otcze__qfk]:
            alias_map[vusy__nezu] = alias_map[otcze__qfk]
    return alias_map, arg_aliases


if _check_numba_change:
    lines = inspect.getsource(ir_utils.find_potential_aliases)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'e6cf3e0f502f903453eb98346fc6854f87dc4ea1ac62f65c2d6aef3bf690b6c5':
        warnings.warn('ir_utils.find_potential_aliases has changed')
ir_utils.find_potential_aliases = find_potential_aliases
numba.parfors.array_analysis.find_potential_aliases = find_potential_aliases
if _check_numba_change:
    lines = inspect.getsource(ir_utils.dead_code_elimination)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '40a8626300a1a17523944ec7842b093c91258bbc60844bbd72191a35a4c366bf':
        warnings.warn('ir_utils.dead_code_elimination has changed')


def mini_dce(func_ir, typemap=None, alias_map=None, arg_aliases=None):
    from numba.core.analysis import compute_cfg_from_blocks, compute_live_map, compute_use_defs
    okz__sbg = compute_cfg_from_blocks(func_ir.blocks)
    uas__xdo = compute_use_defs(func_ir.blocks)
    swf__wtj = compute_live_map(okz__sbg, func_ir.blocks, uas__xdo.usemap,
        uas__xdo.defmap)
    ajd__hno = True
    while ajd__hno:
        ajd__hno = False
        for wsr__ata, block in func_ir.blocks.items():
            lives = {otcze__qfk.name for otcze__qfk in block.terminator.
                list_vars()}
            for nacc__lbx, gnk__xnd in okz__sbg.successors(wsr__ata):
                lives |= swf__wtj[nacc__lbx]
            kpz__hdz = [block.terminator]
            for stmt in reversed(block.body[:-1]):
                if isinstance(stmt, ir.Assign):
                    bqjwa__wqmxx = stmt.target
                    chg__mnb = stmt.value
                    if bqjwa__wqmxx.name not in lives:
                        if isinstance(chg__mnb, ir.Expr
                            ) and chg__mnb.op == 'make_function':
                            continue
                        if isinstance(chg__mnb, ir.Expr
                            ) and chg__mnb.op == 'getattr':
                            continue
                        if isinstance(chg__mnb, ir.Const):
                            continue
                        if typemap and isinstance(typemap.get(bqjwa__wqmxx,
                            None), types.Function):
                            continue
                        if isinstance(chg__mnb, ir.Expr
                            ) and chg__mnb.op == 'build_map':
                            continue
                    if isinstance(chg__mnb, ir.Var
                        ) and bqjwa__wqmxx.name == chg__mnb.name:
                        continue
                if isinstance(stmt, ir.Del):
                    if stmt.value not in lives:
                        continue
                if type(stmt) in analysis.ir_extension_usedefs:
                    koqnt__dsmvd = analysis.ir_extension_usedefs[type(stmt)]
                    rysp__iwp, hyyrn__ukm = koqnt__dsmvd(stmt)
                    lives -= hyyrn__ukm
                    lives |= rysp__iwp
                else:
                    lives |= {otcze__qfk.name for otcze__qfk in stmt.
                        list_vars()}
                    if isinstance(stmt, ir.Assign):
                        lives.remove(bqjwa__wqmxx.name)
                kpz__hdz.append(stmt)
            kpz__hdz.reverse()
            if len(block.body) != len(kpz__hdz):
                ajd__hno = True
            block.body = kpz__hdz


ir_utils.dead_code_elimination = mini_dce
numba.core.typed_passes.dead_code_elimination = mini_dce
numba.core.inline_closurecall.dead_code_elimination = mini_dce
from numba.core.cpu_options import InlineOptions


def make_overload_template(func, overload_func, jit_options, strict, inline,
    prefer_literal=False, **kwargs):
    bef__tyxhk = getattr(func, '__name__', str(func))
    name = 'OverloadTemplate_%s' % (bef__tyxhk,)
    no_unliteral = kwargs.pop('no_unliteral', False)
    base = numba.core.typing.templates._OverloadFunctionTemplate
    mihkk__gmgrx = dict(key=func, _overload_func=staticmethod(overload_func
        ), _impl_cache={}, _compiled_overloads={}, _jit_options=jit_options,
        _strict=strict, _inline=staticmethod(InlineOptions(inline)),
        _inline_overloads={}, prefer_literal=prefer_literal, _no_unliteral=
        no_unliteral, metadata=kwargs)
    return type(base)(name, (base,), mihkk__gmgrx)


if _check_numba_change:
    lines = inspect.getsource(numba.core.typing.templates.
        make_overload_template)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '7f6974584cb10e49995b652827540cc6732e497c0b9f8231b44fd83fcc1c0a83':
        warnings.warn(
            'numba.core.typing.templates.make_overload_template has changed')
numba.core.typing.templates.make_overload_template = make_overload_template


def _resolve(self, typ, attr):
    if self._attr != attr:
        return None
    if isinstance(typ, types.TypeRef):
        assert typ == self.key
    else:
        assert isinstance(typ, self.key)


    class MethodTemplate(AbstractTemplate):
        key = self.key, attr
        _inline = self._inline
        _no_unliteral = getattr(self, '_no_unliteral', False)
        _overload_func = staticmethod(self._overload_func)
        _inline_overloads = self._inline_overloads
        prefer_literal = self.prefer_literal

        def generic(_, args, kws):
            args = (typ,) + tuple(args)
            fnty = self._get_function_type(self.context, typ)
            sig = self._get_signature(self.context, fnty, args, kws)
            sig = sig.replace(pysig=numba.core.utils.pysignature(self.
                _overload_func))
            for gafz__qfv in fnty.templates:
                self._inline_overloads.update(gafz__qfv._inline_overloads)
            if sig is not None:
                return sig.as_method()
    return types.BoundFunction(MethodTemplate, typ)


if _check_numba_change:
    lines = inspect.getsource(numba.core.typing.templates.
        _OverloadMethodTemplate._resolve)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'ce8e0935dc939d0867ef969e1ed2975adb3533a58a4133fcc90ae13c4418e4d6':
        warnings.warn(
            'numba.core.typing.templates._OverloadMethodTemplate._resolve has changed'
            )
numba.core.typing.templates._OverloadMethodTemplate._resolve = _resolve


def make_overload_attribute_template(typ, attr, overload_func, inline,
    prefer_literal=False, base=_OverloadAttributeTemplate, **kwargs):
    assert isinstance(typ, types.Type) or issubclass(typ, types.Type)
    name = 'OverloadAttributeTemplate_%s_%s' % (typ, attr)
    no_unliteral = kwargs.pop('no_unliteral', False)
    mihkk__gmgrx = dict(key=typ, _attr=attr, _impl_cache={}, _inline=
        staticmethod(InlineOptions(inline)), _inline_overloads={},
        _no_unliteral=no_unliteral, _overload_func=staticmethod(
        overload_func), prefer_literal=prefer_literal, metadata=kwargs)
    obj = type(base)(name, (base,), mihkk__gmgrx)
    return obj


if _check_numba_change:
    lines = inspect.getsource(numba.core.typing.templates.
        make_overload_attribute_template)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'f066c38c482d6cf8bf5735a529c3264118ba9b52264b24e58aad12a6b1960f5d':
        warnings.warn(
            'numba.core.typing.templates.make_overload_attribute_template has changed'
            )
numba.core.typing.templates.make_overload_attribute_template = (
    make_overload_attribute_template)


def generic(self, args, kws):
    from numba.core.typed_passes import PreLowerStripPhis
    jes__oxrni, pdce__yvha = self._get_impl(args, kws)
    if jes__oxrni is None:
        return
    scay__zzp = types.Dispatcher(jes__oxrni)
    if not self._inline.is_never_inline:
        from numba.core import compiler, typed_passes
        from numba.core.inline_closurecall import InlineWorker
        iafv__bjewe = jes__oxrni._compiler
        flags = compiler.Flags()
        rvyh__xdlk = iafv__bjewe.targetdescr.typing_context
        npcv__mxmhk = iafv__bjewe.targetdescr.target_context
        ysu__udwl = iafv__bjewe.pipeline_class(rvyh__xdlk, npcv__mxmhk,
            None, None, None, flags, None)
        imt__ljwk = InlineWorker(rvyh__xdlk, npcv__mxmhk, iafv__bjewe.
            locals, ysu__udwl, flags, None)
        utk__cyh = scay__zzp.dispatcher.get_call_template
        gafz__qfv, blp__qbk, uwqs__dmq, kws = utk__cyh(pdce__yvha, kws)
        if uwqs__dmq in self._inline_overloads:
            return self._inline_overloads[uwqs__dmq]['iinfo'].signature
        ir = imt__ljwk.run_untyped_passes(scay__zzp.dispatcher.py_func,
            enable_ssa=True)
        typemap, return_type, calltypes, _ = typed_passes.type_inference_stage(
            self.context, npcv__mxmhk, ir, uwqs__dmq, None)
        ir = PreLowerStripPhis()._strip_phi_nodes(ir)
        ir._definitions = numba.core.ir_utils.build_definitions(ir.blocks)
        sig = Signature(return_type, uwqs__dmq, None)
        self._inline_overloads[sig.args] = {'folded_args': uwqs__dmq}
        ydqcn__ftc = _EmptyImplementationEntry('always inlined')
        self._compiled_overloads[sig.args] = ydqcn__ftc
        if not self._inline.is_always_inline:
            sig = scay__zzp.get_call_type(self.context, pdce__yvha, kws)
            self._compiled_overloads[sig.args] = scay__zzp.get_overload(sig)
        vyt__ses = _inline_info(ir, typemap, calltypes, sig)
        self._inline_overloads[sig.args] = {'folded_args': uwqs__dmq,
            'iinfo': vyt__ses}
    else:
        sig = scay__zzp.get_call_type(self.context, pdce__yvha, kws)
        if sig is None:
            return None
        self._compiled_overloads[sig.args] = scay__zzp.get_overload(sig)
    return sig


if _check_numba_change:
    lines = inspect.getsource(numba.core.typing.templates.
        _OverloadFunctionTemplate.generic)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '5d453a6d0215ebf0bab1279ff59eb0040b34938623be99142ce20acc09cdeb64':
        warnings.warn(
            'numba.core.typing.templates._OverloadFunctionTemplate.generic has changed'
            )
numba.core.typing.templates._OverloadFunctionTemplate.generic = generic


def bound_function(template_key, no_unliteral=False):

    def wrapper(method_resolver):

        @functools.wraps(method_resolver)
        def attribute_resolver(self, ty):


            class MethodTemplate(AbstractTemplate):
                key = template_key

                def generic(_, args, kws):
                    sig = method_resolver(self, ty, args, kws)
                    if sig is not None and sig.recvr is None:
                        sig = sig.replace(recvr=ty)
                    return sig
            MethodTemplate._no_unliteral = no_unliteral
            return types.BoundFunction(MethodTemplate, ty)
        return attribute_resolver
    return wrapper


if _check_numba_change:
    lines = inspect.getsource(numba.core.typing.templates.bound_function)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'a2feefe64eae6a15c56affc47bf0c1d04461f9566913442d539452b397103322':
        warnings.warn('numba.core.typing.templates.bound_function has changed')
numba.core.typing.templates.bound_function = bound_function


def get_call_type(self, context, args, kws):
    from numba.core import utils
    jxvrz__hqi = [True, False]
    mniws__dgnsz = [False, True]
    mjmhd__xwh = _ResolutionFailures(context, self, args, kws, depth=self.
        _depth)
    from numba.core.target_extension import get_local_target
    ypwji__pdu = get_local_target(context)
    vacr__jaxe = utils.order_by_target_specificity(ypwji__pdu, self.
        templates, fnkey=self.key[0])
    self._depth += 1
    for thqv__hyahd in vacr__jaxe:
        xofhx__uut = thqv__hyahd(context)
        jgj__ucftz = jxvrz__hqi if xofhx__uut.prefer_literal else mniws__dgnsz
        jgj__ucftz = [True] if getattr(xofhx__uut, '_no_unliteral', False
            ) else jgj__ucftz
        for spdhf__kbl in jgj__ucftz:
            try:
                if spdhf__kbl:
                    sig = xofhx__uut.apply(args, kws)
                else:
                    oqw__eqp = tuple([_unlit_non_poison(a) for a in args])
                    agx__izldr = {bfxes__thqc: _unlit_non_poison(otcze__qfk
                        ) for bfxes__thqc, otcze__qfk in kws.items()}
                    sig = xofhx__uut.apply(oqw__eqp, agx__izldr)
            except Exception as e:
                from numba.core import utils
                if utils.use_new_style_errors() and not isinstance(e,
                    errors.NumbaError):
                    raise e
                else:
                    sig = None
                    mjmhd__xwh.add_error(xofhx__uut, False, e, spdhf__kbl)
            else:
                if sig is not None:
                    self._impl_keys[sig.args] = xofhx__uut.get_impl_key(sig)
                    self._depth -= 1
                    return sig
                else:
                    hvx__vht = getattr(xofhx__uut, 'cases', None)
                    if hvx__vht is not None:
                        msg = 'No match for registered cases:\n%s'
                        msg = msg % '\n'.join(' * {}'.format(x) for x in
                            hvx__vht)
                    else:
                        msg = 'No match.'
                    mjmhd__xwh.add_error(xofhx__uut, True, msg, spdhf__kbl)
    mjmhd__xwh.raise_error()


if _check_numba_change:
    lines = inspect.getsource(numba.core.types.functions.BaseFunction.
        get_call_type)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '25f038a7216f8e6f40068ea81e11fd9af8ad25d19888f7304a549941b01b7015':
        warnings.warn(
            'numba.core.types.functions.BaseFunction.get_call_type has changed'
            )
numba.core.types.functions.BaseFunction.get_call_type = get_call_type
bodo_typing_error_info = """
This is often caused by the use of unsupported features or typing issues.
See https://docs.bodo.ai/
"""


def get_call_type2(self, context, args, kws):
    gafz__qfv = self.template(context)
    jbw__zqtdf = None
    lruci__cafjv = None
    tabns__rqez = None
    jgj__ucftz = [True, False] if gafz__qfv.prefer_literal else [False, True]
    jgj__ucftz = [True] if getattr(gafz__qfv, '_no_unliteral', False
        ) else jgj__ucftz
    for spdhf__kbl in jgj__ucftz:
        if spdhf__kbl:
            try:
                tabns__rqez = gafz__qfv.apply(args, kws)
            except Exception as vaitz__sovap:
                if isinstance(vaitz__sovap, errors.ForceLiteralArg):
                    raise vaitz__sovap
                jbw__zqtdf = vaitz__sovap
                tabns__rqez = None
            else:
                break
        else:
            nvxef__dcm = tuple([_unlit_non_poison(a) for a in args])
            zbh__vvrny = {bfxes__thqc: _unlit_non_poison(otcze__qfk) for 
                bfxes__thqc, otcze__qfk in kws.items()}
            vjhhb__uqg = nvxef__dcm == args and kws == zbh__vvrny
            if not vjhhb__uqg and tabns__rqez is None:
                try:
                    tabns__rqez = gafz__qfv.apply(nvxef__dcm, zbh__vvrny)
                except Exception as vaitz__sovap:
                    from numba.core import utils
                    if utils.use_new_style_errors() and not isinstance(
                        vaitz__sovap, errors.NumbaError):
                        raise vaitz__sovap
                    if isinstance(vaitz__sovap, errors.ForceLiteralArg):
                        if gafz__qfv.prefer_literal:
                            raise vaitz__sovap
                    lruci__cafjv = vaitz__sovap
                else:
                    break
    if tabns__rqez is None and (lruci__cafjv is not None or jbw__zqtdf is not
        None):
        hpg__wvvh = '- Resolution failure for {} arguments:\n{}\n'
        wahch__lmaw = _termcolor.highlight(hpg__wvvh)
        if numba.core.config.DEVELOPER_MODE:
            zqkfn__rfb = ' ' * 4

            def add_bt(error):
                if isinstance(error, BaseException):
                    vmou__sjaun = traceback.format_exception(type(error),
                        error, error.__traceback__)
                else:
                    vmou__sjaun = ['']
                jozxe__kwv = '\n{}'.format(2 * zqkfn__rfb)
                xel__llvp = _termcolor.reset(jozxe__kwv + jozxe__kwv.join(
                    _bt_as_lines(vmou__sjaun)))
                return _termcolor.reset(xel__llvp)
        else:
            add_bt = lambda X: ''

        def nested_msg(literalness, e):
            fbzw__oks = str(e)
            fbzw__oks = fbzw__oks if fbzw__oks else str(repr(e)) + add_bt(e)
            zucx__itvss = errors.TypingError(textwrap.dedent(fbzw__oks))
            return wahch__lmaw.format(literalness, str(zucx__itvss))
        import bodo
        if isinstance(jbw__zqtdf, bodo.utils.typing.BodoError):
            raise jbw__zqtdf
        if numba.core.config.DEVELOPER_MODE:
            raise errors.TypingError(nested_msg('literal', jbw__zqtdf) +
                nested_msg('non-literal', lruci__cafjv))
        else:
            msg = 'Compilation error for '
            if isinstance(self.this, bodo.hiframes.pd_dataframe_ext.
                DataFrameType):
                msg += 'DataFrame.'
            elif isinstance(self.this, bodo.hiframes.pd_series_ext.SeriesType):
                msg += 'Series.'
            msg += f'{self.typing_key[1]}().{bodo_typing_error_info}'
            raise errors.TypingError(msg)
    return tabns__rqez


if _check_numba_change:
    lines = inspect.getsource(numba.core.types.functions.BoundFunction.
        get_call_type)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '502cd77c0084452e903a45a0f1f8107550bfbde7179363b57dabd617ce135f4a':
        warnings.warn(
            'numba.core.types.functions.BoundFunction.get_call_type has changed'
            )
numba.core.types.functions.BoundFunction.get_call_type = get_call_type2


def string_from_string_and_size(self, string, size):
    from llvmlite.llvmpy.core import Type
    fnty = Type.function(self.pyobj, [self.cstring, self.py_ssize_t])
    arwdx__escdo = 'PyUnicode_FromStringAndSize'
    fn = self._get_function(fnty, name=arwdx__escdo)
    return self.builder.call(fn, [string, size])


numba.core.pythonapi.PythonAPI.string_from_string_and_size = (
    string_from_string_and_size)


def _compile_for_args(self, *args, **kws):
    assert not kws
    self._compilation_chain_init_hook()
    import bodo

    def error_rewrite(e, issue_type):
        if numba.core.config.SHOW_HELP:
            xmb__rdc = errors.error_extras[issue_type]
            e.patch_message('\n'.join((str(e).rstrip(), xmb__rdc)))
        if numba.core.config.FULL_TRACEBACKS:
            raise e
        else:
            raise e.with_traceback(None)
    xxgzj__asf = []
    for a in args:
        if isinstance(a, numba.core.dispatcher.OmittedArg):
            xxgzj__asf.append(types.Omitted(a.value))
        else:
            xxgzj__asf.append(self.typeof_pyval(a))
    snmx__valsr = None
    try:
        error = None
        snmx__valsr = self.compile(tuple(xxgzj__asf))
    except errors.ForceLiteralArg as e:
        nkfk__yji = [xgedl__blg for xgedl__blg in e.requested_args if 
            isinstance(args[xgedl__blg], types.Literal) and not isinstance(
            args[xgedl__blg], types.LiteralStrKeyDict)]
        if nkfk__yji:
            hyoh__pddj = """Repeated literal typing request.
{}.
This is likely caused by an error in typing. Please see nested and suppressed exceptions."""
            polom__mqrum = ', '.join('Arg #{} is {}'.format(xgedl__blg,
                args[xgedl__blg]) for xgedl__blg in sorted(nkfk__yji))
            raise errors.CompilerError(hyoh__pddj.format(polom__mqrum))
        pdce__yvha = []
        try:
            for xgedl__blg, otcze__qfk in enumerate(args):
                if xgedl__blg in e.requested_args:
                    if xgedl__blg in e.file_infos:
                        pdce__yvha.append(types.FilenameType(args[
                            xgedl__blg], e.file_infos[xgedl__blg]))
                    else:
                        pdce__yvha.append(types.literal(args[xgedl__blg]))
                else:
                    pdce__yvha.append(args[xgedl__blg])
            args = pdce__yvha
        except (OSError, FileNotFoundError) as guige__vcomc:
            error = FileNotFoundError(str(guige__vcomc) + '\n' + e.loc.
                strformat() + '\n')
        except bodo.utils.typing.BodoError as e:
            error = bodo.utils.typing.BodoError(str(e))
        if error is None:
            try:
                snmx__valsr = self._compile_for_args(*args)
            except TypingError as e:
                error = errors.TypingError(str(e))
            except bodo.utils.typing.BodoError as e:
                error = bodo.utils.typing.BodoError(str(e))
    except errors.TypingError as e:
        ldfuy__xvs = []
        for xgedl__blg, bqf__rjzof in enumerate(args):
            val = bqf__rjzof.value if isinstance(bqf__rjzof, numba.core.
                dispatcher.OmittedArg) else bqf__rjzof
            try:
                fgd__mtpr = typeof(val, Purpose.argument)
            except ValueError as orck__iypzk:
                ldfuy__xvs.append((xgedl__blg, str(orck__iypzk)))
            else:
                if fgd__mtpr is None:
                    ldfuy__xvs.append((xgedl__blg,
                        f'cannot determine Numba type of value {val}'))
        if ldfuy__xvs:
            xnti__vib = '\n'.join(f'- argument {xgedl__blg}: {jfbo__nziw}' for
                xgedl__blg, jfbo__nziw in ldfuy__xvs)
            msg = f"""{str(e).rstrip()} 

This error may have been caused by the following argument(s):
{xnti__vib}
"""
            e.patch_message(msg)
        if "Cannot determine Numba type of <class 'numpy.ufunc'>" in e.msg:
            msg = 'Unsupported Numpy ufunc encountered in JIT code'
            error = bodo.utils.typing.BodoError(msg, loc=e.loc)
        elif not numba.core.config.DEVELOPER_MODE:
            if bodo_typing_error_info not in e.msg:
                dhk__woo = ['Failed in nopython mode pipeline',
                    'Failed in bodo mode pipeline', 'numba', 'Overload',
                    'lowering']
                dxpv__yezcj = False
                for hmkbg__wnsk in dhk__woo:
                    if hmkbg__wnsk in e.msg:
                        msg = 'Compilation error. '
                        msg += f'{bodo_typing_error_info}'
                        dxpv__yezcj = True
                        break
                if not dxpv__yezcj:
                    msg = f'{str(e)}'
                msg += '\n' + e.loc.strformat() + '\n'
                e.patch_message(msg)
        error_rewrite(e, 'typing')
    except errors.UnsupportedError as e:
        error_rewrite(e, 'unsupported_error')
    except (errors.NotDefinedError, errors.RedefinedError, errors.
        VerificationError) as e:
        error_rewrite(e, 'interpreter')
    except errors.ConstantInferenceError as e:
        error_rewrite(e, 'constant_inference')
    except bodo.utils.typing.BodoError as e:
        error = bodo.utils.typing.BodoError(str(e))
    except Exception as e:
        if numba.core.config.SHOW_HELP:
            if hasattr(e, 'patch_message'):
                xmb__rdc = errors.error_extras['reportable']
                e.patch_message('\n'.join((str(e).rstrip(), xmb__rdc)))
        raise e
    finally:
        self._types_active_call = []
        del args
        if error:
            raise error
    return snmx__valsr


if _check_numba_change:
    lines = inspect.getsource(numba.core.dispatcher._DispatcherBase.
        _compile_for_args)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '5cdfbf0b13a528abf9f0408e70f67207a03e81d610c26b1acab5b2dc1f79bf06':
        warnings.warn(
            'numba.core.dispatcher._DispatcherBase._compile_for_args has changed'
            )
numba.core.dispatcher._DispatcherBase._compile_for_args = _compile_for_args


def resolve_gb_agg_funcs(cres):
    from bodo.ir.aggregate import gb_agg_cfunc_addr
    for uvzay__qcj in cres.library._codegen._engine._defined_symbols:
        if uvzay__qcj.startswith('cfunc'
            ) and 'get_agg_udf_addr' not in uvzay__qcj and (
            'bodo_gb_udf_update_local' in uvzay__qcj or 
            'bodo_gb_udf_combine' in uvzay__qcj or 'bodo_gb_udf_eval' in
            uvzay__qcj or 'bodo_gb_apply_general_udfs' in uvzay__qcj):
            gb_agg_cfunc_addr[uvzay__qcj
                ] = cres.library.get_pointer_to_function(uvzay__qcj)


def resolve_join_general_cond_funcs(cres):
    from bodo.ir.join import join_gen_cond_cfunc_addr
    for uvzay__qcj in cres.library._codegen._engine._defined_symbols:
        if uvzay__qcj.startswith('cfunc') and ('get_join_cond_addr' not in
            uvzay__qcj or 'bodo_join_gen_cond' in uvzay__qcj):
            join_gen_cond_cfunc_addr[uvzay__qcj
                ] = cres.library.get_pointer_to_function(uvzay__qcj)


def compile(self, sig):
    import numba.core.event as ev
    from numba.core import sigutils
    from numba.core.compiler_lock import global_compiler_lock
    jes__oxrni = self._get_dispatcher_for_current_target()
    if jes__oxrni is not self:
        return jes__oxrni.compile(sig)
    with ExitStack() as scope:
        cres = None

        def cb_compiler(dur):
            if cres is not None:
                self._callback_add_compiler_timer(dur, cres)

        def cb_llvm(dur):
            if cres is not None:
                self._callback_add_llvm_timer(dur, cres)
        scope.enter_context(ev.install_timer('numba:compiler_lock',
            cb_compiler))
        scope.enter_context(ev.install_timer('numba:llvm_lock', cb_llvm))
        scope.enter_context(global_compiler_lock)
        if not self._can_compile:
            raise RuntimeError('compilation disabled')
        with self._compiling_counter:
            args, return_type = sigutils.normalize_signature(sig)
            vguxm__kudmb = self.overloads.get(tuple(args))
            if vguxm__kudmb is not None:
                return vguxm__kudmb.entry_point
            cres = self._cache.load_overload(sig, self.targetctx)
            if cres is not None:
                resolve_gb_agg_funcs(cres)
                resolve_join_general_cond_funcs(cres)
                self._cache_hits[sig] += 1
                if not cres.objectmode:
                    self.targetctx.insert_user_function(cres.entry_point,
                        cres.fndesc, [cres.library])
                self.add_overload(cres)
                return cres.entry_point
            self._cache_misses[sig] += 1
            gqx__hsv = dict(dispatcher=self, args=args, return_type=return_type
                )
            with ev.trigger_event('numba:compile', data=gqx__hsv):
                try:
                    cres = self._compiler.compile(args, return_type)
                except errors.ForceLiteralArg as e:

                    def folded(args, kws):
                        return self._compiler.fold_argument_types(args, kws)[1]
                    raise e.bind_fold_arguments(folded)
                self.add_overload(cres)
            self._cache.save_overload(sig, cres)
            return cres.entry_point


if _check_numba_change:
    lines = inspect.getsource(numba.core.dispatcher.Dispatcher.compile)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '934ec993577ea3b1c7dd2181ac02728abf8559fd42c17062cc821541b092ff8f':
        warnings.warn('numba.core.dispatcher.Dispatcher.compile has changed')
numba.core.dispatcher.Dispatcher.compile = compile


def _get_module_for_linking(self):
    import llvmlite.binding as ll
    self._ensure_finalized()
    if self._shared_module is not None:
        return self._shared_module
    nemru__llv = self._final_module
    ncbci__upgp = []
    xyqd__dfrv = 0
    for fn in nemru__llv.functions:
        xyqd__dfrv += 1
        if not fn.is_declaration and fn.linkage == ll.Linkage.external:
            if 'get_agg_udf_addr' not in fn.name:
                if 'bodo_gb_udf_update_local' in fn.name:
                    continue
                if 'bodo_gb_udf_combine' in fn.name:
                    continue
                if 'bodo_gb_udf_eval' in fn.name:
                    continue
                if 'bodo_gb_apply_general_udfs' in fn.name:
                    continue
            if 'get_join_cond_addr' not in fn.name:
                if 'bodo_join_gen_cond' in fn.name:
                    continue
            ncbci__upgp.append(fn.name)
    if xyqd__dfrv == 0:
        raise RuntimeError(
            'library unfit for linking: no available functions in %s' % (self,)
            )
    if ncbci__upgp:
        nemru__llv = nemru__llv.clone()
        for name in ncbci__upgp:
            nemru__llv.get_function(name).linkage = 'linkonce_odr'
    self._shared_module = nemru__llv
    return nemru__llv


if _check_numba_change:
    lines = inspect.getsource(numba.core.codegen.CPUCodeLibrary.
        _get_module_for_linking)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '56dde0e0555b5ec85b93b97c81821bce60784515a1fbf99e4542e92d02ff0a73':
        warnings.warn(
            'numba.core.codegen.CPUCodeLibrary._get_module_for_linking has changed'
            )
numba.core.codegen.CPUCodeLibrary._get_module_for_linking = (
    _get_module_for_linking)


def propagate(self, typeinfer):
    import bodo
    errors = []
    for qapd__tgn in self.constraints:
        loc = qapd__tgn.loc
        with typeinfer.warnings.catch_warnings(filename=loc.filename,
            lineno=loc.line):
            try:
                qapd__tgn(typeinfer)
            except numba.core.errors.ForceLiteralArg as e:
                errors.append(e)
            except numba.core.errors.TypingError as e:
                numba.core.typeinfer._logger.debug('captured error', exc_info=e
                    )
                ilbif__dydlv = numba.core.errors.TypingError(str(e), loc=
                    qapd__tgn.loc, highlighting=False)
                errors.append(numba.core.utils.chain_exception(ilbif__dydlv, e)
                    )
            except bodo.utils.typing.BodoError as e:
                if loc not in e.locs_in_msg:
                    errors.append(bodo.utils.typing.BodoError(str(e.msg) +
                        '\n' + loc.strformat() + '\n', locs_in_msg=e.
                        locs_in_msg + [loc]))
                else:
                    errors.append(bodo.utils.typing.BodoError(e.msg,
                        locs_in_msg=e.locs_in_msg))
            except Exception as e:
                from numba.core import utils
                if utils.use_old_style_errors():
                    numba.core.typeinfer._logger.debug('captured error',
                        exc_info=e)
                    msg = """Internal error at {con}.
{err}
Enable logging at debug level for details."""
                    ilbif__dydlv = numba.core.errors.TypingError(msg.format
                        (con=qapd__tgn, err=str(e)), loc=qapd__tgn.loc,
                        highlighting=False)
                    errors.append(utils.chain_exception(ilbif__dydlv, e))
                elif utils.use_new_style_errors():
                    raise e
                else:
                    msg = (
                        f"Unknown CAPTURED_ERRORS style: '{numba.core.config.CAPTURED_ERRORS}'."
                        )
                    assert 0, msg
    return errors


if _check_numba_change:
    lines = inspect.getsource(numba.core.typeinfer.ConstraintNetwork.propagate)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '1e73635eeba9ba43cb3372f395b747ae214ce73b729fb0adba0a55237a1cb063':
        warnings.warn(
            'numba.core.typeinfer.ConstraintNetwork.propagate has changed')
numba.core.typeinfer.ConstraintNetwork.propagate = propagate


def raise_error(self):
    import bodo
    for fjz__misp in self._failures.values():
        for aqpjc__qnc in fjz__misp:
            if isinstance(aqpjc__qnc.error, ForceLiteralArg):
                raise aqpjc__qnc.error
            if isinstance(aqpjc__qnc.error, bodo.utils.typing.BodoError):
                raise aqpjc__qnc.error
    raise TypingError(self.format())


if _check_numba_change:
    lines = inspect.getsource(numba.core.types.functions.
        _ResolutionFailures.raise_error)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '84b89430f5c8b46cfc684804e6037f00a0f170005cd128ad245551787b2568ea':
        warnings.warn(
            'numba.core.types.functions._ResolutionFailures.raise_error has changed'
            )
numba.core.types.functions._ResolutionFailures.raise_error = raise_error


def bodo_remove_dead_block(block, lives, call_table, arg_aliases, alias_map,
    alias_set, func_ir, typemap):
    from bodo.transforms.distributed_pass import saved_array_analysis
    from bodo.utils.utils import is_array_typ, is_expr
    ydeh__sjpiv = False
    kpz__hdz = [block.terminator]
    for stmt in reversed(block.body[:-1]):
        ljsr__cipw = set()
        dtjon__ihev = lives & alias_set
        for otcze__qfk in dtjon__ihev:
            ljsr__cipw |= alias_map[otcze__qfk]
        lives_n_aliases = lives | ljsr__cipw | arg_aliases
        if type(stmt) in remove_dead_extensions:
            ovuka__mos = remove_dead_extensions[type(stmt)]
            stmt = ovuka__mos(stmt, lives, lives_n_aliases, arg_aliases,
                alias_map, func_ir, typemap)
            if stmt is None:
                ydeh__sjpiv = True
                continue
        if isinstance(stmt, ir.Assign):
            bqjwa__wqmxx = stmt.target
            chg__mnb = stmt.value
            if bqjwa__wqmxx.name not in lives and has_no_side_effect(chg__mnb,
                lives_n_aliases, call_table):
                ydeh__sjpiv = True
                continue
            if saved_array_analysis and bqjwa__wqmxx.name in lives and is_expr(
                chg__mnb, 'getattr'
                ) and chg__mnb.attr == 'shape' and is_array_typ(typemap[
                chg__mnb.value.name]) and chg__mnb.value.name not in lives:
                lgcpa__fvcxm = {otcze__qfk: bfxes__thqc for bfxes__thqc,
                    otcze__qfk in func_ir.blocks.items()}
                if block in lgcpa__fvcxm:
                    wsr__ata = lgcpa__fvcxm[block]
                    vmjq__pko = saved_array_analysis.get_equiv_set(wsr__ata)
                    dlx__raxel = vmjq__pko.get_equiv_set(chg__mnb.value)
                    if dlx__raxel is not None:
                        for otcze__qfk in dlx__raxel:
                            if otcze__qfk.endswith('#0'):
                                otcze__qfk = otcze__qfk[:-2]
                            if otcze__qfk in typemap and is_array_typ(typemap
                                [otcze__qfk]) and otcze__qfk in lives:
                                chg__mnb.value = ir.Var(chg__mnb.value.
                                    scope, otcze__qfk, chg__mnb.value.loc)
                                ydeh__sjpiv = True
                                break
            if isinstance(chg__mnb, ir.Var
                ) and bqjwa__wqmxx.name == chg__mnb.name:
                ydeh__sjpiv = True
                continue
        if isinstance(stmt, ir.Del):
            if stmt.value not in lives:
                ydeh__sjpiv = True
                continue
        if isinstance(stmt, ir.SetItem):
            name = stmt.target.name
            if name not in lives_n_aliases:
                continue
        if type(stmt) in analysis.ir_extension_usedefs:
            koqnt__dsmvd = analysis.ir_extension_usedefs[type(stmt)]
            rysp__iwp, hyyrn__ukm = koqnt__dsmvd(stmt)
            lives -= hyyrn__ukm
            lives |= rysp__iwp
        else:
            lives |= {otcze__qfk.name for otcze__qfk in stmt.list_vars()}
            if isinstance(stmt, ir.Assign):
                ecdn__zpzsu = set()
                if isinstance(chg__mnb, ir.Expr):
                    ecdn__zpzsu = {otcze__qfk.name for otcze__qfk in
                        chg__mnb.list_vars()}
                if bqjwa__wqmxx.name not in ecdn__zpzsu:
                    lives.remove(bqjwa__wqmxx.name)
        kpz__hdz.append(stmt)
    kpz__hdz.reverse()
    block.body = kpz__hdz
    return ydeh__sjpiv


ir_utils.remove_dead_block = bodo_remove_dead_block


@infer_global(set)
class SetBuiltin(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        if args:
            isbq__begih, = args
            if isinstance(isbq__begih, types.IterableType):
                dtype = isbq__begih.iterator_type.yield_type
                if isinstance(dtype, types.Hashable
                    ) or dtype == numba.core.types.unicode_type:
                    return signature(types.Set(dtype), isbq__begih)
        else:
            return signature(types.Set(types.undefined))


def Set__init__(self, dtype, reflected=False):
    assert isinstance(dtype, (types.Hashable, types.Undefined)
        ) or dtype == numba.core.types.unicode_type
    self.dtype = dtype
    self.reflected = reflected
    whz__tifl = 'reflected set' if reflected else 'set'
    name = '%s(%s)' % (whz__tifl, self.dtype)
    super(types.Set, self).__init__(name=name)


types.Set.__init__ = Set__init__


@lower_builtin(operator.eq, types.UnicodeType, types.UnicodeType)
def eq_str(context, builder, sig, args):
    func = numba.cpython.unicode.unicode_eq(*sig.args)
    return context.compile_internal(builder, func, sig, args)


numba.parfors.parfor.push_call_vars = (lambda blocks, saved_globals,
    saved_getattrs, typemap, nested=False: None)


def maybe_literal(value):
    if isinstance(value, (list, dict, pytypes.FunctionType)):
        return
    if isinstance(value, tuple):
        try:
            return types.Tuple([literal(x) for x in value])
        except LiteralTypingError as rnigj__cbzwr:
            return
    try:
        return literal(value)
    except LiteralTypingError as rnigj__cbzwr:
        return


if _check_numba_change:
    lines = inspect.getsource(types.maybe_literal)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '8fb2fd93acf214b28e33e37d19dc2f7290a42792ec59b650553ac278854b5081':
        warnings.warn('types.maybe_literal has changed')
types.maybe_literal = maybe_literal
types.misc.maybe_literal = maybe_literal


def CacheImpl__init__(self, py_func):
    self._lineno = py_func.__code__.co_firstlineno
    try:
        puc__eluw = py_func.__qualname__
    except AttributeError as rnigj__cbzwr:
        puc__eluw = py_func.__name__
    zcrbu__upt = inspect.getfile(py_func)
    for cls in self._locator_classes:
        zrpj__szux = cls.from_function(py_func, zcrbu__upt)
        if zrpj__szux is not None:
            break
    else:
        raise RuntimeError(
            'cannot cache function %r: no locator available for file %r' %
            (puc__eluw, zcrbu__upt))
    self._locator = zrpj__szux
    nqs__woq = inspect.getfile(py_func)
    naev__aabgv = os.path.splitext(os.path.basename(nqs__woq))[0]
    if zcrbu__upt.startswith('<ipython-'):
        qhx__vcc = re.sub('(ipython-input)(-\\d+)(-[0-9a-fA-F]+)', '\\1\\3',
            naev__aabgv, count=1)
        if qhx__vcc == naev__aabgv:
            warnings.warn(
                'Did not recognize ipython module name syntax. Caching might not work'
                )
        naev__aabgv = qhx__vcc
    lbm__ngijo = '%s.%s' % (naev__aabgv, puc__eluw)
    ddsra__vqnrm = getattr(sys, 'abiflags', '')
    self._filename_base = self.get_filename_base(lbm__ngijo, ddsra__vqnrm)


if _check_numba_change:
    lines = inspect.getsource(numba.core.caching._CacheImpl.__init__)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'b46d298146e3844e9eaeef29d36f5165ba4796c270ca50d2b35f9fcdc0fa032a':
        warnings.warn('numba.core.caching._CacheImpl.__init__ has changed')
numba.core.caching._CacheImpl.__init__ = CacheImpl__init__


def _analyze_broadcast(self, scope, equiv_set, loc, args, fn):
    from numba.parfors.array_analysis import ArrayAnalysis
    rqoul__mylhg = list(filter(lambda a: self._istuple(a.name), args))
    if len(rqoul__mylhg) == 2 and fn.__name__ == 'add':
        qgne__ykjm = self.typemap[rqoul__mylhg[0].name]
        enstg__vbaf = self.typemap[rqoul__mylhg[1].name]
        if qgne__ykjm.count == 0:
            return ArrayAnalysis.AnalyzeResult(shape=equiv_set.get_shape(
                rqoul__mylhg[1]))
        if enstg__vbaf.count == 0:
            return ArrayAnalysis.AnalyzeResult(shape=equiv_set.get_shape(
                rqoul__mylhg[0]))
        try:
            yaklt__fqxn = [equiv_set.get_shape(x) for x in rqoul__mylhg]
            if None in yaklt__fqxn:
                return None
            bznxu__itir = sum(yaklt__fqxn, ())
            return ArrayAnalysis.AnalyzeResult(shape=bznxu__itir)
        except GuardException as rnigj__cbzwr:
            return None
    dish__vdjb = list(filter(lambda a: self._isarray(a.name), args))
    require(len(dish__vdjb) > 0)
    kpw__inuqv = [x.name for x in dish__vdjb]
    nezep__rrvra = [self.typemap[x.name].ndim for x in dish__vdjb]
    ikoh__dhuqx = max(nezep__rrvra)
    require(ikoh__dhuqx > 0)
    yaklt__fqxn = [equiv_set.get_shape(x) for x in dish__vdjb]
    if any(a is None for a in yaklt__fqxn):
        return ArrayAnalysis.AnalyzeResult(shape=dish__vdjb[0], pre=self.
            _call_assert_equiv(scope, loc, equiv_set, dish__vdjb))
    return self._broadcast_assert_shapes(scope, equiv_set, loc, yaklt__fqxn,
        kpw__inuqv)


if _check_numba_change:
    lines = inspect.getsource(numba.parfors.array_analysis.ArrayAnalysis.
        _analyze_broadcast)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '6c91fec038f56111338ea2b08f5f0e7f61ebdab1c81fb811fe26658cc354e40f':
        warnings.warn(
            'numba.parfors.array_analysis.ArrayAnalysis._analyze_broadcast has changed'
            )
numba.parfors.array_analysis.ArrayAnalysis._analyze_broadcast = (
    _analyze_broadcast)


def slice_size(self, index, dsize, equiv_set, scope, stmts):
    return None, None


numba.parfors.array_analysis.ArrayAnalysis.slice_size = slice_size


def convert_code_obj_to_function(code_obj, caller_ir):
    import bodo
    biql__nem = code_obj.code
    fbgi__vbvl = len(biql__nem.co_freevars)
    saj__pagj = biql__nem.co_freevars
    if code_obj.closure is not None:
        assert isinstance(code_obj.closure, ir.Var)
        dbn__xemw, op = ir_utils.find_build_sequence(caller_ir, code_obj.
            closure)
        assert op == 'build_tuple'
        saj__pagj = [otcze__qfk.name for otcze__qfk in dbn__xemw]
    aqpl__oqeow = caller_ir.func_id.func.__globals__
    try:
        aqpl__oqeow = getattr(code_obj, 'globals', aqpl__oqeow)
    except KeyError as rnigj__cbzwr:
        pass
    msg = (
        "Inner function is using non-constant variable '{}' from outer function. Please pass as argument if possible. See https://docs.bodo.ai/latest/source/programming_with_bodo/bodo_api_reference/udfs.html"
        )
    uca__zub = []
    for x in saj__pagj:
        try:
            crr__lykl = caller_ir.get_definition(x)
        except KeyError as rnigj__cbzwr:
            raise bodo.utils.typing.BodoError(msg.format(x), loc=code_obj.loc)
        from numba.core.registry import CPUDispatcher
        if isinstance(crr__lykl, (ir.Const, ir.Global, ir.FreeVar)):
            val = crr__lykl.value
            if isinstance(val, str):
                val = "'{}'".format(val)
            if isinstance(val, pytypes.FunctionType):
                bef__tyxhk = ir_utils.mk_unique_var('nested_func').replace('.',
                    '_')
                aqpl__oqeow[bef__tyxhk] = bodo.jit(distributed=False)(val)
                aqpl__oqeow[bef__tyxhk].is_nested_func = True
                val = bef__tyxhk
            if isinstance(val, CPUDispatcher):
                bef__tyxhk = ir_utils.mk_unique_var('nested_func').replace('.',
                    '_')
                aqpl__oqeow[bef__tyxhk] = val
                val = bef__tyxhk
            uca__zub.append(val)
        elif isinstance(crr__lykl, ir.Expr
            ) and crr__lykl.op == 'make_function':
            uupxd__drwqs = convert_code_obj_to_function(crr__lykl, caller_ir)
            bef__tyxhk = ir_utils.mk_unique_var('nested_func').replace('.', '_'
                )
            aqpl__oqeow[bef__tyxhk] = bodo.jit(distributed=False)(uupxd__drwqs)
            aqpl__oqeow[bef__tyxhk].is_nested_func = True
            uca__zub.append(bef__tyxhk)
        else:
            raise bodo.utils.typing.BodoError(msg.format(x), loc=code_obj.loc)
    zwauk__ruu = '\n'.join([('\tc_%d = %s' % (xgedl__blg, x)) for 
        xgedl__blg, x in enumerate(uca__zub)])
    jxf__dgfsq = ','.join([('c_%d' % xgedl__blg) for xgedl__blg in range(
        fbgi__vbvl)])
    apcw__yotc = list(biql__nem.co_varnames)
    pencl__rvqry = 0
    pzn__rvygl = biql__nem.co_argcount
    cmswh__nfk = caller_ir.get_definition(code_obj.defaults)
    if cmswh__nfk is not None:
        if isinstance(cmswh__nfk, tuple):
            wtpfw__pcjs = [caller_ir.get_definition(x).value for x in
                cmswh__nfk]
            ezki__mnoqs = tuple(wtpfw__pcjs)
        else:
            wtpfw__pcjs = [caller_ir.get_definition(x).value for x in
                cmswh__nfk.items]
            ezki__mnoqs = tuple(wtpfw__pcjs)
        pencl__rvqry = len(ezki__mnoqs)
    tsj__btfs = pzn__rvygl - pencl__rvqry
    vvdxi__bkqbx = ','.join([('%s' % apcw__yotc[xgedl__blg]) for xgedl__blg in
        range(tsj__btfs)])
    if pencl__rvqry:
        bpv__tcpke = [('%s = %s' % (apcw__yotc[xgedl__blg + tsj__btfs],
            ezki__mnoqs[xgedl__blg])) for xgedl__blg in range(pencl__rvqry)]
        vvdxi__bkqbx += ', '
        vvdxi__bkqbx += ', '.join(bpv__tcpke)
    return _create_function_from_code_obj(biql__nem, zwauk__ruu,
        vvdxi__bkqbx, jxf__dgfsq, aqpl__oqeow)


if _check_numba_change:
    lines = inspect.getsource(numba.core.ir_utils.convert_code_obj_to_function)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'b840769812418d589460e924a15477e83e7919aac8a3dcb0188ff447344aa8ac':
        warnings.warn(
            'numba.core.ir_utils.convert_code_obj_to_function has changed')
numba.core.ir_utils.convert_code_obj_to_function = convert_code_obj_to_function
numba.core.untyped_passes.convert_code_obj_to_function = (
    convert_code_obj_to_function)


def passmanager_run(self, state):
    from numba.core.compiler import _EarlyPipelineCompletion
    if not self.finalized:
        raise RuntimeError('Cannot run non-finalised pipeline')
    from numba.core.compiler_machinery import CompilerPass, _pass_registry
    import bodo
    for oyge__gsj, (latid__zgg, axkw__tebm) in enumerate(self.passes):
        try:
            numba.core.tracing.event('-- %s' % axkw__tebm)
            obiq__yjxf = _pass_registry.get(latid__zgg).pass_inst
            if isinstance(obiq__yjxf, CompilerPass):
                self._runPass(oyge__gsj, obiq__yjxf, state)
            else:
                raise BaseException('Legacy pass in use')
        except _EarlyPipelineCompletion as e:
            raise e
        except bodo.utils.typing.BodoError as e:
            raise
        except Exception as e:
            if numba.core.config.DEVELOPER_MODE:
                from numba.core import utils
                if utils.use_new_style_errors() and not isinstance(e,
                    errors.NumbaError):
                    raise e
                msg = 'Failed in %s mode pipeline (step: %s)' % (self.
                    pipeline_name, axkw__tebm)
                fkp__otkq = self._patch_error(msg, e)
                raise fkp__otkq
            else:
                raise e


if _check_numba_change:
    lines = inspect.getsource(numba.core.compiler_machinery.PassManager.run)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '43505782e15e690fd2d7e53ea716543bec37aa0633502956864edf649e790cdb':
        warnings.warn(
            'numba.core.compiler_machinery.PassManager.run has changed')
numba.core.compiler_machinery.PassManager.run = passmanager_run
if _check_numba_change:
    lines = inspect.getsource(numba.np.ufunc.parallel._launch_threads)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'a57ef28c4168fdd436a5513bba4351ebc6d9fba76c5819f44046431a79b9030f':
        warnings.warn('numba.np.ufunc.parallel._launch_threads has changed')
numba.np.ufunc.parallel._launch_threads = lambda : None


def get_reduce_nodes(reduction_node, nodes, func_ir):
    vbum__uiol = None
    hyyrn__ukm = {}

    def lookup(var, already_seen, varonly=True):
        val = hyyrn__ukm.get(var.name, None)
        if isinstance(val, ir.Var):
            if val.name in already_seen:
                return var
            already_seen.add(val.name)
            return lookup(val, already_seen, varonly)
        else:
            return var if varonly or val is None else val
    name = reduction_node.name
    ekcs__bzu = reduction_node.unversioned_name
    for xgedl__blg, stmt in enumerate(nodes):
        bqjwa__wqmxx = stmt.target
        chg__mnb = stmt.value
        hyyrn__ukm[bqjwa__wqmxx.name] = chg__mnb
        if isinstance(chg__mnb, ir.Var) and chg__mnb.name in hyyrn__ukm:
            chg__mnb = lookup(chg__mnb, set())
        if isinstance(chg__mnb, ir.Expr):
            tlhrh__ohkh = set(lookup(otcze__qfk, set(), True).name for
                otcze__qfk in chg__mnb.list_vars())
            if name in tlhrh__ohkh:
                args = [(x.name, lookup(x, set(), True)) for x in
                    get_expr_args(chg__mnb)]
                jibai__tarue = [x for x, tqjwz__nxec in args if tqjwz__nxec
                    .name != name]
                args = [(x, tqjwz__nxec) for x, tqjwz__nxec in args if x !=
                    tqjwz__nxec.name]
                huugv__kxco = dict(args)
                if len(jibai__tarue) == 1:
                    huugv__kxco[jibai__tarue[0]] = ir.Var(bqjwa__wqmxx.
                        scope, name + '#init', bqjwa__wqmxx.loc)
                replace_vars_inner(chg__mnb, huugv__kxco)
                vbum__uiol = nodes[xgedl__blg:]
                break
    return vbum__uiol


if _check_numba_change:
    lines = inspect.getsource(numba.parfors.parfor.get_reduce_nodes)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'a05b52aff9cb02e595a510cd34e973857303a71097fc5530567cb70ca183ef3b':
        warnings.warn('numba.parfors.parfor.get_reduce_nodes has changed')
numba.parfors.parfor.get_reduce_nodes = get_reduce_nodes


def _can_reorder_stmts(stmt, next_stmt, func_ir, call_table, alias_map,
    arg_aliases):
    from numba.parfors.parfor import Parfor, expand_aliases, is_assert_equiv
    if isinstance(stmt, Parfor) and not isinstance(next_stmt, Parfor
        ) and not isinstance(next_stmt, ir.Print) and (not isinstance(
        next_stmt, ir.Assign) or has_no_side_effect(next_stmt.value, set(),
        call_table) or guard(is_assert_equiv, func_ir, next_stmt.value)):
        yovf__vexke = expand_aliases({otcze__qfk.name for otcze__qfk in
            stmt.list_vars()}, alias_map, arg_aliases)
        nuwj__kkt = expand_aliases(get_parfor_writes(stmt, func_ir),
            alias_map, arg_aliases)
        jpk__keqem = expand_aliases({otcze__qfk.name for otcze__qfk in
            next_stmt.list_vars()}, alias_map, arg_aliases)
        lfaz__vhsze = expand_aliases(get_stmt_writes(next_stmt, func_ir),
            alias_map, arg_aliases)
        if len(nuwj__kkt & jpk__keqem | lfaz__vhsze & yovf__vexke) == 0:
            return True
    return False


if _check_numba_change:
    lines = inspect.getsource(numba.parfors.parfor._can_reorder_stmts)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '18caa9a01b21ab92b4f79f164cfdbc8574f15ea29deedf7bafdf9b0e755d777c':
        warnings.warn('numba.parfors.parfor._can_reorder_stmts has changed')
numba.parfors.parfor._can_reorder_stmts = _can_reorder_stmts


def get_parfor_writes(parfor, func_ir):
    from numba.parfors.parfor import Parfor
    assert isinstance(parfor, Parfor)
    inf__scfzp = set()
    blocks = parfor.loop_body.copy()
    blocks[-1] = parfor.init_block
    for block in blocks.values():
        for stmt in block.body:
            inf__scfzp.update(get_stmt_writes(stmt, func_ir))
            if isinstance(stmt, Parfor):
                inf__scfzp.update(get_parfor_writes(stmt, func_ir))
    return inf__scfzp


if _check_numba_change:
    lines = inspect.getsource(numba.parfors.parfor.get_parfor_writes)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'a7b29cd76832b6f6f1f2d2397ec0678c1409b57a6eab588bffd344b775b1546f':
        warnings.warn('numba.parfors.parfor.get_parfor_writes has changed')


def get_stmt_writes(stmt, func_ir):
    import bodo
    from bodo.utils.utils import is_call_assign
    inf__scfzp = set()
    if isinstance(stmt, (ir.Assign, ir.SetItem, ir.StaticSetItem)):
        inf__scfzp.add(stmt.target.name)
    if isinstance(stmt, bodo.ir.aggregate.Aggregate):
        inf__scfzp = {otcze__qfk.name for otcze__qfk in stmt.df_out_vars.
            values()}
        if stmt.out_key_vars is not None:
            inf__scfzp.update({otcze__qfk.name for otcze__qfk in stmt.
                out_key_vars})
    if isinstance(stmt, (bodo.ir.csv_ext.CsvReader, bodo.ir.parquet_ext.
        ParquetReader)):
        inf__scfzp = {otcze__qfk.name for otcze__qfk in stmt.out_vars}
    if isinstance(stmt, bodo.ir.join.Join):
        inf__scfzp = {otcze__qfk.name for otcze__qfk in stmt.out_data_vars.
            values()}
    if isinstance(stmt, bodo.ir.sort.Sort):
        if not stmt.inplace:
            inf__scfzp.update({otcze__qfk.name for otcze__qfk in stmt.
                out_key_arrs})
            inf__scfzp.update({otcze__qfk.name for otcze__qfk in stmt.
                df_out_vars.values()})
    if is_call_assign(stmt):
        xuh__movn = guard(find_callname, func_ir, stmt.value)
        if xuh__movn in (('setitem_str_arr_ptr', 'bodo.libs.str_arr_ext'),
            ('setna', 'bodo.libs.array_kernels'), (
            'str_arr_item_to_numeric', 'bodo.libs.str_arr_ext'), (
            'str_arr_setitem_int_to_str', 'bodo.libs.str_arr_ext'), (
            'str_arr_setitem_NA_str', 'bodo.libs.str_arr_ext'), (
            'str_arr_set_not_na', 'bodo.libs.str_arr_ext'), (
            'get_str_arr_item_copy', 'bodo.libs.str_arr_ext'), (
            'set_bit_to_arr', 'bodo.libs.int_arr_ext')):
            inf__scfzp.add(stmt.value.args[0].name)
    return inf__scfzp


if _check_numba_change:
    lines = inspect.getsource(numba.core.ir_utils.get_stmt_writes)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '1a7a80b64c9a0eb27e99dc8eaae187bde379d4da0b74c84fbf87296d87939974':
        warnings.warn('numba.core.ir_utils.get_stmt_writes has changed')


def patch_message(self, new_message):
    self.msg = new_message
    self.args = (new_message,) + self.args[1:]


if _check_numba_change:
    lines = inspect.getsource(numba.core.errors.NumbaError.patch_message)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'ed189a428a7305837e76573596d767b6e840e99f75c05af6941192e0214fa899':
        warnings.warn('numba.core.errors.NumbaError.patch_message has changed')
numba.core.errors.NumbaError.patch_message = patch_message


def add_context(self, msg):
    if numba.core.config.DEVELOPER_MODE:
        self.contexts.append(msg)
        ovuka__mos = _termcolor.errmsg('{0}') + _termcolor.filename(
            'During: {1}')
        tiu__jqqxv = ovuka__mos.format(self, msg)
        self.args = tiu__jqqxv,
    else:
        ovuka__mos = _termcolor.errmsg('{0}')
        tiu__jqqxv = ovuka__mos.format(self)
        self.args = tiu__jqqxv,
    return self


if _check_numba_change:
    lines = inspect.getsource(numba.core.errors.NumbaError.add_context)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '6a388d87788f8432c2152ac55ca9acaa94dbc3b55be973b2cf22dd4ee7179ab8':
        warnings.warn('numba.core.errors.NumbaError.add_context has changed')
numba.core.errors.NumbaError.add_context = add_context


def _get_dist_spec_from_options(spec, **options):
    from bodo.transforms.distributed_analysis import Distribution
    dist_spec = {}
    if 'distributed' in options:
        for qcpe__hhpm in options['distributed']:
            dist_spec[qcpe__hhpm] = Distribution.OneD_Var
    if 'distributed_block' in options:
        for qcpe__hhpm in options['distributed_block']:
            dist_spec[qcpe__hhpm] = Distribution.OneD
    return dist_spec


def register_class_type(cls, spec, class_ctor, builder, **options):
    import typing as pt
    from numba.core.typing.asnumbatype import as_numba_type
    import bodo
    dist_spec = _get_dist_spec_from_options(spec, **options)
    ear__ypx = options.get('returns_maybe_distributed', True)
    if spec is None:
        spec = OrderedDict()
    elif isinstance(spec, Sequence):
        spec = OrderedDict(spec)
    for attr, pzky__iub in pt.get_type_hints(cls).items():
        if attr not in spec:
            spec[attr] = as_numba_type(pzky__iub)
    jitclass_base._validate_spec(spec)
    spec = jitclass_base._fix_up_private_attr(cls.__name__, spec)
    sum__vibno = {}
    for zen__euc in reversed(inspect.getmro(cls)):
        sum__vibno.update(zen__euc.__dict__)
    gfdyj__fwid, ulii__vsrvv, pzgcm__uua, lxjff__jwe = {}, {}, {}, {}
    for bfxes__thqc, otcze__qfk in sum__vibno.items():
        if isinstance(otcze__qfk, pytypes.FunctionType):
            gfdyj__fwid[bfxes__thqc] = otcze__qfk
        elif isinstance(otcze__qfk, property):
            ulii__vsrvv[bfxes__thqc] = otcze__qfk
        elif isinstance(otcze__qfk, staticmethod):
            pzgcm__uua[bfxes__thqc] = otcze__qfk
        else:
            lxjff__jwe[bfxes__thqc] = otcze__qfk
    ynua__atfu = (set(gfdyj__fwid) | set(ulii__vsrvv) | set(pzgcm__uua)) & set(
        spec)
    if ynua__atfu:
        raise NameError('name shadowing: {0}'.format(', '.join(ynua__atfu)))
    vos__hpb = lxjff__jwe.pop('__doc__', '')
    jitclass_base._drop_ignored_attrs(lxjff__jwe)
    if lxjff__jwe:
        msg = 'class members are not yet supported: {0}'
        kqa__gch = ', '.join(lxjff__jwe.keys())
        raise TypeError(msg.format(kqa__gch))
    for bfxes__thqc, otcze__qfk in ulii__vsrvv.items():
        if otcze__qfk.fdel is not None:
            raise TypeError('deleter is not supported: {0}'.format(bfxes__thqc)
                )
    jit_methods = {bfxes__thqc: bodo.jit(returns_maybe_distributed=ear__ypx
        )(otcze__qfk) for bfxes__thqc, otcze__qfk in gfdyj__fwid.items()}
    jit_props = {}
    for bfxes__thqc, otcze__qfk in ulii__vsrvv.items():
        mihkk__gmgrx = {}
        if otcze__qfk.fget:
            mihkk__gmgrx['get'] = bodo.jit(otcze__qfk.fget)
        if otcze__qfk.fset:
            mihkk__gmgrx['set'] = bodo.jit(otcze__qfk.fset)
        jit_props[bfxes__thqc] = mihkk__gmgrx
    jit_static_methods = {bfxes__thqc: bodo.jit(otcze__qfk.__func__) for 
        bfxes__thqc, otcze__qfk in pzgcm__uua.items()}
    mdg__tuuzk = class_ctor(cls, jitclass_base.ConstructorTemplate, spec,
        jit_methods, jit_props, jit_static_methods, dist_spec)
    mwdy__mqi = dict(class_type=mdg__tuuzk, __doc__=vos__hpb)
    mwdy__mqi.update(jit_static_methods)
    cls = jitclass_base.JitClassType(cls.__name__, (cls,), mwdy__mqi)
    typingctx = numba.core.registry.cpu_target.typing_context
    typingctx.insert_global(cls, mdg__tuuzk)
    targetctx = numba.core.registry.cpu_target.target_context
    builder(mdg__tuuzk, typingctx, targetctx).register()
    as_numba_type.register(cls, mdg__tuuzk.instance_type)
    return cls


if _check_numba_change:
    lines = inspect.getsource(jitclass_base.register_class_type)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '005e6e2e89a47f77a19ba86305565050d4dbc2412fc4717395adf2da348671a9':
        warnings.warn('jitclass_base.register_class_type has changed')
jitclass_base.register_class_type = register_class_type


def ClassType__init__(self, class_def, ctor_template_cls, struct,
    jit_methods, jit_props, jit_static_methods, dist_spec=None):
    if dist_spec is None:
        dist_spec = {}
    self.class_name = class_def.__name__
    self.class_doc = class_def.__doc__
    self._ctor_template_class = ctor_template_cls
    self.jit_methods = jit_methods
    self.jit_props = jit_props
    self.jit_static_methods = jit_static_methods
    self.struct = struct
    self.dist_spec = dist_spec
    zref__yuki = ','.join('{0}:{1}'.format(bfxes__thqc, otcze__qfk) for 
        bfxes__thqc, otcze__qfk in struct.items())
    ruyyg__bei = ','.join('{0}:{1}'.format(bfxes__thqc, otcze__qfk) for 
        bfxes__thqc, otcze__qfk in dist_spec.items())
    name = '{0}.{1}#{2:x}<{3}><{4}>'.format(self.name_prefix, self.
        class_name, id(self), zref__yuki, ruyyg__bei)
    super(types.misc.ClassType, self).__init__(name)


if _check_numba_change:
    lines = inspect.getsource(types.misc.ClassType.__init__)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '2b848ea82946c88f540e81f93ba95dfa7cd66045d944152a337fe2fc43451c30':
        warnings.warn('types.misc.ClassType.__init__ has changed')
types.misc.ClassType.__init__ = ClassType__init__


def jitclass(cls_or_spec=None, spec=None, **options):
    if cls_or_spec is not None and spec is None and not isinstance(cls_or_spec,
        type):
        spec = cls_or_spec
        cls_or_spec = None

    def wrap(cls):
        if numba.core.config.DISABLE_JIT:
            return cls
        else:
            from numba.experimental.jitclass.base import ClassBuilder
            return register_class_type(cls, spec, types.ClassType,
                ClassBuilder, **options)
    if cls_or_spec is None:
        return wrap
    else:
        return wrap(cls_or_spec)


if _check_numba_change:
    lines = inspect.getsource(jitclass_decorators.jitclass)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '265f1953ee5881d1a5d90238d3c932cd300732e41495657e65bf51e59f7f4af5':
        warnings.warn('jitclass_decorators.jitclass has changed')


def CallConstraint_resolve(self, typeinfer, typevars, fnty):
    assert fnty
    context = typeinfer.context
    tti__ixjw = numba.core.typeinfer.fold_arg_vars(typevars, self.args,
        self.vararg, self.kws)
    if tti__ixjw is None:
        return
    qnhqq__zppny, iyh__mht = tti__ixjw
    for a in itertools.chain(qnhqq__zppny, iyh__mht.values()):
        if not a.is_precise() and not isinstance(a, types.Array):
            return
    if isinstance(fnty, types.TypeRef):
        fnty = fnty.instance_type
    try:
        sig = typeinfer.resolve_call(fnty, qnhqq__zppny, iyh__mht)
    except ForceLiteralArg as e:
        fffob__cgmf = (fnty.this,) + tuple(self.args) if isinstance(fnty,
            types.BoundFunction) else self.args
        folded = e.fold_arguments(fffob__cgmf, self.kws)
        unfnq__afau = set()
        uvavr__oux = set()
        uqpx__kef = {}
        for oyge__gsj in e.requested_args:
            ycv__bpukl = typeinfer.func_ir.get_definition(folded[oyge__gsj])
            if isinstance(ycv__bpukl, ir.Arg):
                unfnq__afau.add(ycv__bpukl.index)
                if ycv__bpukl.index in e.file_infos:
                    uqpx__kef[ycv__bpukl.index] = e.file_infos[ycv__bpukl.index
                        ]
            else:
                uvavr__oux.add(oyge__gsj)
        if uvavr__oux:
            raise TypingError('Cannot request literal type.', loc=self.loc)
        elif unfnq__afau:
            raise ForceLiteralArg(unfnq__afau, loc=self.loc, file_infos=
                uqpx__kef)
    if sig is None:
        ajl__kgf = 'Invalid use of {0} with parameters ({1})'
        args = [str(a) for a in qnhqq__zppny]
        args += [('%s=%s' % (bfxes__thqc, otcze__qfk)) for bfxes__thqc,
            otcze__qfk in sorted(iyh__mht.items())]
        bzkmf__rrj = ajl__kgf.format(fnty, ', '.join(map(str, args)))
        cwi__zhr = context.explain_function_type(fnty)
        msg = '\n'.join([bzkmf__rrj, cwi__zhr])
        raise TypingError(msg)
    typeinfer.add_type(self.target, sig.return_type, loc=self.loc)
    if isinstance(fnty, types.BoundFunction
        ) and sig.recvr is not None and sig.recvr != fnty.this:
        rhd__macsk = context.unify_pairs(sig.recvr, fnty.this)
        if rhd__macsk is None and fnty.this.is_precise(
            ) and sig.recvr.is_precise():
            msg = 'Cannot refine type {} to {}'.format(sig.recvr, fnty.this)
            raise TypingError(msg, loc=self.loc)
        if rhd__macsk is not None and rhd__macsk.is_precise():
            okop__vtlb = fnty.copy(this=rhd__macsk)
            typeinfer.propagate_refined_type(self.func, okop__vtlb)
    if not sig.return_type.is_precise():
        gfp__qprpp = typevars[self.target]
        if gfp__qprpp.defined:
            zyv__tqk = gfp__qprpp.getone()
            if context.unify_pairs(zyv__tqk, sig.return_type) == zyv__tqk:
                sig = sig.replace(return_type=zyv__tqk)
    self.signature = sig
    self._add_refine_map(typeinfer, typevars, sig)


if _check_numba_change:
    lines = inspect.getsource(numba.core.typeinfer.CallConstraint.resolve)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'c78cd8ffc64b836a6a2ddf0362d481b52b9d380c5249920a87ff4da052ce081f':
        warnings.warn('numba.core.typeinfer.CallConstraint.resolve has changed'
            )
numba.core.typeinfer.CallConstraint.resolve = CallConstraint_resolve


def ForceLiteralArg__init__(self, arg_indices, fold_arguments=None, loc=
    None, file_infos=None):
    super(ForceLiteralArg, self).__init__(
        'Pseudo-exception to force literal arguments in the dispatcher',
        loc=loc)
    self.requested_args = frozenset(arg_indices)
    self.fold_arguments = fold_arguments
    if file_infos is None:
        self.file_infos = {}
    else:
        self.file_infos = file_infos


if _check_numba_change:
    lines = inspect.getsource(numba.core.errors.ForceLiteralArg.__init__)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'b241d5e36a4cf7f4c73a7ad3238693612926606c7a278cad1978070b82fb55ef':
        warnings.warn('numba.core.errors.ForceLiteralArg.__init__ has changed')
numba.core.errors.ForceLiteralArg.__init__ = ForceLiteralArg__init__


def ForceLiteralArg_bind_fold_arguments(self, fold_arguments):
    e = ForceLiteralArg(self.requested_args, fold_arguments, loc=self.loc,
        file_infos=self.file_infos)
    return numba.core.utils.chain_exception(e, self)


if _check_numba_change:
    lines = inspect.getsource(numba.core.errors.ForceLiteralArg.
        bind_fold_arguments)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '1e93cca558f7c604a47214a8f2ec33ee994104cb3e5051166f16d7cc9315141d':
        warnings.warn(
            'numba.core.errors.ForceLiteralArg.bind_fold_arguments has changed'
            )
numba.core.errors.ForceLiteralArg.bind_fold_arguments = (
    ForceLiteralArg_bind_fold_arguments)


def ForceLiteralArg_combine(self, other):
    if not isinstance(other, ForceLiteralArg):
        hyoh__pddj = '*other* must be a {} but got a {} instead'
        raise TypeError(hyoh__pddj.format(ForceLiteralArg, type(other)))
    return ForceLiteralArg(self.requested_args | other.requested_args, {**
        self.file_infos, **other.file_infos})


if _check_numba_change:
    lines = inspect.getsource(numba.core.errors.ForceLiteralArg.combine)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '49bf06612776f5d755c1c7d1c5eb91831a57665a8fed88b5651935f3bf33e899':
        warnings.warn('numba.core.errors.ForceLiteralArg.combine has changed')
numba.core.errors.ForceLiteralArg.combine = ForceLiteralArg_combine


def _get_global_type(self, gv):
    from bodo.utils.typing import FunctionLiteral
    ty = self._lookup_global(gv)
    if ty is not None:
        return ty
    if isinstance(gv, pytypes.ModuleType):
        return types.Module(gv)
    if isinstance(gv, pytypes.FunctionType):
        return FunctionLiteral(gv)


if _check_numba_change:
    lines = inspect.getsource(numba.core.typing.context.BaseContext.
        _get_global_type)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '8ffe6b81175d1eecd62a37639b5005514b4477d88f35f5b5395041ac8c945a4a':
        warnings.warn(
            'numba.core.typing.context.BaseContext._get_global_type has changed'
            )
numba.core.typing.context.BaseContext._get_global_type = _get_global_type


def _legalize_args(self, func_ir, args, kwargs, loc, func_globals,
    func_closures):
    from numba.core import sigutils
    from bodo.utils.transform import get_const_value_inner
    if args:
        raise errors.CompilerError(
            "objectmode context doesn't take any positional arguments")
    wwa__yhqg = {}

    def report_error(varname, msg, loc):
        raise errors.CompilerError(
            f'Error handling objmode argument {varname!r}. {msg}', loc=loc)
    for bfxes__thqc, otcze__qfk in kwargs.items():
        iwl__ubbuk = None
        try:
            rpqy__yznfg = ir.Var(ir.Scope(None, loc), ir_utils.
                mk_unique_var('dummy'), loc)
            func_ir._definitions[rpqy__yznfg.name] = [otcze__qfk]
            iwl__ubbuk = get_const_value_inner(func_ir, rpqy__yznfg)
            func_ir._definitions.pop(rpqy__yznfg.name)
            if isinstance(iwl__ubbuk, str):
                iwl__ubbuk = sigutils._parse_signature_string(iwl__ubbuk)
            if isinstance(iwl__ubbuk, types.abstract._TypeMetaclass):
                raise BodoError(
                    f"""objmode type annotations require full data types, not just data type classes. For example, 'bodo.DataFrameType((bodo.float64[::1],), bodo.RangeIndexType(), ('A',))' is a valid data type but 'bodo.DataFrameType' is not.
Variable {bfxes__thqc} is annotated as type class {iwl__ubbuk}."""
                    )
            assert isinstance(iwl__ubbuk, types.Type)
            if isinstance(iwl__ubbuk, (types.List, types.Set)):
                iwl__ubbuk = iwl__ubbuk.copy(reflected=False)
            wwa__yhqg[bfxes__thqc] = iwl__ubbuk
        except BodoError as rnigj__cbzwr:
            raise
        except:
            msg = (
                'The value must be a compile-time constant either as a non-local variable or an expression that refers to a Bodo type.'
                )
            if isinstance(iwl__ubbuk, ir.UndefinedType):
                msg = f'not defined.'
                if isinstance(otcze__qfk, ir.Global):
                    msg = f'Global {otcze__qfk.name!r} is not defined.'
                if isinstance(otcze__qfk, ir.FreeVar):
                    msg = f'Freevar {otcze__qfk.name!r} is not defined.'
            if isinstance(otcze__qfk, ir.Expr) and otcze__qfk.op == 'getattr':
                msg = 'Getattr cannot be resolved at compile-time.'
            report_error(varname=bfxes__thqc, msg=msg, loc=loc)
    for name, typ in wwa__yhqg.items():
        self._legalize_arg_type(name, typ, loc)
    return wwa__yhqg


if _check_numba_change:
    lines = inspect.getsource(numba.core.withcontexts._ObjModeContextType.
        _legalize_args)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '867c9ba7f1bcf438be56c38e26906bb551f59a99f853a9f68b71208b107c880e':
        warnings.warn(
            'numba.core.withcontexts._ObjModeContextType._legalize_args has changed'
            )
numba.core.withcontexts._ObjModeContextType._legalize_args = _legalize_args


def op_FORMAT_VALUE_byteflow(self, state, inst):
    flags = inst.arg
    if flags & 3 != 0:
        msg = 'str/repr/ascii conversion in f-strings not supported yet'
        raise errors.UnsupportedError(msg, loc=self.get_debug_loc(inst.lineno))
    format_spec = None
    if flags & 4 == 4:
        format_spec = state.pop()
    value = state.pop()
    fmtvar = state.make_temp()
    res = state.make_temp()
    state.append(inst, value=value, res=res, fmtvar=fmtvar, format_spec=
        format_spec)
    state.push(res)


def op_BUILD_STRING_byteflow(self, state, inst):
    dvu__yogbc = inst.arg
    assert dvu__yogbc > 0, 'invalid BUILD_STRING count'
    strings = list(reversed([state.pop() for _ in range(dvu__yogbc)]))
    tmps = [state.make_temp() for _ in range(dvu__yogbc - 1)]
    state.append(inst, strings=strings, tmps=tmps)
    state.push(tmps[-1])


numba.core.byteflow.TraceRunner.op_FORMAT_VALUE = op_FORMAT_VALUE_byteflow
numba.core.byteflow.TraceRunner.op_BUILD_STRING = op_BUILD_STRING_byteflow


def op_FORMAT_VALUE_interpreter(self, inst, value, res, fmtvar, format_spec):
    value = self.get(value)
    ulca__pay = ir.Global('format', format, loc=self.loc)
    self.store(value=ulca__pay, name=fmtvar)
    args = (value, self.get(format_spec)) if format_spec else (value,)
    cqhi__dowk = ir.Expr.call(self.get(fmtvar), args, (), loc=self.loc)
    self.store(value=cqhi__dowk, name=res)


def op_BUILD_STRING_interpreter(self, inst, strings, tmps):
    dvu__yogbc = inst.arg
    assert dvu__yogbc > 0, 'invalid BUILD_STRING count'
    hnt__hvbl = self.get(strings[0])
    for other, twliw__oatkl in zip(strings[1:], tmps):
        other = self.get(other)
        betea__ovk = ir.Expr.binop(operator.add, lhs=hnt__hvbl, rhs=other,
            loc=self.loc)
        self.store(betea__ovk, twliw__oatkl)
        hnt__hvbl = self.get(twliw__oatkl)


numba.core.interpreter.Interpreter.op_FORMAT_VALUE = (
    op_FORMAT_VALUE_interpreter)
numba.core.interpreter.Interpreter.op_BUILD_STRING = (
    op_BUILD_STRING_interpreter)


def object_hasattr_string(self, obj, attr):
    from llvmlite.llvmpy.core import Type
    ldaps__ppmp = self.context.insert_const_string(self.module, attr)
    fnty = Type.function(Type.int(), [self.pyobj, self.cstring])
    fn = self._get_function(fnty, name='PyObject_HasAttrString')
    return self.builder.call(fn, [obj, ldaps__ppmp])


numba.core.pythonapi.PythonAPI.object_hasattr_string = object_hasattr_string


def _created_inlined_var_name(function_name, var_name):
    kug__irpi = mk_unique_var(f'{var_name}')
    mcs__mwkt = kug__irpi.replace('<', '_').replace('>', '_')
    mcs__mwkt = mcs__mwkt.replace('.', '_').replace('$', '_v')
    return mcs__mwkt


if _check_numba_change:
    lines = inspect.getsource(numba.core.inline_closurecall.
        _created_inlined_var_name)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '0d91aac55cd0243e58809afe9d252562f9ae2899cde1112cc01a46804e01821e':
        warnings.warn(
            'numba.core.inline_closurecall._created_inlined_var_name has changed'
            )
numba.core.inline_closurecall._created_inlined_var_name = (
    _created_inlined_var_name)


def resolve_number___call__(self, classty):
    import numpy as np
    from numba.core.typing.templates import make_callable_template
    ty = classty.instance_type

    def typer(val):
        if isinstance(val, (types.BaseTuple, types.Sequence)):
            fnty = self.context.resolve_value_type(np.array)
            sig = fnty.get_call_type(self.context, (val, types.DType(ty)), {})
            return sig.return_type
        elif isinstance(val, (types.Number, types.Boolean, types.IntEnumMember)
            ):
            return ty
        elif val == types.unicode_type:
            return ty
        elif isinstance(val, (types.NPDatetime, types.NPTimedelta)):
            if ty.bitwidth == 64:
                return ty
            else:
                msg = f'Cannot cast {val} to {ty} as {ty} is not 64 bits wide.'
                raise errors.TypingError(msg)
        elif isinstance(val, types.Array
            ) and val.ndim == 0 and val.dtype == ty:
            return ty
        else:
            msg = f'Casting {val} to {ty} directly is unsupported.'
            if isinstance(val, types.Array):
                msg += f" Try doing '<array>.astype(np.{ty})' instead"
            raise errors.TypingError(msg)
    return types.Function(make_callable_template(key=ty, typer=typer))


if _check_numba_change:
    lines = inspect.getsource(numba.core.typing.builtins.
        NumberClassAttribute.resolve___call__)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'fdaf0c7d0820130481bb2bd922985257b9281b670f0bafffe10e51cabf0d5081':
        warnings.warn(
            'numba.core.typing.builtins.NumberClassAttribute.resolve___call__ has changed'
            )
numba.core.typing.builtins.NumberClassAttribute.resolve___call__ = (
    resolve_number___call__)


def on_assign(self, states, assign):
    if assign.target.name == states['varname']:
        scope = states['scope']
        cdp__fxuy = states['defmap']
        if len(cdp__fxuy) == 0:
            sxzv__ivkaa = assign.target
            numba.core.ssa._logger.debug('first assign: %s', sxzv__ivkaa)
            if sxzv__ivkaa.name not in scope.localvars:
                sxzv__ivkaa = scope.define(assign.target.name, loc=assign.loc)
        else:
            sxzv__ivkaa = scope.redefine(assign.target.name, loc=assign.loc)
        assign = ir.Assign(target=sxzv__ivkaa, value=assign.value, loc=
            assign.loc)
        cdp__fxuy[states['label']].append(assign)
    return assign


if _check_numba_change:
    lines = inspect.getsource(numba.core.ssa._FreshVarHandler.on_assign)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '922c4f9807455f81600b794bbab36f9c6edfecfa83fda877bf85f465db7865e8':
        warnings.warn('_FreshVarHandler on_assign has changed')
numba.core.ssa._FreshVarHandler.on_assign = on_assign


def get_np_ufunc_typ_lst(func):
    from numba.core import typing
    yks__urrqy = []
    for bfxes__thqc, otcze__qfk in typing.npydecl.registry.globals:
        if bfxes__thqc == func:
            yks__urrqy.append(otcze__qfk)
    for bfxes__thqc, otcze__qfk in typing.templates.builtin_registry.globals:
        if bfxes__thqc == func:
            yks__urrqy.append(otcze__qfk)
    if len(yks__urrqy) == 0:
        raise RuntimeError('type for func ', func, ' not found')
    return yks__urrqy


def canonicalize_array_math(func_ir, typemap, calltypes, typingctx):
    import numpy
    from numba.core.ir_utils import arr_math, find_topo_order, mk_unique_var
    blocks = func_ir.blocks
    brwqd__wqq = {}
    vcw__diovy = find_topo_order(blocks)
    mwh__sua = {}
    for wsr__ata in vcw__diovy:
        block = blocks[wsr__ata]
        kpz__hdz = []
        for stmt in block.body:
            if isinstance(stmt, ir.Assign) and isinstance(stmt.value, ir.Expr):
                bqjwa__wqmxx = stmt.target.name
                chg__mnb = stmt.value
                if (chg__mnb.op == 'getattr' and chg__mnb.attr in arr_math and
                    isinstance(typemap[chg__mnb.value.name], types.npytypes
                    .Array)):
                    chg__mnb = stmt.value
                    teomv__tsft = chg__mnb.value
                    brwqd__wqq[bqjwa__wqmxx] = teomv__tsft
                    scope = teomv__tsft.scope
                    loc = teomv__tsft.loc
                    aua__cajmy = ir.Var(scope, mk_unique_var('$np_g_var'), loc)
                    typemap[aua__cajmy.name] = types.misc.Module(numpy)
                    mpluo__muwv = ir.Global('np', numpy, loc)
                    uhub__eafg = ir.Assign(mpluo__muwv, aua__cajmy, loc)
                    chg__mnb.value = aua__cajmy
                    kpz__hdz.append(uhub__eafg)
                    func_ir._definitions[aua__cajmy.name] = [mpluo__muwv]
                    func = getattr(numpy, chg__mnb.attr)
                    mrfei__vflp = get_np_ufunc_typ_lst(func)
                    mwh__sua[bqjwa__wqmxx] = mrfei__vflp
                if chg__mnb.op == 'call' and chg__mnb.func.name in brwqd__wqq:
                    teomv__tsft = brwqd__wqq[chg__mnb.func.name]
                    xrge__gxcit = calltypes.pop(chg__mnb)
                    fjlfs__jqe = xrge__gxcit.args[:len(chg__mnb.args)]
                    dbkih__oum = {name: typemap[otcze__qfk.name] for name,
                        otcze__qfk in chg__mnb.kws}
                    bkpur__zip = mwh__sua[chg__mnb.func.name]
                    hml__amra = None
                    for lhqpt__vaux in bkpur__zip:
                        try:
                            hml__amra = lhqpt__vaux.get_call_type(typingctx,
                                [typemap[teomv__tsft.name]] + list(
                                fjlfs__jqe), dbkih__oum)
                            typemap.pop(chg__mnb.func.name)
                            typemap[chg__mnb.func.name] = lhqpt__vaux
                            calltypes[chg__mnb] = hml__amra
                            break
                        except Exception as rnigj__cbzwr:
                            pass
                    if hml__amra is None:
                        raise TypeError(
                            f'No valid template found for {chg__mnb.func.name}'
                            )
                    chg__mnb.args = [teomv__tsft] + chg__mnb.args
            kpz__hdz.append(stmt)
        block.body = kpz__hdz


if _check_numba_change:
    lines = inspect.getsource(numba.core.ir_utils.canonicalize_array_math)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'b2200e9100613631cc554f4b640bc1181ba7cea0ece83630122d15b86941be2e':
        warnings.warn('canonicalize_array_math has changed')
numba.core.ir_utils.canonicalize_array_math = canonicalize_array_math
numba.parfors.parfor.canonicalize_array_math = canonicalize_array_math
numba.core.inline_closurecall.canonicalize_array_math = canonicalize_array_math


def _Numpy_Rules_ufunc_handle_inputs(cls, ufunc, args, kws):
    slm__erqbn = ufunc.nin
    qejtv__rxe = ufunc.nout
    tsj__btfs = ufunc.nargs
    assert tsj__btfs == slm__erqbn + qejtv__rxe
    if len(args) < slm__erqbn:
        msg = "ufunc '{0}': not enough arguments ({1} found, {2} required)"
        raise TypingError(msg=msg.format(ufunc.__name__, len(args), slm__erqbn)
            )
    if len(args) > tsj__btfs:
        msg = "ufunc '{0}': too many arguments ({1} found, {2} maximum)"
        raise TypingError(msg=msg.format(ufunc.__name__, len(args), tsj__btfs))
    args = [(a.as_array if isinstance(a, types.ArrayCompatible) and not
        isinstance(a, types.Bytes) else a) for a in args]
    gbhex__vcdt = [(a.ndim if isinstance(a, types.ArrayCompatible) and not
        isinstance(a, types.Bytes) else 0) for a in args]
    moa__bqqtn = max(gbhex__vcdt)
    cnr__mgi = args[slm__erqbn:]
    if not all(wtpfw__pcjs == moa__bqqtn for wtpfw__pcjs in gbhex__vcdt[
        slm__erqbn:]):
        msg = "ufunc '{0}' called with unsuitable explicit output arrays."
        raise TypingError(msg=msg.format(ufunc.__name__))
    if not all(isinstance(oov__owit, types.ArrayCompatible) and not
        isinstance(oov__owit, types.Bytes) for oov__owit in cnr__mgi):
        msg = "ufunc '{0}' called with an explicit output that is not an array"
        raise TypingError(msg=msg.format(ufunc.__name__))
    if not all(oov__owit.mutable for oov__owit in cnr__mgi):
        msg = "ufunc '{0}' called with an explicit output that is read-only"
        raise TypingError(msg=msg.format(ufunc.__name__))
    fvv__wdt = [(x.dtype if isinstance(x, types.ArrayCompatible) and not
        isinstance(x, types.Bytes) else x) for x in args]
    kpzsn__sfibc = None
    if moa__bqqtn > 0 and len(cnr__mgi) < ufunc.nout:
        kpzsn__sfibc = 'C'
        hlok__zzuyf = [(x.layout if isinstance(x, types.ArrayCompatible) and
            not isinstance(x, types.Bytes) else '') for x in args]
        if 'C' not in hlok__zzuyf and 'F' in hlok__zzuyf:
            kpzsn__sfibc = 'F'
    return fvv__wdt, cnr__mgi, moa__bqqtn, kpzsn__sfibc


if _check_numba_change:
    lines = inspect.getsource(numba.core.typing.npydecl.Numpy_rules_ufunc.
        _handle_inputs)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '4b97c64ad9c3d50e082538795054f35cf6d2fe962c3ca40e8377a4601b344d5c':
        warnings.warn('Numpy_rules_ufunc._handle_inputs has changed')
numba.core.typing.npydecl.Numpy_rules_ufunc._handle_inputs = (
    _Numpy_Rules_ufunc_handle_inputs)
numba.np.ufunc.dufunc.npydecl.Numpy_rules_ufunc._handle_inputs = (
    _Numpy_Rules_ufunc_handle_inputs)


def DictType__init__(self, keyty, valty, initial_value=None):
    from numba.types import DictType, InitialValue, NoneType, Optional, Tuple, TypeRef, unliteral
    assert not isinstance(keyty, TypeRef)
    assert not isinstance(valty, TypeRef)
    keyty = unliteral(keyty)
    valty = unliteral(valty)
    if isinstance(keyty, (Optional, NoneType)):
        jms__twlx = 'Dict.key_type cannot be of type {}'
        raise TypingError(jms__twlx.format(keyty))
    if isinstance(valty, (Optional, NoneType)):
        jms__twlx = 'Dict.value_type cannot be of type {}'
        raise TypingError(jms__twlx.format(valty))
    self.key_type = keyty
    self.value_type = valty
    self.keyvalue_type = Tuple([keyty, valty])
    name = '{}[{},{}]<iv={}>'.format(self.__class__.__name__, keyty, valty,
        initial_value)
    super(DictType, self).__init__(name)
    InitialValue.__init__(self, initial_value)


if _check_numba_change:
    lines = inspect.getsource(numba.core.types.containers.DictType.__init__)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '475acd71224bd51526750343246e064ff071320c0d10c17b8b8ac81d5070d094':
        warnings.warn('DictType.__init__ has changed')
numba.core.types.containers.DictType.__init__ = DictType__init__


def _legalize_arg_types(self, args):
    for xgedl__blg, a in enumerate(args, start=1):
        if isinstance(a, types.Dispatcher):
            msg = (
                'Does not support function type inputs into with-context for arg {}'
                )
            raise errors.TypingError(msg.format(xgedl__blg))


if _check_numba_change:
    lines = inspect.getsource(numba.core.dispatcher.ObjModeLiftedWith.
        _legalize_arg_types)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '4793f44ebc7da8843e8f298e08cd8a5428b4b84b89fd9d5c650273fdb8fee5ee':
        warnings.warn('ObjModeLiftedWith._legalize_arg_types has changed')
numba.core.dispatcher.ObjModeLiftedWith._legalize_arg_types = (
    _legalize_arg_types)


def _overload_template_get_impl(self, args, kws):
    atzx__tvrml = self.context, tuple(args), tuple(kws.items())
    try:
        nhdw__vpm, args = self._impl_cache[atzx__tvrml]
        return nhdw__vpm, args
    except KeyError as rnigj__cbzwr:
        pass
    nhdw__vpm, args = self._build_impl(atzx__tvrml, args, kws)
    return nhdw__vpm, args


if _check_numba_change:
    lines = inspect.getsource(numba.core.typing.templates.
        _OverloadFunctionTemplate._get_impl)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '4e27d07b214ca16d6e8ed88f70d886b6b095e160d8f77f8df369dd4ed2eb3fae':
        warnings.warn(
            'numba.core.typing.templates._OverloadFunctionTemplate._get_impl has changed'
            )
numba.core.typing.templates._OverloadFunctionTemplate._get_impl = (
    _overload_template_get_impl)


def remove_dead_parfor(parfor, lives, lives_n_aliases, arg_aliases,
    alias_map, func_ir, typemap):
    from numba.core.analysis import compute_cfg_from_blocks, compute_live_map, compute_use_defs
    from numba.core.ir_utils import find_topo_order
    from numba.parfors.parfor import _add_liveness_return_block, _update_parfor_get_setitems, dummy_return_in_loop_body, get_index_var, remove_dead_parfor_recursive, simplify_parfor_body_CFG
    with dummy_return_in_loop_body(parfor.loop_body):
        vtk__lnjoa = find_topo_order(parfor.loop_body)
    xssd__mwd = vtk__lnjoa[0]
    pvl__ezaqt = {}
    _update_parfor_get_setitems(parfor.loop_body[xssd__mwd].body, parfor.
        index_var, alias_map, pvl__ezaqt, lives_n_aliases)
    jyh__yayne = set(pvl__ezaqt.keys())
    for tqdsq__oskhd in vtk__lnjoa:
        if tqdsq__oskhd == xssd__mwd:
            continue
        for stmt in parfor.loop_body[tqdsq__oskhd].body:
            if (isinstance(stmt, ir.Assign) and isinstance(stmt.value, ir.
                Expr) and stmt.value.op == 'getitem' and stmt.value.index.
                name == parfor.index_var.name):
                continue
            wml__mqygs = set(otcze__qfk.name for otcze__qfk in stmt.list_vars()
                )
            thcg__dce = wml__mqygs & jyh__yayne
            for a in thcg__dce:
                pvl__ezaqt.pop(a, None)
    for tqdsq__oskhd in vtk__lnjoa:
        if tqdsq__oskhd == xssd__mwd:
            continue
        block = parfor.loop_body[tqdsq__oskhd]
        tznrv__gol = pvl__ezaqt.copy()
        _update_parfor_get_setitems(block.body, parfor.index_var, alias_map,
            tznrv__gol, lives_n_aliases)
    blocks = parfor.loop_body.copy()
    tsnr__zid = max(blocks.keys())
    sndl__mejin, nngx__usqlq = _add_liveness_return_block(blocks,
        lives_n_aliases, typemap)
    xjxny__mnnnr = ir.Jump(sndl__mejin, ir.Loc('parfors_dummy', -1))
    blocks[tsnr__zid].body.append(xjxny__mnnnr)
    okz__sbg = compute_cfg_from_blocks(blocks)
    uas__xdo = compute_use_defs(blocks)
    swf__wtj = compute_live_map(okz__sbg, blocks, uas__xdo.usemap, uas__xdo
        .defmap)
    alias_set = set(alias_map.keys())
    for wsr__ata, block in blocks.items():
        kpz__hdz = []
        mqh__kylo = {otcze__qfk.name for otcze__qfk in block.terminator.
            list_vars()}
        for nacc__lbx, gnk__xnd in okz__sbg.successors(wsr__ata):
            mqh__kylo |= swf__wtj[nacc__lbx]
        for stmt in reversed(block.body):
            ljsr__cipw = mqh__kylo & alias_set
            for otcze__qfk in ljsr__cipw:
                mqh__kylo |= alias_map[otcze__qfk]
            if (isinstance(stmt, (ir.StaticSetItem, ir.SetItem)) and 
                get_index_var(stmt).name == parfor.index_var.name and stmt.
                target.name not in mqh__kylo and stmt.target.name not in
                arg_aliases):
                continue
            elif isinstance(stmt, ir.Assign) and isinstance(stmt.value, ir.Expr
                ) and stmt.value.op == 'call':
                xuh__movn = guard(find_callname, func_ir, stmt.value)
                if xuh__movn == ('setna', 'bodo.libs.array_kernels'
                    ) and stmt.value.args[0
                    ].name not in mqh__kylo and stmt.value.args[0
                    ].name not in arg_aliases:
                    continue
            mqh__kylo |= {otcze__qfk.name for otcze__qfk in stmt.list_vars()}
            kpz__hdz.append(stmt)
        kpz__hdz.reverse()
        block.body = kpz__hdz
    typemap.pop(nngx__usqlq.name)
    blocks[tsnr__zid].body.pop()

    def trim_empty_parfor_branches(parfor):
        ajd__hno = False
        blocks = parfor.loop_body.copy()
        for wsr__ata, block in blocks.items():
            if len(block.body):
                djd__liml = block.body[-1]
                if isinstance(djd__liml, ir.Branch):
                    if len(blocks[djd__liml.truebr].body) == 1 and len(blocks
                        [djd__liml.falsebr].body) == 1:
                        gcrq__moih = blocks[djd__liml.truebr].body[0]
                        mtryd__tysrn = blocks[djd__liml.falsebr].body[0]
                        if isinstance(gcrq__moih, ir.Jump) and isinstance(
                            mtryd__tysrn, ir.Jump
                            ) and gcrq__moih.target == mtryd__tysrn.target:
                            parfor.loop_body[wsr__ata].body[-1] = ir.Jump(
                                gcrq__moih.target, djd__liml.loc)
                            ajd__hno = True
                    elif len(blocks[djd__liml.truebr].body) == 1:
                        gcrq__moih = blocks[djd__liml.truebr].body[0]
                        if isinstance(gcrq__moih, ir.Jump
                            ) and gcrq__moih.target == djd__liml.falsebr:
                            parfor.loop_body[wsr__ata].body[-1] = ir.Jump(
                                gcrq__moih.target, djd__liml.loc)
                            ajd__hno = True
                    elif len(blocks[djd__liml.falsebr].body) == 1:
                        mtryd__tysrn = blocks[djd__liml.falsebr].body[0]
                        if isinstance(mtryd__tysrn, ir.Jump
                            ) and mtryd__tysrn.target == djd__liml.truebr:
                            parfor.loop_body[wsr__ata].body[-1] = ir.Jump(
                                mtryd__tysrn.target, djd__liml.loc)
                            ajd__hno = True
        return ajd__hno
    ajd__hno = True
    while ajd__hno:
        """
        Process parfor body recursively.
        Note that this is the only place in this function that uses the
        argument lives instead of lives_n_aliases.  The former does not
        include the aliases of live variables but only the live variable
        names themselves.  See a comment in this function for how that
        is used.
        """
        remove_dead_parfor_recursive(parfor, lives, arg_aliases, alias_map,
            func_ir, typemap)
        simplify_parfor_body_CFG(func_ir.blocks)
        ajd__hno = trim_empty_parfor_branches(parfor)
    uunn__kjuq = len(parfor.init_block.body) == 0
    for block in parfor.loop_body.values():
        uunn__kjuq &= len(block.body) == 0
    if uunn__kjuq:
        return None
    return parfor


if _check_numba_change:
    lines = inspect.getsource(numba.parfors.parfor.remove_dead_parfor)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '1c9b008a7ead13e988e1efe67618d8f87f0b9f3d092cc2cd6bfcd806b1fdb859':
        warnings.warn('remove_dead_parfor has changed')
numba.parfors.parfor.remove_dead_parfor = remove_dead_parfor
numba.core.ir_utils.remove_dead_extensions[numba.parfors.parfor.Parfor
    ] = remove_dead_parfor


def simplify_parfor_body_CFG(blocks):
    from numba.core.analysis import compute_cfg_from_blocks
    from numba.core.ir_utils import simplify_CFG
    from numba.parfors.parfor import Parfor
    wzwfy__ljrz = 0
    for block in blocks.values():
        for stmt in block.body:
            if isinstance(stmt, Parfor):
                wzwfy__ljrz += 1
                parfor = stmt
                uloe__mmn = parfor.loop_body[max(parfor.loop_body.keys())]
                scope = uloe__mmn.scope
                loc = ir.Loc('parfors_dummy', -1)
                yzk__hlyd = ir.Var(scope, mk_unique_var('$const'), loc)
                uloe__mmn.body.append(ir.Assign(ir.Const(0, loc), yzk__hlyd,
                    loc))
                uloe__mmn.body.append(ir.Return(yzk__hlyd, loc))
                okz__sbg = compute_cfg_from_blocks(parfor.loop_body)
                for qrk__coqv in okz__sbg.dead_nodes():
                    del parfor.loop_body[qrk__coqv]
                parfor.loop_body = simplify_CFG(parfor.loop_body)
                uloe__mmn = parfor.loop_body[max(parfor.loop_body.keys())]
                uloe__mmn.body.pop()
                uloe__mmn.body.pop()
                simplify_parfor_body_CFG(parfor.loop_body)
    return wzwfy__ljrz


if _check_numba_change:
    lines = inspect.getsource(numba.parfors.parfor.simplify_parfor_body_CFG)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '437ae96a5e8ec64a2b69a4f23ba8402a1d170262a5400aa0aa7bfe59e03bf726':
        warnings.warn('simplify_parfor_body_CFG has changed')
numba.parfors.parfor.simplify_parfor_body_CFG = simplify_parfor_body_CFG


def _lifted_compile(self, sig):
    import numba.core.event as ev
    from numba.core import compiler, sigutils
    from numba.core.compiler_lock import global_compiler_lock
    from numba.core.ir_utils import remove_dels
    with ExitStack() as scope:
        cres = None

        def cb_compiler(dur):
            if cres is not None:
                self._callback_add_compiler_timer(dur, cres)

        def cb_llvm(dur):
            if cres is not None:
                self._callback_add_llvm_timer(dur, cres)
        scope.enter_context(ev.install_timer('numba:compiler_lock',
            cb_compiler))
        scope.enter_context(ev.install_timer('numba:llvm_lock', cb_llvm))
        scope.enter_context(global_compiler_lock)
        with self._compiling_counter:
            flags = self.flags
            args, return_type = sigutils.normalize_signature(sig)
            vguxm__kudmb = self.overloads.get(tuple(args))
            if vguxm__kudmb is not None:
                return vguxm__kudmb.entry_point
            self._pre_compile(args, return_type, flags)
            vvtnp__bgn = self.func_ir
            gqx__hsv = dict(dispatcher=self, args=args, return_type=return_type
                )
            with ev.trigger_event('numba:compile', data=gqx__hsv):
                cres = compiler.compile_ir(typingctx=self.typingctx,
                    targetctx=self.targetctx, func_ir=vvtnp__bgn, args=args,
                    return_type=return_type, flags=flags, locals=self.
                    locals, lifted=(), lifted_from=self.lifted_from,
                    is_lifted_loop=True)
                if cres.typing_error is not None and not flags.enable_pyobject:
                    raise cres.typing_error
                self.add_overload(cres)
            remove_dels(self.func_ir.blocks)
            return cres.entry_point


if _check_numba_change:
    lines = inspect.getsource(numba.core.dispatcher.LiftedCode.compile)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '1351ebc5d8812dc8da167b30dad30eafb2ca9bf191b49aaed6241c21e03afff1':
        warnings.warn('numba.core.dispatcher.LiftedCode.compile has changed')
numba.core.dispatcher.LiftedCode.compile = _lifted_compile


def compile_ir(typingctx, targetctx, func_ir, args, return_type, flags,
    locals, lifted=(), lifted_from=None, is_lifted_loop=False, library=None,
    pipeline_class=Compiler):
    if is_lifted_loop:
        sexa__olmfw = copy.deepcopy(flags)
        sexa__olmfw.no_rewrites = True

        def compile_local(the_ir, the_flags):
            azl__cql = pipeline_class(typingctx, targetctx, library, args,
                return_type, the_flags, locals)
            return azl__cql.compile_ir(func_ir=the_ir, lifted=lifted,
                lifted_from=lifted_from)
        qlz__suie = compile_local(func_ir, sexa__olmfw)
        kae__cedi = None
        if not flags.no_rewrites:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', errors.NumbaWarning)
                try:
                    kae__cedi = compile_local(func_ir, flags)
                except Exception as rnigj__cbzwr:
                    pass
        if kae__cedi is not None:
            cres = kae__cedi
        else:
            cres = qlz__suie
        return cres
    else:
        azl__cql = pipeline_class(typingctx, targetctx, library, args,
            return_type, flags, locals)
        return azl__cql.compile_ir(func_ir=func_ir, lifted=lifted,
            lifted_from=lifted_from)


if _check_numba_change:
    lines = inspect.getsource(numba.core.compiler.compile_ir)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'c48ce5493f4c43326e8cbdd46f3ea038b2b9045352d9d25894244798388e5e5b':
        warnings.warn('numba.core.compiler.compile_ir has changed')
numba.core.compiler.compile_ir = compile_ir


def make_constant_array(self, builder, typ, ary):
    import math
    from llvmlite import ir as lir
    from llvmlite.llvmpy.core import Constant, Type
    wxda__uzf = self.get_data_type(typ.dtype)
    zqxt__pzl = 10 ** 7
    if self.allow_dynamic_globals and (typ.layout not in 'FC' or ary.nbytes >
        zqxt__pzl):
        uxq__ldt = ary.ctypes.data
        zvm__cje = self.add_dynamic_addr(builder, uxq__ldt, info=str(type(
            uxq__ldt)))
        ock__cjuvz = self.add_dynamic_addr(builder, id(ary), info=str(type(
            ary)))
        self.global_arrays.append(ary)
    else:
        lglxd__gykh = ary.flatten(order=typ.layout)
        if isinstance(typ.dtype, (types.NPDatetime, types.NPTimedelta)):
            lglxd__gykh = lglxd__gykh.view('int64')
        wzh__ychqv = Constant.array(Type.int(8), bytearray(lglxd__gykh.data))
        zvm__cje = cgutils.global_constant(builder, '.const.array.data',
            wzh__ychqv)
        zvm__cje.align = self.get_abi_alignment(wxda__uzf)
        ock__cjuvz = None
    pnv__kawah = self.get_value_type(types.intp)
    xqy__zfubn = [self.get_constant(types.intp, yjv__jaehb) for yjv__jaehb in
        ary.shape]
    saju__qnu = Constant.array(pnv__kawah, xqy__zfubn)
    enf__byl = [self.get_constant(types.intp, yjv__jaehb) for yjv__jaehb in
        ary.strides]
    szh__hyl = Constant.array(pnv__kawah, enf__byl)
    dma__ftjx = self.get_constant(types.intp, ary.dtype.itemsize)
    nlxfh__ifnxw = self.get_constant(types.intp, math.prod(ary.shape))
    return lir.Constant.literal_struct([self.get_constant_null(types.
        MemInfoPointer(typ.dtype)), self.get_constant_null(types.pyobject),
        nlxfh__ifnxw, dma__ftjx, zvm__cje.bitcast(self.get_value_type(types
        .CPointer(typ.dtype))), saju__qnu, szh__hyl])


if _check_numba_change:
    lines = inspect.getsource(numba.core.base.BaseContext.make_constant_array)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '5721b5360b51f782f79bd794f7bf4d48657911ecdc05c30db22fd55f15dad821':
        warnings.warn(
            'numba.core.base.BaseContext.make_constant_array has changed')
numba.core.base.BaseContext.make_constant_array = make_constant_array


def _define_atomic_inc_dec(module, op, ordering):
    from llvmlite import ir as lir
    from numba.core.runtime.nrtdynmod import _word_type
    xgn__jigba = lir.FunctionType(_word_type, [_word_type.as_pointer()])
    grfwb__gjrv = lir.Function(module, xgn__jigba, name='nrt_atomic_{0}'.
        format(op))
    [dqhe__fusro] = grfwb__gjrv.args
    pyf__ygete = grfwb__gjrv.append_basic_block()
    builder = lir.IRBuilder(pyf__ygete)
    eaqe__fckn = lir.Constant(_word_type, 1)
    if False:
        sjfyt__uzu = builder.atomic_rmw(op, dqhe__fusro, eaqe__fckn,
            ordering=ordering)
        res = getattr(builder, op)(sjfyt__uzu, eaqe__fckn)
        builder.ret(res)
    else:
        sjfyt__uzu = builder.load(dqhe__fusro)
        zapqp__dtqkp = getattr(builder, op)(sjfyt__uzu, eaqe__fckn)
        hxyhw__vfvc = builder.icmp_signed('!=', sjfyt__uzu, lir.Constant(
            sjfyt__uzu.type, -1))
        with cgutils.if_likely(builder, hxyhw__vfvc):
            builder.store(zapqp__dtqkp, dqhe__fusro)
        builder.ret(zapqp__dtqkp)
    return grfwb__gjrv


if _check_numba_change:
    lines = inspect.getsource(numba.core.runtime.nrtdynmod.
        _define_atomic_inc_dec)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '9cc02c532b2980b6537b702f5608ea603a1ff93c6d3c785ae2cf48bace273f48':
        warnings.warn(
            'numba.core.runtime.nrtdynmod._define_atomic_inc_dec has changed')
numba.core.runtime.nrtdynmod._define_atomic_inc_dec = _define_atomic_inc_dec


def NativeLowering_run_pass(self, state):
    from llvmlite import binding as llvm
    from numba.core import funcdesc, lowering
    from numba.core.typed_passes import fallback_context
    if state.library is None:
        kygz__ttxs = state.targetctx.codegen()
        state.library = kygz__ttxs.create_library(state.func_id.func_qualname)
        state.library.enable_object_caching()
    library = state.library
    targetctx = state.targetctx
    aurw__dux = state.func_ir
    typemap = state.typemap
    absw__yscgu = state.return_type
    calltypes = state.calltypes
    flags = state.flags
    onlo__cnji = state.metadata
    ndncr__qdbj = llvm.passmanagers.dump_refprune_stats()
    msg = 'Function %s failed at nopython mode lowering' % (state.func_id.
        func_name,)
    with fallback_context(state, msg):
        bglw__nvty = (funcdesc.PythonFunctionDescriptor.
            from_specialized_function(aurw__dux, typemap, absw__yscgu,
            calltypes, mangler=targetctx.mangler, inline=flags.forceinline,
            noalias=flags.noalias, abi_tags=[flags.get_mangle_string()]))
        targetctx.global_arrays = []
        with targetctx.push_code_library(library):
            fivwx__nka = lowering.Lower(targetctx, library, bglw__nvty,
                aurw__dux, metadata=onlo__cnji)
            fivwx__nka.lower()
            if not flags.no_cpython_wrapper:
                fivwx__nka.create_cpython_wrapper(flags.release_gil)
            if not flags.no_cfunc_wrapper:
                for t in state.args:
                    if isinstance(t, (types.Omitted, types.Generator)):
                        break
                else:
                    if isinstance(absw__yscgu, (types.Optional, types.
                        Generator)):
                        pass
                    else:
                        fivwx__nka.create_cfunc_wrapper()
            vvnla__ketjd = fivwx__nka.env
            uyjzw__szmoo = fivwx__nka.call_helper
            del fivwx__nka
        from numba.core.compiler import _LowerResult
        if flags.no_compile:
            state['cr'] = _LowerResult(bglw__nvty, uyjzw__szmoo, cfunc=None,
                env=vvnla__ketjd)
        else:
            qixts__xbjqu = targetctx.get_executable(library, bglw__nvty,
                vvnla__ketjd)
            targetctx.insert_user_function(qixts__xbjqu, bglw__nvty, [library])
            state['cr'] = _LowerResult(bglw__nvty, uyjzw__szmoo, cfunc=
                qixts__xbjqu, env=vvnla__ketjd)
        onlo__cnji['global_arrs'] = targetctx.global_arrays
        targetctx.global_arrays = []
        wmqgj__tgb = llvm.passmanagers.dump_refprune_stats()
        onlo__cnji['prune_stats'] = wmqgj__tgb - ndncr__qdbj
        onlo__cnji['llvm_pass_timings'] = library.recorded_timings
    return True


if _check_numba_change:
    lines = inspect.getsource(numba.core.typed_passes.NativeLowering.run_pass)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'a777ce6ce1bb2b1cbaa3ac6c2c0e2adab69a9c23888dff5f1cbb67bfb176b5de':
        warnings.warn(
            'numba.core.typed_passes.NativeLowering.run_pass has changed')
numba.core.typed_passes.NativeLowering.run_pass = NativeLowering_run_pass


def _python_list_to_native(typ, obj, c, size, listptr, errorptr):
    from llvmlite import ir as lir
    from numba.core.boxing import _NumbaTypeHelper
    from numba.cpython import listobj

    def check_element_type(nth, itemobj, expected_typobj):
        mzhy__yfsoe = nth.typeof(itemobj)
        with c.builder.if_then(cgutils.is_null(c.builder, mzhy__yfsoe),
            likely=False):
            c.builder.store(cgutils.true_bit, errorptr)
            loop.do_break()
        heste__rcb = c.builder.icmp_signed('!=', mzhy__yfsoe, expected_typobj)
        if not isinstance(typ.dtype, types.Optional):
            with c.builder.if_then(heste__rcb, likely=False):
                c.builder.store(cgutils.true_bit, errorptr)
                c.pyapi.err_format('PyExc_TypeError',
                    "can't unbox heterogeneous list: %S != %S",
                    expected_typobj, mzhy__yfsoe)
                c.pyapi.decref(mzhy__yfsoe)
                loop.do_break()
        c.pyapi.decref(mzhy__yfsoe)
    jbgxa__lmis, list = listobj.ListInstance.allocate_ex(c.context, c.
        builder, typ, size)
    with c.builder.if_else(jbgxa__lmis, likely=True) as (if_ok, if_not_ok):
        with if_ok:
            list.size = size
            hrvpj__qii = lir.Constant(size.type, 0)
            with c.builder.if_then(c.builder.icmp_signed('>', size,
                hrvpj__qii), likely=True):
                with _NumbaTypeHelper(c) as nth:
                    expected_typobj = nth.typeof(c.pyapi.list_getitem(obj,
                        hrvpj__qii))
                    with cgutils.for_range(c.builder, size) as loop:
                        itemobj = c.pyapi.list_getitem(obj, loop.index)
                        check_element_type(nth, itemobj, expected_typobj)
                        lhz__rxhxo = c.unbox(typ.dtype, itemobj)
                        with c.builder.if_then(lhz__rxhxo.is_error, likely=
                            False):
                            c.builder.store(cgutils.true_bit, errorptr)
                            loop.do_break()
                        list.setitem(loop.index, lhz__rxhxo.value, incref=False
                            )
                    c.pyapi.decref(expected_typobj)
            if typ.reflected:
                list.parent = obj
            with c.builder.if_then(c.builder.not_(c.builder.load(errorptr)),
                likely=False):
                c.pyapi.object_set_private_data(obj, list.meminfo)
            list.set_dirty(False)
            c.builder.store(list.value, listptr)
        with if_not_ok:
            c.builder.store(cgutils.true_bit, errorptr)
    with c.builder.if_then(c.builder.load(errorptr)):
        c.context.nrt.decref(c.builder, typ, list.value)


if _check_numba_change:
    lines = inspect.getsource(numba.core.boxing._python_list_to_native)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'f8e546df8b07adfe74a16b6aafb1d4fddbae7d3516d7944b3247cc7c9b7ea88a':
        warnings.warn('numba.core.boxing._python_list_to_native has changed')
numba.core.boxing._python_list_to_native = _python_list_to_native


def make_string_from_constant(context, builder, typ, literal_string):
    from llvmlite import ir as lir
    from numba.cpython.hashing import _Py_hash_t
    from numba.cpython.unicode import compile_time_get_string_data
    sxycj__dshl, ivb__wqhsg, kmced__jllzq, hwomu__zwhgu, ptfy__qqqof = (
        compile_time_get_string_data(literal_string))
    nemru__llv = builder.module
    gv = context.insert_const_bytes(nemru__llv, sxycj__dshl)
    return lir.Constant.literal_struct([gv, context.get_constant(types.intp,
        ivb__wqhsg), context.get_constant(types.int32, kmced__jllzq),
        context.get_constant(types.uint32, hwomu__zwhgu), context.
        get_constant(_Py_hash_t, -1), context.get_constant_null(types.
        MemInfoPointer(types.voidptr)), context.get_constant_null(types.
        pyobject)])


if _check_numba_change:
    lines = inspect.getsource(numba.cpython.unicode.make_string_from_constant)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '525bd507383e06152763e2f046dae246cd60aba027184f50ef0fc9a80d4cd7fa':
        warnings.warn(
            'numba.cpython.unicode.make_string_from_constant has changed')
numba.cpython.unicode.make_string_from_constant = make_string_from_constant


def parse_shape(shape):
    yrad__khu = None
    if isinstance(shape, types.Integer):
        yrad__khu = 1
    elif isinstance(shape, (types.Tuple, types.UniTuple)):
        if all(isinstance(yjv__jaehb, (types.Integer, types.IntEnumMember)) for
            yjv__jaehb in shape):
            yrad__khu = len(shape)
    return yrad__khu


if _check_numba_change:
    lines = inspect.getsource(numba.core.typing.npydecl.parse_shape)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'e62e3ff09d36df5ac9374055947d6a8be27160ce32960d3ef6cb67f89bd16429':
        warnings.warn('numba.core.typing.npydecl.parse_shape has changed')
numba.core.typing.npydecl.parse_shape = parse_shape


def _get_names(self, obj):
    if isinstance(obj, ir.Var) or isinstance(obj, str):
        name = obj if isinstance(obj, str) else obj.name
        if name not in self.typemap:
            return name,
        typ = self.typemap[name]
        if isinstance(typ, (types.BaseTuple, types.ArrayCompatible)):
            yrad__khu = typ.ndim if isinstance(typ, types.ArrayCompatible
                ) else len(typ)
            if yrad__khu == 0:
                return name,
            else:
                return tuple('{}#{}'.format(name, xgedl__blg) for
                    xgedl__blg in range(yrad__khu))
        else:
            return name,
    elif isinstance(obj, ir.Const):
        if isinstance(obj.value, tuple):
            return obj.value
        else:
            return obj.value,
    elif isinstance(obj, tuple):

        def get_names(x):
            kpw__inuqv = self._get_names(x)
            if len(kpw__inuqv) != 0:
                return kpw__inuqv[0]
            return kpw__inuqv
        return tuple(get_names(x) for x in obj)
    elif isinstance(obj, int):
        return obj,
    return ()


def get_equiv_const(self, obj):
    kpw__inuqv = self._get_names(obj)
    if len(kpw__inuqv) != 1:
        return None
    return super(numba.parfors.array_analysis.ShapeEquivSet, self
        ).get_equiv_const(kpw__inuqv[0])


def get_equiv_set(self, obj):
    kpw__inuqv = self._get_names(obj)
    if len(kpw__inuqv) != 1:
        return None
    return super(numba.parfors.array_analysis.ShapeEquivSet, self
        ).get_equiv_set(kpw__inuqv[0])


if _check_numba_change:
    for name, orig, new, hash in ((
        'numba.parfors.array_analysis.ShapeEquivSet._get_names', numba.
        parfors.array_analysis.ShapeEquivSet._get_names, _get_names,
        '8c9bf136109028d5445fd0a82387b6abeb70c23b20b41e2b50c34ba5359516ee'),
        ('numba.parfors.array_analysis.ShapeEquivSet.get_equiv_const',
        numba.parfors.array_analysis.ShapeEquivSet.get_equiv_const,
        get_equiv_const,
        'bef410ca31a9e29df9ee74a4a27d339cc332564e4a237828b8a4decf625ce44e'),
        ('numba.parfors.array_analysis.ShapeEquivSet.get_equiv_set', numba.
        parfors.array_analysis.ShapeEquivSet.get_equiv_set, get_equiv_set,
        'ec936d340c488461122eb74f28a28b88227cb1f1bca2b9ba3c19258cfe1eb40a')):
        lines = inspect.getsource(orig)
        if hashlib.sha256(lines.encode()).hexdigest() != hash:
            warnings.warn(f'{name} has changed')
numba.parfors.array_analysis.ShapeEquivSet._get_names = _get_names
numba.parfors.array_analysis.ShapeEquivSet.get_equiv_const = get_equiv_const
numba.parfors.array_analysis.ShapeEquivSet.get_equiv_set = get_equiv_set


def raise_on_unsupported_feature(func_ir, typemap):
    import numpy
    ejx__ppov = []
    for wxkpn__arh in func_ir.arg_names:
        if wxkpn__arh in typemap and isinstance(typemap[wxkpn__arh], types.
            containers.UniTuple) and typemap[wxkpn__arh].count > 1000:
            msg = (
                """Tuple '{}' length must be smaller than 1000.
Large tuples lead to the generation of a prohibitively large LLVM IR which causes excessive memory pressure and large compile times.
As an alternative, the use of a 'list' is recommended in place of a 'tuple' as lists do not suffer from this problem."""
                .format(wxkpn__arh))
            raise errors.UnsupportedError(msg, func_ir.loc)
    for gmke__vcft in func_ir.blocks.values():
        for stmt in gmke__vcft.find_insts(ir.Assign):
            if isinstance(stmt.value, ir.Expr):
                if stmt.value.op == 'make_function':
                    val = stmt.value
                    fapp__pnpm = getattr(val, 'code', None)
                    if fapp__pnpm is not None:
                        if getattr(val, 'closure', None) is not None:
                            yatsq__vft = '<creating a function from a closure>'
                            betea__ovk = ''
                        else:
                            yatsq__vft = fapp__pnpm.co_name
                            betea__ovk = '(%s) ' % yatsq__vft
                    else:
                        yatsq__vft = '<could not ascertain use case>'
                        betea__ovk = ''
                    msg = (
                        'Numba encountered the use of a language feature it does not support in this context: %s (op code: make_function not supported). If the feature is explicitly supported it is likely that the result of the expression %sis being used in an unsupported manner.'
                         % (yatsq__vft, betea__ovk))
                    raise errors.UnsupportedError(msg, stmt.value.loc)
            if isinstance(stmt.value, (ir.Global, ir.FreeVar)):
                val = stmt.value
                val = getattr(val, 'value', None)
                if val is None:
                    continue
                elep__jdkmx = False
                if isinstance(val, pytypes.FunctionType):
                    elep__jdkmx = val in {numba.gdb, numba.gdb_init}
                if not elep__jdkmx:
                    elep__jdkmx = getattr(val, '_name', '') == 'gdb_internal'
                if elep__jdkmx:
                    ejx__ppov.append(stmt.loc)
            if isinstance(stmt.value, ir.Expr):
                if stmt.value.op == 'getattr' and stmt.value.attr == 'view':
                    var = stmt.value.value.name
                    if isinstance(typemap[var], types.Array):
                        continue
                    rcj__khw = func_ir.get_definition(var)
                    pwq__fgnp = guard(find_callname, func_ir, rcj__khw)
                    if pwq__fgnp and pwq__fgnp[1] == 'numpy':
                        ty = getattr(numpy, pwq__fgnp[0])
                        if numpy.issubdtype(ty, numpy.integer
                            ) or numpy.issubdtype(ty, numpy.floating):
                            continue
                    llze__jrf = '' if var.startswith('$') else "'{}' ".format(
                        var)
                    raise TypingError(
                        "'view' can only be called on NumPy dtypes, try wrapping the variable {}with 'np.<dtype>()'"
                        .format(llze__jrf), loc=stmt.loc)
            if isinstance(stmt.value, ir.Global):
                ty = typemap[stmt.target.name]
                msg = (
                    "The use of a %s type, assigned to variable '%s' in globals, is not supported as globals are considered compile-time constants and there is no known way to compile a %s type as a constant."
                    )
                if isinstance(ty, types.ListType):
                    raise TypingError(msg % (ty, stmt.value.name, ty), loc=
                        stmt.loc)
            if isinstance(stmt.value, ir.Yield) and not func_ir.is_generator:
                msg = 'The use of generator expressions is unsupported.'
                raise errors.UnsupportedError(msg, loc=stmt.loc)
    if len(ejx__ppov) > 1:
        msg = """Calling either numba.gdb() or numba.gdb_init() more than once in a function is unsupported (strange things happen!), use numba.gdb_breakpoint() to create additional breakpoints instead.

Relevant documentation is available here:
https://numba.pydata.org/numba-doc/latest/user/troubleshoot.html/troubleshoot.html#using-numba-s-direct-gdb-bindings-in-nopython-mode

Conflicting calls found at:
 %s"""
        ibkq__hpmmx = '\n'.join([x.strformat() for x in ejx__ppov])
        raise errors.UnsupportedError(msg % ibkq__hpmmx)


if _check_numba_change:
    lines = inspect.getsource(numba.core.ir_utils.raise_on_unsupported_feature)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '237a4fe8395a40899279c718bc3754102cd2577463ef2f48daceea78d79b2d5e':
        warnings.warn(
            'numba.core.ir_utils.raise_on_unsupported_feature has changed')
numba.core.ir_utils.raise_on_unsupported_feature = raise_on_unsupported_feature
numba.core.typed_passes.raise_on_unsupported_feature = (
    raise_on_unsupported_feature)


@typeof_impl.register(dict)
def _typeof_dict(val, c):
    if len(val) == 0:
        raise ValueError('Cannot type empty dict')
    bfxes__thqc, otcze__qfk = next(iter(val.items()))
    nza__kkfg = typeof_impl(bfxes__thqc, c)
    hxtyp__xyg = typeof_impl(otcze__qfk, c)
    if nza__kkfg is None or hxtyp__xyg is None:
        raise ValueError(
            f'Cannot type dict element type {type(bfxes__thqc)}, {type(otcze__qfk)}'
            )
    return types.DictType(nza__kkfg, hxtyp__xyg)


def unbox_dicttype(typ, val, c):
    from llvmlite import ir as lir
    from numba.typed import dictobject
    from numba.typed.typeddict import Dict
    context = c.context
    eeq__veo = cgutils.alloca_once_value(c.builder, val)
    chs__ekrus = c.pyapi.object_hasattr_string(val, '_opaque')
    yryn__suage = c.builder.icmp_unsigned('==', chs__ekrus, lir.Constant(
        chs__ekrus.type, 0))
    embxy__omzs = typ.key_type
    jlgu__jyp = typ.value_type

    def make_dict():
        return numba.typed.Dict.empty(embxy__omzs, jlgu__jyp)

    def copy_dict(out_dict, in_dict):
        for bfxes__thqc, otcze__qfk in in_dict.items():
            out_dict[bfxes__thqc] = otcze__qfk
    with c.builder.if_then(yryn__suage):
        spw__qizty = c.pyapi.unserialize(c.pyapi.serialize_object(make_dict))
        oyaua__rfbc = c.pyapi.call_function_objargs(spw__qizty, [])
        uciyp__qqek = c.pyapi.unserialize(c.pyapi.serialize_object(copy_dict))
        c.pyapi.call_function_objargs(uciyp__qqek, [oyaua__rfbc, val])
        c.builder.store(oyaua__rfbc, eeq__veo)
    val = c.builder.load(eeq__veo)
    jem__kwf = c.pyapi.unserialize(c.pyapi.serialize_object(Dict))
    jkgnq__zjoeu = c.pyapi.object_type(val)
    vjho__aox = c.builder.icmp_unsigned('==', jkgnq__zjoeu, jem__kwf)
    with c.builder.if_else(vjho__aox) as (then, orelse):
        with then:
            rvmhx__obaf = c.pyapi.object_getattr_string(val, '_opaque')
            wol__mtv = types.MemInfoPointer(types.voidptr)
            lhz__rxhxo = c.unbox(wol__mtv, rvmhx__obaf)
            mi = lhz__rxhxo.value
            xxgzj__asf = wol__mtv, typeof(typ)

            def convert(mi, typ):
                return dictobject._from_meminfo(mi, typ)
            sig = signature(typ, *xxgzj__asf)
            mtbqr__yoaby = context.get_constant_null(xxgzj__asf[1])
            args = mi, mtbqr__yoaby
            zvulb__yte, hslu__yxapx = c.pyapi.call_jit_code(convert, sig, args)
            c.context.nrt.decref(c.builder, typ, hslu__yxapx)
            c.pyapi.decref(rvmhx__obaf)
            osy__uev = c.builder.basic_block
        with orelse:
            c.pyapi.err_format('PyExc_TypeError',
                "can't unbox a %S as a %S", jkgnq__zjoeu, jem__kwf)
            lpywx__pdlbd = c.builder.basic_block
    bkrpi__vdf = c.builder.phi(hslu__yxapx.type)
    qoyf__mzytr = c.builder.phi(zvulb__yte.type)
    bkrpi__vdf.add_incoming(hslu__yxapx, osy__uev)
    bkrpi__vdf.add_incoming(hslu__yxapx.type(None), lpywx__pdlbd)
    qoyf__mzytr.add_incoming(zvulb__yte, osy__uev)
    qoyf__mzytr.add_incoming(cgutils.true_bit, lpywx__pdlbd)
    c.pyapi.decref(jem__kwf)
    c.pyapi.decref(jkgnq__zjoeu)
    with c.builder.if_then(yryn__suage):
        c.pyapi.decref(val)
    return NativeValue(bkrpi__vdf, is_error=qoyf__mzytr)


import numba.typed.typeddict
if _check_numba_change:
    lines = inspect.getsource(numba.core.pythonapi._unboxers.functions[
        numba.core.types.DictType])
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '5f6f183b94dc57838538c668a54c2476576c85d8553843f3219f5162c61e7816':
        warnings.warn('unbox_dicttype has changed')
numba.core.pythonapi._unboxers.functions[types.DictType] = unbox_dicttype


def mul_list_generic(self, args, kws):
    a, hury__twm = args
    if isinstance(a, types.List) and isinstance(hury__twm, types.Integer):
        return signature(a, a, types.intp)
    elif isinstance(a, types.Integer) and isinstance(hury__twm, types.List):
        return signature(hury__twm, types.intp, hury__twm)


if _check_numba_change:
    lines = inspect.getsource(numba.core.typing.listdecl.MulList.generic)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '95882385a8ffa67aa576e8169b9ee6b3197e0ad3d5def4b47fa65ce8cd0f1575':
        warnings.warn('numba.core.typing.listdecl.MulList.generic has changed')
numba.core.typing.listdecl.MulList.generic = mul_list_generic


@lower_builtin(operator.mul, types.Integer, types.List)
def list_mul(context, builder, sig, args):
    from llvmlite import ir as lir
    from numba.core.imputils import impl_ret_new_ref
    from numba.cpython.listobj import ListInstance
    if isinstance(sig.args[0], types.List):
        vzfwu__tasps, oqrdj__qrsqj = 0, 1
    else:
        vzfwu__tasps, oqrdj__qrsqj = 1, 0
    dgtwh__bkxqn = ListInstance(context, builder, sig.args[vzfwu__tasps],
        args[vzfwu__tasps])
    lvxm__yfb = dgtwh__bkxqn.size
    jwzmr__okwty = args[oqrdj__qrsqj]
    hrvpj__qii = lir.Constant(jwzmr__okwty.type, 0)
    jwzmr__okwty = builder.select(cgutils.is_neg_int(builder, jwzmr__okwty),
        hrvpj__qii, jwzmr__okwty)
    nlxfh__ifnxw = builder.mul(jwzmr__okwty, lvxm__yfb)
    cel__pqhue = ListInstance.allocate(context, builder, sig.return_type,
        nlxfh__ifnxw)
    cel__pqhue.size = nlxfh__ifnxw
    with cgutils.for_range_slice(builder, hrvpj__qii, nlxfh__ifnxw,
        lvxm__yfb, inc=True) as (dest_offset, _):
        with cgutils.for_range(builder, lvxm__yfb) as loop:
            value = dgtwh__bkxqn.getitem(loop.index)
            cel__pqhue.setitem(builder.add(loop.index, dest_offset), value,
                incref=True)
    return impl_ret_new_ref(context, builder, sig.return_type, cel__pqhue.value
        )
