import hashlib
import inspect
import warnings
import snowflake.sqlalchemy
import sqlalchemy.types as sqltypes
from sqlalchemy import exc as sa_exc
from sqlalchemy import util as sa_util
from sqlalchemy.sql import text
_check_snowflake_sqlalchemy_change = True


def _get_schema_columns(self, connection, schema, **kw):
    bxta__puqnv = {}
    nlgl__xwha, filj__oboh = self._current_database_schema(connection, **kw)
    dorm__yis = self._denormalize_quote_join(nlgl__xwha, schema)
    try:
        qhzh__xdpq = self._get_schema_primary_keys(connection, dorm__yis, **kw)
        tvoxf__khcot = connection.execute(text(
            """
        SELECT /* sqlalchemy:_get_schema_columns */
                ic.table_name,
                ic.column_name,
                ic.data_type,
                ic.character_maximum_length,
                ic.numeric_precision,
                ic.numeric_scale,
                ic.is_nullable,
                ic.column_default,
                ic.is_identity,
                ic.comment
            FROM information_schema.columns ic
            WHERE ic.table_schema=:table_schema
            ORDER BY ic.ordinal_position"""
            ), {'table_schema': self.denormalize_name(schema)})
    except sa_exc.ProgrammingError as jko__oaffk:
        if jko__oaffk.orig.errno == 90030:
            return None
        raise
    for table_name, bdah__eps, txp__kpssv, hpdi__eycj, jbjji__focoo, vvrig__mxkff, rqlj__ftif, luj__jycpg, azaoc__ateuk, bssv__letr in tvoxf__khcot:
        table_name = self.normalize_name(table_name)
        bdah__eps = self.normalize_name(bdah__eps)
        if table_name not in bxta__puqnv:
            bxta__puqnv[table_name] = list()
        if bdah__eps.startswith('sys_clustering_column'):
            continue
        oti__xgqv = self.ischema_names.get(txp__kpssv, None)
        ieq__tfq = {}
        if oti__xgqv is None:
            sa_util.warn("Did not recognize type '{}' of column '{}'".
                format(txp__kpssv, bdah__eps))
            oti__xgqv = sqltypes.NULLTYPE
        elif issubclass(oti__xgqv, sqltypes.FLOAT):
            ieq__tfq['precision'] = jbjji__focoo
            ieq__tfq['decimal_return_scale'] = vvrig__mxkff
        elif issubclass(oti__xgqv, sqltypes.Numeric):
            ieq__tfq['precision'] = jbjji__focoo
            ieq__tfq['scale'] = vvrig__mxkff
        elif issubclass(oti__xgqv, (sqltypes.String, sqltypes.BINARY)):
            ieq__tfq['length'] = hpdi__eycj
        wgsm__jaihv = oti__xgqv if isinstance(oti__xgqv, sqltypes.NullType
            ) else oti__xgqv(**ieq__tfq)
        amfhp__yjhgc = qhzh__xdpq.get(table_name)
        bxta__puqnv[table_name].append({'name': bdah__eps, 'type':
            wgsm__jaihv, 'nullable': rqlj__ftif == 'YES', 'default':
            luj__jycpg, 'autoincrement': azaoc__ateuk == 'YES', 'comment':
            bssv__letr, 'primary_key': bdah__eps in qhzh__xdpq[table_name][
            'constrained_columns'] if amfhp__yjhgc else False})
    return bxta__puqnv


if _check_snowflake_sqlalchemy_change:
    lines = inspect.getsource(snowflake.sqlalchemy.snowdialect.
        SnowflakeDialect._get_schema_columns)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != 'fdf39af1ac165319d3b6074e8cf9296a090a21f0e2c05b644ff8ec0e56e2d769':
        warnings.warn(
            'snowflake.sqlalchemy.snowdialect.SnowflakeDialect._get_schema_columns has changed'
            )
snowflake.sqlalchemy.snowdialect.SnowflakeDialect._get_schema_columns = (
    _get_schema_columns)


def _get_table_columns(self, connection, table_name, schema=None, **kw):
    bxta__puqnv = []
    nlgl__xwha, filj__oboh = self._current_database_schema(connection, **kw)
    dorm__yis = self._denormalize_quote_join(nlgl__xwha, schema)
    qhzh__xdpq = self._get_schema_primary_keys(connection, dorm__yis, **kw)
    tvoxf__khcot = connection.execute(text(
        """
    SELECT /* sqlalchemy:get_table_columns */
            ic.table_name,
            ic.column_name,
            ic.data_type,
            ic.character_maximum_length,
            ic.numeric_precision,
            ic.numeric_scale,
            ic.is_nullable,
            ic.column_default,
            ic.is_identity,
            ic.comment
        FROM information_schema.columns ic
        WHERE ic.table_schema=:table_schema
        AND ic.table_name=:table_name
        ORDER BY ic.ordinal_position"""
        ), {'table_schema': self.denormalize_name(schema), 'table_name':
        self.denormalize_name(table_name)})
    for table_name, bdah__eps, txp__kpssv, hpdi__eycj, jbjji__focoo, vvrig__mxkff, rqlj__ftif, luj__jycpg, azaoc__ateuk, bssv__letr in tvoxf__khcot:
        table_name = self.normalize_name(table_name)
        bdah__eps = self.normalize_name(bdah__eps)
        if bdah__eps.startswith('sys_clustering_column'):
            continue
        oti__xgqv = self.ischema_names.get(txp__kpssv, None)
        ieq__tfq = {}
        if oti__xgqv is None:
            sa_util.warn("Did not recognize type '{}' of column '{}'".
                format(txp__kpssv, bdah__eps))
            oti__xgqv = sqltypes.NULLTYPE
        elif issubclass(oti__xgqv, sqltypes.FLOAT):
            ieq__tfq['precision'] = jbjji__focoo
            ieq__tfq['decimal_return_scale'] = vvrig__mxkff
        elif issubclass(oti__xgqv, sqltypes.Numeric):
            ieq__tfq['precision'] = jbjji__focoo
            ieq__tfq['scale'] = vvrig__mxkff
        elif issubclass(oti__xgqv, (sqltypes.String, sqltypes.BINARY)):
            ieq__tfq['length'] = hpdi__eycj
        wgsm__jaihv = oti__xgqv if isinstance(oti__xgqv, sqltypes.NullType
            ) else oti__xgqv(**ieq__tfq)
        amfhp__yjhgc = qhzh__xdpq.get(table_name)
        bxta__puqnv.append({'name': bdah__eps, 'type': wgsm__jaihv,
            'nullable': rqlj__ftif == 'YES', 'default': luj__jycpg,
            'autoincrement': azaoc__ateuk == 'YES', 'comment': bssv__letr if
            bssv__letr != '' else None, 'primary_key': bdah__eps in
            qhzh__xdpq[table_name]['constrained_columns'] if amfhp__yjhgc else
            False})
    return bxta__puqnv


if _check_snowflake_sqlalchemy_change:
    lines = inspect.getsource(snowflake.sqlalchemy.snowdialect.
        SnowflakeDialect._get_table_columns)
    if hashlib.sha256(lines.encode()).hexdigest(
        ) != '9ecc8a2425c655836ade4008b1b98a8fd1819f3be43ba77b0fbbfc1f8740e2be':
        warnings.warn(
            'snowflake.sqlalchemy.snowdialect.SnowflakeDialect._get_table_columns has changed'
            )
snowflake.sqlalchemy.snowdialect.SnowflakeDialect._get_table_columns = (
    _get_table_columns)
