"""helper functions for code generation with llvmlite
"""
import llvmlite.binding as ll
from llvmlite import ir as lir
from numba.core import cgutils, types
import bodo
from bodo.libs import array_ext, hdist
ll.add_symbol('array_getitem', array_ext.array_getitem)
ll.add_symbol('seq_getitem', array_ext.seq_getitem)
ll.add_symbol('list_check', array_ext.list_check)
ll.add_symbol('dict_keys', array_ext.dict_keys)
ll.add_symbol('dict_values', array_ext.dict_values)
ll.add_symbol('dict_merge_from_seq2', array_ext.dict_merge_from_seq2)
ll.add_symbol('is_na_value', array_ext.is_na_value)


def set_bitmap_bit(builder, null_bitmap_ptr, ind, val):
    cvu__zww = builder.lshr(ind, lir.Constant(lir.IntType(64), 3))
    yifzt__kvub = builder.urem(ind, lir.Constant(lir.IntType(64), 8))
    scf__lml = builder.gep(null_bitmap_ptr, [cvu__zww], inbounds=True)
    kocmw__wepu = builder.load(scf__lml)
    kuozd__qmkz = lir.ArrayType(lir.IntType(8), 8)
    bxhlo__oaskr = cgutils.alloca_once_value(builder, lir.Constant(
        kuozd__qmkz, (1, 2, 4, 8, 16, 32, 64, 128)))
    hryb__bfx = builder.load(builder.gep(bxhlo__oaskr, [lir.Constant(lir.
        IntType(64), 0), yifzt__kvub], inbounds=True))
    if val:
        builder.store(builder.or_(kocmw__wepu, hryb__bfx), scf__lml)
    else:
        hryb__bfx = builder.xor(hryb__bfx, lir.Constant(lir.IntType(8), -1))
        builder.store(builder.and_(kocmw__wepu, hryb__bfx), scf__lml)


def get_bitmap_bit(builder, null_bitmap_ptr, ind):
    cvu__zww = builder.lshr(ind, lir.Constant(lir.IntType(64), 3))
    yifzt__kvub = builder.urem(ind, lir.Constant(lir.IntType(64), 8))
    kocmw__wepu = builder.load(builder.gep(null_bitmap_ptr, [cvu__zww],
        inbounds=True))
    kuozd__qmkz = lir.ArrayType(lir.IntType(8), 8)
    bxhlo__oaskr = cgutils.alloca_once_value(builder, lir.Constant(
        kuozd__qmkz, (1, 2, 4, 8, 16, 32, 64, 128)))
    hryb__bfx = builder.load(builder.gep(bxhlo__oaskr, [lir.Constant(lir.
        IntType(64), 0), yifzt__kvub], inbounds=True))
    return builder.and_(kocmw__wepu, hryb__bfx)


def pyarray_getitem(builder, context, arr_obj, ind):
    qyyo__mgm = context.get_argument_type(types.pyobject)
    zkzu__usvzs = context.get_value_type(types.intp)
    wpmhr__lswy = lir.FunctionType(lir.IntType(8).as_pointer(), [qyyo__mgm,
        zkzu__usvzs])
    dsm__crxr = cgutils.get_or_insert_function(builder.module, wpmhr__lswy,
        name='array_getptr1')
    dbye__cbfvv = lir.FunctionType(qyyo__mgm, [qyyo__mgm, lir.IntType(8).
        as_pointer()])
    uxbsi__elrk = cgutils.get_or_insert_function(builder.module,
        dbye__cbfvv, name='array_getitem')
    ftdxr__pbl = builder.call(dsm__crxr, [arr_obj, ind])
    return builder.call(uxbsi__elrk, [arr_obj, ftdxr__pbl])


def pyarray_setitem(builder, context, arr_obj, ind, val_obj):
    qyyo__mgm = context.get_argument_type(types.pyobject)
    zkzu__usvzs = context.get_value_type(types.intp)
    wpmhr__lswy = lir.FunctionType(lir.IntType(8).as_pointer(), [qyyo__mgm,
        zkzu__usvzs])
    dsm__crxr = cgutils.get_or_insert_function(builder.module, wpmhr__lswy,
        name='array_getptr1')
    bgn__cgvn = lir.FunctionType(lir.VoidType(), [qyyo__mgm, lir.IntType(8)
        .as_pointer(), qyyo__mgm])
    mfip__txqkh = cgutils.get_or_insert_function(builder.module, bgn__cgvn,
        name='array_setitem')
    ftdxr__pbl = builder.call(dsm__crxr, [arr_obj, ind])
    builder.call(mfip__txqkh, [arr_obj, ftdxr__pbl, val_obj])


def seq_getitem(builder, context, obj, ind):
    qyyo__mgm = context.get_argument_type(types.pyobject)
    zkzu__usvzs = context.get_value_type(types.intp)
    qcs__bzu = lir.FunctionType(qyyo__mgm, [qyyo__mgm, zkzu__usvzs])
    oxtaj__djaz = cgutils.get_or_insert_function(builder.module, qcs__bzu,
        name='seq_getitem')
    return builder.call(oxtaj__djaz, [obj, ind])


def is_na_value(builder, context, val, C_NA):
    qyyo__mgm = context.get_argument_type(types.pyobject)
    cgdsu__ahl = lir.FunctionType(lir.IntType(32), [qyyo__mgm, qyyo__mgm])
    tfo__hup = cgutils.get_or_insert_function(builder.module, cgdsu__ahl,
        name='is_na_value')
    return builder.call(tfo__hup, [val, C_NA])


def list_check(builder, context, obj):
    qyyo__mgm = context.get_argument_type(types.pyobject)
    qce__szys = context.get_value_type(types.int32)
    xmdgk__nys = lir.FunctionType(qce__szys, [qyyo__mgm])
    uxot__ebb = cgutils.get_or_insert_function(builder.module, xmdgk__nys,
        name='list_check')
    return builder.call(uxot__ebb, [obj])


def dict_keys(builder, context, obj):
    qyyo__mgm = context.get_argument_type(types.pyobject)
    xmdgk__nys = lir.FunctionType(qyyo__mgm, [qyyo__mgm])
    uxot__ebb = cgutils.get_or_insert_function(builder.module, xmdgk__nys,
        name='dict_keys')
    return builder.call(uxot__ebb, [obj])


def dict_values(builder, context, obj):
    qyyo__mgm = context.get_argument_type(types.pyobject)
    xmdgk__nys = lir.FunctionType(qyyo__mgm, [qyyo__mgm])
    uxot__ebb = cgutils.get_or_insert_function(builder.module, xmdgk__nys,
        name='dict_values')
    return builder.call(uxot__ebb, [obj])


def dict_merge_from_seq2(builder, context, dict_obj, seq2_obj):
    qyyo__mgm = context.get_argument_type(types.pyobject)
    xmdgk__nys = lir.FunctionType(lir.VoidType(), [qyyo__mgm, qyyo__mgm])
    uxot__ebb = cgutils.get_or_insert_function(builder.module, xmdgk__nys,
        name='dict_merge_from_seq2')
    builder.call(uxot__ebb, [dict_obj, seq2_obj])


def to_arr_obj_if_list_obj(c, context, builder, val, typ):
    if not (isinstance(typ, types.List) or bodo.utils.utils.is_array_typ(
        typ, False)):
        return val
    enxgd__qdjaj = cgutils.alloca_once_value(builder, val)
    jttrl__hnho = list_check(builder, context, val)
    wfjvr__nzc = builder.icmp_unsigned('!=', jttrl__hnho, lir.Constant(
        jttrl__hnho.type, 0))
    with builder.if_then(wfjvr__nzc):
        vyx__qktu = context.insert_const_string(builder.module, 'numpy')
        bxrxx__zmb = c.pyapi.import_module_noblock(vyx__qktu)
        mywt__twtag = 'object_'
        if isinstance(typ, types.Array) or isinstance(typ.dtype, types.Float):
            mywt__twtag = str(typ.dtype)
        zjqg__vkvpe = c.pyapi.object_getattr_string(bxrxx__zmb, mywt__twtag)
        qlt__qfkov = builder.load(enxgd__qdjaj)
        epy__vyty = c.pyapi.call_method(bxrxx__zmb, 'asarray', (qlt__qfkov,
            zjqg__vkvpe))
        builder.store(epy__vyty, enxgd__qdjaj)
        c.pyapi.decref(bxrxx__zmb)
        c.pyapi.decref(zjqg__vkvpe)
    val = builder.load(enxgd__qdjaj)
    return val


def get_array_elem_counts(c, builder, context, arr_obj, typ):
    from bodo.libs.array_item_arr_ext import ArrayItemArrayType
    from bodo.libs.map_arr_ext import MapArrayType
    from bodo.libs.str_arr_ext import get_utf8_size, string_array_type
    from bodo.libs.struct_arr_ext import StructArrayType, StructType
    from bodo.libs.tuple_arr_ext import TupleArrayType
    if typ == bodo.string_type:
        ozqw__hsqf = c.pyapi.to_native_value(bodo.string_type, arr_obj).value
        ayag__ismud, ijgz__parhw = c.pyapi.call_jit_code(lambda a:
            get_utf8_size(a), types.int64(bodo.string_type), [ozqw__hsqf])
        context.nrt.decref(builder, typ, ozqw__hsqf)
        return cgutils.pack_array(builder, [ijgz__parhw])
    if isinstance(typ, (StructType, types.BaseTuple)):
        vyx__qktu = context.insert_const_string(builder.module, 'pandas')
        zxixf__jfyr = c.pyapi.import_module_noblock(vyx__qktu)
        C_NA = c.pyapi.object_getattr_string(zxixf__jfyr, 'NA')
        twd__qwpl = bodo.utils.transform.get_type_alloc_counts(typ)
        ohj__ooxgr = context.make_tuple(builder, types.Tuple(twd__qwpl * [
            types.int64]), twd__qwpl * [context.get_constant(types.int64, 0)])
        diuge__gotx = cgutils.alloca_once_value(builder, ohj__ooxgr)
        wwfbz__wbmje = 0
        zplg__mzs = typ.data if isinstance(typ, StructType) else typ.types
        for vemmb__vgi, t in enumerate(zplg__mzs):
            yoj__dssy = bodo.utils.transform.get_type_alloc_counts(t)
            if yoj__dssy == 0:
                continue
            if isinstance(typ, StructType):
                val_obj = c.pyapi.dict_getitem_string(arr_obj, typ.names[
                    vemmb__vgi])
            else:
                val_obj = c.pyapi.tuple_getitem(arr_obj, vemmb__vgi)
            bbxts__qyjsf = is_na_value(builder, context, val_obj, C_NA)
            jcfh__jhs = builder.icmp_unsigned('!=', bbxts__qyjsf, lir.
                Constant(bbxts__qyjsf.type, 1))
            with builder.if_then(jcfh__jhs):
                ohj__ooxgr = builder.load(diuge__gotx)
                smkr__uyclq = get_array_elem_counts(c, builder, context,
                    val_obj, t)
                for vemmb__vgi in range(yoj__dssy):
                    whaif__gtmj = builder.extract_value(ohj__ooxgr, 
                        wwfbz__wbmje + vemmb__vgi)
                    vsok__aim = builder.extract_value(smkr__uyclq, vemmb__vgi)
                    ohj__ooxgr = builder.insert_value(ohj__ooxgr, builder.
                        add(whaif__gtmj, vsok__aim), wwfbz__wbmje + vemmb__vgi)
                builder.store(ohj__ooxgr, diuge__gotx)
            wwfbz__wbmje += yoj__dssy
        c.pyapi.decref(zxixf__jfyr)
        c.pyapi.decref(C_NA)
        return builder.load(diuge__gotx)
    if not bodo.utils.utils.is_array_typ(typ, False):
        return cgutils.pack_array(builder, [], lir.IntType(64))
    n = bodo.utils.utils.object_length(c, arr_obj)
    if not (isinstance(typ, (ArrayItemArrayType, StructArrayType,
        TupleArrayType, MapArrayType)) or typ == string_array_type):
        return cgutils.pack_array(builder, [n])
    vyx__qktu = context.insert_const_string(builder.module, 'pandas')
    zxixf__jfyr = c.pyapi.import_module_noblock(vyx__qktu)
    C_NA = c.pyapi.object_getattr_string(zxixf__jfyr, 'NA')
    twd__qwpl = bodo.utils.transform.get_type_alloc_counts(typ)
    ohj__ooxgr = context.make_tuple(builder, types.Tuple(twd__qwpl * [types
        .int64]), [n] + (twd__qwpl - 1) * [context.get_constant(types.int64,
        0)])
    diuge__gotx = cgutils.alloca_once_value(builder, ohj__ooxgr)
    with cgutils.for_range(builder, n) as loop:
        ueh__ijz = loop.index
        fvs__lbz = seq_getitem(builder, context, arr_obj, ueh__ijz)
        bbxts__qyjsf = is_na_value(builder, context, fvs__lbz, C_NA)
        jcfh__jhs = builder.icmp_unsigned('!=', bbxts__qyjsf, lir.Constant(
            bbxts__qyjsf.type, 1))
        with builder.if_then(jcfh__jhs):
            if isinstance(typ, ArrayItemArrayType) or typ == string_array_type:
                ohj__ooxgr = builder.load(diuge__gotx)
                smkr__uyclq = get_array_elem_counts(c, builder, context,
                    fvs__lbz, typ.dtype)
                for vemmb__vgi in range(twd__qwpl - 1):
                    whaif__gtmj = builder.extract_value(ohj__ooxgr, 
                        vemmb__vgi + 1)
                    vsok__aim = builder.extract_value(smkr__uyclq, vemmb__vgi)
                    ohj__ooxgr = builder.insert_value(ohj__ooxgr, builder.
                        add(whaif__gtmj, vsok__aim), vemmb__vgi + 1)
                builder.store(ohj__ooxgr, diuge__gotx)
            elif isinstance(typ, (StructArrayType, TupleArrayType)):
                wwfbz__wbmje = 1
                for vemmb__vgi, t in enumerate(typ.data):
                    yoj__dssy = bodo.utils.transform.get_type_alloc_counts(t
                        .dtype)
                    if yoj__dssy == 0:
                        continue
                    if isinstance(typ, TupleArrayType):
                        val_obj = c.pyapi.tuple_getitem(fvs__lbz, vemmb__vgi)
                    else:
                        val_obj = c.pyapi.dict_getitem_string(fvs__lbz, typ
                            .names[vemmb__vgi])
                    bbxts__qyjsf = is_na_value(builder, context, val_obj, C_NA)
                    jcfh__jhs = builder.icmp_unsigned('!=', bbxts__qyjsf,
                        lir.Constant(bbxts__qyjsf.type, 1))
                    with builder.if_then(jcfh__jhs):
                        ohj__ooxgr = builder.load(diuge__gotx)
                        smkr__uyclq = get_array_elem_counts(c, builder,
                            context, val_obj, t.dtype)
                        for vemmb__vgi in range(yoj__dssy):
                            whaif__gtmj = builder.extract_value(ohj__ooxgr,
                                wwfbz__wbmje + vemmb__vgi)
                            vsok__aim = builder.extract_value(smkr__uyclq,
                                vemmb__vgi)
                            ohj__ooxgr = builder.insert_value(ohj__ooxgr,
                                builder.add(whaif__gtmj, vsok__aim), 
                                wwfbz__wbmje + vemmb__vgi)
                        builder.store(ohj__ooxgr, diuge__gotx)
                    wwfbz__wbmje += yoj__dssy
            else:
                assert isinstance(typ, MapArrayType), typ
                ohj__ooxgr = builder.load(diuge__gotx)
                uvij__hfj = dict_keys(builder, context, fvs__lbz)
                uxn__awds = dict_values(builder, context, fvs__lbz)
                mya__guvi = get_array_elem_counts(c, builder, context,
                    uvij__hfj, typ.key_arr_type)
                cgcvk__wuw = bodo.utils.transform.get_type_alloc_counts(typ
                    .key_arr_type)
                for vemmb__vgi in range(1, cgcvk__wuw + 1):
                    whaif__gtmj = builder.extract_value(ohj__ooxgr, vemmb__vgi)
                    vsok__aim = builder.extract_value(mya__guvi, vemmb__vgi - 1
                        )
                    ohj__ooxgr = builder.insert_value(ohj__ooxgr, builder.
                        add(whaif__gtmj, vsok__aim), vemmb__vgi)
                ilc__brnp = get_array_elem_counts(c, builder, context,
                    uxn__awds, typ.value_arr_type)
                for vemmb__vgi in range(cgcvk__wuw + 1, twd__qwpl):
                    whaif__gtmj = builder.extract_value(ohj__ooxgr, vemmb__vgi)
                    vsok__aim = builder.extract_value(ilc__brnp, vemmb__vgi -
                        cgcvk__wuw)
                    ohj__ooxgr = builder.insert_value(ohj__ooxgr, builder.
                        add(whaif__gtmj, vsok__aim), vemmb__vgi)
                builder.store(ohj__ooxgr, diuge__gotx)
                c.pyapi.decref(uvij__hfj)
                c.pyapi.decref(uxn__awds)
        c.pyapi.decref(fvs__lbz)
    c.pyapi.decref(zxixf__jfyr)
    c.pyapi.decref(C_NA)
    return builder.load(diuge__gotx)


def gen_allocate_array(context, builder, arr_type, n_elems, c=None):
    bnoou__vtvip = n_elems.type.count
    assert bnoou__vtvip >= 1
    adbzo__mwrq = builder.extract_value(n_elems, 0)
    if bnoou__vtvip != 1:
        oqklb__vpgqe = cgutils.pack_array(builder, [builder.extract_value(
            n_elems, vemmb__vgi) for vemmb__vgi in range(1, bnoou__vtvip)])
        oac__zubrr = types.Tuple([types.int64] * (bnoou__vtvip - 1))
    else:
        oqklb__vpgqe = context.get_dummy_value()
        oac__zubrr = types.none
    khou__ovk = types.TypeRef(arr_type)
    sstr__ylvh = arr_type(types.int64, khou__ovk, oac__zubrr)
    args = [adbzo__mwrq, context.get_dummy_value(), oqklb__vpgqe]
    twjit__xpcc = lambda n, t, s: bodo.utils.utils.alloc_type(n, t, s)
    if c:
        ayag__ismud, ojg__vjit = c.pyapi.call_jit_code(twjit__xpcc,
            sstr__ylvh, args)
    else:
        ojg__vjit = context.compile_internal(builder, twjit__xpcc,
            sstr__ylvh, args)
    return ojg__vjit


def is_ll_eq(builder, val1, val2):
    qbti__yvhrp = val1.type.pointee
    ayg__nsob = val2.type.pointee
    assert qbti__yvhrp == ayg__nsob, 'invalid llvm value comparison'
    if isinstance(qbti__yvhrp, (lir.BaseStructType, lir.ArrayType)):
        n_elems = len(qbti__yvhrp.elements) if isinstance(qbti__yvhrp, lir.
            BaseStructType) else qbti__yvhrp.count
        pqalh__vzlgv = lir.Constant(lir.IntType(1), 1)
        for vemmb__vgi in range(n_elems):
            lgh__umo = lir.IntType(32)(0)
            hkz__mihx = lir.IntType(32)(vemmb__vgi)
            omr__bpb = builder.gep(val1, [lgh__umo, hkz__mihx], inbounds=True)
            fdc__expg = builder.gep(val2, [lgh__umo, hkz__mihx], inbounds=True)
            pqalh__vzlgv = builder.and_(pqalh__vzlgv, is_ll_eq(builder,
                omr__bpb, fdc__expg))
        return pqalh__vzlgv
    rkxdp__rvdma = builder.load(val1)
    ywght__gafm = builder.load(val2)
    if rkxdp__rvdma.type in (lir.FloatType(), lir.DoubleType()):
        menb__nkrj = 32 if rkxdp__rvdma.type == lir.FloatType() else 64
        rkxdp__rvdma = builder.bitcast(rkxdp__rvdma, lir.IntType(menb__nkrj))
        ywght__gafm = builder.bitcast(ywght__gafm, lir.IntType(menb__nkrj))
    return builder.icmp_unsigned('==', rkxdp__rvdma, ywght__gafm)
