"""
helper data structures and functions for shuffle (alltoall).
"""
import os
from collections import namedtuple
import numba
import numpy as np
from numba import generated_jit
from numba.core import types
from numba.extending import overload
import bodo
from bodo.libs.array_item_arr_ext import offset_type
from bodo.libs.binary_arr_ext import binary_array_type, bytes_type
from bodo.libs.bool_arr_ext import BooleanArrayType, boolean_array
from bodo.libs.int_arr_ext import IntegerArrayType
from bodo.libs.str_arr_ext import convert_len_arr_to_offset, convert_len_arr_to_offset32, get_bit_bitmap, get_data_ptr, get_null_bitmap_ptr, get_offset_ptr, get_str_arr_item_length, num_total_chars, print_str_arr, set_bit_to, string_array_type
from bodo.libs.str_ext import string_type
from bodo.libs.timsort import getitem_arr_tup, setitem_arr_tup
from bodo.utils.utils import alloc_arr_tup, get_ctypes_ptr, numba_to_c_type
PreShuffleMeta = namedtuple('PreShuffleMeta',
    'send_counts, send_counts_char_tup, send_arr_lens_tup, send_arr_nulls_tup')
ShuffleMeta = namedtuple('ShuffleMeta',
    'send_counts, recv_counts, n_send, n_out, send_disp, recv_disp, send_disp_nulls, recv_disp_nulls, tmp_offset, send_buff_tup, out_arr_tup, send_counts_char_tup, recv_counts_char_tup, send_arr_lens_tup, send_arr_nulls_tup, send_arr_chars_tup, send_disp_char_tup, recv_disp_char_tup, tmp_offset_char_tup, send_arr_chars_arr_tup'
    )


def alloc_pre_shuffle_metadata(arr, data, n_pes, is_contig):
    return PreShuffleMeta(np.zeros(n_pes, np.int32), ())


@overload(alloc_pre_shuffle_metadata, no_unliteral=True)
def alloc_pre_shuffle_metadata_overload(key_arrs, data, n_pes, is_contig):
    qmmj__wxceb = 'def f(key_arrs, data, n_pes, is_contig):\n'
    qmmj__wxceb += '  send_counts = np.zeros(n_pes, np.int32)\n'
    mzh__vbz = len(key_arrs.types)
    umas__mevtc = mzh__vbz + len(data.types)
    for i, wwodu__wnl in enumerate(key_arrs.types + data.types):
        qmmj__wxceb += '  arr = key_arrs[{}]\n'.format(i
            ) if i < mzh__vbz else """  arr = data[{}]
""".format(i - mzh__vbz)
        if wwodu__wnl in [string_array_type, binary_array_type]:
            qmmj__wxceb += (
                '  send_counts_char_{} = np.zeros(n_pes, np.int32)\n'.format(i)
                )
            qmmj__wxceb += ('  send_arr_lens_{} = np.empty(0, np.uint32)\n'
                .format(i))
            qmmj__wxceb += '  if is_contig:\n'
            qmmj__wxceb += (
                '    send_arr_lens_{} = np.empty(len(arr), np.uint32)\n'.
                format(i))
        else:
            qmmj__wxceb += '  send_counts_char_{} = None\n'.format(i)
            qmmj__wxceb += '  send_arr_lens_{} = None\n'.format(i)
        if is_null_masked_type(wwodu__wnl):
            qmmj__wxceb += ('  send_arr_nulls_{} = np.empty(0, np.uint8)\n'
                .format(i))
            qmmj__wxceb += '  if is_contig:\n'
            qmmj__wxceb += '    n_bytes = (len(arr) + 7) >> 3\n'
            qmmj__wxceb += (
                '    send_arr_nulls_{} = np.full(n_bytes + 2 * n_pes, 255, np.uint8)\n'
                .format(i))
        else:
            qmmj__wxceb += '  send_arr_nulls_{} = None\n'.format(i)
    fdvl__jixyx = ', '.join('send_counts_char_{}'.format(i) for i in range(
        umas__mevtc))
    jxkhp__fpse = ', '.join('send_arr_lens_{}'.format(i) for i in range(
        umas__mevtc))
    usno__qhbov = ', '.join('send_arr_nulls_{}'.format(i) for i in range(
        umas__mevtc))
    gyrn__hzhrt = ',' if umas__mevtc == 1 else ''
    qmmj__wxceb += (
        '  return PreShuffleMeta(send_counts, ({}{}), ({}{}), ({}{}))\n'.
        format(fdvl__jixyx, gyrn__hzhrt, jxkhp__fpse, gyrn__hzhrt,
        usno__qhbov, gyrn__hzhrt))
    yecb__mgi = {}
    exec(qmmj__wxceb, {'np': np, 'PreShuffleMeta': PreShuffleMeta}, yecb__mgi)
    xmmm__xtap = yecb__mgi['f']
    return xmmm__xtap


def update_shuffle_meta(pre_shuffle_meta, node_id, ind, key_arrs, data,
    is_contig=True, padded_bits=0):
    pre_shuffle_meta.send_counts[node_id] += 1


@overload(update_shuffle_meta, no_unliteral=True)
def update_shuffle_meta_overload(pre_shuffle_meta, node_id, ind, key_arrs,
    data, is_contig=True, padded_bits=0):
    llb__dacu = 'BODO_DEBUG_LEVEL'
    avc__kak = 0
    try:
        avc__kak = int(os.environ[llb__dacu])
    except:
        pass
    qmmj__wxceb = """def f(pre_shuffle_meta, node_id, ind, key_arrs, data, is_contig=True, padded_bits=0):
"""
    qmmj__wxceb += '  pre_shuffle_meta.send_counts[node_id] += 1\n'
    if avc__kak > 0:
        qmmj__wxceb += ('  if pre_shuffle_meta.send_counts[node_id] >= {}:\n'
            .format(bodo.libs.distributed_api.INT_MAX))
        qmmj__wxceb += "    print('large shuffle error')\n"
    mzh__vbz = len(key_arrs.types)
    for i, wwodu__wnl in enumerate(key_arrs.types + data.types):
        if wwodu__wnl in (string_type, string_array_type, bytes_type,
            binary_array_type):
            arr = 'key_arrs[{}]'.format(i
                ) if i < mzh__vbz else 'data[{}]'.format(i - mzh__vbz)
            qmmj__wxceb += ('  n_chars = get_str_arr_item_length({}, ind)\n'
                .format(arr))
            qmmj__wxceb += (
                '  pre_shuffle_meta.send_counts_char_tup[{}][node_id] += n_chars\n'
                .format(i))
            if avc__kak > 0:
                qmmj__wxceb += (
                    '  if pre_shuffle_meta.send_counts_char_tup[{}][node_id] >= {}:\n'
                    .format(i, bodo.libs.distributed_api.INT_MAX))
                qmmj__wxceb += "    print('large shuffle error')\n"
            qmmj__wxceb += '  if is_contig:\n'
            qmmj__wxceb += (
                '    pre_shuffle_meta.send_arr_lens_tup[{}][ind] = n_chars\n'
                .format(i))
        if is_null_masked_type(wwodu__wnl):
            qmmj__wxceb += '  if is_contig:\n'
            qmmj__wxceb += (
                '    out_bitmap = pre_shuffle_meta.send_arr_nulls_tup[{}].ctypes\n'
                .format(i))
            if i < mzh__vbz:
                qmmj__wxceb += (
                    '    bit_val = get_mask_bit(key_arrs[{}], ind)\n'.format(i)
                    )
            else:
                qmmj__wxceb += ('    bit_val = get_mask_bit(data[{}], ind)\n'
                    .format(i - mzh__vbz))
            qmmj__wxceb += (
                '    set_bit_to(out_bitmap, padded_bits + ind, bit_val)\n')
    yecb__mgi = {}
    exec(qmmj__wxceb, {'set_bit_to': set_bit_to, 'get_bit_bitmap':
        get_bit_bitmap, 'get_null_bitmap_ptr': get_null_bitmap_ptr,
        'getitem_arr_tup': getitem_arr_tup, 'get_mask_bit': get_mask_bit,
        'get_str_arr_item_length': get_str_arr_item_length}, yecb__mgi)
    lxfe__kyy = yecb__mgi['f']
    return lxfe__kyy


@numba.njit
def calc_disp_nulls(arr):
    ntrt__ufpcm = np.empty_like(arr)
    ntrt__ufpcm[0] = 0
    for i in range(1, len(arr)):
        bcfgx__sbuje = arr[i - 1] + 7 >> 3
        ntrt__ufpcm[i] = ntrt__ufpcm[i - 1] + bcfgx__sbuje
    return ntrt__ufpcm


def finalize_shuffle_meta(arrs, data, pre_shuffle_meta, n_pes, is_contig,
    init_vals=()):
    return ShuffleMeta()


@overload(finalize_shuffle_meta, no_unliteral=True)
def finalize_shuffle_meta_overload(key_arrs, data, pre_shuffle_meta, n_pes,
    is_contig, init_vals=()):
    qmmj__wxceb = (
        'def f(key_arrs, data, pre_shuffle_meta, n_pes, is_contig, init_vals=()):\n'
        )
    qmmj__wxceb += '  send_counts = pre_shuffle_meta.send_counts\n'
    qmmj__wxceb += '  recv_counts = np.empty(n_pes, np.int32)\n'
    qmmj__wxceb += '  tmp_offset = np.zeros(n_pes, np.int32)\n'
    qmmj__wxceb += (
        '  bodo.libs.distributed_api.alltoall(send_counts, recv_counts, 1)\n')
    qmmj__wxceb += '  n_out = recv_counts.sum()\n'
    qmmj__wxceb += '  n_send = send_counts.sum()\n'
    qmmj__wxceb += '  send_disp = bodo.ir.join.calc_disp(send_counts)\n'
    qmmj__wxceb += '  recv_disp = bodo.ir.join.calc_disp(recv_counts)\n'
    qmmj__wxceb += '  send_disp_nulls = calc_disp_nulls(send_counts)\n'
    qmmj__wxceb += '  recv_disp_nulls = calc_disp_nulls(recv_counts)\n'
    mzh__vbz = len(key_arrs.types)
    umas__mevtc = len(key_arrs.types + data.types)
    for i, wwodu__wnl in enumerate(key_arrs.types + data.types):
        qmmj__wxceb += '  arr = key_arrs[{}]\n'.format(i
            ) if i < mzh__vbz else """  arr = data[{}]
""".format(i - mzh__vbz)
        if wwodu__wnl in [string_array_type, binary_array_type]:
            if wwodu__wnl == string_array_type:
                baxc__jfqz = 'bodo.libs.str_arr_ext.pre_alloc_string_array'
            else:
                baxc__jfqz = 'bodo.libs.binary_arr_ext.pre_alloc_binary_array'
            qmmj__wxceb += '  send_buff_{} = None\n'.format(i)
            qmmj__wxceb += (
                '  send_counts_char_{} = pre_shuffle_meta.send_counts_char_tup[{}]\n'
                .format(i, i))
            qmmj__wxceb += (
                '  recv_counts_char_{} = np.empty(n_pes, np.int32)\n'.format(i)
                )
            qmmj__wxceb += (
                """  bodo.libs.distributed_api.alltoall(send_counts_char_{}, recv_counts_char_{}, 1)
"""
                .format(i, i))
            qmmj__wxceb += ('  n_all_chars = recv_counts_char_{}.sum()\n'.
                format(i))
            qmmj__wxceb += '  out_arr_{} = {}(n_out, n_all_chars)\n'.format(i,
                baxc__jfqz)
            qmmj__wxceb += (
                '  send_disp_char_{} = bodo.ir.join.calc_disp(send_counts_char_{})\n'
                .format(i, i))
            qmmj__wxceb += (
                '  recv_disp_char_{} = bodo.ir.join.calc_disp(recv_counts_char_{})\n'
                .format(i, i))
            qmmj__wxceb += (
                '  tmp_offset_char_{} = np.zeros(n_pes, np.int32)\n'.format(i))
            qmmj__wxceb += (
                '  send_arr_lens_{} = pre_shuffle_meta.send_arr_lens_tup[{}]\n'
                .format(i, i))
            qmmj__wxceb += ('  send_arr_chars_arr_{} = np.empty(0, np.uint8)\n'
                .format(i))
            qmmj__wxceb += (
                '  send_arr_chars_{} = get_ctypes_ptr(get_data_ptr(arr))\n'
                .format(i))
            qmmj__wxceb += '  if not is_contig:\n'
            qmmj__wxceb += (
                '    send_arr_lens_{} = np.empty(n_send, np.uint32)\n'.
                format(i))
            qmmj__wxceb += ('    s_n_all_chars = send_counts_char_{}.sum()\n'
                .format(i))
            qmmj__wxceb += (
                '    send_arr_chars_arr_{} = np.empty(s_n_all_chars, np.uint8)\n'
                .format(i))
            qmmj__wxceb += (
                '    send_arr_chars_{} = get_ctypes_ptr(send_arr_chars_arr_{}.ctypes)\n'
                .format(i, i))
        else:
            assert isinstance(wwodu__wnl, (types.Array, IntegerArrayType,
                BooleanArrayType, bodo.CategoricalArrayType))
            qmmj__wxceb += (
                '  out_arr_{} = bodo.utils.utils.alloc_type(n_out, arr)\n'.
                format(i))
            qmmj__wxceb += '  send_buff_{} = arr\n'.format(i)
            qmmj__wxceb += '  if not is_contig:\n'
            if i >= mzh__vbz and init_vals != ():
                qmmj__wxceb += (
                    """    send_buff_{} = bodo.utils.utils.full_type(n_send, init_vals[{}], arr)
"""
                    .format(i, i - mzh__vbz))
            else:
                qmmj__wxceb += (
                    '    send_buff_{} = bodo.utils.utils.alloc_type(n_send, arr)\n'
                    .format(i))
            qmmj__wxceb += '  send_counts_char_{} = None\n'.format(i)
            qmmj__wxceb += '  recv_counts_char_{} = None\n'.format(i)
            qmmj__wxceb += '  send_arr_lens_{} = None\n'.format(i)
            qmmj__wxceb += '  send_arr_chars_{} = None\n'.format(i)
            qmmj__wxceb += '  send_disp_char_{} = None\n'.format(i)
            qmmj__wxceb += '  recv_disp_char_{} = None\n'.format(i)
            qmmj__wxceb += '  tmp_offset_char_{} = None\n'.format(i)
            qmmj__wxceb += '  send_arr_chars_arr_{} = None\n'.format(i)
        if is_null_masked_type(wwodu__wnl):
            qmmj__wxceb += (
                '  send_arr_nulls_{} = pre_shuffle_meta.send_arr_nulls_tup[{}]\n'
                .format(i, i))
            qmmj__wxceb += '  if not is_contig:\n'
            qmmj__wxceb += '    n_bytes = (n_send + 7) >> 3\n'
            qmmj__wxceb += (
                '    send_arr_nulls_{} = np.full(n_bytes + 2 * n_pes, 255, np.uint8)\n'
                .format(i))
        else:
            qmmj__wxceb += '  send_arr_nulls_{} = None\n'.format(i)
    knol__veok = ', '.join('send_buff_{}'.format(i) for i in range(umas__mevtc)
        )
    vih__wif = ', '.join('out_arr_{}'.format(i) for i in range(umas__mevtc))
    ubi__yfjni = ',' if umas__mevtc == 1 else ''
    xvfcc__hdyj = ', '.join('send_counts_char_{}'.format(i) for i in range(
        umas__mevtc))
    fiz__urfvo = ', '.join('recv_counts_char_{}'.format(i) for i in range(
        umas__mevtc))
    zdle__srb = ', '.join('send_arr_lens_{}'.format(i) for i in range(
        umas__mevtc))
    oxjm__rue = ', '.join('send_arr_nulls_{}'.format(i) for i in range(
        umas__mevtc))
    khv__rnwhh = ', '.join('send_arr_chars_{}'.format(i) for i in range(
        umas__mevtc))
    uhge__mrbk = ', '.join('send_disp_char_{}'.format(i) for i in range(
        umas__mevtc))
    atmhq__fjnth = ', '.join('recv_disp_char_{}'.format(i) for i in range(
        umas__mevtc))
    ttyad__nqqx = ', '.join('tmp_offset_char_{}'.format(i) for i in range(
        umas__mevtc))
    xlg__ypf = ', '.join('send_arr_chars_arr_{}'.format(i) for i in range(
        umas__mevtc))
    qmmj__wxceb += (
        """  return ShuffleMeta(send_counts, recv_counts, n_send, n_out, send_disp, recv_disp, send_disp_nulls, recv_disp_nulls, tmp_offset, ({}{}), ({}{}), ({}{}), ({}{}), ({}{}), ({}{}), ({}{}), ({}{}), ({}{}), ({}{}), ({}{}), )
"""
        .format(knol__veok, ubi__yfjni, vih__wif, ubi__yfjni, xvfcc__hdyj,
        ubi__yfjni, fiz__urfvo, ubi__yfjni, zdle__srb, ubi__yfjni,
        oxjm__rue, ubi__yfjni, khv__rnwhh, ubi__yfjni, uhge__mrbk,
        ubi__yfjni, atmhq__fjnth, ubi__yfjni, ttyad__nqqx, ubi__yfjni,
        xlg__ypf, ubi__yfjni))
    yecb__mgi = {}
    exec(qmmj__wxceb, {'np': np, 'bodo': bodo, 'num_total_chars':
        num_total_chars, 'get_data_ptr': get_data_ptr, 'ShuffleMeta':
        ShuffleMeta, 'get_ctypes_ptr': get_ctypes_ptr, 'calc_disp_nulls':
        calc_disp_nulls}, yecb__mgi)
    lxhzv__hhfd = yecb__mgi['f']
    return lxhzv__hhfd


def alltoallv_tup(arrs, shuffle_meta, key_arrs):
    return arrs


@overload(alltoallv_tup, no_unliteral=True)
def alltoallv_tup_overload(arrs, meta, key_arrs):
    mzh__vbz = len(key_arrs.types)
    qmmj__wxceb = 'def f(arrs, meta, key_arrs):\n'
    if any(is_null_masked_type(t) for t in arrs.types):
        qmmj__wxceb += (
            '  send_counts_nulls = np.empty(len(meta.send_counts), np.int32)\n'
            )
        qmmj__wxceb += '  for i in range(len(meta.send_counts)):\n'
        qmmj__wxceb += (
            '    send_counts_nulls[i] = (meta.send_counts[i] + 7) >> 3\n')
        qmmj__wxceb += (
            '  recv_counts_nulls = np.empty(len(meta.recv_counts), np.int32)\n'
            )
        qmmj__wxceb += '  for i in range(len(meta.recv_counts)):\n'
        qmmj__wxceb += (
            '    recv_counts_nulls[i] = (meta.recv_counts[i] + 7) >> 3\n')
        qmmj__wxceb += (
            '  tmp_null_bytes = np.empty(recv_counts_nulls.sum(), np.uint8)\n')
    qmmj__wxceb += '  lens = np.empty(meta.n_out, np.uint32)\n'
    for i, wwodu__wnl in enumerate(arrs.types):
        if isinstance(wwodu__wnl, (types.Array, IntegerArrayType,
            BooleanArrayType, bodo.CategoricalArrayType)):
            qmmj__wxceb += (
                """  bodo.libs.distributed_api.alltoallv(meta.send_buff_tup[{}], meta.out_arr_tup[{}], meta.send_counts,meta.recv_counts, meta.send_disp, meta.recv_disp)
"""
                .format(i, i))
        else:
            assert wwodu__wnl in [string_array_type, binary_array_type]
            qmmj__wxceb += (
                '  offset_ptr_{} = get_offset_ptr(meta.out_arr_tup[{}])\n'.
                format(i, i))
            if offset_type.bitwidth == 32:
                qmmj__wxceb += (
                    """  bodo.libs.distributed_api.c_alltoallv(meta.send_arr_lens_tup[{}].ctypes, offset_ptr_{}, meta.send_counts.ctypes, meta.recv_counts.ctypes, meta.send_disp.ctypes, meta.recv_disp.ctypes, int32_typ_enum)
"""
                    .format(i, i))
            else:
                qmmj__wxceb += (
                    """  bodo.libs.distributed_api.c_alltoallv(meta.send_arr_lens_tup[{}].ctypes, lens.ctypes, meta.send_counts.ctypes, meta.recv_counts.ctypes, meta.send_disp.ctypes, meta.recv_disp.ctypes, int32_typ_enum)
"""
                    .format(i))
            qmmj__wxceb += (
                """  bodo.libs.distributed_api.c_alltoallv(meta.send_arr_chars_tup[{}], get_data_ptr(meta.out_arr_tup[{}]), meta.send_counts_char_tup[{}].ctypes,meta.recv_counts_char_tup[{}].ctypes, meta.send_disp_char_tup[{}].ctypes,meta.recv_disp_char_tup[{}].ctypes, char_typ_enum)
"""
                .format(i, i, i, i, i, i))
            if offset_type.bitwidth == 32:
                qmmj__wxceb += (
                    '  convert_len_arr_to_offset32(offset_ptr_{}, meta.n_out)\n'
                    .format(i))
            else:
                qmmj__wxceb += (
                    """  convert_len_arr_to_offset(lens.ctypes, offset_ptr_{}, meta.n_out)
"""
                    .format(i))
        if is_null_masked_type(wwodu__wnl):
            qmmj__wxceb += (
                '  null_bitmap_ptr_{} = get_arr_null_ptr(meta.out_arr_tup[{}])\n'
                .format(i, i))
            qmmj__wxceb += (
                """  bodo.libs.distributed_api.c_alltoallv(meta.send_arr_nulls_tup[{}].ctypes, tmp_null_bytes.ctypes, send_counts_nulls.ctypes, recv_counts_nulls.ctypes, meta.send_disp_nulls.ctypes, meta.recv_disp_nulls.ctypes, char_typ_enum)
"""
                .format(i))
            qmmj__wxceb += (
                """  copy_gathered_null_bytes(null_bitmap_ptr_{}, tmp_null_bytes, recv_counts_nulls, meta.recv_counts)
"""
                .format(i))
    qmmj__wxceb += '  return ({}{})\n'.format(','.join([
        'meta.out_arr_tup[{}]'.format(i) for i in range(arrs.count)]), ',' if
        arrs.count == 1 else '')
    dovsd__rzjc = np.int32(numba_to_c_type(types.int32))
    zlzs__ugli = np.int32(numba_to_c_type(types.uint8))
    yecb__mgi = {}
    exec(qmmj__wxceb, {'np': np, 'bodo': bodo, 'get_offset_ptr':
        get_offset_ptr, 'get_data_ptr': get_data_ptr, 'int32_typ_enum':
        dovsd__rzjc, 'char_typ_enum': zlzs__ugli,
        'convert_len_arr_to_offset': convert_len_arr_to_offset,
        'convert_len_arr_to_offset32': convert_len_arr_to_offset32,
        'copy_gathered_null_bytes': bodo.libs.distributed_api.
        copy_gathered_null_bytes, 'get_arr_null_ptr': get_arr_null_ptr,
        'print_str_arr': print_str_arr}, yecb__mgi)
    orgdf__xsev = yecb__mgi['f']
    return orgdf__xsev


def shuffle_with_index_impl(key_arrs, node_arr, data):
    n_pes = bodo.libs.distributed_api.get_size()
    pre_shuffle_meta = alloc_pre_shuffle_metadata(key_arrs, data, n_pes, False)
    kqu__vgc = len(key_arrs[0])
    orig_indices = np.arange(kqu__vgc)
    cogsv__sbt = np.empty(kqu__vgc, np.int32)
    for i in range(kqu__vgc):
        val = getitem_arr_tup_single(key_arrs, i)
        node_id = node_arr[i]
        cogsv__sbt[i] = node_id
        update_shuffle_meta(pre_shuffle_meta, node_id, i, key_arrs, data, False
            )
    shuffle_meta = finalize_shuffle_meta(key_arrs, data, pre_shuffle_meta,
        n_pes, False)
    for i in range(kqu__vgc):
        val = getitem_arr_tup_single(key_arrs, i)
        node_id = cogsv__sbt[i]
        olrs__nfy = bodo.ir.join.write_send_buff(shuffle_meta, node_id, i,
            key_arrs, data)
        orig_indices[olrs__nfy] = i
        shuffle_meta.tmp_offset[node_id] += 1
    recvs = alltoallv_tup(key_arrs + data, shuffle_meta, key_arrs)
    jauj__sqbv = _get_keys_tup(recvs, key_arrs)
    ozqla__nlch = _get_data_tup(recvs, key_arrs)
    return jauj__sqbv, ozqla__nlch, orig_indices, shuffle_meta


@generated_jit(nopython=True, cache=True)
def shuffle_with_index(key_arrs, node_arr, data):
    return shuffle_with_index_impl


@numba.njit(cache=True)
def reverse_shuffle(data, orig_indices, shuffle_meta):
    vih__wif = alloc_arr_tup(shuffle_meta.n_send, data)
    mdbsc__vczg = ShuffleMeta(shuffle_meta.recv_counts, shuffle_meta.
        send_counts, shuffle_meta.n_out, shuffle_meta.n_send, shuffle_meta.
        recv_disp, shuffle_meta.send_disp, shuffle_meta.recv_disp_nulls,
        shuffle_meta.send_disp_nulls, shuffle_meta.tmp_offset, data,
        vih__wif, shuffle_meta.recv_counts_char_tup, shuffle_meta.
        send_counts_char_tup, shuffle_meta.send_arr_lens_tup, shuffle_meta.
        send_arr_nulls_tup, shuffle_meta.send_arr_chars_tup, shuffle_meta.
        recv_disp_char_tup, shuffle_meta.send_disp_char_tup, shuffle_meta.
        tmp_offset_char_tup, shuffle_meta.send_arr_chars_arr_tup)
    vih__wif = alltoallv_tup(data, mdbsc__vczg, ())
    yub__tzcrq = alloc_arr_tup(shuffle_meta.n_send, data)
    for i in range(len(orig_indices)):
        setitem_arr_tup(yub__tzcrq, orig_indices[i], getitem_arr_tup(
            vih__wif, i))
    return yub__tzcrq


def _get_keys_tup(recvs, key_arrs):
    return recvs[:len(key_arrs)]


@overload(_get_keys_tup, no_unliteral=True)
def _get_keys_tup_overload(recvs, key_arrs):
    mzh__vbz = len(key_arrs.types)
    qmmj__wxceb = 'def f(recvs, key_arrs):\n'
    edd__rjiq = ','.join('recvs[{}]'.format(i) for i in range(mzh__vbz))
    qmmj__wxceb += '  return ({}{})\n'.format(edd__rjiq, ',' if mzh__vbz ==
        1 else '')
    yecb__mgi = {}
    exec(qmmj__wxceb, {}, yecb__mgi)
    izbd__eouny = yecb__mgi['f']
    return izbd__eouny


def _get_data_tup(recvs, key_arrs):
    return recvs[len(key_arrs):]


@overload(_get_data_tup, no_unliteral=True)
def _get_data_tup_overload(recvs, key_arrs):
    mzh__vbz = len(key_arrs.types)
    umas__mevtc = len(recvs.types)
    lrdfn__ioih = umas__mevtc - mzh__vbz
    qmmj__wxceb = 'def f(recvs, key_arrs):\n'
    edd__rjiq = ','.join('recvs[{}]'.format(i) for i in range(mzh__vbz,
        umas__mevtc))
    qmmj__wxceb += '  return ({}{})\n'.format(edd__rjiq, ',' if lrdfn__ioih ==
        1 else '')
    yecb__mgi = {}
    exec(qmmj__wxceb, {}, yecb__mgi)
    izbd__eouny = yecb__mgi['f']
    return izbd__eouny


def getitem_arr_tup_single(arrs, i):
    return arrs[0][i]


@overload(getitem_arr_tup_single, no_unliteral=True)
def getitem_arr_tup_single_overload(arrs, i):
    if len(arrs.types) == 1:
        return lambda arrs, i: arrs[0][i]
    return lambda arrs, i: getitem_arr_tup(arrs, i)


def val_to_tup(val):
    return val,


@overload(val_to_tup, no_unliteral=True)
def val_to_tup_overload(val):
    if isinstance(val, types.BaseTuple):
        return lambda val: val
    return lambda val: (val,)


def is_null_masked_type(t):
    return t in (string_type, string_array_type, bytes_type,
        binary_array_type, boolean_array) or isinstance(t, IntegerArrayType)


@numba.generated_jit(nopython=True, no_cpython_wrapper=True)
def get_mask_bit(arr, i):
    if arr in [string_array_type, binary_array_type]:
        return lambda arr, i: get_bit_bitmap(get_null_bitmap_ptr(arr), i)
    assert isinstance(arr, IntegerArrayType) or arr == boolean_array
    return lambda arr, i: bodo.libs.int_arr_ext.get_bit_bitmap_arr(arr.
        _null_bitmap, i)


@numba.generated_jit(nopython=True, no_cpython_wrapper=True)
def get_arr_null_ptr(arr):
    if arr in [string_array_type, binary_array_type]:
        return lambda arr: get_null_bitmap_ptr(arr)
    assert isinstance(arr, IntegerArrayType) or arr == boolean_array
    return lambda arr: arr._null_bitmap.ctypes
